// ESAF : Euso Simulation and Analysis Framework
// $Id: AtmosphereData.hh 2749 2006-07-04 13:16:23Z moreggia $
// Sylvain Moreggia created Dec,  1 2003

#ifndef __ATMOSPHEREDATA_HH_
#define __ATMOSPHEREDATA_HH_

#include "euso.hh"
#include "Config.hh"
#include "EsafConfigurable.hh"
#include "EsafMsgSource.hh"
#include "EConst.hh"

/*******************************************************************************
 *
 * AtmosphereData : class description
 *
 * Interface for atmospheric data storage
 * Every atmosphere holds its own version of AtmosphereData
 *
 ******************************************************************************/


class AtmosphereData : public EsafConfigurable, public EsafMsgSource{
public:
    // ctor
    AtmosphereData();

    // dtor
    virtual ~AtmosphereData();

    // Following methods return relevant tables
    inline const Double_t* GetAltitudeTable() const {return fAltitudeTable;}
    inline const Double_t* GetPressureTable() const {return fPressureTable;}
    inline const Double_t* GetTemperatureTable() const {return fTemperatureTable;}
    inline const Double_t* GetAir_DensityTable() const {return fAir_DensityTable;}
    inline const Double_t* GetO_DensityTable() const {return fO_DensityTable;}
    inline const Double_t* GetO2_DensityTable() const {return fO2_DensityTable;}
    inline const Double_t* GetH2O_DensityTable() const {return fH2O_DensityTable;}
    inline const Double_t* GetO3_DensityTable() const {return fO3_DensityTable;}
    inline const Double_t* GetO3_DensityPPMVTable() const {return fO3_DensityPPMVTable;}
    inline const Double_t* GetN2_DensityTable() const {return fN2_DensityTable;}
    inline const Double_t* GetAerosol_DensityTable(string& type) const {return fAerosols[type];}
    inline const Double_t* GetIndexTable() const {return fIndexTable;}

    // within [-90,90] (positive <==> north)
    virtual Double_t GetLatitude() const = 0;
    
    // within [0,360[ (toward east)
    virtual Double_t GetLongitude() const = 0;
    
    // get day of year
    virtual Int_t GetDoy() const = 0;
    
    // NB : day starts from 1 and can reach 365
    // in sec UT, 1sec <==> 1st Jan 00:00:01 (year not taken into account)
    virtual Double_t GetDate() const = 0;
    

   
    EsafConfigClass(Atmosphere,AtmosphereData)

protected:
    // Set tables
    virtual void SetTables() = 0;
    
    // fill missing data with default values
    void GetDefault(string&, Int_t);
    
    // fill Lwtrn_Tape5 with atmospheric values, used by lowtran calculations
    void WriteUserModelFile(Int_t);

    // to achieve constant steps when GetDefault is called
    Double_t LinearInterpolateLowtran(Int_t h, const vector<Double_t>&, const vector<Double_t>&);
    
    // Plot the implemented tables
    //void DebugPlots(Int_t) const;
    //Double_t Smooth(string& , Int_t, Double_t) const;

    Double_t* fAltitudeTable;               // tables ->..
    Double_t* fPressureTable;
    Double_t* fTemperatureTable;
    Double_t* fAir_DensityTable;
    Double_t* fO_DensityTable;
    Double_t* fO2_DensityTable;
    Double_t* fO3_DensityTable;
    Double_t* fO3_DensityPPMVTable;
    Double_t* fN2_DensityTable;
    Double_t* fCO2_DensityTable;
    Double_t* fH2O_DensityTable;
    mutable map<string,Double_t*> fAerosols;
    Double_t* fIndexTable;                   // ..<- tables


    ClassDef(AtmosphereData,0)
};



#endif  /* __ATMOSPHEREDATA_HH_ */

