// $Id: LowtranAerosol.hh 2676 2006-04-28 15:22:49Z moreggia $
// Author: Sylvain Moreggia   2006/04/24

/*****************************************************************************
 * ESAF: Euso Simulation and Analysis Framework                              *
 *                                                                           *
 *  Id: LowtranAerosol                                                           *
 *  Package: <packagename>                                                   *
 *  Coordinator: <coordinator>                                               *
 *                                                                           *
 *****************************************************************************/

#ifndef __LOWTRANAEROSOL_HH__
#define __LOWTRANAEROSOL_HH__

#include "euso.hh"
#include "Aerosol.hh"

////////////////////////////////////////////////////////////////////////////////
//                                                                            //
// LowtranAerosol                                                             //
//                                                                            //  
// Copies of lowtran aerosol models :                                         //
//                                                                            //  
// - only rurals, urban and maritime models                                   //  
//  (desert and other models more difficult to extract fom the fortan code)   //  
//                                                                            //  
// - change of aerosol properties as function of local humidity along path    //  
//   not taken into account. As std models give rel. humidity ~ 50-70%        //  
//   within boudary layers, models at rel. hum. = 70% are copied              //  
//                                                                            //  
// - Profiles are not dependent of ground altitude. They're always            //  
//   given w.r.t. sea level                                                   // 
//                                                                            //  
// - Only properties within [200nm,1536nm]  are copied                        //  
//                                                                            //  
////////////////////////////////////////////////////////////////////////////////

enum AerZone{BOUND1=1, BOUND2=2, ALL=3};
// BOUND1   Boundary layer in [0-1km]
// BOUND2   Boundary layer in [1-2km]
// ALL      whole boundary layer in [0-2km]

class LowtranAerosol : public Aerosol {
public:
    // ctor
    LowtranAerosol(string);
    //dtor
    virtual ~LowtranAerosol();

    // Return impact for given position and direction
    // Convention :    - if already within aerosol, returns pos
    //                 - no impact, returns (0,0,HUGE)
    virtual EarthVector GetImpact(const EarthVector& pos, const EarthVector& dir,string opt = "default") const;

    // to know if a position is within aerosol
    virtual Bool_t IsInAerosol(const EarthVector&) const;

    // Returns the transmission at given wl, along given track
    // set "scat" to the scattering rate along a direction
    virtual Double_t Trans(const EarthVector&,const EarthVector&,Double_t, Double_t& scat) const;
    
    // get a scattering position along a given track
    // depends on  wavelength
    // return kFALSE if absorption process occurs
    virtual Bool_t RandomScatPos(const EarthVector&,const EarthVector&,Double_t,EarthVector&) const;
    
    // set incoming direction to the outgoing direction after scattering
    virtual void RandomDir(EarthVector&,Double_t) const;
    
    // aerosol scattering phase function - Normalized when integrated over full solid angle
    // depends on  wavelength
    virtual Double_t PhaseFunction(const EarthVector&,const EarthVector&,Double_t) const;
    
    // aerosol scattering phase function - Normalized when integrated over full solid angle
    // depends on  wavelength
    Double_t PhaseFunction(Double_t,Double_t) const;
    
    // returns maximum value of currently used scattering phase function
    // depends on  wavelength range
    virtual Double_t GetMaxPhaseFunction(Double_t wlmin,Double_t wlmax) const;
    
    // interpolate fKwl coeff.
    // if abs=kTRUE, return value for absorption, else return value for extinction
    virtual Double_t Kwl(Double_t,Bool_t abs = kFALSE) const;
    
    // reset the instance
    virtual void Reset();

    EsafConfigClass(Atmosphere,LowtranAerosol)

private:
    
    // read config files
    virtual void Configure();

    // build the configured object
    virtual void Build();
    
    // true if track going locally downward
    Bool_t GoingDownward(const EarthVector&,const EarthVector&) const;
    
    // Like IsInAerosol, but deals with aerosol layers zone
    Bool_t IsInAerosolZone(const EarthVector&,AerZone) const;
    
    // check if wavelength is within bounds
    Bool_t CheckWlRange(Double_t) const;

    // Like GetImpact, but deals with aerosol layers zone
    EarthVector GetZoneImpact(const EarthVector& pos, const EarthVector& dir,AerZone zone,string opt = "default") const;
    
    string fType;                      // lowtran aerosol model
    Double_t fSz[3];                   // used for aerosol altitude dependence. Values at 0, 1 and 2km
    Double_t fWlrange[7];              // wavelength ranges used in lowtran
    Double_t fKwl_ext[7];              // used for extinction wavelength dependence
    Double_t fKwl_abs[7];              // used for absorption wavelength dependence
    Double_t fG1[7];                   // Assymetry coeff1. for DHG phase function parametrization
    Double_t fG2[7];                   // Assymetry coeff2. for DHG phase function parametrization
    Double_t fF[7];                    // Forward component for DHG phase function parametrization
   


    ClassDef(LowtranAerosol,0)
};

#endif  /* __LOWTRANAEROSOL_HH__ */

