// $Id: Atmosphere.cc 2819 2009-03-13 11:48:16Z naumov $
// S. Moreggia created 27 October 2003

/*****************************************************************************
 * ESAF: Euso Simulation and Analysis Framework                              *
 *                                                                           *
 *  Id: Atmosphere                                                           *
 *  Package: atmosphere                                                      *
 *  Coordinator: S. Moreggia                                                 *
 *                                                                           *
 *****************************************************************************/

//_____________________________________________________________________________
//
// Atmosphere
//
// <extensive class description>
//
//   Config file parameters
//   ======================
//
//   <parameter name>: <parameter description>
//   -Valid options: <available options>
//

#include "Atmosphere.hh"
#include "AtmosphereFactory.hh"
#include "LinsleyAtmosphere.hh"
#include "LowtranAtmosphere.hh"
#include "MSISE_00Atmosphere.hh"
#include "Config.hh"
#include "EConst.hh"
#include "utils.hh"
#include "NumbersFileParser.hh"

#include <TF1.h>
#include <TSpline.h>

ClassImp(Atmosphere)

using namespace sou;
using namespace TMath;
using namespace EConst;

Atmosphere* Atmosphere::fChild = NULL;

//______________________________________________________________________________
Double_t densityIntegral(Double_t *x, Double_t *par)
{
    // 
    Double_t Altitude = par[0],  Angle = par[1];
    Double_t H0=8.43*km; // An auxiulary constant, km
    Double_t h = Altitude - H0*TMath::Log(x[0]);
    Double_t f = ((EarthRadius() + Altitude)/(EarthRadius() + h))*TMath::Sin(Angle);
    return H0*Atmosphere::Get()->Air_Density(h)/TMath::Sqrt(1.-TMath::Power(f,2))/x[0];
}

//______________________________________________________________________________
Atmosphere::Atmosphere() : EsafConfigurable(), EsafMsgSource() {
    //
    // ctor
    //
    
    // builds clouds and aerosol if required
    fClouds = AtmosphereFactory::Get()->GetClouds();
    if(!fClouds) Msg(EsafMsg::Panic) <<"No clouds built : memory problems"<< MsgDispatch;
    fAerosol = AtmosphereFactory::Get()->GetAerosol();
    if(!fAerosol) Msg(EsafMsg::Panic) <<"No aerosol built : memory problems"<< MsgDispatch;
    
    // init
    fDepthCalculationPrecision = 1.e-5;
    fTempInterpol = 0;
    fDensInterpol = 0;
    fPressInterpol = 0;
    fMSISErandom = kFALSE;
    fTOA = 100*km;
    
    // read ozone coeff.
    fWvNBmin = 27370.;
    fWvNBmax = 40800.;
    NumbersFileParser nf0("config/Atmosphere/O3Crossec/O3CrossecT0",1);
    vector<Double_t> temp0   = nf0.GetCol(0);
    NumbersFileParser nf1("config/Atmosphere/O3Crossec/O3CrossecT1",1);
    vector<Double_t> temp1   = nf1.GetCol(0);
    NumbersFileParser nf2("config/Atmosphere/O3Crossec/O3CrossecT2",1);
    vector<Double_t> temp2   = nf2.GetCol(0);
    for(Int_t i=0; i < 2690; i++) {
        if(i < 2687) fC0[i] = temp0[i];
        fC1[i] = temp1[i];
        fC2[i] = temp2[i];
    }
}

//______________________________________________________________________________
Atmosphere::~Atmosphere() {
    //
    // dtor
    //
    SafeDelete(fClouds);
    SafeDelete(fAerosol);
    SafeDelete(fTempInterpol);
    SafeDelete(fDensInterpol);
    SafeDelete(fPressInterpol);
}

//______________________________________________________________________________
void Atmosphere::Delete() {
    //
    // staic method to call dtor
    //
    
    SafeDelete(fChild);
}

//______________________________________________________________________________
const Atmosphere* Atmosphere::Get() {
    //
    // Static method to get the right atmosphere singleton from anywhere in the code
    //
    if(fChild == 0) {
        ConfigFileParser *pConfig = Config::Get()->GetCF("Atmosphere","AtmosphereFactory");
        string type = pConfig->GetStr("AtmosphereFactory.Atmosphere.type");
        
	if(type == "linsley")
            LinsleyAtmosphere::CreateInstance();
        
	else if(type == "msise00")
	    MSISE_00Atmosphere::CreateInstance();

	else if(type == "RandomMsise00") {
	    MSISE_00Atmosphere::CreateInstance();
	    fChild->SetRandomMode();
	}

        else if(type == "lowtran")
            LowtranAtmosphere::CreateInstance(); 

        else cout<<"Wrong type of atmosphere\n";
        fChild->Msg(EsafMsg::Info) << "Atmosphere built" << MsgDispatch;
    }
    
     return fChild;
}
    
//______________________________________________________________________________
void Atmosphere::Reset() {
    //
    // Static method to reset atmosphere : usefull only in MSISErandom mode
    //
    
    if(fChild) {
        if(fChild->IsRandomMode()) {
	    MSISE_00Atmosphere* atmo = (MSISE_00Atmosphere*)fChild;
	    atmo->ResetInstance();
	}
    }
}

//______________________________________________________________________________
Double_t Atmosphere::GetOzoneCoeff(Double_t wl, UInt_t i) const {
    //
    // get Ozone C coeff. using fC tables (i=0 C0, i=1 C1, i=2 C2)
    // Lowtran tables have been generated using a WvNB stepwidth of 5
    //
    
    Double_t rtn = 0.;
    Double_t WvNB = TranslateInWvNB(wl);
    if(WvNB < fWvNBmin || WvNB > fWvNBmax) return 0.;
    Int_t index = Int_t((WvNB - fWvNBmin) / 5.);
    if(i == 0)      rtn = fC0[index];
    else if(i == 1) rtn = fC1[index];
    else if(i == 2) rtn = fC2[index];
    else Msg(EsafMsg::Panic) << "<GetOzoneCoeff> Wrong argument asked" << MsgDispatch;
    
    return rtn;
}

//______________________________________________________________________________
Double_t Atmosphere::Index(Double_t h, Double_t wl) const {
    //
    // Returns air index value at a given altitude
    // By default, it is taken CONSTANT with wavelength (lambda=350nm)
    // (formula drawn from Reinhard Beer, Hanbook of Optics, Chap.2)
    //
    Double_t rtn;
    rtn = Index_Minus1(h,wl);
    return 1 + rtn;
}

//______________________________________________________________________________
Double_t Atmosphere::Index_Minus1(Double_t h, Double_t wl) const {
    //
    // Returns (n-1) at a given altitude
    // By default, it is taken CONSTANT with wavelength (lambda=350nm)
    // (formula drawn from Reinhard Beer, Hanbook of Optics, Chap.2)
    //
    Double_t rtn;
    rtn = 88.43 + 185.08/(1 - 1/pow(wl/cm*1.14e5,2)) + 4.11/(1 - 1/pow(wl/cm*6.24e4,2));
    rtn *= (Pressure(h) - WaterVaporPartialPressure(h))/Temperature(h);
    rtn *= 296.15*kelvin/atmosphere;
    rtn += (43.49 - 1/pow(wl/cm*1.7e4,2))*WaterVaporPartialPressure(h)/atmosphere;
    return rtn*1e-6; 
}

//______________________________________________________________________________
Double_t Atmosphere::Grammage(const EarthVector& V_1, const EarthVector& V_2, string opt, Double_t maxalt) const {
    //
    // Calculate grammage
    // - opt = pos : between two positions V1, V2 in the atmosphere
    // - opt = dir : from a point V1 until atmosphere top, along a track of given angles (V2 thus used for direction)
    //
    Double_t dl = 500*m;
    Double_t dX, rho2, h2;
    EarthVector U, inter;
    EarthVector V1(V_1), V2(V_2);
    Double_t X = 0;
    Double_t rho1;
    Bool_t IsTooHigh = false;
    if(maxalt == -HUGE) maxalt = fTOA;
    
    if(opt == "pos") {
	EarthVector Vmin;
	EarthVector Vmax;
	if(V1.IsUnderSeaLevel() || V2.IsUnderSeaLevel()) {
	    if(V1.IsUnderSeaLevel() && V2.IsUnderSeaLevel()) {
	        Msg(EsafMsg::Warning) << "<Grammage> Both V1 and V2 are under sea level --> SHOULD NOT OCCUR, zero returned"<<MsgDispatch;
	        return 0.;
	    }
	    Msg(EsafMsg::Warning) << "<Grammage> V1 or V2 is under sea level --> search for the true track impact"<<MsgDispatch;
	    EarthVector impact(1);
	    impact = ImpactASL(V1,(V2 - V1).Unit());
	    if(impact.Z() == -HUGE) {
	        impact = ImpactASL(V2,(V1 - V2).Unit());
	        if(impact.Z() == HUGE || impact.Z() == -HUGE) {
	            Msg(EsafMsg::Warning) << "<Grammage> Failed to find an impact, zero returned"<<MsgDispatch;
	            return 0.;
	        }
	        else V1 = impact;
	    }
	    else V2 = impact;
	}
	
	if(V1.Zv() <= V2.Zv()) {
            Vmin = V1;
	    Vmax = V2;
	}
	else {
            Vmin = V2;
	    Vmax = V1;
	}

	if(Vmax.Zv() > maxalt) IsTooHigh = true;
	U = (Vmax - Vmin).Unit();
	inter = Vmin;
	rho1 = Air_Density(Vmin.Zv());
	
        while((Vmax - Vmin).Mag() > (inter - Vmin).Mag()) {
	    if( (inter + U*dl - Vmin).Mag() >= (Vmax - Vmin).Mag() ) {
	        rho2 = Air_Density(Vmax.Zv());
	        dX = 0.5*(rho1 + rho2) * (Vmax - inter).Mag();
	        X += dX;
	        break;
	    }
	    inter += U*dl;
	    h2 = inter.Zv();
	    if(h2 > maxalt && IsTooHigh) break;
	    rho2 = Air_Density(h2);
	    dX = 0.5*(rho1 + rho2) * dl;
	    X += dX;
	    rho1 = rho2;
        }
    }
	
	    
    else if(opt == "dir") {
	if(V1.IsUnderSeaLevel()) Msg(EsafMsg::Warning) << "<Grammage> V1 is under sea level : it SHOULD NOT"<<MsgDispatch;
        EarthVector TOA = ImpactAtTOA(V1,V2,fTOA);
	if(TOA.Z() == -HUGE) {
	    Msg(EsafMsg::Warning) << "<Grammage(\"dir\")> : starting position is UNDERSEALEVEL, O returned" << MsgDispatch;
	    return 0.;
	}
	if(TOA.Z() == HUGE) {
	    Msg(EsafMsg::Warning) << "<Grammage(\"dir\")> : track DOES NOT GO OUT atmosphere, O returned" << MsgDispatch;
	    return 0.;
	}
	U = V2.Unit();
	inter = V1;
	rho1 =  Air_Density(V1.Zv());
	while(true) {
	    inter += U*dl;
	    h2 = inter.Zv();
	    rho2 = Air_Density(h2);
	    dX = 0.5*(rho1 + rho2) * dl;
	    X += dX;
	    rho1 = rho2;
	    if(inter.Mag() > (TOA - V1).Mag()) break;
        }
   }
   
    else  Msg(EsafMsg::Warning) << "Wrong option for grammage calculation" << MsgDispatch;

    return X;
}

//______________________________________________________________________________
Double_t Atmosphere::OzoneAmountAlongPath(const EarthVector& V_1, const EarthVector& V_2, Double_t wl) const {
    //
    // Get ozone amount along the path, at given wavelength
    // Integrate    X_path(s,wl) = Crossec(s,wl) * OzDens(s) * ds     along the defined path
    //
    
    // init
    Double_t dl = 500*m;
    Double_t maxalt = 60*km;  // Ozone amount negligible above this altitude
    Double_t dX, rho2, h2;
    EarthVector U, inter;
    EarthVector V1(V_1), V2(V_2);
    Double_t X = 0;
    Double_t rho1;
    Bool_t IsTooHigh = false;
    
    // find wavelength dependent Ozone coeff.
    Double_t C0 = GetOzoneCoeff(wl,0);
    Double_t C1 = GetOzoneCoeff(wl,1);
    Double_t C2 = GetOzoneCoeff(wl,2);
    
    EarthVector Vmin;
    EarthVector Vmax;
    if(V1.IsUnderSeaLevel() || V2.IsUnderSeaLevel()) {
	if(V1.IsUnderSeaLevel() && V2.IsUnderSeaLevel()) {
	    Msg(EsafMsg::Warning) << "<OzoneAmountAlongPath> Both V1 and V2 are under sea level --> SHOULD NOT OCCUR, zero returned"<<MsgDispatch;
	    return 0.;
	}
	Msg(EsafMsg::Warning) << "<OzoneAmountAlongPath> V1 or V2 is under sea level --> search for the true track impact"<<MsgDispatch;
	EarthVector impact(1);
	impact = ImpactASL(V1,(V2 - V1).Unit());
	if(impact.Z() == -HUGE) {
	    impact = ImpactASL(V2,(V1 - V2).Unit());
	    if(impact.Z() == HUGE || impact.Z() == -HUGE) {
	        Msg(EsafMsg::Warning) << "<OzoneAmountAlongPath> Failed to find an impact, zero returned"<<MsgDispatch;
	        return 0.;
	    }
	    else V1 = impact;
	}
	else V2 = impact;
    }

    if(V1.Zv() <= V2.Zv()) {
        Vmin = V1;
	Vmax = V2;
    }
    else {
        Vmin = V2;
	Vmax = V1;
    }

    if(Vmax.Zv() > maxalt) IsTooHigh = true;
    U = (Vmax - Vmin).Unit();
    inter = Vmin;
    Double_t DeltaT = Temperature(Vmin.Zv()) - STP_Temperature();
    rho1 = O3_DensityPPMV(Vmin.Zv()) * C0 *(1 + C1*DeltaT + C2*DeltaT*DeltaT) * Pressure(Vmin.Zv()) / Temperature(Vmin.Zv());


    while((Vmax - Vmin).Mag() > (inter - Vmin).Mag()) {
	if( (inter + U*dl - Vmin).Mag() >= (Vmax - Vmin).Mag() ) {
            DeltaT = Temperature(Vmax.Zv()) - STP_Temperature();
	    rho2 = O3_DensityPPMV(Vmax.Zv()) * C0 *(1 + C1*DeltaT + C2*DeltaT*DeltaT) * Pressure(Vmax.Zv()) / Temperature(Vmax.Zv());
	    dX = 0.5*(rho1 + rho2) * (Vmax - inter).Mag()/km;
	    X += dX;
	    break;
	}
	inter += U*dl;
	h2 = inter.Zv();
	if(h2 > maxalt && IsTooHigh) break;
        DeltaT = Temperature(h2) - STP_Temperature();
	rho2 = O3_DensityPPMV(h2) * C0 *(1 + C1*DeltaT + C2*DeltaT*DeltaT) * Pressure(h2) / Temperature(h2);
	dX = 0.5*(rho1 + rho2) * dl/km;
	X += dX;
	rho1 = rho2;
    }
	
    return X*0.0269*STP_Temperature()/STP_Pressure();
}

//______________________________________________________________________________
Int_t Atmosphere::InvertGrammage(const EarthVector& pos1, const EarthVector& direc, Double_t depth, EarthVector& rtn, Double_t TOA_alt, Double_t maxtof) const {
    //
    // calculate final position for given pos1,direc and air depth
    //
    // if 0 < TOA_alt < fTOA  --> TOA_alt is used to make present calculation (to optimize CPU time)
    // 
    // it returns : -1 if pos1 is under sea level or if none impact (latter should not occur)
    //               0 if position found
    //               1 if TOF cut or infinite loop
    //               2 if TOA reached before
    //               3 if sea level reached before
    //
    
    // init
    Int_t status(-1);
    if(pos1.IsUnderSeaLevel()) return status;
    if(rtn.Mag()) rtn.SetMag(0);
    EarthVector dir = direc.Unit();
    EarthVector impact(1);
    Double_t TOA(0.);
    
    // TOA settings
    if( (0 < TOA_alt) && (TOA_alt < fTOA) ) TOA = TOA_alt;
    else TOA = fTOA;
    
    // in case depth is null
    if(!depth) {
        rtn = pos1;
	return 0;
    }
    
    // check if sea level reached before foreseen air depth been travelled
    // check if TOA reached before the foreseen air depth has been travelled
    impact = ImpactASL(pos1,dir);
    // if no impact at sea level
    if(impact.Z() == HUGE) {
        impact = ImpactAtTOA(pos1,dir,TOA);
	if(impact.Z() == HUGE) {
	    Msg(EsafMsg::Warning) << "<InvertGrammage> track reaches NEITHER sea level NOR TOA -> SHOUD NOT HAPPEN"<<MsgDispatch;
            return -1;
	}
        if(Grammage(pos1,impact) < depth) {
	    rtn = impact;
	    return 2;
        }
    }
    else {
        if(Grammage(pos1,impact) < depth) {
	    rtn = impact;
	    return 3;
        }
    }
    
    
    // step by step process until it reaches depth value
    Double_t dl = 500*m;
    Double_t dX, rho2, h2;
    EarthVector U, inter;
    Double_t X = 0.;
    Double_t rho1;
    U = dir;
    inter = pos1;
    rho1 = Air_Density(pos1.Zv());
    if((inter - impact).Mag() < dl) dl = (inter - impact).Mag();

    Int_t counter(0), tofcut(0);
    if(maxtof < 0) tofcut = Int_t(2*(EarthRadius() + fTOA) / dl) + 1;
    else tofcut = Int_t(maxtof*Clight() / dl);
    while(true) {
        if(counter++ > tofcut) {
#ifdef DEBUG
            //Msg(EsafMsg::Debug) <<"<atmoInvertGrammage> TOF cut reached : photon dumped here "<<MsgDispatch;
#endif
            if(rtn.Mag() == 0) rtn = pos1;
	    return 1;
	}
        if(counter > Int_t(2*(EarthRadius() + fTOA) / dl)) {
            Msg(EsafMsg::Warning) <<"<atmoInvertGrammage> \"infinite\" loop (>2*EarthRadius travelled) broken by hand "<<MsgDispatch;
            if(rtn.Mag() == 0) rtn = pos1;
            return 1;
        }
	inter += U*dl;
	h2 = inter.Zv();
	rho2 = Air_Density(h2);
	dX = 0.5*(rho1 + rho2) * dl;
	if((X + dX) >= depth) break;
	X += dX;
	rho1 = rho2;
	if((inter - impact).Mag() < dl) dl = (inter - impact).Mag();
    }
    // come back to last iteration for fine tuning
    rtn = inter - U*dl;
    Double_t slope = (rho2 - rho1) / dl;  // density profile considered linear locally
    if(slope) dl = sign(slope) * sqrt( (rho1/slope) * (rho1/slope) + 2.*(depth - X) / slope) - rho1/slope;
    else dl = (depth - X) / rho1;  // if constant density
    rtn += U*dl;

    if(rtn.Mag() == 0) rtn = pos1;
    
    return 0;
}

//______________________________________________________________________________
Int_t Atmosphere::InvertOzoneAmount(const EarthVector& pos1, const EarthVector& direc, Double_t OzAmount, Double_t wl, EarthVector& rtn, Double_t TOA_alt, Double_t maxtof) const {
    //
    // calculate position corresponding to ozone amount, using given pos1,direc
    //
    // if 0 < TOA_alt < fTOA  --> TOA_alt is used to make present calculation (to optimize CPU time)
    // 
    // it returns : -1 if pos1 is under sea level or if none impact (latter should not occur)
    //               0 if position found
    //               1 if TOF cut or infinite loop
    //               2 if TOA reached before
    //               3 if sea level reached before
    //
    
    // init
    Int_t status(-1);
    if(pos1.IsUnderSeaLevel()) return status;
    if(rtn.Mag()) rtn.SetMag(0);
    EarthVector dir = direc.Unit();
    EarthVector impact(1);
    Double_t TOA(0.);
    
    // TOA settings
    if( (0 < TOA_alt) && (TOA_alt < fTOA) ) TOA = TOA_alt;
    else TOA = fTOA;
    
    // in case ozone amount is null
    if(OzAmount == 0.) {
        rtn = pos1;
	return 0;
    }

    // find wavelength dependent Ozone coeff.
    Double_t C0 = GetOzoneCoeff(wl,0);
    Double_t C1 = GetOzoneCoeff(wl,1);
    Double_t C2 = GetOzoneCoeff(wl,2);
    
    
    // check if sea level reached before foreseen ozone amount been travelled
    // check if TOA reached before the foreseen ozone amount has been travelled
    impact = ImpactASL(pos1,dir);
    // if no impact at sea level
    if(impact.Z() == HUGE) {
        impact = ImpactAtTOA(pos1,dir,TOA);
	if(impact.Z() == HUGE) {
	    Msg(EsafMsg::Warning) << "<InvertOzoneAmount> track reaches NEITHER sea level NOR TOA -> SHOUD NOT HAPPEN"<<MsgDispatch;
            return -1;
	}
        if(OzoneAmountAlongPath(pos1,impact,wl) < OzAmount) {
	    rtn = impact;
	    return 2;
        }
    }
    else {
        if(OzoneAmountAlongPath(pos1,impact,wl) < OzAmount) {
	    rtn = impact;
	    return 3;
        }
    }
    
    
    // step by step process until it reaches OzAmount value
    Double_t dl = 500*m;
    Double_t dX, rho2, h2;
    EarthVector U, inter;
    Double_t X = 0.;
    Double_t rho1;
    U = dir;
    inter = pos1;
    Double_t DeltaT = Temperature(pos1.Zv()) - STP_Temperature();
    rho1 = O3_DensityPPMV(pos1.Zv()) * C0 *(1 + C1*DeltaT + C2*DeltaT*DeltaT) * Pressure(pos1.Zv()) / Temperature(pos1.Zv());
    if((inter - impact).Mag() < dl) dl = (inter - impact).Mag();

    Int_t counter(0), tofcut(0);
    Int_t stopvalue = Int_t(2*(EarthRadius() + TOA) / dl);
    if(maxtof < 0) tofcut = stopvalue + 1;
    else tofcut = Int_t(maxtof*Clight() / dl);
    while(true) {
        if(counter++ > tofcut) {
#ifdef DEBUG
            //Msg(EsafMsg::Debug) <<"<InvertOzoneAmount> TOF cut reached : photon dumped here "<<MsgDispatch;
#endif
            if(rtn.Mag() == 0) rtn = pos1;
	    return 1;
	}
        if(counter > stopvalue) {
            Msg(EsafMsg::Warning) <<"<InvertOzoneAmount> \"infinite\" loop (>2*EarthRadius travelled) broken by hand"<<MsgDispatch;
            Msg(EsafMsg::Warning) <<"<InvertOzoneAmount> counter = "<<counter <<MsgDispatch;
            Msg(EsafMsg::Warning) <<"<InvertOzoneAmount> Stop value =  "<<stopvalue <<MsgDispatch;
            Msg(EsafMsg::Warning) <<"<InvertOzoneAmount> maxtof =  "<<maxtof <<MsgDispatch;
            Msg(EsafMsg::Warning) <<"<InvertOzoneAmount> MCRT TOA = " << TOA <<MsgDispatch;
            Msg(EsafMsg::Warning) <<"<InvertOzoneAmount> initpos = " << pos1 <<MsgDispatch;
            Msg(EsafMsg::Warning) <<"<InvertOzoneAmount> initpos.Zv = " << pos1.Zv() <<MsgDispatch;
            Msg(EsafMsg::Warning) <<"<InvertOzoneAmount> finalpos = " << EarthVector(pos1+inter) <<MsgDispatch;
            Msg(EsafMsg::Warning) <<"<InvertOzoneAmount> finalpos.Zv = " << EarthVector(pos1+inter).Zv() <<MsgDispatch;
            Msg(EsafMsg::Warning) <<"<InvertOzoneAmount> direc = " << direc <<MsgDispatch;
            Msg(EsafMsg::Warning) <<"<InvertOzoneAmount> direc.theta = " << direc.Theta()*RadToDeg() <<MsgDispatch;
            Msg(EsafMsg::Warning) <<"<InvertOzoneAmount> direc.phi = " << direc.Phi()*RadToDeg() <<MsgDispatch;
            if(rtn.Mag() == 0) rtn = pos1;
            return 1;
        }
	inter += U*dl;
	h2 = inter.Zv();
        DeltaT = Temperature(h2) - STP_Temperature();
        rho2 = O3_DensityPPMV(h2) * C0 *(1 + C1*DeltaT + C2*DeltaT*DeltaT) * Pressure(h2) / Temperature(h2);
	dX = 0.5*(rho1 + rho2) * dl/km * 0.0269*STP_Temperature()/STP_Pressure();
	if((X + dX) >= OzAmount) break;
	X += dX;
	rho1 = rho2;
	if((inter - impact).Mag() < dl) dl = (inter - impact).Mag();
    }

    // come back to last iteration for fine tuning
    rtn = inter - U*dl;
    Double_t slope = (rho2 - rho1) / dl;  // ozone amount profile considered linear locally
    if(slope) dl = sign(slope) * sqrt( (rho1/slope) * (rho1/slope) + 2.*(OzAmount - X) / (0.0269/km*STP_Temperature()/STP_Pressure()) / slope) - rho1/slope;
    else dl = (OzAmount - X) / (0.0269/km*STP_Temperature()/STP_Pressure()) / rho1;  // if locally constant ozone amount profile
    rtn += U*dl;


    if(rtn.Mag() == 0) rtn = pos1;
    
    return 0;
}

//______________________________________________________________________________
EarthVector Atmosphere::ImpactASL(const EarthVector& pos, const EarthVector& dir) const {
    //
    // returns impact at sea level of a track defined by starting position and direction
    //

    // if pos is under sea level
    EarthVector rtn;
    if(pos.IsUnderSeaLevel()) {
        rtn.SetXYZ(0,0,-HUGE);
        return rtn;
    }


    // to know if dir is locally going upward
    Double_t angle;
    EarthVector temp(pos);
    temp(2) += EarthRadius();  // pos expressed in earth-centered frame -> gives local vertical direction expressed in MES frame
    angle = dir.Angle(temp);   // angle between dir and local vertical
    if(angle < (PiOver2() - kTolerance)) {
        rtn.SetXYZ(0,0,HUGE);
        return rtn;
    }

    // now, impact calculation
    Double_t mag(0);
    EarthVector direc = dir.Unit();

/*
//  flat earth
if(fabs(direc.Z()) < kTolerance) rtn.SetXYZ(0,0,HUGE);
else {
    mag = -pos.Z() / direc.Z();
    if(mag < -kAltitudeTolerance) rtn.SetXYZ(0,0,HUGE);
    else rtn = pos + mag*direc;
}
*/
    // spherical earth
    Double_t b = pos*direc + direc.Z()*EarthRadius();
    Double_t c = pos.Mag2() + 2*EarthRadius()*pos.Z();
    pair<Int_t,Double_t*>& p = findRoots(1.,2*b,c);
    if(p.first == 0) {
        rtn.SetXYZ(0,0,HUGE);
        return rtn;
    }
    if(p.first == 1) mag = p.second[0];
    else if(p.first ==2) mag = min(p.second[0],p.second[1]);
    
    if(mag < -kAltitudeTolerance) Msg(EsafMsg::Debug) <<"SHOULD NOT HAVE IMPACT AT SEA LEVEL, mag = "<<mag << MsgDispatch;
    
    rtn = pos + mag*direc;
    return rtn;    
}

//______________________________________________________________________________
EarthVector Atmosphere::ImpactAtTOA(const EarthVector& pos, const EarthVector& dir, Double_t TOA_alt) const {
    //
    // returns impact at Top Of Atmosphere of a track defined by starting position and direction
    // if 0 < TOA_alt < fTOA  --> TOA_alt is used to make present calculation (to optimize CPU time, cf. InvertGrammage())
    //

    // if pos is under sea level
    EarthVector rtn;
    if(pos.IsUnderSeaLevel()) {
        rtn.SetXYZ(0,0,-HUGE);
        return rtn;
    }
    
    // TOA settings
    Double_t altTOA = 0.;
    if( (0 < TOA_alt) && (TOA_alt < fTOA) ) altTOA = TOA_alt;
    else altTOA = fTOA;

    // now, impact calculation
    Double_t mag(0);
    EarthVector direc = dir.Unit();

/*
//  flat earth
if(fabs(direc.Z()) < kTolerance) rtn.SetXYZ(0,0,HUGE);
else {
    mag = (altTOA - pos.Z()) / direc.Z();
    if(mag < -kAltitudeTolerance) rtn.SetXYZ(0,0,HUGE);
    else rtn = pos + mag*direc;
}
*/

    // spherical earth
    Double_t b = pos*direc + direc.Z()*EarthRadius();
    Double_t c = pos.Mag2() + 2*EarthRadius()*(pos.Z() - altTOA) - pow(altTOA,2);
    pair<Int_t,Double_t*>& p = findRoots(1.,2*b,c);
    if(p.first == 0) {
        rtn.SetXYZ(0,0,HUGE);
        return rtn;
    }
    if(p.first == 1) mag = p.second[0];
    else if(p.first == 2) {
	if((p.second[0]+kAltitudeTolerance) * (p.second[1]+kAltitudeTolerance) < 0) mag = max(p.second[0],p.second[1]);
	else mag = min(p.second[0],p.second[1]);
    }
    if(mag < -kAltitudeTolerance) {
        Msg(EsafMsg::Warning) << "PB in impact calculation ImpactAtTOA(), mag[min,max] = ["<<mag<<", "<< max(p.second[0],p.second[1])<<"]" << MsgDispatch;
        Msg(EsafMsg::Warning) << "pos = "<<pos << MsgDispatch;
        Msg(EsafMsg::Warning) << "dir = "<<direc << MsgDispatch;
    }
    rtn = pos + mag*direc;
    return rtn;    
}



// used in reco only : obsolete, should use Grammage method
//______________________________________________________________________________
Double_t Atmosphere::Depth(const Double_t H, const Double_t Theta) const {
    // Compute the atmosphere depth between the given point with (h,theta) coordinates and infinity
    //
    Double_t eps=1.e-5;
    TF1 *depthIntegral = new TF1("depthIntegral",densityIntegral,0.,1.,2);
    
    if (Theta <= TMath::PiOver2()) {
	Double_t pars[2] = {H,Theta};
        Double_t depth = depthIntegral->Integral(eps,1-eps,pars,fDepthCalculationPrecision);
        delete depthIntegral;
	return depth;
    }
    else { 
	Double_t Hstar = ( EarthRadius() + H)*TMath::Sin(TMath::Pi() - Theta ) - 
	    EarthRadius();
	Double_t pars1[2] = {Hstar,TMath::PiOver2()};
	Double_t pars2[2] = {H,TMath::Pi() - Theta};
	Double_t depth=2*depthIntegral->Integral(eps,1-eps,pars1) - depthIntegral->Integral(eps,1-eps,pars2,
            fDepthCalculationPrecision);
	delete depthIntegral;
	return depth;
    }
}

