// ESAF : Euso Simulation and Analysis Framework
// $Id: LowtranAtmosphere.cc 2749 2006-07-04 13:16:23Z moreggia $
// Corinne Berat created Apr,  2 2004

/*****************************************************************************
 * ESAF: Euso Simulation and Analysis Framework                              *
 *                                                                           *
 *  Id: LowtranAtmosphere                                                    *
 *  Package: atmosphere                                                      *
 *  Coordinator: S. Moreggia, C. Berat                                       *
 *                                                                           *
 *****************************************************************************/

//_____________________________________________________________________________
//
// LowtranAtmosphere
//
// <extensive class description>
//
//   Config file parameters
//   ======================
//
//   <parameter name>: <parameter description>
//   -Valid options: <available options>
//

#include "LowtranAtmosphere.hh"
#include "Config.hh"
#include "LowtranAtmosphereData.hh"
#include <TH1F.h>
#include <TROOT.h>
#include <math.h>
#include <TSpline.h>

using namespace sou;
using namespace TMath;

ClassImp(LowtranAtmosphere)

//_____________________________________________________________________________
LowtranAtmosphere::LowtranAtmosphere(string type) : Atmosphere(),  fData(0) {
    //
    // ctor
    //
    Msg(EsafMsg::Info) << "Building with the configuration:"<< MsgDispatch;
    fType = type;
    Build();
}

//_____________________________________________________________________________
LowtranAtmosphere::~LowtranAtmosphere() {
    //
    // ctor
    //
    SafeDelete(fData);
}

//___________________________________________________________________________________
void LowtranAtmosphere::CreateInstance() {
    //
    // Create the single atmosphere instance as being a LowtranAtmosphere object
    //
    if(!Atmosphere::fChild) Atmosphere::fChild = new LowtranAtmosphere("lowtran");
    else Atmosphere::fChild->Msg(EsafMsg::Warning) << "Trying to create two atmospheres !!"<< MsgDispatch;
}

//_____________________________________________________________________________
void LowtranAtmosphere::Build() {
    //
    // Build atmosphere according to config parameters
    //
    fData =  new LowtranAtmosphereData();
    if(!fData)  Msg(EsafMsg::Panic) << "No new LowtranData, memory pb" << MsgDispatch ;
    
    const Double_t* data = 0;
    // interpolators initialization
    data =  fData->GetTemperatureTable();
    TArrayD t_arr(fData->NumberOfElements(),data);
    fTempInterpol = new TSpline3("fTempInterpol",0,fData->NumberOfElements()*km,t_arr.GetArray(),fData->NumberOfElements());    
    
    data =  fData->GetAir_DensityTable();
    TArrayD rho_arr(fData->NumberOfElements(),data);
    fDensInterpol = new TSpline3("fDensInterpol",0,fData->NumberOfElements()*km,rho_arr.GetArray(),fData->NumberOfElements());    
    
    data =  fData->GetPressureTable();
    TArrayD press_arr(fData->NumberOfElements(),data);
    fPressInterpol = new TSpline3("fPressInterpol",0,fData->NumberOfElements()*km,press_arr.GetArray(),fData->NumberOfElements());    
}

//_____________________________________________________________________________
Double_t LowtranAtmosphere::Interpolate(string& val,string& mod,Double_t h) const {
    //
    // Calculate simple interpolation for secondary data
    //
    const Double_t* data = 0;

    if(val == "Odens")            data = fData->GetO_DensityTable();
    else if(val == "O2dens")      data = fData->GetO2_DensityTable();
    else if(val == "H2Odens")     data = fData->GetH2O_DensityTable();
    else if(val == "N2dens")      data = fData->GetN2_DensityTable();
    else if(val == "O3dens")      data = fData->GetO3_DensityTable();
    else if(val == "O3densPPMV")  data = fData->GetO3_DensityPPMVTable();
    else   Msg(EsafMsg::Panic)<< "Wrong val argument in LOWTRAN_00Atmosphere::Interpolate" << MsgDispatch ;
    
    
    if(h <0) return data[0];
    Int_t nb = fData->NumberOfElements();
    Double_t step = 1*km;
    if(h > step*(nb-1)) return data[nb-1];
    Double_t val1, val2;
    Double_t h1;
    Int_t i = 0;
    i = Int_t(floor(h/step));
    h1 = i * step;
    val1 = data[i];
    val2 = data[i+1];
    
    if(mod == "linear" || val1 == 0 || val2 == 0)  return (val2 - val1)*(h - h1)/step + val1;
    else if(mod == "expon")                        return val1 * exp(-(h - h1) * log(val1/val2)/step);
    else Msg(EsafMsg::Panic)<<"Wrong mod argument in LowtranAtmosphere::Interpolate" << MsgDispatch ;
    
    return 0;
}

//_____________________________________________________________________________
Double_t LowtranAtmosphere::Pressure(Double_t h) const {
    //
    // Pressure at a given altitude
    //
    if(h > fTOA)  return 0.;
    // may occur with interfaces with external codes
    if(IsNaN(h)) Msg(EsafMsg::Panic)<<"<Pressure> Given altitude is NaN" << MsgDispatch ;
    return fPressInterpol->Eval(h);
}

//_____________________________________________________________________________
Double_t LowtranAtmosphere::Temperature(Double_t h) const {
    //
    // Temperature at a given altitude
    //
    if(h > fTOA)  return 0.;
    // may occur with interfaces with external codes
    if(IsNaN(h)) Msg(EsafMsg::Panic)<<"<Pressure> Given altitude is NaN" << MsgDispatch ;
    return fTempInterpol->Eval(h);
}

//_____________________________________________________________________________
Double_t LowtranAtmosphere::AbsoluteHumidity(Double_t h) const {
    //
    // Absolute humidity at a given altitude (water vapor density)
    //
    
    if(h > fTOA)  return 0.;
    string val = "H2Odens";
    string mod = "linear";
    
    return Interpolate(val,mod,h);
}

//_____________________________________________________________________________
Double_t LowtranAtmosphere::Air_Density(Double_t h) const {
    //
    // Atmospheric air density at a given altitude
    //
    if(h > fTOA)  return 0.;
    // may occur with interfaces with external codes
    if(IsNaN(h)) Msg(EsafMsg::Panic)<<"<Pressure> Given altitude is NaN" << MsgDispatch ;
    return fDensInterpol->Eval(h);
}

//_____________________________________________________________________________
Double_t LowtranAtmosphere::O_Density(Double_t h) const {
    //
    // Oxygen density at a given altitude
    //
    if(h > fTOA) return 0.;

    // exponential interpolation between table values
    string val = "Odens";
    string mod = "expon";
    
    return Interpolate(val,mod,h);
}

//_____________________________________________________________________________
Double_t LowtranAtmosphere::O2_Density(Double_t h) const {
    //
    // Di-oxygen density at a given altitude
    //
    if(h > fTOA) return 0.;

        // exponential interpolation between table values
        string val = "O2dens";
	string mod = "expon";
	return Interpolate(val,mod,h);
}

//_____________________________________________________________________________
Double_t LowtranAtmosphere::O3_Density(Double_t h) const {
    //
    // Ozone density at a given altitude
    //
    if(h > fTOA) return 0.;

    // linear interpolation between table values
    string val = "O3dens";
    string mod = "linear";
    
    return Interpolate(val,mod,h);
}

//_____________________________________________________________________________
Double_t LowtranAtmosphere::O3_DensityPPMV(Double_t h) const {
    //
    // Ozone PPMV density at a given altitude
    //
    if(h > fTOA) return 0.;

    // linear interpolation between table values
    string val = "O3densPPMV";
    string mod = "linear";
    
    return Interpolate(val,mod,h);
}

//_____________________________________________________________________________
Double_t LowtranAtmosphere::N2_Density(Double_t h) const {
    //
    // Di-nitrogen density at a given altitude
    //
    if(h > fTOA) return 0.;

    // exponential interpolation between table values
    string val = "N2dens";
    string mod = "expon";
    
    return Interpolate(val,mod,h);
}

//_____________________________________________________________________________
Double_t LowtranAtmosphere::GetLatitude() const {
    //
    // used to identify the model
    //
    
    return fData->GetLatitude();
}

//_____________________________________________________________________________
Double_t LowtranAtmosphere::GetLongitude() const {
    //
    // dummy
    //
    
    return fData->GetLongitude();
}

//_____________________________________________________________________________
Double_t LowtranAtmosphere::GetDate() const {
    //
    // dummy
    //
    
    return fData->GetDate();
}




