// ESAF : Euso Simulation and Analysis Framework
// $Id: MSISE_00Atmosphere.cc 2794 2008-02-11 07:56:34Z naumov $
// S. Moreggia created 18 November 2003

/*****************************************************************************
 * ESAF: Euso Simulation and Analysis Framework                              *
 *                                                                           *
 *  Id: MSISE_00Atmosphere                                                   *
 *  Package: atmosphere                                                      *
 *  Coordinator: S. Moreggia                                                 *
 *                                                                           *
 *****************************************************************************/

//_____________________________________________________________________________
//
// MSISE_00Atmosphere
//
// <extensive class description>
//
//   Config file parameters
//   ======================
//
//   <parameter name>: <parameter description>
//   -Valid options: <available options>
//

#include <math.h>
#include "MSISE_00Atmosphere.hh"
#include "MSISEtool.hh"
#include "Config.hh"
#include "MSISE_00AtmosphereData.hh"
#include "EConst.hh"
#include <TSpline.h>
#include <TArrayD.h>
ClassImp(MSISE_00Atmosphere)

using EConst::Avogadro;
using EConst::R_ideal;
using EConst::AtomicMass;
using namespace sou;
using namespace TMath;

//____________________________________________________________________________
MSISE_00Atmosphere::MSISE_00Atmosphere(string type) : Atmosphere(), fData(0) {
    //
    // ctor
    //
    fType = type;
    Msg(EsafMsg::Info) << "*** MSISE_00Atmosphere ***"<< MsgDispatch;
    Build();
}

//____________________________________________________________________________
MSISE_00Atmosphere::~MSISE_00Atmosphere() {
    //
    // dtor
    //
    SafeDelete(fData);
}

//___________________________________________________________________________________
void MSISE_00Atmosphere::CreateInstance() {
    //
    // Create the single atmosphere instance as being a MSISE_00Atmosphere object
    //
    string type = "msise00";
    if(!Atmosphere::fChild) Atmosphere::fChild = new MSISE_00Atmosphere(type);
    else Atmosphere::fChild->Msg(EsafMsg::Panic) << "Trying to create two atmospheres !!"<< MsgDispatch;
}

//___________________________________________________________________________________
void MSISE_00Atmosphere::ResetInstance() {
    //
    // in random mode, reset tabulated profiles and randomly build a new atmosphere
    //
    
    SafeDelete(fData);
    Msg(EsafMsg::Info) << "*** Build a new MSISE_00Atmosphere ***"<< MsgDispatch;
    Build();
}

//____________________________________________________________________________
void MSISE_00Atmosphere::Build() {
    //
    // Build atmosphere according to config parameters
    // Data arrays are built
    //

    if(!fData) {
        fData =  new MSISE_00AtmosphereData(IsRandomMode());
        if(!fData) Msg(EsafMsg::Panic)<<"No new MSISEData, memory pb"<<MsgDispatch;
    }

    const Double_t* data = 0;
    // interpolators initialization
    data =  fData->GetTemperatureTable();
    TArrayD t_arr(fData->NumberOfElements(),data);
    fTempInterpol = new TSpline3("fTempInterpol",0,fData->NumberOfElements()*km,t_arr.GetArray(),fData->NumberOfElements());    

    data =  fData->GetAir_DensityTable();
    TArrayD rho_arr(fData->NumberOfElements(),data);
    fDensInterpol = new TSpline3("fDensInterpol",0,fData->NumberOfElements()*km,rho_arr.GetArray(),fData->NumberOfElements());    

    data =  fData->GetPressureTable();
    TArrayD press_arr(fData->NumberOfElements(),data);
    fPressInterpol = new TSpline3("fPressInterpol",0,fData->NumberOfElements()*km,press_arr.GetArray(),fData->NumberOfElements());    
}

//____________________________________________________________________________
Double_t MSISE_00Atmosphere::Interpolate(string& val,string& mod,Double_t h) const {
    //
    // Calculate interpolation
    //    
    const Double_t* data = 0;
    if(val == "Odens")            data = fData->GetO_DensityTable();
    else if(val == "O2dens")      data = fData->GetO2_DensityTable();
    else if(val == "H2Odens")     data = fData->GetH2O_DensityTable();
    else if(val == "N2dens")      data = fData->GetN2_DensityTable();
    else if(val == "O3dens")      data = fData->GetO3_DensityTable();
    else if(val == "O3densPPMV")  data = fData->GetO3_DensityPPMVTable();
    else  Msg(EsafMsg::Panic)<< "Wrong val argument in MSISE_00Atmosphere::Interpolate ==> "<<val <<  MsgDispatch;
    
    if(h < 0) return data[0];
    Int_t nb = fData->NumberOfElements();
    Double_t step = fData->LayerSize();
    if(h > step*(nb-1)) return data[nb-1];
    Double_t val1, val2;
    Double_t h1;
    Int_t i = 0;
    i = Int_t(floor(h/step));
    h1 = i * step;
    val1 = data[i];
    val2 = data[i+1];
    
    if(mod == "linear" || val1 == 0 || val2 == 0)  return (val2 - val1)*(h - h1)/step + val1;
    else if(mod == "expon")                        return val1 * exp(-(h - h1) * log(val1/val2)/step);
    else Msg(EsafMsg::Panic)<<"Wrong mod argument in MSISE_00Atmosphere::Interpolate" << MsgDispatch ;
    
    return 0;
}

//____________________________________________________________________________
Double_t MSISE_00Atmosphere::Pressure(Double_t h) const {
    //
    // Pressure at a given altitude
    //
    if(h > fTOA)  return 0.;

    // may occur with interfaces with external codes
    if(IsNaN(h)) Msg(EsafMsg::Panic)<<"<Pressure> Given altitude is NaN" << MsgDispatch ;
    return fPressInterpol->Eval(h);
}

//____________________________________________________________________________
Double_t MSISE_00Atmosphere::Temperature(Double_t h) const {
    //
    // Temperature at a given altitude
    //
    if(h > fTOA)  return 0.;

    // may occur with interfaces with external codes
    if(IsNaN(h)) Msg(EsafMsg::Panic)<<"<Pressure> Given altitude is NaN" << MsgDispatch ;
    return fTempInterpol->Eval(h);
}

//____________________________________________________________________________
Double_t MSISE_00Atmosphere::AbsoluteHumidity(Double_t h) const {
    //
    // Absolute humidity at a given altitude (water vapor density)
    //
    
    if(h > fTOA)  return 0.;
    string val = "H2Odens";
    string mod = "linear";
    
    return Interpolate(val,mod,h);
}

//____________________________________________________________________________
Double_t MSISE_00Atmosphere::Air_Density(Double_t h) const {    
    //
    // Atmospheric air density at a given altitude
    //
    if(h > fTOA)  return 0.;

    // may occur with interfaces with external codes
    if(IsNaN(h)) Msg(EsafMsg::Panic)<<"<Pressure> Given altitude is NaN" << MsgDispatch ;
    return fDensInterpol->Eval(h);
}

//____________________________________________________________________________
Double_t MSISE_00Atmosphere::O_Density(Double_t h) const {
    //
    // Oxygen density at a given altitude
    //
    if(h > fTOA)
        return 0.;

    // exponential interpolation between table values
    string val = "Odens";
    string mod = "expon";
    return Interpolate(val,mod,h);
}

//____________________________________________________________________________
Double_t MSISE_00Atmosphere::O2_Density(Double_t h) const {
    //
    // Di-oxygen density at a given altitude
    //
    if(h > fTOA)
        return 0.;

    // exponential interpolation between table values
    string val = "O2dens";
    string mod = "expon";
    return Interpolate(val,mod,h);
}

//____________________________________________________________________________
Double_t MSISE_00Atmosphere::O3_Density(Double_t h) const {
    //
    // Ozone density at a given altitude
    //
    
    // exponential interpolation between table values
    string val = "O3dens";
    string mod = "linear";
    return Interpolate(val,mod,h);
}

//____________________________________________________________________________
Double_t MSISE_00Atmosphere::O3_DensityPPMV(Double_t h) const {
    //
    // Ozone PPMV density at a given altitude
    //
    
    // exponential interpolation between table values
    string val = "O3densPPMV";
    string mod = "linear";
    return Interpolate(val,mod,h);
}

//____________________________________________________________________________
Double_t MSISE_00Atmosphere::N2_Density(Double_t h) const {
    //
    // Di-nitrogen density at a given altitude
    //
    if(h > fTOA)
        return 0.;

    // exponential interpolation between table values
    string val = "N2dens";
    string mod = "expon";
    return Interpolate(val,mod,h);
}

//____________________________________________________________________________
Double_t MSISE_00Atmosphere::Aerosols_Density(string& type,Double_t h) const {
    //
    // Aerosols density for a given type of aerosols at a given altitude
    //
    Msg(EsafMsg::Panic)<<"Aerosol_Density not available for MSISE_00Atmosphere"
                         << MsgDispatch ;
    return 0;
}


/*
// For MSISEscan Macro --- WARNING : Build() call in ctor MUST BE DISABLED
//____________________________________________________________________________
void MSISE_00Atmosphere::Build(Int_t doy,Double_t lst, Double_t lat, Double_t longi) {
    //
    // Build atmosphere according to config parameters
    // Data arrays are built
    //

    if(!fData) {
        fData =  new MSISE_00AtmosphereData(IsRandomMode());
        if(!fData) Msg(EsafMsg::Panic)<<"No new MSISEData, memory pb"<<MsgDispatch;
    }
    
    fData->InitForMacro(doy,lst,lat,longi);
    fData->SetTables();

    const Double_t* data = 0;
    // interpolators initialization
    data =  fData->GetTemperatureTable();
    TArrayD t_arr(fData->NumberOfElements(),data);
    fTempInterpol = new TSpline3("fTempInterpol",0,fData->NumberOfElements()*km,t_arr.GetArray(),fData->NumberOfElements());    

    data =  fData->GetAir_DensityTable();
    TArrayD rho_arr(fData->NumberOfElements(),data);
    fDensInterpol = new TSpline3("fDensInterpol",0,fData->NumberOfElements()*km,rho_arr.GetArray(),fData->NumberOfElements());    

    data =  fData->GetPressureTable();
    TArrayD press_arr(fData->NumberOfElements(),data);
    fPressInterpol = new TSpline3("fPressInterpol",0,fData->NumberOfElements()*km,press_arr.GetArray(),fData->NumberOfElements());    
}
*/

