// ESAF : Euso Simulation and Analysis Framework
// $Id: MSISE_00AtmosphereData.cc 2695 2006-05-27 19:09:42Z moreggia $
// Sylvain Moreggia created Dec, 12 2003

#include "MSISE_00AtmosphereData.hh"
#include "Config.hh"
#include <TH1F.h>
#include <TROOT.h>
#include <TMath.h>
#include "EConst.hh"
#include "EsafRandom.hh"

ClassImp(MSISE_00AtmosphereData)

using EConst::Avogadro;
using EConst::AtomicMass;
using EConst::R_ideal;
using namespace sou;
using namespace TMath;

//________________________________________________________________________________________
MSISE_00AtmosphereData::MSISE_00AtmosphereData(Bool_t rndm) : AtmosphereData(), fTool(0) {
    //
    // ctor
    //
    
    fRandomMode = rndm;
    fNb = 100;
    fStep = 100*km / fNb;
    Init();
    SetTables();
}

//________________________________________________________________________________________
MSISE_00AtmosphereData::~MSISE_00AtmosphereData() {
    //
    // dtor
    //
    SafeDelete(fTool);
}

//________________________________________________________________________________________
void MSISE_00AtmosphereData::Init() {
    //
    // Initialize msise parameters to work out tables claculations
    //
    
    fTool = new MSISEtool();
    if(!fTool)  Msg(EsafMsg::Panic) << "No new MSISEtool, memory pb" << MsgDispatch ;
    TRandom* rndm = EsafRandom::Get();
    
    if(fRandomMode) {
        //
        // SCAN MODE ---> get randomly :
	// - latitude, longitude
	// - day of year
	// NB : local solar time always taken at midnight
	// NB : positions with longitude < 180 are considered in advance in the day (the day after if needed)
	// NB : positions with longitude > 180 are considered as being late in the day (the day before if needed)
	//
	ConfigFileParser* pConf = Config::Get()->GetCF("Atmosphere","MSISE_00Atmosphere");
	fInput.year = (Int_t)pConf->GetNum("MSISE_00Atmosphere.year");  // without effects with current MSISE version
	fInput.doy = Int_t(floor(364.*rndm->Rndm() + 1.5));  // 0.5 for correct round-off
	
	
	
	// randomly get latitude, longitude as function of telescope orbit
	ConfigFileParser* pConf2 = Config::Get()->GetCF("Atmosphere","AtmosphereFactory");
	Double_t th_slope = pConf2->GetNum("AtmosphereFactory.MSISE_00AtmosphereData.Slope")*DegToRad();
	Double_t phi_localframe = TwoPi()*rndm->Rndm();
	
	// after change of frame, to come back to std earth-centered frame
	fInput.g_lat  = -ASin(Cos(phi_localframe)*Sin(th_slope))*RadToDeg();
	
	// longitude is actually uniform because of earth self rotation
	fInput.g_long  = 360.*rndm->Rndm();
	    //fInput.g_long = (ATan2(Sin(phi_localframe),Cos(phi_localframe) * Cos(th_slope)) + Pi())*RadToDeg(); // +pi to achieve [0,2pi[
		
		
	// local midnight translated into UT time
	fInput.lst = 0.;  // locally midnight : negligible effect on atmosphere profiles w.r.t. latitude and day
	// positions with longitude < 180 are considered in advance in the day (the day after if needed)
	// positions with longitude > 180 are considered as being late in the day (the day before if needed)
	if(fInput.g_long <= 180.) fInput.sec = -fInput.g_long / 15. * 3600.;
	else if(fInput.g_long < 360.) fInput.sec = (360. - fInput.g_long) / 15. * 3600.;
	// some checks
	if(fInput.sec < 0.) {
	    if(fInput.doy > 1) fInput.doy--;
	    else fInput.doy = 365;
	    fInput.sec += 24.*3600.;
	    if(fInput.sec < 0.) Msg(EsafMsg::Warning) <<"Second in day is negative. SHOULD not occur " <<fInput.sec <<MsgDispatch;
	}
	else if(fInput.sec > 24.*3600.) {
	    if(fInput.doy < 365) fInput.doy++;
	    else fInput.doy = 1;
	    fInput.sec -= 24.*3600.;
	    if(fInput.sec > 24.*3600.) Msg(EsafMsg::Warning) <<"Second in day is > 24*3600. SHOULD not occur " <<fInput.sec <<MsgDispatch;
	}
	
	// printings
	Msg(EsafMsg::Info) <<"LAT. = " <<fInput.g_lat <<MsgDispatch;
	Msg(EsafMsg::Info) <<"LONG. = " <<fInput.g_long <<MsgDispatch;
	Msg(EsafMsg::Info) <<"DAY = " <<fInput.doy <<MsgDispatch;
	Msg(EsafMsg::Info) <<"Local Solar Time = " <<fInput.lst <<" h"<<MsgDispatch;
        Msg(EsafMsg::Info) <<"UT time = " <<fInput.sec/3600.<<" h"<<MsgDispatch;
	fInput.f107A = pConf->GetNum("MSISE_00Atmosphere.f107A");
	fInput.f107 = pConf->GetNum("MSISE_00Atmosphere.f107");
	fInput.ap = pConf->GetNum("MSISE_00Atmosphere.ap");
	fTuning.switches[0] = 0;
	for(Int_t i=1; i<24; i++) fTuning.switches[i] = 1;  //TOFIX : needed to allow better tuning
    }
    else {
        //
        // atmosphere does NOT vary event by event
	//
	
	// load parameters bounds : latitude, longitude, day and time in day
	ConfigFileParser* pConf = Config::Get()->GetCF("Atmosphere","MSISE_00Atmosphere");
        string TimeMode     = pConf->GetStr("MSISE_00Atmosphere.TimeMode");
        Double_t sec    = pConf->GetNum("MSISE_00Atmosphere.sec");
        Double_t lst  = pConf->GetNum("MSISE_00Atmosphere.lst");
        Int_t doy       = (Int_t)pConf->GetNum("MSISE_00Atmosphere.doy");
        Double_t g_lat = pConf->GetNum("MSISE_00Atmosphere.g_lat");
        Double_t g_long = pConf->GetNum("MSISE_00Atmosphere.g_long");
	
	// then set parameters randomly between bounds
	fInput.g_lat  = g_lat;
	fInput.g_long = g_long;
	fInput.doy    = doy;
	if(fInput.doy <= 0) {
	    Msg(EsafMsg::Warning) <<"Day of year must be > 0 --> Set to 1"<<MsgDispatch;
	    fInput.doy = 1;
	}
	else if(fInput.doy > 365) {
	    Msg(EsafMsg::Warning) <<"Day of year must be < 366 --> Set to 365"<<MsgDispatch;
	    fInput.doy = 365;
	}
	
	// set UT or LST first according to TimeMode configured
	// then the other one is deduced according to longitude
	if(TimeMode == "UT") {
	    fInput.sec = sec;
	    if(fInput.sec < 0.) {
		Msg(EsafMsg::Warning) <<"Second in day must be >= 0 --> Set to 0"<<MsgDispatch;
		fInput.sec = 0.;
	    }
	    else if(fInput.sec > 24.*3600.) {
		Msg(EsafMsg::Warning) <<"Second in day must be <= 24*3600 --> Set to 24*3600"<<MsgDispatch;
		fInput.sec = 24.*3600.;
	    }
	    fInput.lst = fInput.sec/3600 + fInput.g_long/15;
	    if(fInput.lst > 24.) fInput.lst -= 24.;
	    if(fInput.lst < 0.) fInput.lst += 24.;
	    if(fInput.lst < 0. || fInput.lst > 24.) Msg(EsafMsg::Warning) <<"Local solar time NOT CORRECT = "<< fInput.lst <<MsgDispatch;
	    
        }
	else if(TimeMode == "LST") {
	    if(lst < 24.) {
	        fInput.lst = lst;
	        if(fInput.g_long <= 180.) fInput.sec = (fInput.lst - fInput.g_long / 15.) * 3600.;
	        else if(fInput.g_long < 360.) fInput.sec = (fInput.lst + (360. - fInput.g_long) / 15.) * 3600.;
	    }
	    else {
	        fInput.lst = lst - 24.;
	        if(fInput.g_long <= 180.) fInput.sec = (lst - fInput.g_long / 15.) * 3600.;
	        else if(fInput.g_long < 360.) fInput.sec = (lst + (360. - fInput.g_long) / 15.) * 3600.;
	    }
	    // correct UT time if necessary
	    if(fInput.sec < 0.) {
	        if(fInput.doy > 1) fInput.doy--;
	        else fInput.doy = 365;
	        fInput.sec += 24.*3600.;
	    }
	    else if(fInput.sec > 24.*3600.) {
		if(fInput.doy < 365) fInput.doy++;
		else fInput.doy = 1;
		fInput.sec -= 24.*3600.;
	    }
	    // final check
	    if(fInput.sec < 0. || fInput.sec > 24.*3600.) Msg(EsafMsg::Warning) <<"Second in day not in bounds. SHOULD not occur = " <<fInput.sec <<MsgDispatch;
	}
	else Msg(EsafMsg::Panic) <<"Wrong argument for MSISE_00Atmosphere.TimeMode = "<< TimeMode <<MsgDispatch;
	
	// printings
	Msg(EsafMsg::Info) <<"LAT. = " <<fInput.g_lat <<MsgDispatch;
	Msg(EsafMsg::Info) <<"LONG. = " <<fInput.g_long <<MsgDispatch;
	Msg(EsafMsg::Info) <<"DAY = " <<fInput.doy <<MsgDispatch;
	Msg(EsafMsg::Info) <<"Local Solar Time = " <<fInput.lst <<" h"<<MsgDispatch;
        Msg(EsafMsg::Info) <<"UT time = " <<fInput.sec/3600.<<" h"<<MsgDispatch;
	
	// other parameters
	fInput.year = (Int_t)pConf->GetNum("MSISE_00Atmosphere.year");
	fInput.f107A = pConf->GetNum("MSISE_00Atmosphere.f107A");
	fInput.f107 = pConf->GetNum("MSISE_00Atmosphere.f107");
	fInput.ap = pConf->GetNum("MSISE_00Atmosphere.ap");
	fTuning.switches[0] = 0;
	for(Int_t i=1; i<24; i++) fTuning.switches[i] = 1;  //TOFIX : needed to allow better tuning
    } 
}

//________________________________________________________________________________________
void MSISE_00AtmosphereData::SetTables() {
    //
    // Set tables using MSISEtool
    //
    
    fAltitudeTable = new Double_t[fNb];
    fAir_DensityTable = new Double_t[fNb];
    fO_DensityTable = new Double_t[fNb];
    fO2_DensityTable = new Double_t[fNb];
    fN2_DensityTable = new Double_t[fNb];
    fTemperatureTable = new Double_t[fNb];
    fPressureTable = new Double_t[fNb];

    struct nrlmsise_output _out;
    for(Int_t i=0; i<fNb; i++) {
        fInput.alt = i * fStep/km;
	fTool->gtd7(&fInput,&fTuning,&_out);
	fO_DensityTable[i] = _out.d[1] * (AtomicMass("Oxygen")*g/(mole*Avogadro())) / cm3;
	fN2_DensityTable[i] = _out.d[2] * (AtomicMass("DiNitrogen")*g/(mole*Avogadro())) / cm3;
	fO2_DensityTable[i] = _out.d[3] * (AtomicMass("DiOxygen")*g/(mole*Avogadro())) / cm3;
	fAir_DensityTable[i] = _out.d[5] * g/cm3;
	fTemperatureTable[i] = _out.t[1] * kelvin;
	fPressureTable[i] = fAir_DensityTable[i] / (AtomicMass("Air")*g/mole) * 
        (R_ideal()/kelvin/mole) * fTemperatureTable[i];
	fAltitudeTable[i] = i * fStep;
    }
    string atmod = "msise";
    GetDefault(atmod,fNb);
    WriteUserModelFile(fNb);
#ifdef DEBUG
    //DebugPlots(fNb);    
#endif
}

/*
// For MSISEscan Macro --- WARNING : Init and setTables call in ctor MUST BE DISABLED (settable() must be called by msiseatmo)
//________________________________________________________________________________________
void MSISE_00AtmosphereData::InitForMacro(Int_t doy,Double_t lst, Double_t lat, Double_t longi) {
    //
    // Initialize msise parameters by hand
    //
    
    fTool = new MSISEtool();
    if(!fTool)  Msg(EsafMsg::Panic) << "No new MSISEtool, memory pb" << MsgDispatch ;
    ConfigFileParser* pConf = Config::Get()->GetCF("Atmosphere","MSISE_00Atmosphere");
    
	
    // then set parameters randomly between bounds
    fInput.g_lat  = lat;
    fInput.g_long = longi;
    fInput.doy    = doy;
    if(fInput.doy <= 0) {
	Msg(EsafMsg::Warning) <<"Day of year must be > 0 --> Set to 1"<<MsgDispatch;
	fInput.doy = 1;
    }
    else if(fInput.doy > 365) {
	Msg(EsafMsg::Warning) <<"Day of year must be < 366 --> Set to 365"<<MsgDispatch;
	fInput.doy = 365;
    }

    if(lst < 24.) {
	fInput.lst = lst;
	if(fInput.g_long <= 180.) fInput.sec = (fInput.lst - fInput.g_long / 15.) * 3600.;
	else if(fInput.g_long < 360.) fInput.sec = (fInput.lst + (360. - fInput.g_long) / 15.) * 3600.;
    }
    else {
	fInput.lst = lst - 24.;
	if(fInput.g_long <= 180.) fInput.sec = (lst - fInput.g_long / 15.) * 3600.;
	else if(fInput.g_long < 360.) fInput.sec = (lst + (360. - fInput.g_long) / 15.) * 3600.;
    }
    // correct UT time if necessary
    if(fInput.sec < 0.) {
	if(fInput.doy > 1) fInput.doy--;
	else fInput.doy = 365;
	fInput.sec += 24.*3600.;
    }
    else if(fInput.sec > 24.*3600.) {
	if(fInput.doy < 365) fInput.doy++;
	else fInput.doy = 1;
	fInput.sec -= 24.*3600.;
    }
    // final check
    if(fInput.sec < 0. || fInput.sec > 24.*3600.) Msg(EsafMsg::Warning) <<"Second in day not in bounds. SHOULD not occur = " <<fInput.sec <<MsgDispatch;

    // printings
    Msg(EsafMsg::Info) <<"LAT. = " <<fInput.g_lat <<MsgDispatch;
    Msg(EsafMsg::Info) <<"LONG. = " <<fInput.g_long <<MsgDispatch;
    Msg(EsafMsg::Info) <<"DAY = " <<fInput.doy <<MsgDispatch;
    Msg(EsafMsg::Info) <<"Local Solar Time = " <<fInput.lst <<" h"<<MsgDispatch;
    Msg(EsafMsg::Info) <<"UT time = " <<fInput.sec/3600.<<" h"<<MsgDispatch;

    // other parameters
    fInput.year = 0;
    fInput.f107A = pConf->GetNum("MSISE_00Atmosphere.f107A");
    fInput.f107 = pConf->GetNum("MSISE_00Atmosphere.f107");
    fInput.ap = pConf->GetNum("MSISE_00Atmosphere.ap");
    fTuning.switches[0] = 0;
    for(Int_t i=1; i<24; i++) fTuning.switches[i] = 1;  //TOFIX : needed to allow better tuning
}
*/
