// $Id: ETreePainter.hh 2399 2005-12-07 16:28:09Z moreggia $
// Author: Anne Stutz   2005/03/03

/*****************************************************************************
 * ESAF: Euso Simulation and Analysis Framework                              *
 *                                                                           *
 *  Id: ETreePainter                                                         *
 *  Package: <packagename>                                                   *
 *  Coordinator: <coordinator>                                               *
 *                                                                           *
 *****************************************************************************/

#ifndef __ETREEPAINTER_HH__
#define __ETREEPAINTER_HH__

//#include "euso.hh"
#include <TObject.h>
#include <TSeqCollection.h>
#include <TVector3.h>

class TGeoManager;
class TGeoVolume;
class TGeoMedium;
class TTree;
class TChain;
class EAtmosphere;
class ETruth;
class EShower;
class EAtmosphereHistoPainter;
class EEvent;
class TH1F;
class TCut;

////////////////////////////////////////////////////////////////////////////////
//                                                                            
// ETreePainter                                                                
//                                                                              
// <brief class description>                                                    
// * This class is responsible for reading events from the euso simulation tree
// * and doing its graphycal representation. 
// * It can draw the EUSO Field of View, displays 3D tracks,
// *        draw histos from Truth with Energy,Theta, Phi, X1
// *        draw histograms for photons in atmosphere
// *        draw histograms for photons on pupil
// *        fill a simple tree with fundamental parameters                                                   
////////////////////////////////////////////////////////////////////////////////



class ETreePainter : public TObject {
public:
    ETreePainter( TTree* );
    ETreePainter( TChain* );
    virtual ~ETreePainter();

    void Build();
    void SetBranchesStatus( TTree* );
    void BuildTree();                                            // Build fStree 
    // methods getting some info generated from simu data histos
    TH1F* BuildAltHisto( Option_t *opt );
    TH1F* BuildLongitudinalHisto( Option_t *opt );
    TH1F* BuildTofHisto( Option_t *opt );
    TH1F* BuildWlHisto( Option_t *opt );
    
    void FillTrack();                                            // Fill fTrack Data Member
    void FillTree();                                             // Fills the simple tree
    
    // Prepare and draw the histogramms
    void PrepareHistos(Option_t *,Float_t min,Float_t max,UInt_t Nbins);
    void DrawHistos(Option_t *, const TCut& drawcut,Option_t * opt="",Float_t min=0.,Float_t max=0.,UInt_t Nbins=0);
    
    inline TVector3 EUSO() const {return fEUSO;}
    inline Double_t DetectorRadius() const {return fRadius;}
    
    
    // for shower drawings
    //void AddShower();                               // Adds new shower read from the file 
    //void BuildWorld();                                      // Build the geometry
   // virtual void Draw( Option_t * ="");     
    //inline Bool_t IsWorldBuilt() {return fWorldBuild;}      // Checks if the geometry is built 
    

private:
        
    TVector3 fEUSO;                   // detector position
    Double_t fRadius;                 // pupil radius
    Double_t fGTU;                    // GTU length
    TChain      *fChain;              // Pointer for the euso tree chain to read from
    TTree       *fTree;               // Pointer for the euso tree to read from
    EEvent      *fEv;                 // Current Euso Event
    EShower     *fShower;             // shower simu data
    EAtmosphere *fAtmo;               // atmosphere simu data
    ETruth      *fTruth;              // simu truth

    Int_t        fEvTot;              // Total number of events in the fTree
    TTree       *fStree;              // resulting tree

    
    // define a stucture to fill fStree
    typedef struct {
        
	// flags to check the present analysis
	Int_t fGood_fit;           // 1 if all fits ok, 0 otherwise
	Int_t fH_max_last_bin;     // electrons Hmax, distce from maxbin to lastbin : to know if crashed shower
	Int_t fX_max_last_bin;     // idem but for Xmax (longi profile)
	Int_t fL_max_last_bin;     // idem but for Lmax (longi profile in km)
	Int_t fW_max_last_bin;     // idem but for Width right side search (longi profile in km)
	Int_t fH_fluomax_last_bin; // fluo Hmax, distce from maxbin to lastbin : to know if crashed shower
	Int_t fX_fluomax_last_bin; // idem but for Xmax (longi profile)
	Int_t fL_fluomax_last_bin; // idem but for Lmax (longi profile in km)
	Int_t fW_fluomax_last_bin; // idem but for Width right side search (longi profile in km)
	Int_t fH_ckovmax_last_bin; // idem for ckov
	Int_t fX_ckovmax_last_bin; // idem for ckov
	
	// atmosphere conditions
	Float_t fLatitude;       // geodetic latitude   [from -90 (south) to 90 (north)]
	Float_t fLongitude;      // geodetic longitude  [from 0 to 360 (east)]
	Float_t fDate;           // event date          [in sec UT, from 1st Jan 00:00:00] (year not taken into account)
	Float_t fHclouds;        // clouds top altitude (km)
	Float_t fCloudsThick;    // vertical thickness  (km)
	Float_t fCloudsOD;       // vertical Optical Depth
      
	// Truth
        Float_t fEnergy;
        Float_t fTheta;
        Float_t fPhi;
        
	Float_t fH1;             // first interaction point
	Float_t fInitPosX;
	Float_t fInitPosY;
	Float_t fInitPosZ;
        Float_t fX1;
	
        Float_t fTrueImpactX;    // shower impact and maximum
        Float_t fTrueImpactY;
        Float_t fTrueImpactZ;
        Float_t fTrueMaxPosX;
        Float_t fTrueMaxPosY;
        Float_t fTrueMaxPosZ;
        Float_t fTrueHmax;
        Float_t fTrueXmax;
	Float_t fTrueDmax;       // (km) shower max distance from detector


      // Electrons from shower
        Float_t fNe;           // longitudinal profile integration (in g/cm2)
        Float_t fNe2;          // integrated until max only (in g/cm2)
	Float_t fNemax;        // max of electrons longitudinal profile
	Float_t fWidthshow;    // width (at half max) of electrons profile  (in km)
	Float_t fWidth2show;   // width (at half max) of _the first half_ of electrons profile  (in km)
	
	Float_t fXmaxshow;     // Xmax from electrons profile fit  (in g/cm2)
	Float_t fHmaxshow;     // Hmax from electrons profile fit  (in g/cm2)
        Float_t fLmaxshow;     // shower max -> along track coordinate (from X1)
     
     
      // Photons at creation
      /******** FLUO ********/
        Float_t fNph_f_L;      // total fluo created  (longitudinal histo in km)
	Float_t fNph2_f_L;     // integrated until maximum only  (longitudinal histo in km)
	Float_t fNmax_f_L;     // in 1/km  (i.e. per track binlength)
	Float_t fWidth_f;      // width (at half max) of longitudinal fluo profile (in km)
	Float_t fWidth2_f;     // width (at half max) of _the first half_ of longitudinal fluo profile (in km)
	
        Float_t fHmax_f;       // fluo max local height (km)
        Float_t fxmax_f;       // fluo max -> x coordinate
        Float_t fymax_f;       // fluo max -> y coordinate
        Float_t fLmax_f;       // fluo max -> along track coordinate (from X1)
	Float_t fXmaxph_f;     // Xmax for fluo (g/cm2)
	Float_t fYieldmean;    // (per meter, per electron) mean fluo yield, averaged along the track
	Float_t fYieldHmax;    // (per meter, per electron) fluo yield at shower max
	
        //Float_t fNph_f;        // total fluo created  (longitudinal histo in g/cm2)
	//Float_t fNph2_f;       // integrated until maximum only  (longitudinal histo in g/cm2)
	//Float_t fNmax_f;       // in cm2/g (i.e. per grammage binlength)
        
	
      /******** CKOV ********/      // do not have : Lmax, Xmax, Ymax, Nph_L
	Float_t fNph_c_L;        // total ckov created
	Float_t fNph2_c_L;       // integrated until max only
        Float_t fHmax_c;         // ckov max local height (km)
	Float_t fYieldHmax_c;    // (per meter, per electron) ckov yield at shower max
        
        //Float_t fNmax_c;         // in cm2/g (i.e. per grammage binlength)
	//Float_t fXmaxph_c;       // Xmax for ckov (g/cm2)
	//Float_t fNph_c;        // total ckov created
	//Float_t fNph2_c;       // integrated until max only
	
	//Float_t fWlmean_f;     // in nm
	//Float_t fWlmean_c;     // ...
	//Float_t fWlmean_tot;   // ...  // pbic for filling + not really usefull so far
      
      
      // Photons on pupil
	    // Nph
	Float_t fCkovInOmega;     // ckov within solid angle
	Float_t fFluoInOmega;     // fluo within solid angle
	Float_t fNph_f_p;         // total fluo on pupil
	Float_t fNph_cdirect_p;   // total direct ckov on pupil
	Float_t fNph2_f_p;        // fluo on pupil integrated until Tmax
        Float_t fNph_airscat_p;   // air scattered ckov on pupil
        Float_t fNph_cloudscat_p; // clouds scattered ckov on pupil
        Float_t fNph_cr_p;        // reflected ckov on pupil
            
	    // Nmax
	Float_t fNmax_f_p_raw;    // per GTU - fluo, without any fit
	Float_t fNmax_f_p;        // per GTU - fluo, with fit
	Float_t fNmax_f2_p;       // per GTU : after ckov "substraction", i.e. with two fits
	Float_t fNmax_tot_p_raw;  // per GTU -- with air scat ckov, without any fit
	Float_t fNmax_tot_p;      // per GTU -- with air scat ckov, with fit
	Float_t fNmax_cr_p;       // per GTU - reflected ckov
	Float_t fNmax_crtot_p;    // per GTU - reflected + scattered ckov IN THE BIN
        Float_t fWidth_f2_p;      // (in GTU) SIGMA of gaussian fit : HALF Width of fluo profile on pupil, after ckov "substraction"
        Float_t fWidth_f_p;       // (in GTU) SIGMA of gaussian fit : HALF Width of fluo profile on pupil
        Float_t fWidth_tot_p;     // (in GTU) SIGMA of gaussian fit : HALF Width of (fluo+scattered photons) profile on pupil
            
	    // miscellaneous
	Float_t fYieldGTUmax;  // (per meter, per electron) fluo yield (averaged if sev. bunches (fNbBunches) contribute to GTUmax)
	Int_t fNbBunches;      // associated with above data (nb of bunches within GTUmax)
	Float_t fXGTUmax;      // (g/cm2) GTU max corresponding position along track (averaged if sev. bunches (fNbBunches) contribute to GTUmax)
	Float_t fXGTUmax_tot;  // (g/cm2) same as here above, but on (FLUO + CKOV backscattering) profile
	Float_t fTransGTUmax;  // ratio (photons_on_pupil / photons_in_omega) #FLUO# bunch at XGTUmax
	Float_t fTransMax;     // ratio (photons_on_pupil / photons_in_omega) #FLUO# bunch at Xmax_fluo
	Float_t fWlmean_f_p;   // in nm
	Float_t fWlmean_cb_p;  // ...
	Float_t fWlmean_cr_p;  // ...
	Float_t fWlmean_tot_p; // ...
    
    } Track;
    
    Track fTrack;
    
    // for shower drawings
    //Bool_t       fWorldBuild;         // true if the geometry is build, false othewise
    //TGeoVolume  *fTop;                // TOP Volume of the global geometry 
    //TGeoVolume  *fFoVVolume;          // Geometry realization of the Field of View of EUSO
    //TGeoMedium  *fVacuum;             // Geometry Medium
    //Int_t        fCurrentTrack;       // index of the current track read from fTree


    ClassDef(ETreePainter,0)
};

#endif  /* __ETREEPAINTER_HH__ */

