// ESAF : Euso Simulation and Analysis Framework
// Reconstruction Module data container
// $Id: RecoData.cc 2820 2009-03-13 14:34:09Z naumov $
// Marco Pallavicini created Feb, 16 2004
// Description:
// Rather simple and crude data container
// A RecoData can store data in the form of integers, doubles and
// objects of any kind
// Each data is identified with a name
// Name duplication is not allowed
// Objects are stored internally as pointer to void, therefore the user must
// know the type of object he is reading back and apply the right casting
//

#include <iostream>
#include <exception>
#include "RecoData.hh"

ClassImp(RecoData)

//______________________________________________________________________________
RecoData::RecoData() {
    //
    // Constructor
    // 
}

//______________________________________________________________________________
RecoData::~RecoData() {
    //
    // Destructor
    // 
}

//______________________________________________________________________________
Bool_t RecoData::CheckName(const string& name) {
    //
    // returns true if a name is valid (not already existing)
    // 
    if ( (fIntMap.count(name) == 0)    &&
         (fDoubleMap.count(name) == 0) &&
         (fObjMap.count(name) == 0) )
        return true;
    return false;
}

//______________________________________________________________________________
Int_t RecoData::Add(const string& name, const Int_t& val) {
    //
    // Add an integer value to the map
    // returns 0 if OK, 1 if value already existing
    //
    if ( CheckName(name) ) {
        fIntMap[name]=val;
        return 0;
    }
    Msg(EsafMsg::Warning) << "Duplicate name definition in module "+GetName() << MsgDispatch;
    return 1;
}

// add an integer value to the map
// returns 0 if OK, 1 if value already existing
Int_t RecoData::Add(const string& name, const Double_t& val) {
    if ( CheckName(name) ) {
        fDoubleMap[name]=val;
        return 0;
    }
    Msg(EsafMsg::Warning) << "Duplicate name definition in module "+GetName() << MsgDispatch;
    return 1;
}

//______________________________________________________________________________
Int_t RecoData::Add(const string& name, void* p) {
    //
    // Add an integer value to the map returns 0 if OK, 1 if value already
    // existing
    //
    if ( CheckName(name) ) {
        fObjMap[name]=p;
        return 0;
    }
    Msg(EsafMsg::Warning) << "Duplicate name definition in module "+GetName() << MsgDispatch;
    return 1;
}

//______________________________________________________________________________
Int_t RecoData::GetInt(const string& name) const {
    //
    // Get methods
    // Get an integer parameter
    //
    if ( fIntMap.count(name) == 1 )
        return fIntMap.find( name )->second ;
    if ( fIntMap.count(name) > 1 )
        FatalError("Duplicate integer parameter definition in module "+
        GetName());
    Msg(EsafMsg::Warning) << name << " is NOT existing integer parameter in module "+GetName() << MsgDispatch;
    return -999999;
}

//______________________________________________________________________________
Double_t RecoData::GetDouble(const string& name) const {
    //
    // Get a Double_t parameter
    //
    if ( fDoubleMap.count(name) == 1 )
        return fDoubleMap.find( name )->second;
    if ( fDoubleMap.count(name) > 1 )
        FatalError("Duplicate Double_t parameter definition in module"+
        GetName());
    Msg(EsafMsg::Warning) << name << " is NOT existing double parameter in module "+GetName() << MsgDispatch;
    return -999997.;
}


//______________________________________________________________________________
void* RecoData::GetObj(const string& name) const {
    // get any object
    // the user must know the object type and cast it properly
    // this must be used to store:
    // single pointer to objects
    // pointers to collection of objects as vector<> or map<>
    // pointers to TClonesArrays
    // any other type of object ALWAYS as pointers
   
    if ( fObjMap.count( name ) == 1 )
        return fObjMap.find( name )->second;
    if ( fObjMap.count( name ) > 1 )
        FatalError("Duplicate object definition in module "+GetName());
    Msg(EsafMsg::Warning) << name << " is NOT existing object in module "+GetName() << MsgDispatch;
    return NULL;
}

//______________________________________________________________________________
Bool_t RecoData::RemoveObj(const string& name) {
    // 
    // Remove an object
    // object is NOT deleted
    // 
    if ( fObjMap.count( name ) == 1 ) {
        fObjMap.erase( name );
        return true;
    }
    FatalError("Removing Unknown object in module "+GetName());
    return false;
}

//______________________________________________________________________________
void RecoData::CleanMaps() {    
    //
    // clean maps
    //

    Msg(EsafMsg::Debug) << "Cleaning " << fIntMap.size() << " Integers and " << fDoubleMap.size() << " Doubles " << " in the Module " << GetName() << " map" << MsgDispatch;
    fIntMap.clear();
    fDoubleMap.clear();
    if ( fObjMap.size() > 0 ) {
        Msg(EsafMsg::Warning) << "The module "  << GetName() << " the object map has still the following unremoved objects: " << MsgDispatch;
        map<string,void*>::const_iterator it;
        for( it = fObjMap.begin(); it != fObjMap.end(); it++) {
            Msg(EsafMsg::Warning) <<  "Name= "<<it->first<<"   Address="<<(Long_t)it->second<<MsgDispatch;
        }
		Msg(EsafMsg::Panic) <<  "Error in fObjMap handling" << MsgDispatch;
    }
    fObjMap.clear();
}
