// ESAF : Euso Simulation and Analysis Framework
// $Id: BaseClusteringModule.hh 2734 2006-06-09 13:03:09Z pesce $
// R.Pesce created Jan, 19 2004

#ifndef __BASECLUSTERINGMODULE_HH_
#define __BASECLUSTERINGMODULE_HH_

#include <vector>
#include "euso.hh"
#include "RecoModule.hh"

class EusoCluster;
class RecoRootEvent;

class BaseClusteringModule : public RecoModule {
public:
    // ctor
    BaseClusteringModule();

    // dtor
    virtual ~BaseClusteringModule();

    // Init method: called at the beginning of a run
    virtual Bool_t Init();

    // called before each event process
    virtual Bool_t PreProcess();

    // event processing
    virtual Bool_t Process( RecoEvent* );

    // called after processing an event
    virtual Bool_t PostProcess();

    // called at the end of each run
    virtual Bool_t Done();

    // save root output
    virtual Bool_t SaveRootData( RecoRootEvent* );

    // memory clean
    virtual void UserMemoryClean();

    // get a specified cluster
    EusoCluster* GetCluster( Int_t num );

    EsafConfigClass(Reco,BaseClusteringModule)

private:

    RecoEvent *fEv;
    RecoRootEvent *fRecoRootEvent; 
    Double_t fGtuLength;        // GTU length

    Double_t fSignalToNoise;    // signal to noise ratio (used with processonlysignal)
    Float_t fAlpha;             // correcting factor
    Int_t fMainMacroCellId;     // id of the most populous macrocell
    Double_t fDensity;          // density of points in the main macrocell
    Double_t fThreshold;        // threshold on the distance between two points of a cluster
    Int_t fTotalNumPoints;      // total number of points
    Int_t fNumClusters;         // total number of clusters
    Int_t fNumSigClusters;      // number of significant clusters
    Int_t fNumClustersUniform;  // number of clusters expected in an uniform distribution
    Int_t fNumPointsMinimum;    // minimum number of points in a significant cluster
    Int_t fSignificanceLevel;   // level of significance (from config file)
    Double_t fNumHitsMinimum;   // minimum number of points in a significant pixel (from config file)
                                // if is a negative value the number is calculated from the background rate in the pixel 
    Int_t fMeanThreshold;       // mean threshold in pixels hits;
    Double_t fFOVThreshold;     // if threshold type is fov, the clustering threshold is set to
                                // fFOVThreshold * Sqrt(DeltaThetaFOVPixel*DeltaPhiFOVPixel)
                                // using the pixel-angle map
    Double_t fNSigmaToSelect;   // number of sigmas from the main cluster line within the points are selected

    Int_t fGTUDiffThreshold;    // threshold in time difference in GTU between two pixels
    
    vector<Int_t> fHitted;      // references to pixels with minimum number of hits
    Int_t fNumPoints;           // number of pixels with minimum number of hits
    Int_t fNumHits;             // number of hits in points over pixel threshold
    // variables for clustering
    Int_t fNumNodes;            // number of nodes
    Int_t fNumPointsCluster;    // number of points in cluster
    Int_t fCounter;             // point to be processed
    Int_t fBookmark;            // first point of a new cluster
    vector<Int_t> fId;          // identifier of point
    vector<Bool_t> fFlag1; // flag that indicates if a point is already in a cluster
    vector<Bool_t> fFlag2; // flag that indicates if a point is already considered in the
                           // current search
    
    // methods for clustering
    void CalculateDensity();
    void DensityMainMC();
    void DensityEvent();
    void CalculateNaturalThreshold();
    Double_t CalculateDistance(Int_t n1, Int_t n2);
    Int_t    CalculateTimeDifference(Int_t n1, Int_t n2);
    void CalculateNumClustersUniform();
    void CalculateNumPointsMinimum();

    void NewCluster();
    void SearchClusters();
    void WriteCluster();
    void FillCluster();
    
    void FitCluster( EusoCluster *clu );
    void HistoCluster();

    // vector of significant clusters
    vector<EusoCluster*> fClusters;

    // parameters of the main cluster
    Double_t fSlopeMainPhi; Double_t fOffsetMainPhi; Double_t fAbsDevMainPhi;
    Double_t fSlopeMainTheta; Double_t fOffsetMainTheta; Double_t fAbsDevMainTheta;

    // vector of pixel ids to pass to the hough module
    vector<Int_t> fSelectedPixels;
   
    // data organization in GTU, EC,....
    Int_t fGtuMin;
    Int_t fGtuMax;
    void MakeMaps();   // create the maps of pixel data in Gtu,...
    map< Int_t, vector<Int_t> > fPixelGtuMap;

    // quality parameters //FIXME: to check definitions
    Double_t fSignalFraction; // signal fraction (over all signal) in the selected points
    Double_t fSigFracOverThresh; // signal fraction (over the signal in the pixels over threshold)
                                 // in the selected points
    Double_t fContamination;  // percentual of pixel w/o signal in the selected points
    Double_t fContaminationInCluster;  // percentual of pixel w/o signal in the main cluster
    
    friend class AutomatedClustering;
    friend class GTUClusteringModule;
    ClassDef(BaseClusteringModule,0)
};

#endif  /* __BASECLUSTERINGMODULE_HH_ */

