// ESAF : Euso Simulation and Analysis Framework
// $Id: BaseClusteringModule.cc 2734 2006-06-09 13:03:09Z pesce $
// R. Pesce created Jan, 19 2004

#include "BaseClusteringModule.hh"
#include "RecoEvent.hh"
#include "RecoGlobalData.hh"
#include "RecoRootEvent.hh"
#include "RecoCellInfo.hh"
#include "TDirectory.h"
#include "TMath.h"
#include "Config.hh"
#include "RecoPixelData.hh"
#include "EusoCluster.hh"
#include "ERunParameters.hh"
#include "MedianFit.hh"
#include "LeastSquaresFit.hh"
#include "HoughFit.hh"

#include <TH3F.h>
#include <TH2F.h>
#include <TCanvas.h>
#include <TGraph.h>

ClassImp(BaseClusteringModule)
using namespace TMath;
using namespace sou;

//____________________________________________________________________________
BaseClusteringModule::BaseClusteringModule() : 
    RecoModule(string("BaseClustering")) {
    //
    // ctor
    // 
}

// dtor
//____________________________________________________________________________
BaseClusteringModule::~BaseClusteringModule() {
}

//____________________________________________________________________________
Bool_t BaseClusteringModule::Init() {
    //
    // init
    // 
    fEv = (RecoEvent*)0;
    Msg(EsafMsg::Info) << "Initializing " << MsgDispatch;
    fAlpha = 0.55;
    fNumHitsMinimum = Conf()->GetNum("BaseClusteringModule.fNumHitsMinimum");
    fSignificanceLevel = (Int_t)Conf()->GetNum("BaseClusteringModule.fSignificanceLevel");
    if ( fSignificanceLevel <= 0 )
        Msg(EsafMsg::Panic) << "You must specify a positive number in BaseClusteringModule.fSignificanceLevel" << MsgDispatch;  
    fSignalToNoise    = (Double_t)Conf()->GetNum("BaseClusteringModule.fSignalToNoise");    
    fFOVThreshold = (Double_t)Conf()->GetNum("BaseClusteringModule.fFOVThreshold");
    fNSigmaToSelect = (Double_t)Conf()->GetNum("BaseClusteringModule.fNSigmaToSelect");
    fGTUDiffThreshold = (Int_t)Conf()->GetNum("BaseClusteringModule.fGTUDiffThreshold");
    return kTRUE;
}

//____________________________________________________________________________
Bool_t BaseClusteringModule::PreProcess() {
    //
    // pre-process
    // 
    fTotalNumPoints = 0;
    fNumPoints = 0;
    fNumHits = 0;
    fNumClusters = 0;
    fNumSigClusters = 0;
    fNumNodes = 0;
    fNumPointsCluster = 0;
    fBookmark = 0;
    fHitted.clear();
    fId.clear();
    fFlag1.clear();
    fFlag2.clear();
    fClusters.clear();
    fSignalFraction = 0.;
    fContamination = 1.;
    fGtuMin = 100000;
    fGtuMax = -100000;
    fPixelGtuMap.clear();
        
    return kTRUE;
}

//____________________________________________________________________________
Bool_t BaseClusteringModule::Process( RecoEvent* ev ) {
    //
    // process
    // 
    fEv = ev;
    fTotalNumPoints = fEv->GetHeader().GetNumActiveFee();
    fGtuLength = fEv->GetHeader().GetGtuLength();
    

    if ( fTotalNumPoints <= 0 ) {       
        Msg(EsafMsg::Info) << "No points in event! Exit from this event." << MsgDispatch;
        return kFALSE;
    }
    
    // get id of most populous macrocell (main macrocell)
    fMainMacroCellId = fEv->GetMainMacroCell(1);
    if ( fEv->GetRecoCellInfo(fMainMacroCellId) == NULL )
        Msg(EsafMsg::Panic) << "Main macrocell does not correspond to any macrocell simulated in the event." << MsgDispatch;

    // process only signal and filter scattered cherenkov
    Bool_t ProcessOnlySignal = Config::Get()->GetCF("Reco","RootInputModule")->GetBool("RootInputModule.ProcessOnlySignal");
    Bool_t FilterScatteredCherenkov = Config::Get()->GetCF("Reco","RootInputModule")->GetBool("RootInputModule.FilterScatteredCherenkov");

    // check if any previous pattern recognition
    Bool_t isAlreadyPattReco = kFALSE;
    vector<Int_t> selpix; selpix.clear();
    RecoGlobalData* gdPattReco = fEv->FindGlobalData("PatternRecognition");
    if ( gdPattReco ) {
        isAlreadyPattReco = kTRUE;
        selpix = *(vector<Int_t>*) gdPattReco->GetObj("SelectedPixels");
        fTotalNumPoints = selpix.size();
    }
        

    
    // search pixels with num hits > minimum
    // and calculate the total number of signal hits
    // in all the event and in the pixels over threshold
    Int_t numsel(0);
    fMeanThreshold = 0;
    Int_t totalSignals(0);
    Int_t ovThSignals(0);
    for( Int_t i=0; i<fTotalNumPoints; i++ ) {
        RecoPixelData *pix;
        if (!isAlreadyPattReco) pix = (RecoPixelData*)fEv->GetRecoPixelData(i);
        else pix = (RecoPixelData*)fEv->GetRecoPixelData(selpix[i]); 
        if (FilterScatteredCherenkov) {
            pix->SetCountsSignal(pix->GetCountsSignal() - pix->GetCountsCherScatt());
            pix->SetCounts(pix->GetCounts() - pix->GetCountsCherScatt());
        }
        
        totalSignals += pix->GetCountsSignal();
        
        Bool_t thereIsSignal = kTRUE;
        if (ProcessOnlySignal)  {
             Int_t signal_counts = pix->GetCountsSignal();
             Int_t noise_counts  = pix->GetCountsNoise();
             Double_t ratio (-1);
             if      (noise_counts > 0   )  ratio = 1.e0*signal_counts/Sqrt(noise_counts);
             else if ( signal_counts > 0 )  ratio = 1.e10;
             if (ratio <= fSignalToNoise) thereIsSignal = kFALSE;
        }
        if (!thereIsSignal) continue;

        Int_t threshold(0);
        if ( fNumHitsMinimum >=0 ) threshold = (Int_t)fNumHitsMinimum;
        else {
            Int_t uid = pix->GetPixelId();
            Double_t rate = fEv->GetHeader().GetRunPars()->GetNightGlowRateByUId(uid)*fGtuLength/microsecond;
            threshold = (Int_t)Ceil( rate + Abs(fNumHitsMinimum)*Sqrt(rate));
        }
        numsel++; fMeanThreshold += threshold;
        if ( pix->GetCounts() >= threshold ) {
            fHitted.push_back(i);
            fNumPoints++;
            fNumHits += pix->GetCounts();
            ovThSignals += pix->GetCountsSignal();
        }
    }
    if (numsel) fMeanThreshold /= numsel;
    else fMeanThreshold = 100000;
    
    Msg(EsafMsg::Debug) << "Num. of pixels with at least " << fNumHitsMinimum << " hits = " << fNumPoints << MsgDispatch;
    if (fNumPoints<10) {
        Msg(EsafMsg::Warning) << "Number of points "<< fNumPoints << " is less then 10 "<< " Exit from this event." << MsgDispatch;
        fHitted.clear();
        fMainMacroCellId = 0;
        fNumPoints = 0;
        return kFALSE;
    }

    // fill id and flag vectors
    for( Int_t i=0; i<fNumPoints; i++ ) {
        fId.push_back(0);
        fFlag1.push_back(kFALSE);
        fFlag2.push_back(kFALSE);
    }
   
    Msg(EsafMsg::Debug) << "Mean Threshold: " << fMeanThreshold << MsgDispatch;
    
    //make data maps
    MakeMaps();
    
    // clustering method
    if (Conf()->GetStr("BaseClusteringModule.fThresholdType")=="natural") {
        CalculateDensity();
    	CalculateNaturalThreshold();
	    Msg(EsafMsg::Debug) <<"Density: " << fDensity << "  Natural Distance Threshold: " << fThreshold << MsgDispatch;
    	CalculateNumClustersUniform();
        CalculateNumPointsMinimum();
    	Msg(EsafMsg::Debug) << "Number of clusters in uniform distribution: " << fNumClustersUniform << MsgDispatch;
        Msg(EsafMsg::Debug) << "Number of points minimum in a significant cluster: " << fNumPointsMinimum << MsgDispatch;
    } else if (Conf()->GetStr("BaseClusteringModule.fThresholdType")=="fixed") {
    	fThreshold = Conf()->GetNum("BaseClusteringModule.fThreshold");
	    fNumPointsMinimum = (Int_t)Conf()->GetNum("BaseClusteringModule.fNumPointsMinimum");
    	Msg(EsafMsg::Debug) << "Distance Threshold: " << fThreshold << MsgDispatch;
    } else if (Conf()->GetStr("BaseClusteringModule.fThresholdType")=="fov") {
        fThreshold = fEv->GetRecoPixelData(fHitted[0])->GetThetaSigma() * fEv->GetRecoPixelData(fHitted[0])->GetPhiSigma();
        fThreshold = Abs(fFOVThreshold) * Sqrt(fThreshold);
        Msg(EsafMsg::Debug) << "Distance Threshold: " << fThreshold << MsgDispatch;
        fNumPointsMinimum = (Int_t)Conf()->GetNum("BaseClusteringModule.fNumPointsMinimum"); 
    } else {
        Msg(EsafMsg::Panic) << "Wrong value in BaseClusteringModule.fThresholdType" << MsgDispatch;
    }
    
    SearchClusters(); 
    
    //FIXME: just a try
    /*while ( (fNumSigClusters == 0) && (fSignificanceLevel>1) ) {
        fSignificanceLevel /= 2;
        CalculateNumPointsMinimum();
        SearchClusters();
        Msg(EsafMsg::Debug) << "Significance level: " << fSignificanceLevel << MsgDispatch;
        Msg(EsafMsg::Debug) << "Number of significant clusters: " << fNumSigClusters << " in " << fNumClusters << " clusters" << MsgDispatch;
    }*/

    //find the main cluster (cluster with maximum number of points)
    Int_t mainclu_id = -1;
    Int_t maxnumpoints = -1;
    for(Int_t i(0); i<fNumSigClusters; i++) {
        if ( fClusters[i]->GetNumPoints() > maxnumpoints ) {
            maxnumpoints = fClusters[i]->GetNumPoints();
            mainclu_id = i;
        }
    }

    // calculate the background contamination in the main cluster
    // fit the cluster and do the selection of points 
    if ( mainclu_id >=0 ) {
        EusoCluster* mainclu = fClusters[mainclu_id];
        FitCluster(mainclu);
    
        fSlopeMainTheta = mainclu->GetSlopeTheta();
        fOffsetMainTheta = mainclu->GetOffsetTheta();
        fAbsDevMainTheta = mainclu->GetAbsoluteDeviationTheta();
        fSlopeMainPhi = mainclu->GetSlopePhi();
        fOffsetMainPhi = mainclu->GetOffsetPhi();
        fAbsDevMainPhi = mainclu->GetAbsoluteDeviationPhi();

        // contamination in the cluster
        Int_t nNoisePixCluster(0);
        for( Int_t i(0); i<mainclu->GetNumPoints(); i++) {
            RecoPixelData *pix = (RecoPixelData*)fEv->GetRecoPixelData(mainclu->GetPixelId(i));
            if (!pix->GetCountsSignal()) nNoisePixCluster++;
        }
        fContaminationInCluster = (Double_t)nNoisePixCluster/mainclu->GetNumPoints();
        
        // Selection of points
        for( Int_t i(0); i<fNumPoints; i++ ) {
            Double_t theta = fEv->GetRecoPixelData(fHitted[i])->GetTheta(); 
            Double_t phi   = fEv->GetRecoPixelData(fHitted[i])->GetPhi();
            Double_t time   = fEv->GetRecoPixelData(fHitted[i])->GetGtu()*fGtuLength;
            if (Conf()->GetStr("BaseClusteringModule.fFitMethod")=="spacetime") {
                if ( ( Abs(theta -fSlopeMainTheta*time -fOffsetMainTheta) < Abs(fNSigmaToSelect)*fAbsDevMainTheta ) &&
                 ( Abs(phi -fSlopeMainPhi*time -fOffsetMainPhi) < Abs(fNSigmaToSelect)*fAbsDevMainPhi ) ) 
                    fSelectedPixels.push_back(fHitted[i]);
            } else if (Conf()->GetStr("BaseClusteringModule.fFitMethod")=="space") {
                if  ( Abs(theta -fSlopeMainTheta*phi -fOffsetMainTheta) < Abs(fNSigmaToSelect)*fAbsDevMainTheta ) 
                    fSelectedPixels.push_back(fHitted[i]);
            }

        }
    }
        if ( fNumSigClusters != 0 ) {
            for( Int_t i=0; i<fNumSigClusters; i++ ) {
	        MsgForm(EsafMsg::Debug, "Cluster n.%d with %d points",i,fClusters[i]->GetNumPoints());
        }

        // calculate the quality parameters
        Int_t nPixNoise(0); // number of pure noise pixels in selection
        Int_t selectedSignals(0); // number of signal counts in selection
        for( UInt_t i(0); i<fSelectedPixels.size(); i++ ) {
            RecoPixelData *pix = (RecoPixelData*)fEv->GetRecoPixelData(fSelectedPixels[i]);
            if ( pix->GetCountsSignal() ) selectedSignals += pix->GetCountsSignal();
            else nPixNoise++;
        }
        fSignalFraction = (Double_t)selectedSignals/totalSignals;
        fSigFracOverThresh = (Double_t)selectedSignals/ovThSignals;
        fContamination = (Double_t)nPixNoise/fSelectedPixels.size();
        Msg(EsafMsg::Info) << "SignalFraction = " << fSignalFraction << " Contamination = " << fContamination << MsgDispatch;
        Msg(EsafMsg::Info) << "SignalFraction over threshold = " << fSigFracOverThresh << MsgDispatch;
        Msg(EsafMsg::Info) << "Contaminaion in main cluster = " << fContaminationInCluster << MsgDispatch;
    
        if ( Conf()->GetBool("BaseClusteringModule.fDoHistogram") ) {
            HistoCluster();
        }
        
        // add data to event
        vector<EusoCluster*> *pclu = &fClusters;
        MyData()->Add("Clusters", pclu);
        vector<Int_t> *selpix = &fSelectedPixels;
        MyData()->Add("SelectedPixels", selpix);
        MyData()->Add("NumHitsMinimum",fNumHitsMinimum);
        MyData()->Add("MeanThreshold",fMeanThreshold);
        MyData()->Add("SignificanceLevel",fSignificanceLevel);
        MyData()->Add("NumPointsMinimum",fNumPointsMinimum);
        MyData()->Add("DistanceThreshold",fThreshold);
        fSignificanceLevel = (Int_t)Conf()->GetNum("BaseClusteringModule.fSignificanceLevel"); //reset

        return kTRUE;
    }
        
    if ( Conf()->GetBool("BaseClusteringModule.fDoHistogram") ) {
        HistoCluster();
    }   

    fSignificanceLevel = (Int_t)Conf()->GetNum("BaseClusteringModule.fSignificanceLevel"); //reset
    return kFALSE;
}

//___________________________________________________________________________
void BaseClusteringModule::MakeMaps() {
    //
    // Organize the event data to optimize the clustering process
    //

    // sort fHitted by gtu
    /*Int_t n = fHitted.size();
    for (Int_t j=n/2; 0<j; j/=2) {
        for(Int_t i(0); i<n; i++) {
            for(Int_t k=i-j; 0<=k; k-=j) {
                if (fEv->GetRecoPixelData(fHitted[k+j])->GetGtu() < fEv->GetRecoPixelData(fHitted[k])->GetGtu() ) {
                    Int_t dummy = fHitted[k];
                    fHitted[k] = fHitted[k+j];
                    fHitted[k+j] = dummy;
                }
            }
        }
    }*/
    
    // fill the gtu-pixel map
    // for each gtu store the corresponding fHitted entries
    // (fHitted is the vector of the ids of recopixeldata over threshold)
    for(Int_t i(0); i<fNumPoints; i++) {
        Int_t gtu = fEv->GetRecoPixelData(fHitted[i])->GetGtu();
        if ( gtu < fGtuMin ) fGtuMin = gtu;
        if ( gtu > fGtuMax ) fGtuMax = gtu;
        fPixelGtuMap[gtu].push_back(i);
    }
    
    
}

//___________________________________________________________________________
Bool_t BaseClusteringModule::PostProcess() {
    //
    // post-process
    // 

    if ( fTotalNumPoints == 0 ) return kTRUE;
    if ( fNumPoints < 10 ) {
        fEv = (RecoEvent*)0; 
        return kTRUE;
    }
    
    // Update data in PatternRecognition Global data
    RecoGlobalData* data = fEv->FindCreateGlobalData("PatternRecognition");
    data->CleanMaps();
    data->SetIssuer(GetName());
    data->Add("SelectedPixels",&fSelectedPixels);
    data->Add("NumPointsMinimum",fNumPointsMinimum);
    data->Add("MeanThreshold",fMeanThreshold);
    
        
    fEv = (RecoEvent*)0;
    return kTRUE;
      
}

//___________________________________________________________________________
Bool_t BaseClusteringModule::SaveRootData(RecoRootEvent *fRecoRootEvent) {
    fRecoRootEvent->GetRecoClustering().SetNumPoints((Int_t)fSelectedPixels.size());
    fRecoRootEvent->GetRecoClustering().SetSignalFraction(fSignalFraction);
    fRecoRootEvent->GetRecoClustering().SetSigFracOverThresh(fSigFracOverThresh);
    fRecoRootEvent->GetRecoClustering().SetContamination(fContamination);
    fRecoRootEvent->GetRecoClustering().SetContaminationInCluster(fContaminationInCluster);
    fRecoRootEvent->GetRecoClustering().SetAbsDevMainTheta(fAbsDevMainTheta);
    fRecoRootEvent->GetRecoClustering().SetAbsDevMainPhi(fAbsDevMainPhi);
    
    return kTRUE;
}

//___________________________________________________________________________
Bool_t BaseClusteringModule::Done() {
    //
    // done
    // 
    fHitted.clear();
    fId.clear();
    fFlag1.clear();
    fFlag2.clear();
    fClusters.clear();
    fPixelGtuMap.clear();
    Msg(EsafMsg::Info) << "Completed" << MsgDispatch;
    return kTRUE;
}


//___________________________________________________________________________
void BaseClusteringModule::UserMemoryClean() {
    //
    // memory clean
    // 
    if ( fTotalNumPoints == 0 || fNumPoints < 10 || fNumSigClusters == 0 ) return;

    vector<EusoCluster*> *pclu = (vector<EusoCluster*>*)MyData()->GetObj("Clusters");
    EusoCluster *dummy(NULL);
    for(size_t i(0); i<pclu->size(); i++) {
        dummy = (*pclu)[i];
        delete dummy;
    }
    (*pclu).clear();
    MyData()->RemoveObj("Clusters");

    vector<Int_t> *selpix = (vector<Int_t>*)MyData()->GetObj("SelectedPixels");
    if (selpix) {
        selpix->clear();
        MyData()->RemoveObj("SelectedPixels");
        vector<Int_t> dummyV;
        dummyV.swap(*selpix);
    }

}

//___________________________________________________________________________
void BaseClusteringModule::CalculateDensity() {
    //
    // calculate density of points
    // 
    if ( Conf()->GetStr("BaseClusteringModule.fDensityMethod")=="mainmc" ) {
        DensityMainMC();
    } else if ( Conf()->GetStr("BaseClusteringModule.fDensityMethod")=="event" ) {
        DensityEvent();
    } else {
        Msg(EsafMsg::Panic) <<"Wrong config value BaseClusteringModule.fDensityMethod" << MsgDispatch;  
    }
}

//___________________________________________________________________________
void BaseClusteringModule::DensityMainMC() {
    //
    // Calculate density from main macrocell
    // 
    Double_t nph = (Double_t)fEv->GetNumPx(fMainMacroCellId, fMeanThreshold);
    Double_t area = (Double_t)fEv->GetThetaRange(fMainMacroCellId, fMeanThreshold);
    area *= (Double_t)fEv->GetPhiRange(fMainMacroCellId, fMeanThreshold);
    fDensity = nph / area;
    if (fDensity==0) DensityEvent();
}

//___________________________________________________________________________
void BaseClusteringModule::DensityEvent() {
    //
    // Calculate density from complete event
    //
    Float_t thmin = fEv->GetRecoPixelData(fHitted[0])->GetTheta();
    Float_t thmax = fEv->GetRecoPixelData(fHitted[0])->GetTheta();
    Float_t phimin = fEv->GetRecoPixelData(fHitted[0])->GetPhi();
    Float_t phimax = fEv->GetRecoPixelData(fHitted[0])->GetPhi();
    
    for(Int_t i=1; i<fNumPoints; i++) {
        Int_t j = fHitted[i];
        if ( fEv->GetRecoPixelData(j)->GetTheta() < thmin )
            thmin = fEv->GetRecoPixelData(j)->GetTheta();
        if ( fEv->GetRecoPixelData(j)->GetTheta() > thmax )
            thmax = fEv->GetRecoPixelData(j)->GetTheta();
        if ( fEv->GetRecoPixelData(j)->GetPhi() < phimin )
            phimin = fEv->GetRecoPixelData(i)->GetPhi();
        if ( fEv->GetRecoPixelData(j)->GetPhi() > phimax )
            phimax = fEv->GetRecoPixelData(i)->GetPhi();
    }
    Double_t area = Abs(( fEv->GetEndGtu()-fEv->GetStartGtu() )*(thmax-thmin)*(phimax-phimin));
    fDensity = (Double_t) fNumPoints / area;
}

//___________________________________________________________________________
void BaseClusteringModule::CalculateNaturalThreshold() {
    //
    // calculate natural threshold
    //
    fThreshold = Sqrt(Log(2)/
        (fDensity*fAlpha*Pi()));
}

//___________________________________________________________________________
Double_t BaseClusteringModule::CalculateDistance(Int_t n1, Int_t n2) {
    //
    // calculate distance between two points
    // 
    RecoPixelData *px1=fEv->GetRecoPixelData(n1);
    RecoPixelData *px2=fEv->GetRecoPixelData(n2);
    return ACos( Sin(px1->GetTheta())*Sin(px2->GetTheta())*
           Cos(px1->GetPhi())*Cos(px2->GetPhi()) +
           Sin(px1->GetTheta())*Sin(px2->GetTheta())*
           Sin(px1->GetPhi())*Sin(px2->GetPhi()) +
           Cos(px1->GetTheta())*Cos(px2->GetTheta()) );
}

//___________________________________________________________________________
Int_t BaseClusteringModule::CalculateTimeDifference(Int_t n1, Int_t n2) {
    //
    // calculate the time difference between two points in GTU
    // 
    RecoPixelData *px1=fEv->GetRecoPixelData(n1);
    RecoPixelData *px2=fEv->GetRecoPixelData(n2);
    return Abs(px1->GetGtu() - px2->GetGtu());
}

//___________________________________________________________________________
void BaseClusteringModule::CalculateNumClustersUniform() {
    //
    // calculate number of clusters expected in a distribution uniform
    // 
    fNumClustersUniform = (Int_t)( 1 + ( fNumPoints - 1 ) *
        Exp( -fAlpha * fDensity * fThreshold * fThreshold * Pi() ) );
}

//___________________________________________________________________________
void BaseClusteringModule::CalculateNumPointsMinimum() {
    //
    // calculate the minimum number of points in a significant cluster
    // 
    fNumPointsMinimum = (Int_t) ( fNumPoints / fNumClustersUniform 
        + fSignificanceLevel * Sqrt( (Double_t)( fNumPoints / fNumClustersUniform ) ) );
}

//___________________________________________________________________________
void BaseClusteringModule::SearchClusters() {
    //
    // search for clusters
    // 
    while( fNumNodes < fNumPoints ) {
        NewCluster();
        FillCluster();
        WriteCluster();
    }
}

//___________________________________________________________________________
void BaseClusteringModule::NewCluster() {
    //
    // init a new cluster
    // 
    fNumPointsCluster = 1;
    fNumClusters++;
    fNumNodes++;
    fFlag2[fBookmark] = kTRUE;
    fId[0] = fBookmark;
    fCounter = 0;
}

//___________________________________________________________________________
void BaseClusteringModule::FillCluster() {
    //
    // fill a cluster
    // 
    /*while ( fCounter < fNumPointsCluster ) {
    
        Int_t k = fId[fCounter];
        if ( !fFlag1[k] ) {
            fFlag1[k] = kTRUE;
            for (Int_t i=0; i<fNumPoints; i++) {
                if ( !fFlag2[i] ) {
                    if ( (CalculateDistance(fHitted[i], fHitted[k]) <= fThreshold) && 
                            (CalculateTimeDifference(fHitted[i], fHitted[k]) <= fGTUDiffThreshold) ) {   
                        fFlag2[i] = kTRUE;
                        fNumPointsCluster++;
                        fId[fNumPointsCluster-1] = i;
                        fNumNodes++;
                    }
                    fBookmark = i;
                }
            }
        }
        fCounter++;
    }*/
    while ( fCounter < fNumPointsCluster ) {
    
        Int_t k = fId[fCounter];
        if ( !fFlag1[k] ) {
            fFlag1[k] = kTRUE;
            Int_t gtu = fEv->GetRecoPixelData(fHitted[k])->GetGtu();
            for (Int_t j=-1; j<2; j++) {
                if ( (gtu+j)<fGtuMin || (gtu+j)>fGtuMax ) continue;
                for (Int_t i=0; i<(Int_t)(fPixelGtuMap[gtu+j]).size(); i++) {
                    Int_t q = (fPixelGtuMap[gtu+j])[i];
                    if ( !fFlag2[q] ) {
                        if  (CalculateDistance(fHitted[q], fHitted[k]) <= fThreshold) {
                            fFlag2[q] = kTRUE;
                            fNumPointsCluster++;
                            fId[fNumPointsCluster-1] = q;
                            fNumNodes++;
                        }
                        //fBookmark = q;
                    }
                }
            }
        }
        fCounter++;

    }
    fBookmark++;
}

//___________________________________________________________________________
void BaseClusteringModule::WriteCluster() {
    //
    // close and write the current cluster if is significant
    //

    // check cluster significance
    if ( fNumPointsCluster < fNumPointsMinimum )
        return;

    // write the cluster
    fNumSigClusters++;
    EusoCluster *cl = new EusoCluster();
    Double_t thmin, thmax, phimin, phimax;
    Int_t gtumin, gtumax;
    thmin = fEv->GetRecoPixelData(fHitted[fId[0]])->GetTheta(); 
    thmax = fEv->GetRecoPixelData(fHitted[fId[0]])->GetTheta();
    phimin = fEv->GetRecoPixelData(fHitted[fId[0]])->GetPhi();
    phimax = fEv->GetRecoPixelData(fHitted[fId[0]])->GetPhi();
    gtumin = fEv->GetRecoPixelData(fHitted[fId[0]])->GetGtu();
    gtumax = fEv->GetRecoPixelData(fHitted[fId[0]])->GetGtu();
   
    Double_t theta(0),phi(0);
    Int_t gtu(0);
    for( Int_t i=0; i<fNumPointsCluster; i++ ) {
        theta = fEv->GetRecoPixelData(fHitted[fId[i]])->GetTheta();
        phi = fEv->GetRecoPixelData(fHitted[fId[i]])->GetPhi(); 
        gtu = fEv->GetRecoPixelData(fHitted[fId[i]])->GetGtu();
        if ( theta < thmin ) thmin = theta;
        if ( theta > thmax ) thmax = theta;
        if ( phi < phimin ) phimin = phi;
        if ( phi > phimax ) phimax = phi;
        if ( gtu < gtumin ) gtumin = gtu;
        if ( gtu > gtumax ) gtumax = gtu;
        cl->AddPoint( fHitted[fId[i]] );
    }
    cl->SetThetaMin( thmin );
    cl->SetThetaMax( thmax );
    cl->SetPhiMin( phimin );
    cl->SetPhiMax( phimax );
    cl->SetGtuMin( gtumin );
    cl->SetGtuMax( gtumax );
    cl->SetThreshold( (Int_t)fNumHitsMinimum );
    fClusters.push_back(cl);
}


//___________________________________________________________________________
EusoCluster* BaseClusteringModule::GetCluster( Int_t i ) {
    //
    // get a specified cluster
    // 
    if ( i >= 0 && i < (Int_t)fClusters.size() )
        return fClusters[i];
    else
        return NULL;
}

//___________________________________________________________________________
void BaseClusteringModule::FitCluster( EusoCluster *clu ) {
    //
    // fit the points of a cluster; use linear or median or hough fit
    // 

    vector<Double_t> thetavector;
    vector<Double_t> phivector;
    vector<Double_t> timevector;
    for( Int_t i=0; i<clu->GetNumPoints(); i++ ) {
        RecoPixelData *pix = fEv->GetRecoPixelData( clu->GetPixelId(i) );
        thetavector.push_back(pix->GetTheta());
        phivector.push_back(pix->GetPhi());
        timevector.push_back(pix->GetGtu()*fGtuLength);
    }
    
    if ( Conf()->GetStr("BaseClusteringModule.fFitMethod")=="spacetime" ) {
        MedianFit *mftheta = new MedianFit( clu->GetNumPoints(), timevector, thetavector );
        clu->SetFitted( kTRUE );
        clu->SetSlopeTheta( mftheta->GetSlope() );
        clu->SetOffsetTheta( mftheta->GetOffset() );
        clu->SetAbsoluteDeviationTheta( mftheta->GetAbsoluteDeviation() );
        MedianFit *mfphi = new MedianFit( clu->GetNumPoints(), timevector, phivector );
        clu->SetSlopePhi( mfphi->GetSlope() );
        clu->SetOffsetPhi( mfphi->GetOffset() );
        clu->SetAbsoluteDeviationPhi( mfphi->GetAbsoluteDeviation() );
        //cout << mftheta->GetSlope() << " " << mftheta->GetOffset() << " " << mftheta->GetAbsoluteDeviation() << endl;
        //cout << mfphi->GetSlope() << " " << mfphi->GetOffset() << " " << mfphi->GetAbsoluteDeviation() << endl;
    } else if ( Conf()->GetStr("BaseClusteringModule.fFitMethod")=="space" ) {
        MedianFit *mf = new MedianFit( clu->GetNumPoints(), phivector, thetavector );
        clu->SetFitted( kTRUE );
        clu->SetSlopeTheta( mf->GetSlope() );
        clu->SetOffsetTheta( mf->GetOffset() );
        clu->SetAbsoluteDeviationTheta( mf->GetAbsoluteDeviation() );
    } else {
        Msg(EsafMsg::Panic) << "Wrong config value BaseClusteringModule.fFitMethod" << MsgDispatch;  
    }
}

//___________________________________________________________________________
void BaseClusteringModule::HistoCluster() {
    //
    // make an histogram of the cluster
    // 

    gDirectory->cd("/");
    string pathname = Form("cluster%d", fEv->GetHeader().GetNum());
    TDirectory *path = new TDirectory(pathname.c_str(),pathname.c_str());
    path->cd();
    
    Float_t thmin = fEv->GetRecoPixelData(0)->GetTheta();
    Float_t thmax = fEv->GetRecoPixelData(0)->GetTheta();
    Float_t phimin = fEv->GetRecoPixelData(0)->GetPhi();
    Float_t phimax = fEv->GetRecoPixelData(0)->GetPhi();
    Float_t tmin = fEv->GetRecoPixelData(0)->GetGtu();
    Float_t tmax = fEv->GetRecoPixelData(0)->GetGtu();
    
    for( Int_t i=1; i<fTotalNumPoints; i++ ) {
        if ( fEv->GetRecoPixelData(i)->GetTheta() < thmin )
            thmin = fEv->GetRecoPixelData(i)->GetTheta();
        if ( fEv->GetRecoPixelData(i)->GetTheta() > thmax )
            thmax = fEv->GetRecoPixelData(i)->GetTheta();
        if ( fEv->GetRecoPixelData(i)->GetPhi() < phimin )
            phimin = fEv->GetRecoPixelData(i)->GetPhi();
        if ( fEv->GetRecoPixelData(i)->GetPhi() > phimax )
            phimax = fEv->GetRecoPixelData(i)->GetPhi();
        if ( fEv->GetRecoPixelData(i)->GetGtu() < tmin )
            tmin = fEv->GetRecoPixelData(i)->GetGtu();
        if ( fEv->GetRecoPixelData(i)->GetGtu() > tmax )
            tmax = fEv->GetRecoPixelData(i)->GetGtu();
    }
	
    Int_t bin = (Int_t)Conf()->GetNum("BaseClusteringModule.fHistogramBinning");
    TH3F *clhisto = new TH3F("ClHisto","Clustered Points",bin,phimin,phimax,bin,thmin,thmax,bin,tmin,tmax);
    TH3F *hithisto = new TH3F("HitHisto","Points with minimum hits",bin,phimin,phimax,bin,thmin,thmax,
                              bin,tmin,tmax);
    TH3F *histo = new TH3F("Histo","Points",bin,phimin,phimax,bin,thmin,thmax,bin,tmin,tmax);
    TH2F *clhisto2 = new TH2F("ClHisto2","Clustered Points",bin,phimin,phimax,bin,thmin,thmax);
    TH2F *hithisto2 = new TH2F("HitHisto2","Points with minimum hits",bin,phimin,phimax,bin,thmin,thmax);
    TH2F *histo2 = new TH2F("Histo2","Points",bin,phimin,phimax,bin,thmin,thmax);
    TH3F *selhisto = new TH3F("SelHisto","Selected Points",bin,phimin,phimax,bin,thmin,thmax,bin,tmin,tmax);
    TH2F *selhisto2 = new TH2F("SelHisto2","Selected Points",bin,phimin,phimax,bin,thmin,thmax);
 
 
    for( Int_t i=0; i<fTotalNumPoints; i++ ) {
        histo->Fill( fEv->GetRecoPixelData(i)->GetPhi(),fEv->GetRecoPixelData(i)->GetTheta(), 
                     fEv->GetRecoPixelData(i)->GetGtu() );
        histo2->Fill( fEv->GetRecoPixelData(i)->GetPhi(),fEv->GetRecoPixelData(i)->GetTheta());
    }
    histo->GetXaxis()->SetTitle("#varphi (rad)");
    histo->GetYaxis()->SetTitle("#theta (rad)");
    histo->GetZaxis()->SetTitle("GTU");
    histo2->GetXaxis()->SetTitle("#varphi (rad)");
    histo2->GetYaxis()->SetTitle("#theta (rad)");
    
    for( Int_t i=0; i<fNumPoints; i++ ) {
        hithisto->Fill(fEv->GetRecoPixelData(fHitted[i])->GetPhi(),
            fEv->GetRecoPixelData(fHitted[i])->GetTheta(), fEv->GetRecoPixelData(fHitted[i])->GetGtu() );
        hithisto2->Fill(fEv->GetRecoPixelData(fHitted[i])->GetPhi(),
            fEv->GetRecoPixelData(fHitted[i])->GetTheta());
    }
    hithisto->GetXaxis()->SetTitle("#varphi (rad)");
    hithisto->GetYaxis()->SetTitle("#theta (rad)");
    hithisto->GetZaxis()->SetTitle("GTU");
    hithisto2->GetXaxis()->SetTitle("#varphi (rad)");
    hithisto2->GetYaxis()->SetTitle("#theta (rad)");
    
    if (!fClusters.size()) {
        hithisto->Write();
        histo->Write();
        delete hithisto;
        delete histo;
        hithisto2->Write();
        histo2->Write();
        delete hithisto2;
        delete histo2;
        return;
    }
    
    for(Int_t i=0; i<(Int_t)fClusters.size(); i++) {
        EusoCluster *clu = fClusters[i];
        for( Int_t j=0; j<clu->GetNumPoints(); j++) {
            RecoPixelData *pix = fEv->GetRecoPixelData(clu->GetPixelId(j));
            clhisto->Fill( pix->GetPhi(), pix->GetTheta(), pix->GetGtu() );
            clhisto2->Fill( pix->GetPhi(), pix->GetTheta());
        }
    }
    clhisto->GetXaxis()->SetTitle("#varphi (rad)");
    clhisto->GetYaxis()->SetTitle("#theta (rad)");
    clhisto->GetZaxis()->SetTitle("GTU");
    clhisto2->GetXaxis()->SetTitle("#varphi (rad)");
    clhisto2->GetYaxis()->SetTitle("#theta (rad)");
    
    Int_t numsel = fSelectedPixels.size();
    Double_t th[numsel]; Double_t phi[numsel];
    Double_t thline[numsel];
    for( Int_t i(0); i<numsel; i++ ) {
        th[i] = fEv->GetRecoPixelData(fSelectedPixels[i])->GetTheta();
        phi[i] = fEv->GetRecoPixelData(fSelectedPixels[i])->GetPhi();
        selhisto2->Fill(phi[i], th[i]);
        selhisto->Fill(phi[i], th[i], fEv->GetRecoPixelData(fSelectedPixels[i])->GetGtu());
    }
    selhisto2->GetXaxis()->SetTitle("#varphi (rad)");
    selhisto2->GetYaxis()->SetTitle("#theta (rad)");
    selhisto->GetXaxis()->SetTitle("#varphi (rad)");
    selhisto->GetYaxis()->SetTitle("#theta (rad)");
    selhisto->GetZaxis()->SetTitle("GTU");
    
    TGraph *selpoints = new TGraph(numsel,phi,th);
    TGraph *mainline = new TGraph(numsel,phi,thline);
    TCanvas *cv = new TCanvas(); cv->Divide(1,1); cv->cd(1);
    selpoints->Draw("AP");
    mainline->Draw("PLsame");
    cv->Write(); delete cv;
    delete selpoints;

    TCanvas *cv1 = new TCanvas(); cv1->Divide(1,2); cv1->cd(1);
    selhisto2->Draw();
    cv1->cd(2); selhisto->Draw();
    cv1->Write(); delete cv1;

    clhisto->Write();
    hithisto->Write();
    histo->Write();
    delete clhisto;
    delete hithisto;
    delete histo;
    clhisto2->Write();
    hithisto2->Write();
    histo2->Write();
    delete clhisto2;
    delete hithisto2;
    delete histo2;
}
