
// $Id: HoughModule.cc
// Author: Elena Taddei   

/*****************************************************************************
 * ESAF: Euso Simulation and Analysis Framework                              *
 *                                                                           *
 *  Id: HoughModule                                                          *
 *                                                                           *
 *****************************************************************************/

//_____________________________________________________________________________
//
// HoughModule
//
// <extensive class description>
//
//   Config file parameters
//   ======================
//
//   <parameter name>: <parameter description>
//   -Valid options: <available options>
//

#include "HoughModule.hh"
#include "RecoEvent.hh"
#include "RecoRootEvent.hh"
#include "RecoGlobalData.hh"
#include "RecoCellInfo.hh"
#include "TDirectory.h"
#include "TMath.h"
#include "Config.hh"
#include "RecoPixelData.hh"
#include "EusoCluster.hh"
#include "MedianFit.hh"
#include "HoughFit.hh"
#include "ERunParameters.hh"
#include "EDetector.hh"
#include "EConst.hh"

#include <TH3F.h>
#include <TH2F.h>
#include <TH1F.h>
#include <TGraph.h>
#include <TCanvas.h>

using namespace TMath;
using namespace EConst;
using namespace sou;

ClassImp(HoughModule)

// FIXME memory leaks to be fixed asap. D.Naumov

//_____________________________________________________________________________
HoughModule::HoughModule():RecoModule("HoughTransform") {
    //
    // ctor
    //
}

//_____________________________________________________________________________
HoughModule::~HoughModule() {
    //
    // dtor
    //
}

//_____________________________________________________________________________
Bool_t HoughModule::Init() {
    //
    // init
    //
    
    
    fEv = (RecoEvent*)NULL;
    
    fNumPointsMinimum = (Int_t)Conf()->GetNum("HoughModule.fNumPointsMinimum");
    fSignalToNoise    = (Double_t)Conf()->GetNum("HoughModule.fSignalToNoise");
    fDebugPlots       = Conf()->GetBool("HoughModule.fDebugPlots");
    fProcedure        = Conf()->GetStr("HoughModule.fProcedure");
    Msg(EsafMsg::Info) << "Initializing... Procedure: \""<< fProcedure.c_str() << "\"" << MsgDispatch;
    
    fListObj = new TList;

    return kTRUE;
}

//_____________________________________________________________________________
Bool_t HoughModule::PreProcess() {
    //
    // pre-process
    //
	
    fTotalNumPoints = 0;
    fNumPointsAll = 0;
    fClustersWritten = 0;
    
    fPreviousPattReco.clear();
    fHittedFinal.clear();
    fHittedFinalMacrocell.clear();
    fHittedSelHough.clear();
    fHittedSel2Hough.clear();
    fHittedAll.clear();
    fIdMacrocellsSelected.clear();
    fHittedHoughMerged.clear();
       
    fClusters.clear();
        
    return kTRUE;
}

// process
//_____________________________________________________________________________
Bool_t HoughModule::Process( RecoEvent* ev ) {
	
    fEv = ev;
    fTotalNumPoints = fEv->GetHeader().GetNumActiveFee();
    fIdEv = fEv->GetHeader().GetNum();
    fGtuLength = fEv->GetHeader().GetGtuLength();
    
    // check if any previous pattern recognition
    fIsAlreadyPattReco = kFALSE;
    RecoGlobalData* gdPattReco = fEv->FindGlobalData("PatternRecognition");
    if ( gdPattReco ) {
        fIsAlreadyPattReco = kTRUE;
        fPreviousPattReco = *(vector<Int_t>*) gdPattReco->GetObj("SelectedPixels");
        fTotalNumPoints = fPreviousPattReco.size();
    }	
    
    if ( fTotalNumPoints <= fNumPointsMinimum ) {       
        Msg(EsafMsg::Info) << "Not enough points in event! Exit from this event." << MsgDispatch;
        return kFALSE;
    }else{
        Msg(EsafMsg::Debug) << "fTotalNumPoints = " << fTotalNumPoints << MsgDispatch;
    }
    
    fMainMacroCellId = fEv->GetMainMacroCell(1);
    if ( fEv->GetRecoCellInfo(fMainMacroCellId) == NULL )
        Msg(EsafMsg::Info) << "Main macrocell does not correspond to any macrocell hit." << MsgDispatch;
    
      
    
    Bool_t ProcessOnlySignal = Config::Get()->GetCF("Reco","RootInputModule")->GetBool("RootInputModule.ProcessOnlySignal");
    Bool_t FilterScatteredCherenkov = Config::Get()->GetCF("Reco","RootInputModule")->GetBool("RootInputModule.FilterScatteredCherenkov");

    if (ProcessOnlySignal)  {
        for( Int_t i=0; i<fTotalNumPoints; i++ ) {
            RecoPixelData *pix;
            if (!fIsAlreadyPattReco) pix = (RecoPixelData*)fEv->GetRecoPixelData(i);
            else pix = (RecoPixelData*)fEv->GetRecoPixelData(fPreviousPattReco[i]);
            if (FilterScatteredCherenkov) {
                 pix->SetCountsSignal(pix->GetCountsSignal() - pix->GetCountsCherScatt());
                 pix->SetCounts(pix->GetCounts() - pix->GetCountsCherScatt());
             }
             Int_t signal_counts = pix->GetCountsSignal();
             Int_t noise_counts  = pix->GetCountsNoise();
             Double_t ratio (-1);
             if      (noise_counts > 0   )  ratio = 1.e0*signal_counts/Sqrt(noise_counts);
             else if ( signal_counts > 0 )  ratio = 1.e10;
             if (ratio >= fSignalToNoise) fHittedFinal.push_back(i);
        }
        fFinalNumHitsMinimum = 1;
        WriteSingleCluster();
        return kTRUE;
    }
    
    // create a new directory only if debug plots are enabled
    if (fDebugPlots) {
        gDirectory->cd("/");
        string pathnameevent = Form("HoughModule%d", fIdEv);
        TDirectory *path =  new TDirectory(pathnameevent.c_str(),pathnameevent.c_str());
        path->cd();
    }

    Bool_t ret = kFALSE;
    // check if clustering exist
    if (fIsAlreadyPattReco) ret = ProcessHoughAll();
    else if( fProcedure == "single" )	    ret = ProcessHoughSingleMacrocell();
    else if( fProcedure == "velocity" )	ret = ProcessVelocity();
    else if( fProcedure == "all" )	    ret = ProcessHoughAll();
    else Msg(EsafMsg::Panic) << "Wrong procedure in HoughModule.fProcedure. Can be single, velocity or all." << MsgDispatch;
    
    if(!ret){
        Msg(EsafMsg::Info) << "Problems ret=0.. " << MsgDispatch;
    	return kFALSE;
    } else {
    	if(fHittedFinal.size()<10 || fHittedFinal.size()>100000)	return kFALSE;
	    WriteSingleCluster();
    	return kTRUE;
    } 
    return kFALSE; 
    
}

//_____________________________________________________________________________
Bool_t HoughModule::ProcessHoughSingleMacrocell(){
    //
    // Do Hough Transform for each macrocell hitted and trying to select 
    // which are hitted by real signal
    //
	
    Msg(EsafMsg::Info) << " ...ProcessHoughSingleMacrocell()" << MsgDispatch;
    
    Int_t numcountsmin = (Int_t)Conf()->GetNum("HoughModule.fMinCountsHoughSingleMacrocell");
	
    RecoEventHeader header = fEv->GetHeader();
    ERunParameters *runpars = header.GetRunPars();
	
    map< Int_t,vector<Int_t> > mapIdfHittedSingleCell;
    for(Int_t i=0;i<200;i++){
    	RecoCellInfo *info = fEv->GetRecoCellInfo(i);
    	if( info != NULL){
    		vector<Int_t> empty;
    		pair< Int_t,vector<Int_t> > p(i,empty);
    		mapIdfHittedSingleCell.insert(p);
    	}
    }
    Msg(EsafMsg::Info) << "Number of macrocells hitted = " << mapIdfHittedSingleCell.size() << MsgDispatch;

    map< Int_t,vector<Int_t> >::iterator it;
    for( Int_t i=0; i<fTotalNumPoints; i++ ) {
        Int_t threshold(0);
        if ( numcountsmin >=0 ) threshold = numcountsmin;
        else {
            Int_t uid = fEv->GetRecoPixelData(i)->GetPixelId();
            Double_t rate = fEv->GetHeader().GetRunPars()->GetNightGlowRateByUId(uid)*fGtuLength/microsecond; 
            threshold = (Int_t) Ceil(rate + Abs(numcountsmin)*Sqrt(rate));
        }
	    if ( fEv->GetRecoPixelData(i)->GetCounts() >= threshold ){
	        Int_t cellid=fEv->GetRecoPixelData(i)->GetMacroCellId();
	        it = mapIdfHittedSingleCell.find(cellid);
	        it->second.push_back(i);
	    }
    }
   
    if (numcountsmin >= 0) {
        Msg(EsafMsg::Info) << "Selected pixels with counts > " << numcountsmin << MsgDispatch;
    } else {
        Msg(EsafMsg::Info) << "Selectet pixels with counts over " << -numcountsmin << " sigma from background mean" << MsgDispatch;
    }
    for( it = mapIdfHittedSingleCell.begin(); it != mapIdfHittedSingleCell.end(); it++ ) {
        Msg(EsafMsg::Info)<<"\n\t\t---->macrocellid = "<<it->first<<"\t with "<<(it->second).size()<<" points "<<MsgDispatch;
        if((it->second).size()<10) continue;
 
        Int_t idmacrocell=it->first;
        vector<Int_t> fHittedmc=it->second;
        vector<Double_t> xmc;
        vector<Double_t> ymc;
       	vector<Double_t> tmc;
	    vector<Int_t> cmc;
		
        TH1F *h1 = 0;
	    if (fDebugPlots ) fListObj->Add(h1 = new TH1F("h1","h1",10,0,10));
        for(Int_t i=0;i<(Int_t)fHittedmc.size();i++){
    	    if (fDebugPlots) h1->Fill(fEv->GetRecoPixelData(fHittedmc[i])->GetCounts());
    	    Int_t pixid=fEv->GetRecoPixelData(fHittedmc[i])->GetPixelId();
		    TVector3 dummy2 = runpars->PixelCenter(pixid);
    	    xmc.push_back(dummy2.x()/m);//m
    	    ymc.push_back(dummy2.y()/m);//m
    	    Int_t tmcgtu = fEv->GetRecoPixelData(fHittedmc[i])->GetGtu();
    	    tmc.push_back(tmcgtu*fGtuLength /microsecond*0.01); // ????? (RP)
    	    cmc.push_back(fEv->GetRecoPixelData(fHittedmc[i])->GetCounts());
        }
    	
        if (fDebugPlots) {
            string pathname = Form("/HoughModule%d/", fIdEv);
            gDirectory->cd(pathname.c_str());
        	string pathnames = Form("SingleHoughMacrocell%d", idmacrocell);
            TDirectory *path = new TDirectory(pathnames.c_str(),pathnames.c_str());
            path->cd();
            h1->Draw();
        	h1->Write(); 
        }
	    SafeDelete(h1);
        
    	
        Bool_t possiblesignal = HoughTransform(xmc,ymc,tmc,cmc);
    	
        if ( possiblesignal ) {
    	    Msg(EsafMsg::Info) << "Macrocell " << idmacrocell << " probably HAS signal!!!" << MsgDispatch;
    	    fIdMacrocellsSelected.push_back(idmacrocell);
        	fHittedFinalMacrocell.insert(fHittedFinalMacrocell.end(),fHittedmc.begin(),fHittedmc.end());
        } else {
    	    Msg(EsafMsg::Info) << "Macrocell " << idmacrocell << " probably DOESN'T HAVE a signal,";
        	if ( idmacrocell==fMainMacroCellId ){
        		Msg(EsafMsg::Info) << " but is added because is the main!" << MsgDispatch;
    	    	fHittedFinalMacrocell.insert(fHittedFinalMacrocell.end(),fHittedmc.begin(),fHittedmc.end());
    		    fIdMacrocellsSelected.push_back(idmacrocell);
        	} else {
    			Msg(EsafMsg::Info) << " discarded!" << MsgDispatch;
    		}
    	}
    	
    	xmc.clear();
    	ymc.clear();
    	tmc.clear();
    	cmc.clear();
    }
    
    if (fIdMacrocellsSelected.size()<1 || fHittedFinalMacrocell.size()<10)	return kFALSE;
    Msg(EsafMsg::Info) << "Re-process " << fIdMacrocellsSelected.size() << " selected macrocells by ProcessHoughSingleMacrocell()" << MsgDispatch;  
    Bool_t ret = ProcessHoughAll();

    return ret;    
}

//_____________________________________________________________________________
Bool_t HoughModule::ProcessVelocity(){
    //
    // Do histograms of velocities for each macrocell hitted and trying to select 
    // which are hitted by the real signal
    //
	
    Msg(EsafMsg::Info) << " ...ProcessVelocity()" << MsgDispatch;
    
    Int_t numcountsmin = (Int_t)Conf()->GetNum("HoughModule.fMinCountsVelocity");
	
    RecoEventHeader header = fEv->GetHeader();
    ERunParameters *runpars = header.GetRunPars();
	
    map< Int_t,vector<Int_t> > mapIdfHittedSingleCell;
    for(Int_t i=0;i<200;i++) {
    	RecoCellInfo *info = fEv->GetRecoCellInfo(i);
    	if( info != NULL) {
    		vector<Int_t> empty;
    		pair< Int_t,vector<Int_t> > p(i,empty);
    		mapIdfHittedSingleCell.insert(p);
    	}
    }
    Msg(EsafMsg::Info) << "Number of macrocells hitted = " << mapIdfHittedSingleCell.size() << MsgDispatch;
    
    map< Int_t,vector<Int_t> >::iterator it;
    for( Int_t i=0; i<fTotalNumPoints; i++ ) {
        Int_t threshold(0);
        if ( numcountsmin >=0 ) threshold = numcountsmin;
        else {
            Int_t uid = fEv->GetRecoPixelData(i)->GetPixelId();
            Double_t rate = fEv->GetHeader().GetRunPars()->GetNightGlowRateByUId(uid)*fGtuLength/microsecond; 
            threshold = (Int_t) Ceil(rate + Abs(numcountsmin)*Sqrt(rate));
        }
        if( fEv->GetRecoPixelData(i)->GetCounts() >= threshold  ){
	        Int_t cellid=fEv->GetRecoPixelData(i)->GetMacroCellId();
    	    it = mapIdfHittedSingleCell.find(cellid);
	        it->second.push_back(i);
	    }
    }
    
    Int_t binv = 20;
    if (numcountsmin >= 0) {
        Msg(EsafMsg::Info) << "Selected pixels with counts > " << numcountsmin << MsgDispatch;
    } else {
        Msg(EsafMsg::Info) << "Selectet pixels with counts over " << -numcountsmin << " sigma from background mean" << MsgDispatch;
    }
    for( it = mapIdfHittedSingleCell.begin(); it != mapIdfHittedSingleCell.end(); it++ ) {
    	Int_t entriesTH2(0);    	
    	Msg(EsafMsg::Info) << "macrocellid = " << it->first << "\t with " << (it->second).size() << " points" << MsgDispatch;
    	
    	if((it->second).size()<10) continue;
    	
    	TH2F *hsxt(0), *hsyt(0), *s(0), *sxy(0);
        TH1F *sx(0), *sy(0);
        if (fDebugPlots) {
            fListObj->Add(hsxt = new TH2F("hsxt","hsxt;t(mus);x(mm)",1000,0,200,1000,-1000,1000));
        	fListObj->Add(hsyt = new TH2F("hsyt","hsyt;t(mus);y(mm)",1000,0,200,1000,-1000,1000));
        }
        fListObj->Add(s = new TH2F("s","s;velocita;coeffang",200,0,2,200,-4,4)); 
        fListObj->Add(sx = new TH1F("sx","sx;velocita",binv,-2,2)); 
       	fListObj->Add(sy = new TH1F("sy","sy;velocita",binv,-2,2)); 
       	fListObj->Add(sxy = new TH2F("sxy","velocita",binv,-2,2,binv,-2,2)); 
    	Int_t idmacrocell = it->first;
    	vector<Int_t> fHittedmc = it->second;
    	
    	for( Int_t i=0;i<(Int_t)fHittedmc.size();i++ ) {
    	    Int_t pixid = fEv->GetRecoPixelData(fHittedmc[i])->GetPixelId();
	        TVector3 dummy2 = runpars->PixelCenter(pixid);//mm
    	    Int_t tmcgtu = fEv->GetRecoPixelData(fHittedmc[i])->GetGtu();
    	    Int_t counts = fEv->GetRecoPixelData(fHittedmc[i])->GetCounts();
    	    if (fDebugPlots) {
                hsxt->Fill(tmcgtu*fGtuLength/microsecond,dummy2.x(),counts);
        	    hsyt->Fill(tmcgtu*fGtuLength/microsecond,dummy2.y(),counts);
            }
    	    Int_t pointstocouple = fHittedmc.size();
    	    for( Int_t j=i+1;j<pointstocouple;j++ ) {
    	        Int_t pixtocouple = fEv->GetRecoPixelData(fHittedmc[j])->GetPixelId();
		        TVector3 dummycouple = runpars->PixelCenter(pixtocouple);
		        Int_t tmccouple = fEv->GetRecoPixelData(fHittedmc[j])->GetGtu();
    		    if (Abs(dummy2.x()-dummycouple.x())>0 && Abs(dummy2.y()-dummycouple.y())>0 && tmccouple-tmcgtu>0 
                        && tmccouple-tmcgtu<20 && Abs(dummy2.x()-dummycouple.x())<100 && Abs(dummy2.y()-dummycouple.y())<100 ) {
		            Int_t countproduct = (Int_t)Sqrt(1.e0*counts*fEv->GetRecoPixelData(fHittedmc[j])->GetCounts());
		            Double_t m = (dummy2.y()-dummycouple.y())/(dummy2.x()-dummycouple.x());
	                Double_t v = Sqrt((dummy2.y()-dummycouple.y())*(dummy2.y()-dummycouple.y())+(dummy2.x()-dummycouple.x())
                            *(dummy2.x()-dummycouple.x()))/((tmccouple-tmcgtu)*fGtuLength/microsecond);// mm/us
    			    Double_t vx = (dummy2.x()-dummycouple.x())/((tmcgtu-tmccouple)*fGtuLength/microsecond);
	    		    Double_t vy = (dummy2.y()-dummycouple.y())/((tmcgtu-tmccouple)*fGtuLength/microsecond);
		    	    entriesTH2++;
                    s->Fill(v,m,countproduct);
        			sx->Fill(vx,countproduct);
    	        	sy->Fill(vy,countproduct);
	    	        sxy->Fill(vx,vy,countproduct);
		        }
    	    }
    	}    
    
        Int_t binmaxv = sxy->GetMaximumBin();
        Int_t nxv = (sxy->GetXaxis())->GetNbins()+2;
        Int_t binxmaxv = binmaxv % nxv;
        Int_t binymaxv = binmaxv / nxv;
        Float_t vxmax[1], vymax[1];
        vxmax[0] = (sxy->GetXaxis())->GetBinCenter(binxmaxv);
        vymax[0] = (sxy->GetYaxis())->GetBinCenter(binymaxv);
        TGraph *gvmax=new TGraph(1,vxmax,vymax);
        gvmax->SetMarkerColor(2); gvmax->SetMarkerStyle(29); gvmax->SetMarkerSize(1.6);
        //Float_t contmaxv=sxy->GetBinContent(binxmaxv,binymaxv);
        Stat_t countx(0),county(0);
        for(Int_t bin=1;bin<binv;bin++){
    	    countx += sx->GetBinContent(bin);
        	county += sy->GetBinContent(bin);
        }
        Float_t countmaxx = sx->GetMaximum();
        Float_t countmaxy = sy->GetMaximum();
        Float_t countmeanx = (Float_t)((Float_t)countx/binv);
        Float_t countmeany = (Float_t)((Float_t)county/binv);
        Float_t ms = 20.;
        /*cout<<"countmeanx="<<countmeanx<<"\tcontmaxx="<<countmaxx<<"\tcontmaxx-countmeanx="<<countmaxx-countmeanx<<"\tsig="<<ms*TMath::Sqrt(countmeanx)<<endl;
    	cout<<"countmeany="<<countmeany<<"\tcontmaxy="<<countmaxy<<"\tcontmaxy-countmeany="<<countmaxy-countmeany<<"\tsig="<<ms*TMath::Sqrt(countmeany)<<endl;*/
        string a = "a"; string b = "b";
        string pathname = Form("/HoughModule%d/", fIdEv);
        if (fDebugPlots) gDirectory->cd(pathname.c_str());
        string pathnames;
        if( ((countmaxx-countmeanx) > ms*Sqrt(countmeanx) || (countmaxy-countmeany) > ms*Sqrt(countmeany)) && sx->GetEntries()>10 ) {
            fHittedFinalMacrocell.insert(fHittedFinalMacrocell.end(),fHittedmc.begin(),fHittedmc.end());
    	    fIdMacrocellsSelected.push_back(idmacrocell);
            Msg(EsafMsg::Info) << "Macrocell " << idmacrocell << " probably HAS signal!!!" << MsgDispatch;
            if (idmacrocell==fMainMacroCellId) {
    	        pathnames = Form("VelocityMacrocellmain%d", idmacrocell);
            } else {
                pathnames = Form("VelocityMacrocell%d", idmacrocell);
    	    }
        } else{
            Msg(EsafMsg::Info) << "Macrocell " << idmacrocell << " probably DOESN'T HAVE a signal,";
	    if (idmacrocell==fMainMacroCellId) {
	        pathnames = Form("VelocityMacrocellmain%d", idmacrocell);
    	        Msg(EsafMsg::Info) <<" but is added because is the main!"<< MsgDispatch;
    	        fHittedFinalMacrocell.insert(fHittedFinalMacrocell.end(),fHittedmc.begin(),fHittedmc.end());
    	        fIdMacrocellsSelected.push_back(idmacrocell);
            } else {
    	        pathnames = Form("VelocityMacrocelldiscarded%d", idmacrocell);
       	        Msg(EsafMsg::Info) <<" discarded!"<< MsgDispatch;
    	    }
        }
		
        s->SetTitle(pathnames.c_str());
        if (fDebugPlots) {
            TDirectory *path = new TDirectory(pathnames.c_str(),pathnames.c_str());
            path->cd();
            TCanvas *ac;
            fListObj->Add(ac = new TCanvas(a.c_str(),a.c_str(),1)); ac->Divide(2,2);
            ac->cd(1);s->Draw();    ac->cd(2);s->Draw("LEGO");
            ac->cd(3);sx->Draw(); 	ac->cd(4);sy->Draw(); 
            ac->Write();
            SafeDelete(ac);
            TCanvas *bc;
            fListObj->Add(bc = new TCanvas(b.c_str(),b.c_str(),1)); bc->Divide(2,2);
            bc->cd(1);hsxt->Draw(); 
            bc->cd(2);hsyt->Draw(); 
            bc->cd(3);sxy->Draw();gvmax->Draw("same"); 
            bc->cd(4);sxy->Draw("LEGO"); 
            bc->Write();
            SafeDelete(bc);
        }
        SafeDelete(s);	SafeDelete(sx);	SafeDelete(sy);
        SafeDelete(hsxt);	SafeDelete(hsyt);
        SafeDelete(sxy); SafeDelete(gvmax);
    }
    
    if (fIdMacrocellsSelected.size()<1 || fHittedFinalMacrocell.size()<10)	return kFALSE;
	
    Msg(EsafMsg::Info) << "Re-process " << fIdMacrocellsSelected.size() << " selected macrocells by ProcessVelocity()" << MsgDispatch;  
    Bool_t ret = ProcessHoughAll();
    
    return ret;

}

//_____________________________________________________________________________
Bool_t HoughModule::ProcessHoughAll(){
    //
    // Doing Hough Transform to all macrocells selected by previous 
    // methods or by euso trigger
    //
	
    Msg(EsafMsg::Info) << " ...ProcessHoughAll()" << MsgDispatch;
	
    if (fDebugPlots) {
        string pathname = Form("/HoughModule%d/", fIdEv);
        gDirectory->cd(pathname.c_str());
        string pathnametotal = Form("HoughAll");
        TDirectory *pathtotal = new TDirectory(pathnametotal.c_str(),pathnametotal.c_str());
        pathtotal->cd();
    }
	
    RecoEventHeader header = fEv->GetHeader();
    ERunParameters *runpars = header.GetRunPars();
        
        fNumHitsMinimum = (Int_t)Conf()->GetNum("HoughModule.fMinCountsHoughAll");
    if ( fIsAlreadyPattReco && fNumHitsMinimum < 0 ){
        Msg(EsafMsg::Info) << "Using relative threshold at " << Abs(fNumHitsMinimum) << " sigma from the mean background value" << MsgDispatch;  
    }else if (!fIsAlreadyPattReco ) {
     	Msg(EsafMsg::Info) << "Let's try beginning with MinCountsHoughAll = " << fNumHitsMinimum << MsgDispatch;
    }
    
    Int_t regr = 0;
    
    Int_t gtumin = 10000;
    Int_t gtumax = -1;
    Int_t deltagtu = 0;
    Int_t mean_cmin = 0; // mean number of hits minimum in pixels (differs from fNumHitsMinimum using relative threshold)
    
    fNumPointsAll = 0;
    fHittedAll.clear();
    if(!fIsAlreadyPattReco && !(fHittedFinalMacrocell.size())){
	    Msg(EsafMsg::Info) << "processing all macrocells!" << MsgDispatch;
        mean_cmin = 0;
	    for( Int_t i=0; i<fTotalNumPoints; i++ ) {
        	if (fEv->GetRecoPixelData(i)->GetGtu()<gtumin) gtumin=fEv->GetRecoPixelData(i)->GetGtu();
    	    if (fEv->GetRecoPixelData(i)->GetGtu()>gtumax) gtumax=fEv->GetRecoPixelData(i)->GetGtu();
            Int_t countsmin = 0;
            // compute relative threshold if requested
            if ( fNumHitsMinimum >= 0 ) {
                countsmin = fNumHitsMinimum;
                mean_cmin = fNumHitsMinimum;
            } else {
                Int_t uid = fEv->GetRecoPixelData(i)->GetPixelId();
                Double_t rate = fEv->GetHeader().GetRunPars()->GetNightGlowRateByUId(uid) * fGtuLength/microsecond;
                countsmin = (Int_t)Ceil( rate + Abs(fNumHitsMinimum)*Sqrt(rate) );
            }
    	    if (fEv->GetRecoPixelData(i)->GetCounts() >= countsmin) {
                fHittedAll.push_back(i);
                if ( fNumHitsMinimum < 0 ) mean_cmin += countsmin;
            }
        }
        if ( !fHittedAll.size() ) {
            Msg(EsafMsg::Warning) << "No pixels over threshold. Exit from this event." << MsgDispatch;
            fHittedAll.clear();
            fMainMacroCellId = 0;
            fNumPointsAll = 0;
            return kFALSE;
        }
        if ( fNumHitsMinimum < 0 ) mean_cmin /= (Int_t)fHittedAll.size();
    } else if (!fIsAlreadyPattReco){
        Msg(EsafMsg::Info)<<"processing fHittedFinalMacrocell.size()="<<fHittedFinalMacrocell.size()<<MsgDispatch;
        mean_cmin = 0;
        Int_t add = 0;
        for( Int_t j=0; j<(Int_t)(fHittedFinalMacrocell.size()); j++ ){
    	    Int_t i = fHittedFinalMacrocell[j];
    	    if (fEv->GetRecoPixelData(i)->GetGtu()<gtumin) gtumin=fEv->GetRecoPixelData(i)->GetGtu();
    	    if (fEv->GetRecoPixelData(i)->GetGtu()>gtumax) gtumax=fEv->GetRecoPixelData(i)->GetGtu();
            Int_t countsmin = 0;
            // compute relative threshold if requested
            if ( fNumHitsMinimum>=0 ) {
                countsmin = fNumHitsMinimum;
                mean_cmin = fNumHitsMinimum;
            } else {
                Int_t uid = fEv->GetRecoPixelData(i)->GetPixelId();
                Double_t rate = fEv->GetHeader().GetRunPars()->GetNightGlowRateByUId(uid) * fGtuLength/microsecond;
                countsmin = (Int_t)Ceil( rate + Abs(fNumHitsMinimum)*Sqrt(rate) );
                //mean_cmin += countsmin;
            }
    	    if (fEv->GetRecoPixelData(i)->GetCounts() >= countsmin) {
                fHittedAll.push_back(i);
                add++;
                if ( fNumHitsMinimum < 0 ) mean_cmin += countsmin;
            }
        }
        if ( fNumHitsMinimum < 0 && add ) mean_cmin /= add;
    } else if (fIsAlreadyPattReco) {
        fHittedAll = fPreviousPattReco;
        fNumHitsMinimum = fEv->FindGlobalData("PatternRecognition")->GetInt("NumPointsMinimum");
        mean_cmin = fEv->FindGlobalData("PatternRecognition")->GetInt("MeanThreshold");
        for (Int_t i(0); i<(Int_t)fHittedAll.size(); i++) {
            Int_t j = fHittedAll[i];
    	    if (fEv->GetRecoPixelData(j)->GetGtu()<gtumin) gtumin=fEv->GetRecoPixelData(j)->GetGtu();
    	    if (fEv->GetRecoPixelData(j)->GetGtu()>gtumax) gtumax=fEv->GetRecoPixelData(j)->GetGtu();
        }
    }
        
    
    fNumPointsAll = fHittedAll.size();
        
    fTmin = gtumin*fGtuLength/microsecond*0.01; //???? (RP)
    fTmax = gtumax*fGtuLength/microsecond*0.01; //???? (RP)
    deltagtu = gtumax - gtumin;
   
    Msg(EsafMsg::Info) << "Num. of pixels over threshold= " << fNumPointsAll << " over fTotalNumPoints="  << fTotalNumPoints << MsgDispatch;
    if (fNumPointsAll<10 || deltagtu<5) {
        if( fNumPointsAll<10 )   Msg(EsafMsg::Warning) << "Number of points "<< fNumPointsAll << " is less than 10, Exit from this event." << MsgDispatch;
        if ( deltagtu<5 )	Msg(EsafMsg::Warning) << " deltagtu = "<< deltagtu << " is less than 5, Exit from this event." << MsgDispatch;
        fHittedAll.clear();
        fMainMacroCellId = 0;
        fNumPointsAll = 0;
        return kFALSE;
    }
    
    vector<Double_t> x;
    vector<Double_t> y;
    vector<Double_t> t;
    vector<Int_t> c;
	
    TH2F *single2(0), *single2xt(0), *single2yt(0);
    Int_t bin = 2000;
    if (fDebugPlots) {
        fListObj->Add(single2 = new TH2F("single2","single2;x(m);y(m)",bin,-1,1,bin,-1,1));
    	fListObj->Add(single2xt = new TH2F("single2xt","single2xt;t;x(m)",bin,fTmin,fTmax,bin,-1,1));
	    fListObj->Add(single2yt = new TH2F("single2yt","single2yt;t;y(m)",bin,fTmin,fTmax,bin,-1,1));
    }
	
	for( Int_t i=0; i < fNumPointsAll; i++ ) {
    	Int_t pixid=fEv->GetRecoPixelData(fHittedAll[i])->GetPixelId();
		TVector3 dummy2 = runpars->PixelCenter(pixid); // mm
    	x.push_back(dummy2.x()/m);
    	y.push_back(dummy2.y()/m);
    	t.push_back((Double_t)fEv->GetRecoPixelData(fHittedAll[i])->GetGtu()*fGtuLength/microsecond*0.01); // ??? (RP)
    	c.push_back(fEv->GetRecoPixelData(fHittedAll[i])->GetCounts());
    	if (fDebugPlots) {
            single2->Fill(x[i],y[i],c[i]);
            single2xt->Fill(t[i],x[i],c[i]);
       	    single2yt->Fill(t[i],y[i],c[i]);
        }
    }
    
    if (fDebugPlots) {
        TCanvas *HMinit;
        fListObj->Add(HMinit = new TCanvas("HMinit","HMinit",1));
        HMinit->Divide(3,1);
        HMinit->cd(1);  single2->SetMarkerStyle(6);  single2->Draw();
        HMinit->cd(2);  single2xt->SetMarkerStyle(6);  single2xt->Draw();
        HMinit->cd(3);  single2yt->SetMarkerStyle(6);  single2yt->Draw();
        HMinit->Write();    
        SafeDelete(HMinit);
    }
    SafeDelete(single2xt);
    SafeDelete(single2yt);
    SafeDelete(single2);
    
    HoughTransform(x,y,t,c); 
    
    x.clear();	y.clear();  t.clear();  c.clear();

    if( (mean_cmin - regr) > 3)	regr++;
    
    vector<Double_t> xsel;
    vector<Double_t> ysel;
    vector<Double_t> tsel;
    vector<Int_t> csel;
    
    TH2F *single2sel(0), *single2xtsel(0), *single2ytsel(0);
    if (fDebugPlots) {
        fListObj->Add(single2sel = new TH2F("single2sel","single2sel;x(m);y(m)",bin,fXmin,fXmax,bin,fYmin,fYmax));
        fListObj->Add(single2xtsel = new TH2F("single2xtsel","single2xtsel;t;x(m)",bin,fTmin,fTmax,bin,fXmin,fXmax));
        fListObj->Add(single2ytsel = new TH2F("single2ytsel","single2ytsel;t;y(m)",bin,fTmin,fTmax,bin,fYmin,fYmax));
    }
    for( Int_t i=0; i<fTotalNumPoints; i++ ) {
    	Int_t pixid=fEv->GetRecoPixelData(i)->GetPixelId();
	TVector3 dummy2 = runpars->PixelCenter(pixid);
    	Double_t x = dummy2.x()/m; // meters
    	Double_t y = dummy2.y()/m; // meters
    	Double_t t=(Double_t)fEv->GetRecoPixelData(i)->GetGtu()*fGtuLength/microsecond*0.01; // ???
    	Int_t c=fEv->GetRecoPixelData(i)->GetCounts();
        if ( x>=fXmin && x<=fXmax && y>=fYmin && y<=fYmax && c>=(mean_cmin-regr) ) {
            if (fDebugPlots) {
                single2sel->Fill(x,y,c);
                single2xtsel->Fill(t,x,c);
                single2ytsel->Fill(t,y,c);
            }
            xsel.push_back(x);
            ysel.push_back(y);
            tsel.push_back(t);
            csel.push_back(c);
            fHittedSelHough.push_back(i);
        }
    }
    
    Msg(EsafMsg::Info) << "first selected zone: pixels with at least " << mean_cmin-regr << " hits = " << fHittedSelHough.size() << MsgDispatch;
    if(fHittedSelHough.size()<10 && !fIsAlreadyPattReco){
    	Msg(EsafMsg::Info) << "too few! exit from this event" <<MsgDispatch;
    	return kFALSE;
    } else if (fHittedSelHough.size()<10 && fIsAlreadyPattReco) {
        fHittedFinal = fHittedFinal;
        fFinalNumHitsMinimum = mean_cmin - regr;
        return kTRUE;
    } 
    
    if (fDebugPlots) {
        TCanvas *HMsel;
        fListObj->Add(HMsel = new TCanvas("HMsel","HMsel",1));
        HMsel->Divide(3,1);
        HMsel->cd(1);  single2sel->SetMarkerStyle(6);  single2sel->Draw();
        HMsel->cd(2);  single2xtsel->SetMarkerStyle(6);  single2xtsel->Draw();
        HMsel->cd(3);  single2ytsel->SetMarkerStyle(6);  single2ytsel->Draw();
        HMsel->Write();
        SafeDelete(HMsel);
    }
    SafeDelete(single2xtsel);
    SafeDelete(single2ytsel);
    SafeDelete(single2sel);
    
    HoughTransform(xsel,ysel,tsel,csel); 
    
    xsel.clear();  ysel.clear();  tsel.clear();  csel.clear();
    if(mean_cmin-regr > 2)	regr++;
    vector<Double_t> xselsel;
    vector<Double_t> yselsel;
    vector<Double_t> tselsel;
    vector<Int_t> cselsel;
   
    TH2F *single2selsel(0), *single2xtselsel(0), *single2ytselsel(0);
    if (fDebugPlots) { 
        fListObj->Add(single2selsel = new TH2F("single2selsel","single2selsel;x(m);y(m)",bin,fXmin,fXmax,bin,fYmin,fYmax));
        fListObj->Add(single2xtselsel = new TH2F("single2xtselsel","single2xtselsel;t;x(m)",bin,fTmin,fTmax,bin,fXmin,fXmax));
        fListObj->Add(single2ytselsel = new TH2F("single2ytselsel","single2ytselsel;t;y(m)",bin,fTmin,fTmax,bin,fYmin,fYmax));
    }
    for( Int_t i=0; i<fTotalNumPoints; i++ ) {
    	Int_t pixid=fEv->GetRecoPixelData(i)->GetPixelId();
	    TVector3 dummy2 = runpars->PixelCenter(pixid);
    	Double_t x = dummy2.x()/m; //meters
    	Double_t y = dummy2.y()/m; //meters 
    	Double_t t = (Double_t)fEv->GetRecoPixelData(i)->GetGtu()*fGtuLength/microsecond*0.01;
    	Int_t c=fEv->GetRecoPixelData(i)->GetCounts();
        if ( x>=fXmin && x<=fXmax && y>=fYmin && y<=fYmax && c>=(mean_cmin-regr) ) {    
            if (fDebugPlots) {
                single2selsel->Fill(x,y,c);
            	single2xtselsel->Fill(t,x,c);
            	single2ytselsel->Fill(t,y,c);
            }
        	xselsel.push_back(x);
        	yselsel.push_back(y);
        	tselsel.push_back(t);
        	cselsel.push_back(c);
        	fHittedSel2Hough.push_back(i);
        }
    }
    Msg(EsafMsg::Info) << "second selected zone: pixels with at least " << mean_cmin-regr << " hits = " << fHittedSel2Hough.size()<< MsgDispatch;
    if(fHittedSel2Hough.size()<10 && !fIsAlreadyPattReco){
    	Msg(EsafMsg::Info) << "too few! exit from this event" <<MsgDispatch;
    	return kFALSE;
    } else if (fHittedSel2Hough.size()<10 && fIsAlreadyPattReco) {
        fHittedFinal = fHittedSelHough;
        fFinalNumHitsMinimum = mean_cmin - regr;
        return kTRUE;
    } 
    
    if (fDebugPlots) {
        TCanvas *HMselsel;
        fListObj->Add(HMselsel = new TCanvas("HMselsel","HMselsel",1));
        HMselsel->Divide(3,1);
        HMselsel->cd(1);  single2selsel->SetMarkerStyle(6);  single2selsel->Draw();
        HMselsel->cd(2);  single2xtselsel->SetMarkerStyle(6);  single2xtselsel->Draw();
        HMselsel->cd(3);  single2ytselsel->SetMarkerStyle(6);  single2ytselsel->Draw();
        HMselsel->Write();
        SafeDelete(HMselsel);
    }
    SafeDelete(single2xtselsel);
    SafeDelete(single2ytselsel);
    SafeDelete(single2selsel);
    
    
    HoughTransform(xselsel,yselsel,tselsel,cselsel,fHittedSel2Hough); 
    fHittedFinal = fHittedHoughMerged;
    xselsel.clear();  yselsel.clear();  tselsel.clear();  cselsel.clear();
    
    //now not used
    /*vector<Double_t> xend;
    vector<Double_t> yend;
    vector<Double_t> tend;
    vector<Int_t> cend;*/
    
    TH2F *single2end(0), *single2xtend(0), *single2ytend(0);
    if (fDebugPlots) {
        fListObj->Add(single2end = new TH2F("single2end","single2end;x(m);y(m)",bin,fXmin,fXmax,bin,fYmin,fYmax));
        fListObj->Add(single2xtend = new TH2F("single2xtend","single2xtend;t;x(m)",bin,fTmin,fTmax,bin,fXmin,fXmax));
    	fListObj->Add(single2ytend = new TH2F("single2ytend","single2ytend;t;y(m)",bin,fTmin,fTmax,bin,fYmin,fYmax));
	for( Int_t i=0; i<(Int_t)fHittedHoughMerged.size(); i++ ) {
    	Int_t pixid=fEv->GetRecoPixelData(fHittedHoughMerged[i])->GetPixelId();
		TVector3 dummy2 = runpars->PixelCenter(pixid);
    	Double_t x = dummy2.x()/m; // meters
    	Double_t y = dummy2.y()/m; // meters
    	Double_t t = (Double_t)fEv->GetRecoPixelData(fHittedHoughMerged[i])->GetGtu()*fGtuLength/microsecond*0.01; // ???
    	Int_t c = fEv->GetRecoPixelData(fHittedHoughMerged[i])->GetCounts();
        single2end->Fill(x,y,c);
      	single2xtend->Fill(t,x,c);
       	single2ytend->Fill(t,y,c);
       	/*xend.push_back(x);
       	yend.push_back(y);
       	tend.push_back(t);
       	cend.push_back(c);*/
    }
    }
    Msg(EsafMsg::Info) << "final selected zone: pixels with at least " << mean_cmin-regr << " hits = " << fHittedHoughMerged.size()<< MsgDispatch;
    if(fHittedHoughMerged.size()<10 && !fIsAlreadyPattReco){
    	Msg(EsafMsg::Info) << "too few! exit from this event" <<MsgDispatch;
    	return kFALSE;
    } else if (fHittedHoughMerged.size()<10 && fIsAlreadyPattReco) {
        fHittedFinal = fHittedSel2Hough;
        fFinalNumHitsMinimum = mean_cmin - regr;
        return kTRUE;
    } 
    
    if (fDebugPlots) {
        TCanvas *HMfinal;
        fListObj->Add(HMfinal = new TCanvas("HMfinal","HMfinal",1));
        HMfinal->Divide(3,1);
        HMfinal->cd(1);  single2end->SetMarkerStyle(6);  single2end->Draw();
        HMfinal->cd(2);  single2xtend->SetMarkerStyle(6);  single2xtend->Draw();
        HMfinal->cd(3);  single2ytend->SetMarkerStyle(6);  single2ytend->Draw();
        HMfinal->Write();
        SafeDelete(HMfinal);
    }
    SafeDelete(single2xtend);
    SafeDelete(single2ytend);
    SafeDelete(single2end);
    
    //xend.clear(); yend.clear(); tend.clear(); cend.clear();
    fFinalNumHitsMinimum = mean_cmin - regr;
    return kTRUE;
    
}


//_____________________________________________________________________________
Bool_t HoughModule::HoughTransform(vector<Double_t> x,vector<Double_t> y,vector<Double_t> t,vector<Int_t> c,vector<Int_t> id){
    //
    // Implements the Hough Transform
    // Return true if there is a possible signal
    //

    if( x.size()<10 )	return 0;
	
    Int_t optionselection = 1;
	
    Float_t ex = 0.004;
    Float_t ey = 0.004;
    Float_t et = fGtuLength/microsecond*0.01; // ???
	
    HoughFit *xtFit = new HoughFit(t, x, c, optionselection,et,ex,id);
    //Double_t vx = xtFit->GetSlope();//Double_t wx = xtFit->GetOffset();
    fXmin = xtFit->GetMinY();
    fXmax = xtFit->GetMaxY();
    Double_t tmin1 = xtFit->GetMinX();
    Double_t tmax1 = xtFit->GetMaxX();
    fAbsDevX = xtFit->GetAbsDev();
    Bool_t possiblesignalxt = xtFit->IsSignal();
    fCmaxToCmeanX = xtFit->GetMaxToMedium();
    vector<Int_t> idselx = xtFit->GetIdSel();
    delete xtFit;
	
    HoughFit *ytFit = new HoughFit(t, y, c, optionselection,et,ey,id);
    //Double_t vy = ytFit->GetSlope();//Double_t wy = ytFit->GetOffset();
    fYmin = ytFit->GetMinY();
    fYmax = ytFit->GetMaxY();
    Double_t tmin2 = ytFit->GetMinX();
    Double_t tmax2 = ytFit->GetMaxX();
    fAbsDevY = ytFit->GetAbsDev();
    Bool_t possiblesignalyt = ytFit->IsSignal();
    fCmaxToCmeanY = ytFit->GetMaxToMedium();
    vector<Int_t> idsely = ytFit->GetIdSel();
    delete ytFit;
	
    if(id.size()>0) {
	Int_t sumsel=idsely.size()+idselx.size();
	//cout<<"id.size()="<<id.size()<<"\tidselx.size()="<<idselx.size()<<"\tidsely.size()="<<idsely.size()<<endl;
    vector<Int_t> idselectedmerged(sumsel);
	merge(idselx.begin(), idselx.end(),idsely.begin(), idsely.end(),idselectedmerged.begin());
	vector<Int_t>::iterator diversi=unique(idselectedmerged.begin(),idselectedmerged.end());
	idselectedmerged.erase(diversi,idselectedmerged.end());
	//cout<<"idselectedmerged.size()="<<idselectedmerged.size()<<endl;
	fHittedHoughMerged = idselectedmerged;
    }
	
    //Msg(EsafMsg::Info) <<"possiblesignalxt="<<possiblesignalxt<<"\tpossiblesignalyt="<<possiblesignalyt<<MsgDispatch;
    //Msg(EsafMsg::Info) <<"fCmaxToCmeanX="<<fCmaxToCmeanX<<"\tcmaxtocmedioy="<<fCmaxToCmeanY<<MsgDispatch;
    Bool_t possible = kFALSE;
    if( possiblesignalxt || possiblesignalyt ) possible = kTRUE;
	
    fTmin = (tmin1<tmin2) ? tmin1 : tmin2;
    fTmax = (tmax1<tmax2) ? tmax2 : tmax1;
	
    x.clear();  y.clear();  t.clear();  c.clear();
    
    return possible;
    
}

//_____________________________________________________________________________
Bool_t HoughModule::PostProcess() {
    //
    // post-process
    //
    
    if ( !fTotalNumPoints ) return kTRUE;
   
    TmpMemoryClean();
    
    fEv = (RecoEvent*)0;
    return kTRUE;
    	
}

//____________________________________________________________________________
Bool_t HoughModule::SaveRootData(RecoRootEvent *fRecoRootEvent) {
    //
    // Save reco root data
    //
    
    return kTRUE;
}

//____________________________________________________________________________
Bool_t HoughModule::Done() {
    //
    // done
    //
    
    fHittedAll.clear();
    fHittedFinal.clear();
    fHittedSelHough.clear();
    fHittedSel2Hough.clear();
    fIdMacrocellsSelected.clear();
   
    fClusters.clear();
    Msg(EsafMsg::Info) << "Completed" << MsgDispatch;
    return kTRUE;
}

//_____________________________________________________________________________
void HoughModule::UserMemoryClean() {
    //
    // memory clean
    //

    Msg(EsafMsg::Debug) << "UserMemoryClean()" << MsgDispatch;
    //fListObj->Delete();
    //fListObj->Clear();
	
//     Int_t fHittedFinalsize = fHittedFinal.size();

//     if(fHittedFinalsize<10 || !fTotalNumPoints || fHittedFinalsize>100000)		return;
	
    vector<EusoCluster*> *pclu = (vector<EusoCluster*>*)MyData()->GetObj("Clusters");
    if (pclu) {
        EusoCluster *dummy(NULL);
        for(size_t i(0); i<pclu->size(); i++) {
            dummy = (*pclu)[i];
            delete dummy;
        }
        pclu->clear();
        MyData()->RemoveObj("Clusters");
        vector<EusoCluster*> dummyV;
        dummyV.swap(*pclu);

    }
    
    vector<Int_t> *pix = (vector<Int_t>*)MyData()->GetObj("CluPixels");
    if (pix) {
        pix->clear();
        MyData()->RemoveObj("CluPixels");
        vector<Int_t> dummyV;
        dummyV.swap(*pix);
    }
   
}


//_____________________________________________________________________________
void HoughModule::TmpMemoryClean() {
    //
    // Clean all the temporary buffers
    //
    
    vector<Int_t> dummy1;
    dummy1.swap(fHittedAll);
    vector<Int_t> dummy2;
    dummy2.swap(fHittedSelHough);
    vector<Int_t> dummy3;
    dummy3.swap(fHittedSel2Hough);
    vector<Int_t> dummy4;
    dummy4.swap(fHittedFinalMacrocell);
    vector<Int_t> dummy5;
    dummy5.swap(fIdMacrocellsSelected);
    vector<Int_t> dummy6;
    dummy6.swap(fHittedHoughMerged);

}

//_____________________________________________________________________________
void HoughModule::WriteSingleCluster() {
    //
    // close and write the current cluster if is significant
    //
    
    if ( !fHittedFinal.size() ) return; 
    // write the cluster
    EusoCluster *cl = new EusoCluster();
    for( UInt_t i=0; i<fHittedFinal.size(); i++ ) {
        cl->AddPoint( fHittedFinal[i] );
    }
    fClusters.push_back(cl);
    Msg(EsafMsg::Info) << "SingleCluster saved with "<< fHittedFinal.size() << " points" << MsgDispatch;
   
    // add data to event
    vector<Int_t> *pix = &fHittedFinal;
    MyData()->Add("CluPixels", pix);
    vector<EusoCluster*> *pclu = &fClusters;
    MyData()->Add("Clusters", pclu);
    MyData()->Add("NumHitsMinimum",fFinalNumHitsMinimum);

    // add global data to the event
    RecoGlobalData* data = fEv->FindCreateGlobalData("PatternRecognition");
    data->CleanMaps();
    data->SetIssuer(GetName());
    data->Add("SelectedPixels",&fHittedFinal);

}

//_____________________________________________________________________________
EusoCluster* HoughModule::GetCluster( Int_t i ) {
    //
    // get a specified cluster
    //

    if ( i >= 0 && i < (Int_t)fClusters.size() )
        return fClusters[i];
    else
        return NULL;
}



