// $Id: ProfileModule.cc 2759 2006-09-07 08:10:49Z moreggia $
// Author: Anne Stutz   Jun,  6 2005

/*****************************************************************************
 * ESAF: Euso Simulation and Analysis Framework                              *
 *                                                                           *
 *  Id: ProfileModule                                                           *
 *  Package: <packagename>                                                   *
 *  Coordinator: <coordinator>                                               *
 *                                                                           *
 *****************************************************************************/

//_____________________________________________________________________________
//
// ProfileModule
//
// <extensive class description>
//
//   Config file parameters
//   ======================
//
//   <parameter name>: <parameter description>
//   -Valid options: <available options>
//

#include "ProfileModule.hh"
#include "RecoEvent.hh"
#include "RecoPhotonOnPupilData.hh"
#include "RecoShowerTrackData.hh"
#include "RecoShowerStepData.hh"
#include <TMath.h>
#include <TF1.h>
#include <TStyle.h> //DELETE
#include "EConst.hh"
#include "KakimotoFluoCalculator.hh"
#include "EsafSpectrum.hh"
#include "O1_ClearSkyPropagator.hh"  //DELETE
#include "TestGround.hh"
#include "Atmosphere.hh"
#include "RecoRootEvent.hh"
#include "EarthVector.hh"
#include "EsafRandom.hh"

ClassImp(ProfileModule)

using namespace sou;
using namespace EConst;
using namespace TMath;

// parametrization definitions
//_________________________________________________________________________________________ 
Double_t Profile_GFA(Double_t* x, Double_t* par) {
    //
    // number of electrons at given depth according to Gaussian Function in Age parameterization.
    // Formulae from C. Song, Astropart. Physics 22(2004)151
    //
    // par[0] : Nmax
    // par[1] : Xmax
    // par[2] : sigma
    //
    
    Double_t Nmax  = par[0];
    Double_t Xmax  = par[1];
    Double_t sigma = par[2];
    Double_t Age = 3*x[0] / (x[0] + 2*Xmax);

    Double_t Ne = -1;
    
    
    if(Age <= 0) return 0;
    
    Ne = Nmax * exp((-1./(2.*sigma*sigma)) * pow(Age - 1.,2));
    
    return Ne;
}




//_____________________________________________________________________________
ProfileModule::ProfileModule() : RecoModule("Profile") {
    //
    // ctor
    //

}

//_____________________________________________________________________________
ProfileModule::~ProfileModule() {
    //
    // dtor
    //
}

//_____________________________________________________________________________
Bool_t ProfileModule::Init() {
    //
    // initializations
    //

    Msg(EsafMsg::Info) << "Initializing " << MsgDispatch;
    string AnalysisType = Conf()->GetStr( "ProfileModule.fAnalysisType" );
    fThetaMode = Conf()->GetStr( "ProfileModule.fThetaMode" );
    
    fSigmaZerror = Conf()->GetNum("ProfileModule.fSigmaZerror")*km;
    fMeanZerror  = Conf()->GetNum("ProfileModule.fMeanZerror")*km;
    fSigmaThetaerror = Conf()->GetNum("ProfileModule.fSigmaThetaerror")*DegToRad();
    fMeanThetaerror  = Conf()->GetNum("ProfileModule.fMeanThetaerror")*DegToRad();
    
    if(AnalysisType == "Xmaxshift_airdensity") {
        Msg(EsafMsg::Info) << "Analysis type for ProfileModule Xmaxshift_airdensity" << MsgDispatch;
	fAnalysisType = Xmaxshift_airdensity;
    }
    else if(AnalysisType == "EShift_airDensity") {
        Msg(EsafMsg::Info) << "Analysis type for ProfileModule EShift_AirDensity" << MsgDispatch;
	fAnalysisType = EShift_airDensity;
    }
    else Msg(EsafMsg::Panic) << "Wrong config value ProfileModule.fAnalysisType " << fAnalysisType << " Setted to Xmaxshift_airdensity" << MsgDispatch;
    
    return kTRUE;
}

//_____________________________________________________________________________
Bool_t ProfileModule::PreProcess() {
    //
    //
    //
    
    fEvent          = NULL;
    fTheta          = 0.;
    fPhi            = 0.;
    fXmax           = 0.;
    fTrueXmax       = 0.;
    fElecProfile_simu = 0;
    fElecProfile_reco = 0;
    ResetArrays();
    
    return kTRUE;
}

//_____________________________________________________________________________
Bool_t ProfileModule::Process(RecoEvent *ev) {
    //
    //
    //
#ifdef DEBUG
    Msg(EsafMsg::Debug)<<"Process"<< MsgDispatch;
#endif

    // manage the profile reconstruction chain
    fEvent = ev; 
    
    // load information from previous modules
    // amd buil
    LoadModules();



    // Assess the effect of air density profile on the shower dvpt (Xmax shift)
    if(fAnalysisType == Xmaxshift_airdensity) {
        XmaxShift_AirDensity();
    }
    else if(fAnalysisType == EShift_airDensity) {
        EShift_AirDensity();
    }
    
        
            
    return kTRUE;
}

//_____________________________________________________________________________
void ProfileModule::LoadModules()  {
    //
    //
    //
#ifdef DEBUG
          Msg(EsafMsg::Debug)<<"Load Modules"<< MsgDispatch;
#endif
    
    fTheta =  fEvent->GetHeader().GetTrueTheta();
    fPhi =  fEvent->GetHeader().GetTruePhi();
    
    
    if(fAnalysisType == Xmaxshift_airdensity || fAnalysisType == EShift_airDensity) {
        Msg(EsafMsg::Info)<<"************  ANALYSIS of Xmax / Energy shift due to air density profile  ************"<< MsgDispatch;
	
	// get input data
	RecoShowerTrackData* trackdata = fEvent->GetRecoShowerTrackData();
	
	// get Xmax and showermaxpos truth
	fTrueXmax = trackdata->GetXmax();
	fTOAImpact = trackdata->GetTOAImpact();
	fNbsteps_simu = trackdata->GetNumSteps();

        // get true shower max 3D-POS
	fShowerMaxPos = trackdata->GetShowerMaxPos();
	fEarthImpact = trackdata->GetEarthImpact();
    }
    
    else Msg(EsafMsg::Panic) << "Should not occur, analysis type :" <<fAnalysisType << MsgDispatch;
}

//_____________________________________________________________________________
void ProfileModule::XmaxShift_AirDensity()  {
    //
    // Assess Xmax resolution : Simulated max position is used directly to reconstruct Xmax
    //    - Max pos is modified     --> effect on Xmax
    //    - atmosphere is modified  --> effect on Xmax
    // NB : HERE THE PROFILE IS NOT ENTIRELY RECONSTRUCTED !! only the max position is used
    //
    
    Msg(EsafMsg::Info)<<"XmaxShift_AirDensity analysis"<< MsgDispatch;
    
    
    
    // get atmosphere description (usually US-Std) and simulated depth of maximum
    const Atmosphere* atmo = Atmosphere::Get();
    EarthVector maxpos = fShowerMaxPos;
    EarthVector TOAimpact = fTOAImpact;
    fQuality = 0.;
    
    
    
    // introduce an error in Z determination, if Z underground or outside atmo, resp. 0km and TOA values are used
    if(fSigmaZerror > 0. || fabs(fMeanZerror) > 0.) {
	Double_t Zshift = IntroduceZerror();
	Msg(EsafMsg::Info)<<"Zshift = "<<Zshift/km<<" km"<< MsgDispatch;
	maxpos(2) += Zshift;
	fZshift = Zshift;
	if(maxpos.IsUnderSeaLevel()) {
            maxpos = fShowerMaxPos;
	    maxpos(2) = sqrt( pow(EarthRadius(),2) - fShowerMaxPos.Perp2() ) - EarthRadius(); // <=> Zv=0
            Msg(EsafMsg::Info)<<"Zshift PUT POS under sea level, NOW put at Zv = "<<maxpos.Zv()<< MsgDispatch;
            fZshift =  1000.;
	    fQuality = -10.;
	}
	else if(maxpos.Zv() > atmo->GetTOAAltitude()) {
            maxpos = fShowerMaxPos;
	    maxpos(2) = sqrt( pow(EarthRadius() + atmo->GetTOAAltitude(),2) - fShowerMaxPos.Perp2() ) - EarthRadius(); // <=> Zv=TOA
            Msg(EsafMsg::Info)<<"Zshift PUT POS above TOA, NOW put at Zv = "<<maxpos.Zv()<< MsgDispatch;
            fZshift =  1000.;
	    fQuality = 10.;
	}
    }
    
    
    
    // introduce an error in theta determination, TAKEN AT shower MAXIMUM, then TOA impact is calculated
    if(fSigmaThetaerror > 0. || fabs(fMeanThetaerror) > 0.) {
        Double_t ThetaShift = IntroduceThetaerror();
        Msg(EsafMsg::Info)<<"Theta shift = "<<ThetaShift*RadToDeg()<<" deg"<< MsgDispatch;
        fThetashift = ThetaShift;
        EarthVector direction = (fTOAImpact - fShowerMaxPos).Unit();
        direction.SetTheta(direction.Theta() + ThetaShift);
        TOAimpact = atmo->ImpactAtTOA(fShowerMaxPos,direction);
	fTOAImpact = TOAimpact;
	if(fThetaMode == "earthimpact") {
	    if(fEarthImpact.Zv() > kAltitudeTolerance) fQuality = -100.;
	    else {
	        Double_t magmax = (fEarthImpact - fShowerMaxPos).Mag();
                direction = (fTOAImpact - fEarthImpact).Unit();
                direction.SetTheta(direction.Theta() + ThetaShift);
                TOAimpact = atmo->ImpactAtTOA(fEarthImpact,direction);
	        fTOAImpact = TOAimpact;
		maxpos = fEarthImpact + magmax*direction;
		if( fabs((atmo->ImpactASL(maxpos,direction)).Z()) != HUGE) fQuality = -50.;
	    }
	}
    }
    
    
    
    // calculate Grammage between TOA and shower max position --> reco Xmax with current atmosphere config
    fXmax = -10000.;
    if(fQuality == 0.) fXmax = atmo->Grammage(maxpos,TOAimpact);
}

//_____________________________________________________________________________
void ProfileModule::EShift_AirDensity()  {
    //
    // Assess Xmax resolution : Simulated max position is used directly to reconstruct Xmax
    //    - Max pos is modified     --> effect on profile integral
    //    - atmosphere is modified  --> effect on profile integral
    //
    
    Msg(EsafMsg::Info)<<"EShift_AirDensity analysis"<< MsgDispatch;
    
    
    
    // init
    const Atmosphere* atmo = Atmosphere::Get();
    // get input data
    RecoShowerTrackData* trackdata = fEvent->GetRecoShowerTrackData();
    const RecoShowerStepData* recostep = 0;
    // set arrays with new event data
    ResetArrays(fNbsteps_simu);
    for(Int_t i=0; i < fNbsteps_simu; i++) {
	recostep = trackdata->GetStep(i);
	fx_simu[i] = recostep->GetPosi().X();
	fy_simu[i] = recostep->GetPosi().Y();
	fz_simu[i] = recostep->GetPosi().Z();
	fx_reco[i] = recostep->GetPosi().X();
	fy_reco[i] = recostep->GetPosi().Y();
	fz_reco[i] = recostep->GetPosi().Z();
	fX_simu[i] = recostep->GetXi();
	fX_reco[i] = recostep->GetXi();
	fNe_simu[i] = recostep->GetNe();
    }
    fx_simu[fNbsteps_simu] = recostep->GetPosf().X();
    fy_simu[fNbsteps_simu] = recostep->GetPosf().Y();
    fz_simu[fNbsteps_simu] = recostep->GetPosf().Z();
    fx_reco[fNbsteps_simu] = recostep->GetPosf().X();
    fy_reco[fNbsteps_simu] = recostep->GetPosf().Y();
    fz_reco[fNbsteps_simu] = recostep->GetPosf().Z();
    fX_simu[fNbsteps_simu] = recostep->GetXf();
    fX_reco[fNbsteps_simu] = recostep->GetXf();
    fNbsteps_crash = fNbsteps_simu;
    
    
    
    
    // introduce an error in Z determination
    if(fSigmaZerror > 0. || fabs(fMeanZerror) > 0.) {
        Bool_t stop_crash = kFALSE;
	Double_t Zshift = IntroduceZerror();
	Msg(EsafMsg::Info)<<"Zshift = "<<Zshift/km<<" km"<< MsgDispatch;
        fZshift = Zshift;
	fQuality = 0.;
	for(Int_t i=0; i < fNbsteps_simu; i++) {
	    if(stop_crash) {
	        fNe_simu[i] = 0.;
		continue;
	    }
	    EarthVector stepos(fx_reco[i],fy_reco[i],fz_reco[i]+Zshift);
	    if(stepos.IsUnderSeaLevel()) {
		fNbsteps_crash = i-1;
		if(fNbsteps_crash < 0) fNbsteps_crash = 0;
		fQuality = -10.;
		stop_crash = kTRUE;
		fNe_simu[i] = 0.;
		fNe_simu[i-1] = 0.;
	    }
	    else if(stepos.Zv() > atmo->GetTOAAltitude()) {
		fQuality = 10.;
		stop_crash = kTRUE;
		fNe_simu[i] = 0.;
		fNe_simu[i-1] = 0.;
	    }
	    else {
	        //apply Zshift
		fz_reco[i] += Zshift;
		// determine depth of the reco step
		if(i == 0) {
		    EarthVector TOAreco = fTOAImpact;
		    TOAreco(2) += Zshift;
		    fX_reco[i] = atmo->Grammage(stepos,TOAreco);
		}
		else {
		    EarthVector prevpos(fx_reco[i-1],fy_reco[i-1],fz_reco[i-1]);
		    fX_reco[i] = fX_reco[i-1] + atmo->Grammage(prevpos,stepos);
		}
	    }
	}
	if(!stop_crash) {
	    // for last step
	    EarthVector stepos(fx_reco[fNbsteps_simu],fy_reco[fNbsteps_simu],fz_reco[fNbsteps_simu]+Zshift);
	    if(stepos.IsUnderSeaLevel()) {
		fNbsteps_crash = fNbsteps_simu-1;
		if(fNbsteps_crash < 0) fNbsteps_crash = 0;
		fQuality = -10.;
		stop_crash = kTRUE;
		fNe_simu[fNbsteps_simu-1] = 0.;
	    }
	    else if(stepos.Zv() > atmo->GetTOAAltitude()) {
		fQuality = 10.;
		stop_crash = kTRUE;
		fNe_simu[fNbsteps_simu-1] = 0.;
	    }
	    else {
		// apply Zshift
		fz_reco[fNbsteps_simu] += Zshift;
		// determine depth of the reco step
        	EarthVector prevpos(fx_reco[fNbsteps_simu-1],fy_reco[fNbsteps_simu-1],fz_reco[fNbsteps_simu-1]);
		fX_reco[fNbsteps_simu] = fX_reco[fNbsteps_simu-1] + atmo->Grammage(prevpos,stepos);
	    }
	}
    }
    
    


    
    // introduce an error in theta determination, TAKEN AT shower MAXIMUM
    if(fSigmaThetaerror > 0. || fabs(fMeanThetaerror) > 0.) {
	Bool_t stop_crash = kFALSE;
	fQuality = 0.;
        EarthVector REFpos = fShowerMaxPos;
	if(fThetaMode == "earthimpact") {
	    REFpos = fEarthImpact;
	    if(fEarthImpact.Zv() > kAltitudeTolerance) {
	        stop_crash = kTRUE;
		fQuality = -100.;
	    }
	}
	Double_t ThetaShift = IntroduceThetaerror();
	Msg(EsafMsg::Info)<<"Theta shift = "<<ThetaShift*RadToDeg()<<" deg"<< MsgDispatch;
	fThetashift = ThetaShift;
	EarthVector pos_reco = REFpos;
	EarthVector dir_up = (fTOAImpact - REFpos).Unit();
	dir_up.SetTheta(dir_up.Theta() + ThetaShift);
	EarthVector dir = -dir_up;
	EarthVector pos1_simu(fx_simu[0],fy_simu[0],fz_simu[0]);
	Double_t magmax = (pos1_simu - REFpos).Mag();
	pos_reco += magmax*dir_up;
	fx_reco[0] = pos_reco.X();
	fy_reco[0] = pos_reco.Y();
	fz_reco[0] = pos_reco.Z();
	EarthVector TOAreco = atmo->ImpactAtTOA(REFpos,dir_up);
        fX_reco[0] = atmo->Grammage(pos_reco,TOAreco);
	for(Int_t i=1; i < fNbsteps_simu; i++) {
	    if(stop_crash) {
	        fNe_simu[i] = 0.;
		continue;
	    }
	    EarthVector prevpos_simu(fx_simu[i-1],fy_simu[i-1],fz_simu[i-1]);
	    EarthVector nextpos_simu(fx_simu[i],fy_simu[i],fz_simu[i]);
	    pos_reco += (nextpos_simu - prevpos_simu).Mag() * dir;
	    if(pos_reco.IsUnderSeaLevel()) {
		fNbsteps_crash = i-1;
		if(fNbsteps_crash < 0) fNbsteps_crash = 0;
		fQuality = -10.;
		stop_crash = kTRUE;
		fNe_simu[i] = 0.;
		fNe_simu[i-1] = 0.;
	    }
	    else if(pos_reco.Zv() > atmo->GetTOAAltitude()) {
		fQuality = 10.;
		stop_crash = kTRUE;
		fNe_simu[i] = 0.;
		fNe_simu[i-1] = 0.;
	    }
	    else {
	        fx_reco[i] = pos_reco.X();
	        fy_reco[i] = pos_reco.Y();
	        fz_reco[i] = pos_reco.Z();
		// determine depth of the reco step
		EarthVector prevpos(fx_reco[i-1],fy_reco[i-1],fz_reco[i-1]);
		fX_reco[i] = fX_reco[i-1] + atmo->Grammage(prevpos,pos_reco);
	    }
	}
	// for last step
	if(!stop_crash) {
	    EarthVector prevpos_simu(fx_simu[fNbsteps_simu-1],fy_simu[fNbsteps_simu-1],fz_simu[fNbsteps_simu-1]);
	    EarthVector nextpos_simu(fx_simu[fNbsteps_simu],fy_simu[fNbsteps_simu],fz_simu[fNbsteps_simu]);
	    pos_reco += (nextpos_simu - prevpos_simu).Mag() * dir;
	    if(pos_reco.IsUnderSeaLevel()) {
		fNbsteps_crash = fNbsteps_simu-1;
		if(fNbsteps_crash < 0) fNbsteps_crash = 0;
		fQuality = -10.;
		stop_crash = kTRUE;
		fNe_simu[fNbsteps_simu-1] = 0.;
	    }
	    else if(pos_reco.Zv() > atmo->GetTOAAltitude()) {
		fQuality = 10.;
		stop_crash = kTRUE;
		fNe_simu[fNbsteps_simu-1] = 0.;
	    }
	    else {
		fx_reco[fNbsteps_simu] = pos_reco.X();
		fy_reco[fNbsteps_simu] = pos_reco.Y();
		fz_reco[fNbsteps_simu] = pos_reco.Z();
		// determine depth of the reco step
        	EarthVector prevpos(fx_reco[fNbsteps_simu-1],fy_reco[fNbsteps_simu-1],fz_reco[fNbsteps_simu-1]);
		fX_reco[fNbsteps_simu] = fX_reco[fNbsteps_simu-1] + atmo->Grammage(prevpos,pos_reco);
	    }
	}
    }
    
    
        
    
    
    // build histos of simu and reco profile
    fElecProfile_simu = new TH1F("profile_simu","profile_simu",fNbsteps_simu,fX_simu.GetArray());
    fElecProfile_reco = new TH1F("profile_reco","profile_reco",fNbsteps_simu,fX_reco.GetArray());
    
    // set bin content to shower size Ne
    for(Int_t i=0; i < fNbsteps_simu; i++) {
        fElecProfile_reco->SetBinContent(i+1,fNe_simu[i]);
        fElecProfile_simu->SetBinContent(i+1,fNe_simu[i]);
    }
    
    // integrate profiles and store values in outputs
    fRecoProfIntegral = fElecProfile_reco->Integral("width");
    fSimuProfIntegral = fElecProfile_simu->Integral("width");
}

//_____________________________________________________________________________
Double_t ProfileModule::IntroduceZerror() {
    //
    // introduce an error in Z-cordinate, then applied to XmaxShift_AirDensity() reco part
    //
    return EsafRandom::Get()->Gaus(fMeanZerror,fSigmaZerror);
}

//_____________________________________________________________________________
Double_t ProfileModule::IntroduceThetaerror() {
    //
    // introduce an error in Theta, then applied to XmaxShift_AirDensity() reco part
    //
    return EsafRandom::Get()->Gaus(fMeanThetaerror,fSigmaThetaerror);
}

//_____________________________________________________________________________
void ProfileModule::ResetArrays(Int_t n)  {
    //
    // clear TArrayD data, and define new size
    //
    fx_simu.Reset();
    fy_simu.Reset();
    fz_simu.Reset();
    fx_reco.Reset();
    fy_reco.Reset();
    fz_reco.Reset();
    fX_simu.Reset();
    fX_reco.Reset();
    fNe_simu.Reset();
    fx_simu.Set(n+1);
    fy_simu.Set(n+1);
    fz_simu.Set(n+1);
    fx_reco.Set(n+1);
    fy_reco.Set(n+1);
    fz_reco.Set(n+1);
    fX_simu.Set(n+1);
    fX_reco.Set(n+1);
    fNe_simu.Set(n);
    if(fElecProfile_simu) SafeDelete(fElecProfile_simu);
    if(fElecProfile_reco) SafeDelete(fElecProfile_reco);
}

//_____________________________________________________________________________
Bool_t ProfileModule::PostProcess() {
    //
    //
    //
    ResetArrays();
    
    return kTRUE;
}

//_____________________________________________________________________________
Bool_t ProfileModule::Done() {
    //
    //chi
    //

    Msg(EsafMsg::Info)<<"ProfileModule done. "<< MsgDispatch;
    return kTRUE;
}

//_____________________________________________________________________________
void ProfileModule::UserMemoryClean()  {
    //
    // delete user objects
    //
    
    ResetArrays();
}

//_____________________________________________________________________________
Bool_t ProfileModule::SaveRootData(RecoRootEvent *fRecoRootEvent) {
    //
    //
    //
#ifdef DEBUG
    Msg(EsafMsg::Debug)<<"Save reco root data"<< MsgDispatch;
#endif
    
    RecoProfile& output = fRecoRootEvent->GetRecoProfile();
    
    // profile outputs
    output.SetQuality(fQuality);
    output.SetXmax(fXmax*cm2/g);
    output.SetTrueXmax(fTrueXmax*cm2/g);
    output.SetRecoProfIntegral(fRecoProfIntegral*cm2/g);
    output.SetSimuProfIntegral(fSimuProfIntegral*cm2/g);
    output.SetZshift(fZshift/km);
    output.SetThetashift(fThetashift*RadToDeg());
    output.SetTOAImpact(fTOAImpact.X(),fTOAImpact.Y(),fTOAImpact.Z());

    return kTRUE;
}



