// $Id: RecoTreePainter.cc 2797 2008-02-11 13:09:12Z naumov $
// Author: Dmitry.Naumov   2006/02/03

/*****************************************************************************
 * ESAF: Euso Simulation and Analysis Framework                              *
 *                                                                           *
 *  Id: RecoTreePainter                                                      *
 *  Package: RecoViewer                                                      *
 *  Coordinator: Dmitry.Naumov                                               *
 *                                                                           *
 *****************************************************************************/

//_____________________________________________________________________________
//
// RecoTreePainter
//
// <extensive class description>
//
//   Config file parameters
//   ======================
//
//   <parameter name>: <parameter description>
//   -Valid options: <available options>
//

#include "RecoTreePainter.hh"
#include "RecoRootEvent.hh"
#include "EGViewer.hh"
#include "ESystemOfUnits.hh"

#include <TEventList.h>
#include <TH1F.h>
#include <TH1D.h>
#include <TH2F.h>
#include <THStack.h>
#include <TProfile.h>
#include <TProfile2D.h>
#include <TList.h>
#include <TSystem.h>
#include <TStyle.h>
#include <TCanvas.h>
#include <TTree.h>
#include <TMath.h>
#include <TF1.h>
#include <TLegend.h>
#include <TLegend.h>
#include <TGraphErrors.h>
#include <TText.h>
#include <TPaveText.h>
#include <TDirectory.h>

using namespace TMath;

ClassImp(RecoTreePainter)

using namespace sou;
//_____________________________________________________________________________
RecoTreePainter::RecoTreePainter(TTree* tree): EsafMsgSource(), fEvent(NULL), fSaveEpsFiles(kFALSE) {
    //
    //
    //
    if (!tree) Error("RecoTreePainter","ctor must be initialized with valid TTree pointer to the recotree\n");
    fHistoList = new TList;
    fEvent = new RecoRootEvent;
    tree->SetBranchAddress("events",&fEvent);
    fNevents = tree->GetEntries();
    fTree = tree;
    Info("RecoTreePainter",Form("Initialized TTree with %d entries",fNevents));

    fEnergyIndependentFromDir = kTRUE;
    fUseAA1 = kFALSE;

    //
    left_edge["Energy"] = 13;
    left_edge["Theta"]  = 0;
    left_edge["Phi"]    = 0;

    right_edge["Energy"] = 15;
    right_edge["Theta"]  = 102;
    right_edge["Phi"]    = 360;
    fFilename = "EnergyPlots";
    gStyle->SetCanvasColor(0);
    fEpsDir = "./";

    fEventList = new TEventList("recoEventList");
}

//_____________________________________________________________________________
RecoTreePainter::~RecoTreePainter() {
    //
    // Destructor
    //

    fHistoList->Delete();
    delete fHistoList;
}

//______________________________________________________________________________
Long64_t RecoTreePainter::UpdateEventList() {
    //
    // Update the eventlist according the current cut
    //

    return fTree->Draw(">>recoEventList",fCut);
 
}
//______________________________________________________________________________
Long64_t RecoTreePainter::UpdateEventList( const char* selection ) {
    //
    // Update the eventlist according selection
    //

    fCut = selection;
    return UpdateEventList();
    
}


//______________________________________________________________________________
Long64_t RecoTreePainter::GetEntry( Int_t i ) {


    if ( i >= fEventList->GetN() ) {
        MsgForm(EsafMsg::Info,"Entry id %d is greater than the maximum number of selected events (%d)",
             i,fEventList->GetN());
        return kFALSE;
    }

    Long64_t entry = fEventList->GetEntry(i);
    return fTree->GetEntry(entry);
}

//_____________________________________________________________________________
void RecoTreePainter::Loop() {
    //
    //
    //
    Int_t   progress    = 0;
    UpdateEventList();

    if ( !fEventList->GetN() ) return;

    for (Int_t i(0); i<fEventList->GetN(); i++) {
         GetEntry(i);
         FillHistos();

         if (100.*i/fEventList->GetN() >= progress) {
         Msg(EsafMsg::Info).SetProgress(progress);
         Msg(EsafMsg::Info) << "Loop():" << MsgCount;
         progress+=10;
          }
    }
    
    FillAngularResolutions();
    
    Msg(EsafMsg::Info) << "Loop():" << MsgCount;
    if (progress <= 100) Msg(EsafMsg::Info) << MsgDispatch;
}

//_____________________________________________________________________________
void RecoTreePainter::FillHistos() {
    //
    //
    //
    MakeBaseDataMap();
    FillSimHistos();
    FillThetaHistos();
    FillPhiHistos();
    FillDirectionHistos();
    FillEnergyHistos();
}

//______________________________________________________________________________
void RecoTreePainter::FillSimHistos() {
    //
    //
    //
    map <string, Bool_t>::iterator it;
    // BULK of simulated information
    TH1 *hThetaSim = GetHisto("ThetaSimulated_1D", "Theta Simulated", "1D", 100, left_edge["Theta"], right_edge["Theta"],0,0,0);
    hThetaSim->SetXTitle("#theta_{sim} [deg]");
    hThetaSim->SetYTitle("Entries");
    hThetaSim->Fill(fSim["ThetaSim"]*RadToDeg());
    for (it=fCuts.begin(); it != fCuts.end(); it++) {
         TString name  = "ThetaSimulated"  + it->first;
         TString title = "Theta Simulated in " + it->first;
         TH1* h = GetHisto(name+"_1D",title,"1D",100,left_edge["Theta"],right_edge["Theta"],0,0,0);
     if (it->second) h->Fill(fSim["ThetaSim"]*RadToDeg());

    }
    TH1 *hPhiSim = GetHisto("PhiSimulated_1D", "Phi Simulated", "1D", 100, left_edge["Phi"], right_edge["Phi"],0,0,0);
    hPhiSim->SetXTitle("#varphi_{sim} [deg]");
    hPhiSim->SetYTitle("Entries");
    hPhiSim->Fill(fSim["PhiSim"]*RadToDeg());
    for (it=fCuts.begin(); it != fCuts.end(); it++) {
         TString name  = "PhiSimulated"  + it->first;
         TString title = "Phi Simulated in " + it->first;
         TH1* h = GetHisto(name+"_1D",title,"1D",100,left_edge["Phi"],right_edge["Phi"],0,0,0);
     if (it->second) h->Fill(fSim["PhiSim"]*RadToDeg());

    } 
    TH1* h3 = GetHisto("EnergySimulated_1D","Energy Simulated","1D",100,left_edge["Energy"],right_edge["Energy"],0,0,0);
    h3->SetXTitle("Simulated Energy/EeV");
    h3->SetYTitle("Entries");
    h3->Fill(fSim["EnergySim"]);
    for (it=fCuts.begin(); it != fCuts.end(); it++) {
         TString name  = "EnergySimulated"  + it->first;
         TString title = "EnergySimulated in " + it->first;
         TH1* h = GetHisto(name+"_1D",title,"1D",100,left_edge["Energy"],right_edge["Energy"],0,0,0);
     if (it->second) h->Fill(fSim["EnergySim"]);

    }
}

//______________________________________________________________________________
void RecoTreePainter::FillThetaHistos() {
    //
    //
    //
    map <string, Bool_t>::iterator it;
    
    if (!AcceptDirection()) return;

    Int_t nBinsTheta = TMath::FloorNint(right_edge["Theta"]-left_edge["Theta"])-1;
    Int_t nBinsPhi = TMath::FloorNint((right_edge["Phi"]-left_edge["Phi"])/2)-1;
    Double_t errmax = 30;
    Int_t errbins = 2*TMath::FloorNint(errmax)-1;

    // Theta reco vs theta sim
    TH1 *h1 = GetHisto("ThetaSim_vs_Rec_2D","Theta Sim vs Rec","2D",nBinsTheta,left_edge["Theta"],right_edge["Theta"],
                                                                   nBinsTheta,left_edge["Theta"],right_edge["Theta"]);
    h1->SetXTitle("#theta_{sim} [deg]");
    h1->SetYTitle("#theta_{rec} [deg]");
    h1->Fill(fSim["ThetaSim"]*RadToDeg(),fRec["ThetaRec"]*RadToDeg());
    // theta error
    h1 = GetHisto("ThetaError","ThetaError","1D",errbins,-errmax,errmax,0,0,0);
    h1->SetXTitle("#Delta #theta [deg]");
    h1->SetYTitle("Entries");
    h1->Fill(fRec["ThetaError"]*RadToDeg());
    // abs theta error
    h1 = GetHisto("AbsThetaError","AbsThetaError","1D",180,0,errmax,0,0,0);
    h1->SetXTitle("|#Delta #theta| [deg]");
    h1->SetYTitle("Entries");
    h1->Fill(Abs(fRec["ThetaError"]*RadToDeg()));
    
     // Fill Theta Error of vs Theta Simulated
    h1 = GetHisto("ThetaError_vs_Theta_2D","ThetaError vs Theta","2D",nBinsTheta,left_edge["Theta"],right_edge["Theta"],
                                            errbins,-errmax,errmax);
    h1->SetXTitle("#theta_{sim} [deg]");
    h1->SetYTitle("#Delta #theta [deg]");
    h1->Fill(fSim["ThetaSim"]*RadToDeg(),fRec["ThetaError"]*RadToDeg());
    // Fill Theta Error of vs Phi Simulated
    h1 = GetHisto("ThetaError_vs_Phi_2D","ThetaError vs Phi","2D",nBinsPhi,left_edge["Phi"],right_edge["Phi"],
                                            errbins,-errmax,errmax);
    h1->SetXTitle("#varphi_{sim} [deg]");
    h1->SetYTitle("#Delta #theta [deg]");
    h1->Fill(fSim["PhiSim"]*RadToDeg(),fRec["ThetaError"]*RadToDeg());
     // Fill Theta Error of vs signal/noise ratio
    h1 = GetHisto("ThetaError_vs_SignalToNoise_2D","ThetaError vs S/N","2D",100,0,100,errbins,-errmax,errmax);
    h1->SetXTitle("S/N");
    h1->SetYTitle("#Delta #theta [deg]");
    h1->Fill(fRec["SignalToNoise"],fRec["ThetaError"]*RadToDeg());
     // Fill Theta Error vs Energy Simulated
    h1 = GetHisto("ThetaError_vs_EnergySim_2D","ThetaError vs Energy","2D",100,left_edge["Energy"],right_edge["Energy"],errbins,-errmax,errmax);
    h1->SetXTitle("Simulated Energy/EeV");
    h1->SetYTitle("#Delta #theta [deg]");
    h1->Fill(fSim["EnergySim"],fRec["ThetaError"]*RadToDeg());
     // Fill Theta Error vs Distance in FOV
    h1 = GetHisto("ThetaError_vs_Distance_2D","ThetaError vs Distance in FOV","2D",100,0,300,errbins,-errmax,errmax);
    h1->SetXTitle("Distance in FOV [km]");
    h1->SetYTitle("#Delta #theta [deg]");
    h1->Fill(fSim["FOVDistance"],fRec["ThetaError"]*RadToDeg());

    // histos with cuts
    for (it=fCuts.begin(); it != fCuts.end(); it++) {
        TString name  = "ThetaSim_vs_Rec"  + it->first;
        TString title = "ThetaSim_vs_Rec " + it->first;
        TH1* h = NULL;

        // Fill Theta Simulated vs Theta Reconstructed
        h = GetHisto(name+"_2D",title,"2D",nBinsTheta,left_edge["Theta"],right_edge["Theta"],
                nBinsTheta,left_edge["Theta"],right_edge["Theta"]);
        if (it->second) h->Fill(fSim["ThetaSim"]*RadToDeg(),fRec["ThetaRec"]*RadToDeg());
        // theta error
        name = "ThetaError" + it->first;
        title = "ThetaError " + it->first;
        h = GetHisto(name+"_1D",title,"1D",errbins,-errmax,errmax,0,0,0);
        if (it->second) h->Fill(fRec["ThetaError"]*RadToDeg());
        // absolute theta error
        name = "AbsThetaError" + it->first;
        title = "AbsThetaError " + it->first;
        h = GetHisto(name+"_1D",title,"1D",90,0,errmax,0,0,0);
        if (it->second) h->Fill(Abs(fRec["ThetaError"]*RadToDeg()));
        // Fill Theta Error of vs Theta Simulated
        name  = "ThetaError_vs_Theta" + it->first;
        title = "ThetaError vs Theta"+ it->first;
        h = GetHisto(name+"_2D",title,"2D",nBinsTheta,left_edge["Theta"],right_edge["Theta"],
                errbins,-errmax,errmax);
        if (it->second) h->Fill(fSim["ThetaSim"]*RadToDeg(),fRec["ThetaError"]*RadToDeg());
        // Fill Theta Error of vs Phi Simulated
        name  = "ThetaError_vs_Phi" + it->first;
        title = "ThetaError vs Phi"+ it->first;
        h = GetHisto(name+"_2D",title,"2D",nBinsTheta,left_edge["Phi"],right_edge["Phi"],
                errbins,-errmax,errmax);
        if (it->second) h->Fill(fSim["PhiSim"]*RadToDeg(),fRec["ThetaError"]*RadToDeg());
        // Fill Theta Error of vs signal/noise ratio
        name  = "ThetaError_vs_SignalToNoise" + it->first;
        title = "ThetaError vs S/N"+ it->first;
        h = GetHisto(name+"_2D",title,"2D",100,0,100,errbins,-errmax,errmax);
        if (it->second) h->Fill(fRec["SignalToNoise"],fRec["ThetaError"]*RadToDeg());
        // Fill Theta Error vs Energy Simulated
        name  = "ThetaError_vs_EnergySim" + it->first;
        title = "ThetaError vs Energy"+ it->first;
        h = GetHisto(name+"_2D",title,"2D",100,left_edge["Energy"],right_edge["Energy"],errbins,-errmax,errmax);
        if (it->second) h->Fill(fSim["EnergySim"],fRec["ThetaError"]*RadToDeg());
        // Fill Theta Error vs Distance in FOV
        name = "ThetaError_vs_Distance" + it->first;
        title = "ThetaError vs Distance in FOV"+ it->first;
        h = GetHisto(name+"_2D",title,"2D",100,0,300,errbins,-errmax,errmax);
        if (it->second) h->Fill(fSim["FOVDistance"],fRec["ThetaError"]*RadToDeg());
    }
}

//______________________________________________________________________________
void RecoTreePainter::FillPhiHistos() {
    //
    //
    //
    map <string, Bool_t>::iterator it;
    
    if (!AcceptDirection()) return;

    Int_t nBinsTheta = TMath::FloorNint(right_edge["Theta"]-left_edge["Theta"])-1;
    Int_t nBinsPhi = TMath::FloorNint((right_edge["Phi"]-left_edge["Phi"])/2)-1;
    Double_t errmax = 180;
    Int_t errbins = 2*TMath::FloorNint(errmax/2)-1;
    
    TH1* h1 = GetHisto("PhiSim_vs_Rec_2D","Phi Sim vs Rec","2D",nBinsPhi,left_edge["Phi"],right_edge["Phi"],
                                                                   nBinsPhi,left_edge["Phi"],right_edge["Phi"]);
    h1->SetXTitle("#varphi_{sim} [deg]");
    h1->SetYTitle("#varphi_{rec} [deg]");
    h1->Fill(fSim["PhiSim"]*RadToDeg(),fRec["PhiRec"]*RadToDeg());
    // phi error
    h1 = GetHisto("PhiError","PhiError","1D",errbins,-errmax,errmax,0,0,0);
    h1->SetXTitle("#Delta #varphi [deg]");
    h1->SetYTitle("Entries");
    h1->Fill(fRec["PhiError"]*RadToDeg());
    // abs phi error
    h1 = GetHisto("AbsPhiError","AbsPhiError","1D",180,0,errmax,0,0,0);
    h1->SetXTitle("|#Delta #varphi| [deg]");
    h1->SetYTitle("Entries");
    h1->Fill(Abs(fRec["PhiError"]*RadToDeg()));

     // Fill phi Error of vs Theta Simulated
    h1 = GetHisto("PhiError_vs_Theta_2D","PhiError vs Theta","2D",nBinsTheta,left_edge["Theta"],right_edge["Theta"],
                                            errbins,-errmax,errmax);
    h1->SetXTitle("#theta_{sim} [deg]");
    h1->SetYTitle("#Delta #varphi [deg]");
    h1->Fill(fSim["ThetaSim"]*RadToDeg(),fRec["PhiError"]*RadToDeg());
    // Fill phi Error of vs Phi Simulated
    h1 = GetHisto("PhiError_vs_Phi_2D","PhiError vs Phi","2D",nBinsPhi,left_edge["Phi"],right_edge["Phi"],
                                            errbins,-errmax,errmax);
    h1->SetXTitle("#varphi_{sim} [deg]");
    h1->SetYTitle("#Delta #varphi [deg]");
    h1->Fill(fSim["PhiSim"]*RadToDeg(),fRec["PhiError"]*RadToDeg());
     // Fill Phi Error of vs signal/noise ratio
    h1 = GetHisto("PhiError_vs_SignalToNoise_2D","PhiError vs S/N","2D",100,0,100,errbins,-errmax,errmax);
    h1->SetXTitle("S/N");
    h1->SetYTitle("#Delta #varphi [deg]");
    h1->Fill(fRec["SignalToNoise"],fRec["PhiError"]*RadToDeg());
     // Fill phi Error vs Energy Simulated
    h1 = GetHisto("PhiError_vs_EnergySim_2D","PhiError vs Energy","2D",100,left_edge["Energy"],right_edge["Energy"],errbins,-errmax,errmax);
    h1->SetXTitle("Simulated Energy/EeV");
    h1->SetYTitle("#Delta #varphi [deg]");
    h1->Fill(fSim["EnergySim"],fRec["PhiError"]*RadToDeg());
     // Fill phi Error vs Distance in FOV
    h1 = GetHisto("PhiError_vs_Distance_2D","PhiError vs Distance in FOV","2D",100,0,300,60,-180,180);
    h1->SetXTitle("Distance in FOV [km]");
    h1->SetYTitle("#Delta #varphi [deg]");
    h1->Fill(fSim["FOVDistance"],fRec["PhiError"]*RadToDeg());

    // histos with cuts

    for (it=fCuts.begin(); it != fCuts.end(); it++) {
         TString name  = "PhiSim_vs_Rec"  + it->first;
         TString title = "PhiSim_vs_Rec " + it->first;
         TH1* h = NULL;
         
     // Fill Phi Simulated vs Phi Reconstructed
     h = GetHisto(name+"_2D",title,"2D",nBinsPhi,left_edge["Phi"],right_edge["Phi"],
                                            nBinsPhi,left_edge["Phi"],right_edge["Phi"]);
         if (it->second) h->Fill(fSim["PhiSim"]*RadToDeg(),fRec["PhiRec"]*RadToDeg());
    // phi error
        name = "PhiError" + it->first;
        title = "PhiError " + it->first;
        h = GetHisto(name+"_1D",title,"1D",errbins,-errmax,errmax,0,0,0);
        if (it->second) h->Fill(fRec["PhiError"]*RadToDeg());
    // absolute phi error
        name = "AbsPhiError" + it->first;
        title = "AbsPhiError" + it->first;
        h = GetHisto(name+"_1D",title,"1D",90,0,errmax,0,0,0);
        if (it->second) h->Fill(Abs(fRec["PhiError"]*RadToDeg()));
     // Fill Phi Error of vs Theta Simulated
     name  = "PhiError_vs_Theta" + it->first;
     title = "PhiError vs Theta"+ it->first;
     h = GetHisto(name+"_2D",title,"2D",nBinsTheta,left_edge["Theta"],right_edge["Theta"],
                                            errbins,-errmax,errmax);
         if (it->second) h->Fill(fSim["ThetaSim"]*RadToDeg(),fRec["PhiError"]*RadToDeg());
     // Fill Phi Error of vs Phi Simulated
     name  = "PhiError_vs_Phi" + it->first;
     title = "PhiError vs Phi"+ it->first;
     h = GetHisto(name+"_2D",title,"2D",nBinsPhi,left_edge["Phi"],right_edge["Phi"],
                                            errbins,-errmax,errmax);
         if (it->second) h->Fill(fSim["PhiSim"]*RadToDeg(),fRec["PhiError"]*RadToDeg());
     // Fill Phi Error of vs signal/noise ratio
     name  = "PhiError_vs_SignalToNoise" + it->first;
     title = "PhiError vs S/N"+ it->first;
     h = GetHisto(name+"_2D",title,"2D",100,0,100,errbins,-errmax,errmax);
         if (it->second) h->Fill(fRec["SignalToNoise"],fRec["PhiError"]*RadToDeg());
     // Fill Phi Error vs Energy Simulated
     name  = "PhiError_vs_EnergySim" + it->first;
     title = "PhiError vs Energy"+ it->first;
     h = GetHisto(name+"_2D",title,"2D",100,left_edge["Energy"],right_edge["Energy"],errbins,-errmax,errmax);
         if (it->second) h->Fill(fSim["EnergySim"],fRec["PhiError"]*RadToDeg());
     // Fill Phi Error vs Distance in FOV
     name = "PhiError_vs_Distance" + it->first;
     title = "PhiError vs Distance in FOV"+ it->first;
     h = GetHisto(name+"_2D",title,"2D",100,0,300,errbins,-errmax,errmax);
         if (it->second) h->Fill(fSim["FOVDistance"],fRec["PhiError"]*RadToDeg());

   }
}

//______________________________________________________________________________
void RecoTreePainter::FillAngularResolutions() {
    //
    //
    //
    
    map <string, Bool_t>::iterator it;
    
    // search theta68
    TH1 *h1 = GetHisto("AbsThetaError","","1D",0,0,0,0,0,0);
    Double_t integral = h1->Integral();
    Double_t theta68 = -1.;
    Int_t bin68 = -1;
    for( Int_t i(0); i<h1->GetNbinsX(); i++ ) {
        if ( (h1->Integral(1,i+1))/integral >= 0.68 ) {
            bin68 = i+1;
            theta68 = h1->GetBinCenter(i+1);
            break;
        }
    }
    Double_t th68max = h1->GetBinLowEdge(bin68)+h1->GetBinWidth(bin68);
    TH1 *h68 = GetHisto("Theta68","Theta68","1D",bin68,0,th68max,0,0,0);
    for( Int_t i(0); i<bin68; i++ ) h68->SetBinContent(i+1,h1->GetBinContent(i+1));

    for (it=fCuts.begin(); it != fCuts.end(); it++) {
        TString name = "AbsThetaError" + it->first;
        TString title = "AbsThetaError " + it->first;
        TH1 *h = GetHisto(name+"_1D",title,"1D",0,0,0,0,0,0);
    // search theta68
    integral = h->Integral();
    theta68 = -1.;
    bin68 = -1;
    for( Int_t i(0); i<h->GetNbinsX(); i++ ) {
        if ( (h->Integral(1,i+1))/integral >= 0.68 ) {
            bin68 = i+1;
            theta68 = h->GetBinCenter(i+1);
            break;
        }
    }
    th68max = h->GetBinLowEdge(bin68)+h->GetBinWidth(bin68);
    name = "Theta68" + it->first;
    title = "Theta68 " + it->first;
    TH1 *h68 = GetHisto(name+"_1D",title,"1D",bin68,0,th68max,0,0,0);
    for( Int_t i(0); i<bin68; i++ ) h68->SetBinContent(i+1,h->GetBinContent(i+1));
    }
    
    // search phi68
    h1 = GetHisto("AbsPhiError","AbsPhiError","1D",0,0,0,0,0,0);
    integral = h1->Integral();
    Double_t phi68 = -1.;
    bin68 = -1;
    for( Int_t i(0); i<h1->GetNbinsX(); i++ ) {
        if ( (h1->Integral(1,i+1))/integral >= 0.68 ) {
            bin68 = i+1;
            phi68 = h1->GetBinCenter(i+1);
            break;
        }
    }
    Double_t phi68max = h1->GetBinLowEdge(bin68)+h1->GetBinWidth(bin68);
    h68 = GetHisto("Phi68","Phi68","1D",bin68,0,phi68max,0,0,0);
    for( Int_t i(0); i<bin68; i++ ) h68->SetBinContent(i+1,h1->GetBinContent(i+1));

    for (it=fCuts.begin(); it != fCuts.end(); it++) {
        TString name = "AbsPhiError" + it->first;
        TString title = "AbsPhiError" + it->first;
        TH1 *h = GetHisto(name+"_1D",title,"1D",0,0,0,0,0,0);
    // search phi68
    integral = h->Integral();
    phi68 = -1.;
    bin68 = -1;
    for( Int_t i(0); i<h->GetNbinsX(); i++ ) {
        if ( (h->Integral(1,i+1))/integral >= 0.68 ) {
            bin68 = i+1;
            phi68 = h->GetBinCenter(i+1);
            break;
        }
    }
    phi68max = h->GetBinLowEdge(bin68)+h->GetBinWidth(bin68);
    name = "Phi68" + it->first;
    title = "Phi68 " + it->first;
    TH1 *h68 = GetHisto(name+"_1D",title,"1D",bin68,0,phi68max,0,0,0);
    for( Int_t i(0); i<bin68; i++ ) h68->SetBinContent(i+1,h->GetBinContent(i+1));
    }

    // calculate the resolution in the direction
    for (it=fCuts.begin(); it != fCuts.end(); it++) {
        TString name = "DirError" + it->first;
        TString title = "DirError" + it->first;
        TH1 *h = GetHisto(name+"_1D",title,"1D",0,0,0,0,0,0);
    // search psi68
    integral = h->Integral();
    phi68 = -1.;
    bin68 = -1;
    for( Int_t i(0); i<h->GetNbinsX(); i++ ) {
        if ( (h->Integral(1,i+1))/integral >= 0.68 ) {
            bin68 = i+1;
            phi68 = h->GetBinCenter(i+1);
            break;
        }
    }
    phi68max = h->GetBinLowEdge(bin68)+h->GetBinWidth(bin68);
    name = "Psi68" + it->first;
    title = "Psi68 " + it->first;
    TH1 *h68 = GetHisto(name+"_1D",title,"1D",bin68,0,phi68max,0,0,0);
    for( Int_t i(0); i<bin68; i++ ) h68->SetBinContent(i+1,h->GetBinContent(i+1));
    }

}

//______________________________________________________________________________
void RecoTreePainter::FillDirectionHistos() {
    //
    //
    //
    map <string, Bool_t>::iterator it;
    
    if (!AcceptDirection()) return;

    for (it=fCuts.begin(); it != fCuts.end(); it++) {
         TString name  = "DirError_vs_Phi"  + it->first;
         TString title = "DirError_vs_Phi " + it->first;
         TH1* h = NULL;
     // Fill Direction Error vs Phi Simulated
     h = GetHisto(name+"_2D",title,"2D",100,left_edge["Phi"],right_edge["Phi"],
                                            100,0,180);
         if (it->second) h->Fill(fSim["PhiSim"]*RadToDeg(),fRec["DirectionError"]*RadToDeg());
     // Fill Dir Error
     name  = "DirError" + it->first;
     title = "DirError "+ it->first;
     h = GetHisto(name+"_1D",title,"1D",360,0,180,0,0,0);
     if (it->second) h->Fill(fRec["DirectionError"]*RadToDeg());
     // Fill Dir Error of vs Theta Simulated
     name  = "DirError_vs_Theta" + it->first;
     title = "DirError vs Theta"+ it->first;
     h = GetHisto(name+"_2D",title,"2D",100,left_edge["Theta"],right_edge["Theta"],
                                            100,0,180);
         if (it->second) h->Fill(fSim["ThetaSim"]*RadToDeg(),fRec["DirectionError"]*RadToDeg());
     // Fill Theta Error of vs signal/noise ratio
     name  = "DirError_vs_SignalToNoise" + it->first;
     title = "DirError vs S/N"+ it->first;
     h = GetHisto(name+"_2D",title,"2D",100,0,100,100,0,180);
         if (it->second) h->Fill(fRec["SignalToNoise"],fRec["DirectionError"]*RadToDeg());
     // Fill Theta Error vs Energy Simulated
     name  = "DirError_vs_EnergySim" + it->first;
     title = "DirError vs Energy"+ it->first;
     h = GetHisto(name+"_2D",title,"2D",100,left_edge["Energy"],right_edge["Energy"],100,0,180);
         if (it->second) h->Fill(fSim["EnergySim"],fRec["DirectionError"]*RadToDeg());
   }

}

//______________________________________________________________________________
void RecoTreePainter::FillEnergyHistos() {
    //
    //
    //
    
    map <string, Bool_t>::iterator it;
    
    if (!AcceptEnergy()) return;

    TH1* h1 = GetHisto("EnergySim_vs_Rec_2D","Energy Sim vs Rec","2D",100,left_edge["Energy"],right_edge["Energy"],
                                                                   100,left_edge["Energy"],right_edge["Energy"]);
    h1->SetXTitle("Simulated Energy/EeV");
    h1->SetYTitle("Reconstructed Energy/EeV");
    h1->Fill(fSim["EnergySim"],fRec["EnergyRec"]);

    for (it=fCuts.begin(); it != fCuts.end(); it++) {
        TString name  = "EnergySim_vs_Rec"  + it->first;
        TString title = "EnergySim_vs_Rec " + it->first;
        TH1* h = NULL;
        
        // Fill Energy Simulated vs Energy Reconstructed
        h = GetHisto(name+"_2D",title,"2D",100,left_edge["Energy"],right_edge["Energy"],
                100,left_edge["Energy"],right_edge["Energy"]);
        if (it->second) h->Fill(fSim["EnergySim"],fRec["EnergyRec"]);
        
        // Fill Relative Error of vs Energy Simulated
        name  = "EnergyErrorRelative_vs_Energy" + it->first;
        title = "EnergyError Relative vs Energy"+ it->first;
        h = GetHisto(name+"_2D",title,"2D",100,left_edge["Energy"],right_edge["Energy"],
                100,-2,2);
        if (it->second) h->Fill(fSim["EnergySim"],fRec["EnergyErrorRelative"]);
        
        // Fill Relative Error of vs signal/noise ratio
        name  = "EnergyErrorRelative_vs_SignalToNoise" + it->first;
        title = "EnergyError Relative vs S/N"+ it->first;
        h = GetHisto(name+"_2D",title,"2D",100,0,100,100,-2,2);
        if (it->second) h->Fill(fRec["SignalToNoise"],fRec["EnergyErrorRelative"]);
        
        // Fill signal/noise ratio vs Energy Simulated
        name  = "SignalToNoise_vs_EnergySim" + it->first;
        title = "S/N vs Energy"+ it->first;
        h = GetHisto(name+"_2D",title,"2D",100,left_edge["Energy"],right_edge["Energy"],100,0,100);
        if (it->second) h->Fill(fSim["EnergySim"],fRec["SignalToNoise"]);
        
        // Fill simulated Theta angle for this theta bin
        name  = "SimulatedTheta"+it->first;
        title = "Simulated Theta in "+it->first;
        h = GetHisto(name,title,"1D",100,left_edge["Theta"],right_edge["Theta"],0,0,0);
        if (it->second) h->Fill(fSim["ThetaSim"]*RadToDeg());
        
        // Fill simulated Phi angle for this theta bin
        name  = "SimulatedPhi"+it->first;
        title = "Simulated Phi in "+it->first;
        h = GetHisto(name,title,"1D",100,left_edge["Phi"],right_edge["Phi"],0,0,0);
        if (it->second) h->Fill(fSim["PhiSim"]*RadToDeg());
        
        // Fill reconstructed Energy  for this theta bin
        name  = "ReconstructedEnergy:"+it->first;
        title = "Reconstructed Energy in "+it->first;
        h = GetHisto(name,title,"1D",100,left_edge["Energy"],right_edge["Energy"],0,0,0);
        if (it->second) h->Fill(fRec["EnergyRec"]);
        
        // Fill simulated Energy  for this theta bin
        name  = "AcceptedEventsRatio:"+it->first;
        title = "Fraction of accepted events in "+it->first;
        h = GetHisto(name,title,"1D",100,left_edge["Energy"],right_edge["Energy"],0,0,0);
        //   if (it->second && TMath::Abs(fRec["EnergyErrorRelative"])<0.3) h->Fill(fSim["EnergySim"]);
        if (it->second ) h->Fill(fSim["EnergySim"]);
        
        // Fill statistical expectation for the resolution
        name  = "StatisticalExpectation"+it->first;
        title = "Statistical Expectation in "+it->first;
        h = GetHisto(name,title,"1D",100,0,100,0,0,0);
        if (it->second) h->Fill(fRec["SignalToNoise"],1./TMath::Sqrt(fRec["Signal"]));
    }


    TProfile2D* h2 = (TProfile2D*)GetHisto("EnergyError_prof2","Relative Energy Error","prof2",10,left_edge["Energy"],right_edge["Energy"],
                                                         10,left_edge["Theta"],right_edge["Theta"]);
    h2->SetXTitle("Simulated Energy/EeV");
    h2->SetYTitle("Simulated Theta/deg");
    h2->Fill(fSim["EnergySim"],fSim["ThetaSim"]*TMath::RadToDeg(),fRec["EnergyErrorRelative"]);

    TH2F* h21 = (TH2F*)GetHisto("EnergyRatio_2D","Reconstructed/Simulated","2D",100,left_edge["Energy"],right_edge["Energy"],100,0,3);
    h21->SetXTitle("Simulated Energy/EeV");
    h21->SetYTitle("Reconstructed/Simulated Energy");
    h21->Fill(fSim["EnergySim"],fRec["EnergyRec"]/fSim["EnergySim"]);

    TH1 *h4 = GetHisto("Altitude_vs_Theta","Altitude vs Theta","2D",100,left_edge["Theta"],right_edge["Theta"],100,0,25);
    h4->Fill(fSim["ThetaSim"]*RadToDeg(),fRec["Altitude"]);

    TH1 *h5 = GetHisto("Quality_vs_Energy","Quality vs Energy","2D",100,left_edge["Energy"],right_edge["Energy"],100,0,25);
    h5->Fill(fSim["EnergySim"],fRec["EnergyQuality"]);
}
//_____________________________________________________________________________
Bool_t RecoTreePainter::AcceptDirection() {
     //
     //
     //
     if ( fEvent->GetRecoTrackDirection2().GetQuality() < 0 )                                      return kFALSE;
     //if ( Sqrt(Power(fEvent->GetTruth().GetMaxPos().X(),2)+Power(fEvent->GetTruth().GetMaxPos().X(),2)) > 200 ) return kFALSE;
     //if ( fEvent->GetTruth().GetEnergy()<1.e14 ) return kFALSE;
     return kTRUE;
}

//_____________________________________________________________________________
Bool_t RecoTreePainter::AcceptEnergy() {
    //
    //
    //
     if ( !(AcceptDirection() || fEnergyIndependentFromDir ) ) return kFALSE;
     if (fEvent->GetRecoEnergy().GetEnergy() == 5.e14)                                            return kFALSE;
     if (fEvent->GetRecoEnergy().GetQuality() < 0 )     return kFALSE;
//     if (fEvent->GetRecoEnergy().GetQuality() < 0 || fEvent->GetRecoEnergy().GetQuality()> 30)     return kFALSE;
     if (fEvent->GetRecoEnergy().GetEnergy() <= 0 ) return kFALSE;
     return kTRUE;
}

//_____________________________________________________________________________
void RecoTreePainter::MakeBaseDataMap() {
    //
    //
    //
    fSim["EnergySim"] = (Double_t)Log10(fEvent->GetTruth().GetEnergy());
    fSim["ThetaSim"]  = (Double_t)fEvent->GetTruth().GetTheta();
    fSim["PhiSim"]    = (Double_t)fEvent->GetTruth().GetPhi();
    fSim["FOVDistance"] = (Double_t)Sqrt(Power(fEvent->GetTruth().GetMaxPos().X(),2)+Power(fEvent->GetTruth().GetMaxPos().Y(),2));

    if (!fUseAA1) {
        fRec["ThetaRec"]         = (Double_t)fEvent->GetRecoTrackDirection2().GetTheta()*DegToRad();
        fRec["PhiRec"]           = (Double_t)fEvent->GetRecoTrackDirection2().GetPhi()*DegToRad();
        fRec["ThetaError"]       = (Double_t)fEvent->GetRecoTrackDirection2().GetErrorTheta();
        fRec["PhiError"]         = (Double_t)fEvent->GetRecoTrackDirection2().GetErrorPhi();
        fRec["DirectionError"]   = (Double_t)fEvent->GetRecoTrackDirection2().GetErrorDirection()*DegToRad();
    } else {
        fRec["ThetaRec"]         = (Double_t)fEvent->GetRecoTrackDirection2().GetThetaAA1()*DegToRad();
        fRec["PhiRec"]           = (Double_t)fEvent->GetRecoTrackDirection2().GetPhiAA1()*DegToRad();
        fRec["ThetaError"]       = (Double_t)(fEvent->GetRecoTrackDirection2().GetThetaAA1() - fEvent->GetTruth().GetTheta()*RadToDeg());
        fRec["PhiError"]         = (Double_t)(fEvent->GetRecoTrackDirection2().GetPhiAA1() - fEvent->GetTruth().GetPhi()*RadToDeg());
        fRec["DirectionError"]   = (Double_t)fEvent->GetRecoTrackDirection2().GetErrorDirectionAA1()*DegToRad();
    }
    if (fRec["ThetaError"]>180) fRec["ThetaError"] -= 360.;
    if (fRec["ThetaError"]<-180) fRec["ThetaError"] += 360.;
    fRec["ThetaError"] *= DegToRad();
    if (fRec["PhiError"]>180) fRec["PhiError"] -= 360.;
    if (fRec["PhiError"]<-180) fRec["PhiError"] += 360.;
    fRec["PhiError"] *= DegToRad();
    fRec["TDPError"]         = (Double_t)fEvent->GetRecoTrackDirection2().GetErrorTDP()*DegToRad();

    fRec["EnergyRec"]           = (Double_t)TMath::Log10(fEvent->GetRecoEnergy().GetEnergy());
    fRec["EnergyError"]         = (Double_t)fEvent->GetRecoEnergy().GetEnergyMCError();
    fRec["EnergyErrorRelative"] = (Double_t)fEvent->GetRecoEnergy().GetEnergyMCError()/fEvent->GetTruth().GetEnergy();
    fRec["EnergyQuality"]       = (Double_t)fEvent->GetRecoEnergy().GetQuality();
    fRec["SignalToNoise"]       = (Double_t)fEvent->GetRecoEnergy().GetSignalCluster()/TMath::Sqrt(fEvent->GetRecoEnergy().GetNoiseCluster());
    fRec["Signal"]              = (Double_t)fEvent->GetRecoEnergy().GetSignalCluster();
    fRec["Altitude"]            = (Double_t)fEvent->GetRecoEnergy().GetAltitude()/km;

    // seting up the cuts
    fCuts["0.lt.Theta.lt.15"]  = 0*DegToRad()  < fEvent->GetTruth().GetTheta() && fEvent->GetTruth().GetTheta() <= 15*DegToRad();
    fCuts["15.lt.Theta.lt.30"] = 15*DegToRad() < fEvent->GetTruth().GetTheta() && fEvent->GetTruth().GetTheta() <= 30*DegToRad();
    fCuts["30.lt.Theta.lt.45"] = 30*DegToRad() < fEvent->GetTruth().GetTheta() && fEvent->GetTruth().GetTheta() <= 45*DegToRad();
    fCuts["45.lt.Theta.lt.60"] = 45*DegToRad() < fEvent->GetTruth().GetTheta() && fEvent->GetTruth().GetTheta() <= 60*DegToRad();
    fCuts["60.lt.Theta.lt.75"] = 60*DegToRad() < fEvent->GetTruth().GetTheta() && fEvent->GetTruth().GetTheta() <= 75*DegToRad();
    fCuts["75.lt.Theta.lt.102"] = 75*DegToRad() < fEvent->GetTruth().GetTheta() && fEvent->GetTruth().GetTheta() <= 102*DegToRad();
}

//_____________________________________________________________________________
void RecoTreePainter::DrawAll() {
    //
    //
    //
    DrawTheta();
    DrawPhi();
    //DrawDirection();
    DrawEnergy();
}

//_____________________________________________________________________________
void RecoTreePainter::DrawTheta() {
    //
    //
    //
    fThetaViewer = new EGViewer();
    fThetaViewer->SetName("ThetaReconstuction");
    fThetaViewer->SetTitle("Theta Reconstuction");
    fThetaViewer->SetWindowName("Theta Reconstuction");

    map <string, Bool_t>::iterator it;
    TString cname = "Theta ";
    TCanvas *c1 = fThetaViewer->FindCanvas(cname);
    if(!c1 )  c1 = fThetaViewer->AddTab(cname,cname);
    c1->Clear();
    c1->cd();
    c1->Divide(3,3);
    // draw simulated theta distribution: all events and superimposed in the given bin
    c1->cd(1);
    TH1* h1 = GetHisto("ThetaSimulated_1D","Theta Simulated","1D",0,0,0,0,0,0);
    TH1* h2 = GetHisto("ThetaError","","1D",0,0,0,0,0,0);
    h1->Draw();
    TPaveText *summary = new TPaveText(0.65,0.75,0.98,0.98,"brNDC");
    summary->SetBorderSize(1);
    summary->SetTextAlign(12);
    summary->AddText("Events");
    summary->AddText(Form("%d simulated",(Int_t)h1->GetEntries()));
    summary->AddText(Form("%d reconstructed",(Int_t)h2->GetEntries()));
    summary->Draw();




    // draw simulated theta vs reconstructed theta
    c1->cd(2);
    TH1* h3 = GetHisto("ThetaSim_vs_Rec_2D","","2D",0,0,0,0,0,0);
    h3->Draw("colz");
    // draw theta error
    c1->cd(4);
    h2->Draw();
    // draw theta abs error
    TVirtualPad *p51 = c1->cd(5);
    p51->SetLogy();
    TH1* h21 = GetHisto("AbsThetaError","","1D",0,0,0,0,0,0);
    h21->Draw();
    TH1* h22 = GetHisto("Theta68","","1D",0,0,0,0,0,0);
    h22->SetFillColor(kRed);
    h22->Draw("same");

    // draw Theta Error of vs Theta Simulated
    c1->cd(3);
    TH1* h4 = GetHisto("ThetaError_vs_Theta_2D","","2D",0,0,0,0,0,0);
    h4->Draw("colz");
    // draw Theta Error of vs Theta Simulated
    c1->cd(6);
    TH1* h5 = GetHisto("ThetaError_vs_Phi_2D","","2D",0,0,0,0,0,0);
    h5->Draw("colz");
    // draw Theta Error of vs S/N
    c1->cd(7);
    TH1* h6 = GetHisto("ThetaError_vs_SignalToNoise_2D","","2D",0,0,0,0,0,0);
    h6->Draw("colz");
    // draw Theta Error of vs Energy
    c1->cd(8);
    TH1* h7 = GetHisto("ThetaError_vs_EnergySim_2D","","2D",0,0,0,0,0,0);
    h7->Draw("colz");
    // draw Theta Error of vs Energy
    c1->cd(9);
    TH1* h8 = GetHisto("ThetaError_vs_Distance_2D","","2D",0,0,0,0,0,0);
    h8->Draw("colz");

    Int_t cuts_size = fCuts.size();
    Double_t effvstheta[(const Int_t)cuts_size]; 
    Int_t i(0);
    for (it=fCuts.begin(); it != fCuts.end(); it++) {
        TCanvas *c = fThetaViewer->FindCanvas(cname+(it->first).c_str());
        if(!c )  c = fThetaViewer->AddTab(cname+(it->first).c_str(),cname+(it->first).c_str());
        c->Clear();
        c->cd();
        c->Divide(3,3);
        TString name, title;
        // draw simulated theta distribution: all events and superimposed in the given bin
        c->cd(1);
        TH1* h1 = GetHisto("ThetaSimulated_1D","Theta Simulated","1D",0,0,0,0,0,0);
        h1->DrawCopy();

        name = "ThetaSimulated"  + it->first;
        TH1* hTot = GetHisto(name+"_1D","","1D",0,0,0,0,0,0);
        TH1* hCopyTot = hTot->DrawCopy("same");
        hCopyTot->SetFillColor(kBlue);
        hCopyTot->SetLineColor(kBlue);
        
        name  = "SimulatedTheta"+it->first;
        TH1* h2 = GetHisto(name,"","1D",0,0,0,0,0,0);
        TH1* hCopyRe = h2->DrawCopy("same");
        hCopyRe->SetXTitle(" #theta_{sim} [deg]");
        hCopyRe->SetFillColor(kRed);
        hCopyRe->SetLineColor(kRed);
        name = "ThetaError" + it->first;
        TH1 *hT = GetHisto(name+"_1D","","1D",0,0,0,0,0,0);
        
        TLegend *leg = new TLegend(0.65,0.75,0.98,0.98,"Events","brNDC");
        leg->SetBorderSize(1);
        leg->SetTextAlign(12);
        leg->AddEntry(hCopyTot,Form("%d simulated",(Int_t)hCopyTot->GetEntries()),"F");
        leg->AddEntry(hCopyRe,Form("%d reconstructed",(Int_t)hT->GetEntries()),"F");
        leg->SetBit(kCanDelete);
        leg->Draw();

        //reco eff
        effvstheta[i++] = (Double_t)hT->GetEntries()/hCopyTot->GetEntries(); 

        // draw theta error
        c->cd(4);

        hT->Draw();
        hT->SetXTitle("#Delta #theta [deg]");
        hT->SetYTitle("Entries");

        // draw abs theta error
        TVirtualPad *p5 = c->cd(5);
        p5->SetLogy();
        name = "AbsThetaError" + it->first;
        TH1 *hTabs = GetHisto(name+"_1D","","1D",0,0,0,0,0,0);
        hTabs->Draw();
        hTabs->SetXTitle("|#Delta #theta| [deg]");
        hTabs->SetYTitle("Entries");
        name = "Theta68" + it->first;
        TH1 *h68 = GetHisto(name+"_1D","","1D",0,0,0,0,0,0);
        h68->SetFillColor(kRed);
        h68->Draw("same");

        // draw simulated theta vs reconstructed theta
        c->cd(2);
        name  = "ThetaSim_vs_Rec"  + it->first;
        TH1* h3 = GetHisto(name+"_2D","","2D",0,0,0,0,0,0);
        h3->Draw("colz");
        h3->SetXTitle("#theta_{sim} [deg]");
        h3->SetYTitle("#theta_{rec} [deg]");
        // draw Theta Error of vs Theta Simulated
        c->cd(3);
        name  = "ThetaError_vs_Theta" + it->first;
        TH1* h4 = GetHisto(name+"_2D","","2D",0,0,0,0,0,0);
        h4->SetXTitle("#theta_{sim} [deg]");
        h4->SetYTitle("#Delta #theta [deg]");
        h4->Draw("colz");
        // draw Theta Error of vs Theta Simulated
        c->cd(6);
        name  = "ThetaError_vs_Phi" + it->first;
        TH1* h5 = GetHisto(name+"_2D","","2D",0,0,0,0,0,0);
        h5->SetXTitle("#varphi_{sim} [deg]");
        h5->SetYTitle("#Delta #theta [deg]");
        h5->Draw("colz");
        // draw Theta Error of vs S/N
        c->cd(7);
        name  = "ThetaError_vs_SignalToNoise" + it->first;
        TH1* h6 = GetHisto(name+"_2D","","2D",0,0,0,0,0,0);
        h6->SetXTitle("S/N");
        h6->SetYTitle("#Delta #theta [deg]");
        h6->Draw("colz");
        // draw Theta Error of vs Energy
        c->cd(8);
        name  = "ThetaError_vs_EnergySim" + it->first;
        TH1* h7 = GetHisto(name+"_2D","","2D",0,0,0,0,0,0);
        h7->SetXTitle("Simulated log_{10}Energy");
        h7->SetYTitle("#Delta #theta [deg]");
        h7->Draw("colz");
        // draw Theta Error of vs Energy
        c->cd(9);
        name  = "ThetaError_vs_Distance" + it->first;
        TH1* h8 = GetHisto(name+"_2D","","2D",0,0,0,0,0,0);
        h8->SetXTitle("Distance in FOV [km]");
        h8->SetYTitle("#Delta #theta [deg]");
        h8->Draw("colz");

    }

    cname = "Summary";
    TCanvas *c2 = fThetaViewer->FindCanvas(cname);
    if(!c2 )  c2 = fThetaViewer->AddTab(cname,cname);
    c2->Clear();
    c2->cd();
    c2->Divide(2,2);

    Double_t theta[] = {0.,15.,30.,45.,60.,75.};
    Double_t theta68[(const Int_t)cuts_size];
    Double_t psi68[(const Int_t)cuts_size];
    i = 0;
    for (it=fCuts.begin(); it != fCuts.end(); it++) {
        TString hname = "Theta68" + it->first;
        TH1* histo = GetHisto(hname+"_1D","","1D",0,0,0,0,0,0);
        Int_t nbins = histo->GetNbinsX();
        theta68[i] = histo->GetBinCenter(nbins);
        hname = "Psi68" + it->first;
        histo = GetHisto(hname+"_1D","","1D",0,0,0,0,0,0);
        nbins = histo->GetNbinsX();
        psi68[i] = histo->GetBinCenter(nbins);
        i++;
    }

    c2->cd(1);
    TGraph *dirRes = new TGraph(6,theta,psi68);
    dirRes->SetName("DirectionResolution");
    dirRes->SetTitle("Direction Resolution");
    dirRes->GetXaxis()->SetTitle("#theta [deg]");
    dirRes->GetYaxis()->SetTitle("#psi_{68} [deg]");
    dirRes->SetMarkerStyle(22);
    dirRes->SetMarkerColor(kRed);
    dirRes->Draw("AP");

    c2->cd(2);
    TGraph *thRes = new TGraph(6,theta,theta68);
    thRes->SetName("ThetaResolution");
    thRes->SetTitle("Theta Resolution");
    thRes->GetXaxis()->SetTitle("#theta [deg]");
    thRes->GetYaxis()->SetTitle("#theta_{68} [deg]");
    thRes->SetMarkerStyle(21);
    thRes->SetMarkerColor(kBlue);
    thRes->Draw("AP");

    // reconstruction efficiency vs theta
    c2->cd(3);
    TGraph *effTheta = new TGraph(6,theta,effvstheta);
    effTheta->SetName("RecoEfficiency");
    effTheta->SetTitle("Reconstruction Efficiency vs #theta");
    effTheta->GetXaxis()->SetTitle("#theta [deg]");
    effTheta->GetYaxis()->SetTitle("Eff.");
    effTheta->SetMarkerStyle(22);
    effTheta->SetMarkerColor(kBlue);
    effTheta->Draw("AP");

}

//_____________________________________________________________________________
void RecoTreePainter::DrawPhi() {
    //
    //
    //
    fPhiViewer = new EGViewer();
    fPhiViewer->SetName("PhiReconstuction");
    fPhiViewer->SetTitle("Phi Reconstuction");
    fPhiViewer->SetWindowName("Phi Reconstuction");

    map <string, Bool_t>::iterator it;
    TString cname = "Phi ";
    TCanvas *c1 = fPhiViewer->FindCanvas(cname);
    if(!c1 )  c1 = fPhiViewer->AddTab(cname,cname);
    c1->Clear();
    c1->cd();
    c1->Divide(3,3);
    // draw simulated phi distribution: all events and superimposed in the given bin
    c1->cd(1);
    TH1* h1 = GetHisto("PhiSimulated_1D","Phi Simulated","1D",0,0,0,0,0,0);
    h1->Draw();
    TH1* h2 = GetHisto("PhiError","","1D",0,0,0,0,0,0);
    TPaveText *summary = new TPaveText(0.65,0.75,0.98,0.98,"brNDC");
    summary->SetBorderSize(1);
    summary->SetTextAlign(12);
    summary->AddText("Events");
    summary->AddText(Form("%d simulated",(Int_t)h1->GetEntries()));
    summary->AddText(Form("%d reconstructed",(Int_t)h2->GetEntries()));
    summary->Draw();
    // draw simulated phi vs reconstructed theta
    c1->cd(2);
    TH1* h3 = GetHisto("PhiSim_vs_Rec_2D","","2D",0,0,0,0,0,0);
    h3->Draw("colz");
    // draw phi error
    c1->cd(4);
    h2->Draw();
    // draw phi abs error
    TVirtualPad *p51 = c1->cd(5);
    p51->SetLogy();
    TH1* h21 = GetHisto("AbsPhiError","","1D",0,0,0,0,0,0);
    h21->Draw();
    TH1* h22 = GetHisto("Phi68","","1D",0,0,0,0,0,0);
    h22->SetFillColor(kRed);
    h22->Draw("same");

    // draw phi Error of vs Theta Simulated
    c1->cd(3);
    TH1* h4 = GetHisto("PhiError_vs_Theta_2D","","2D",0,0,0,0,0,0);
    h4->Draw("colz");
    // draw phi Error of vs Phi Simulated
    c1->cd(6);
    TH1* h5 = GetHisto("PhiError_vs_Phi_2D","","2D",0,0,0,0,0,0);
    h5->Draw("colz");
    // draw phi Error of vs S/N
    c1->cd(7);
    TH1* h6 = GetHisto("PhiError_vs_SignalToNoise_2D","","2D",0,0,0,0,0,0);
    h6->Draw("colz");
    // draw Phi Error of vs Energy
    c1->cd(8);
    TH1* h7 = GetHisto("PhiError_vs_EnergySim_2D","","2D",0,0,0,0,0,0);
    h7->Draw("colz");
    // draw Phi Error of vs Energy
    c1->cd(9);
    TH1* h8 = GetHisto("PhiError_vs_Distance_2D","","2D",0,0,0,0,0,0);
    h8->Draw("colz");

    for (it=fCuts.begin(); it != fCuts.end(); it++) {
        TCanvas *c = fPhiViewer->FindCanvas((it->first).c_str());
        if(!c )  c = fPhiViewer->AddTab(cname+(it->first).c_str(),cname+(it->first).c_str());
        c->Clear();
        c->cd();
        c->Divide(3,3);
        TString name, title;

        // draw simulated phi distribution: all events and superimposed in the given bin
        c->cd(1);
        TH1* h1 = GetHisto("ThetaSimulated_1D","Theta Simulated","1D",0,0,0,0,0,0);
        h1->DrawCopy();

        name = "ThetaSimulated"  + it->first;
        TH1* hTot = GetHisto(name+"_1D","","1D",0,0,0,0,0,0);
        TH1* hCopyTot = hTot->DrawCopy("same");
        hCopyTot->SetFillColor(kBlue);
        hCopyTot->SetLineColor(kBlue);

        name  = "SimulatedTheta"+it->first;
        TH1* h2 = GetHisto(name,"","1D",0,0,0,0,0,0);
        TH1* hCopyRe = h2->DrawCopy("same");
        hCopyRe->SetXTitle(" #theta_{sim} [deg]");
        hCopyRe->SetFillColor(kRed);
        hCopyRe->SetLineColor(kRed); 

        TLegend *leg = new TLegend(0.65,0.75,0.98,0.98,"Events","brNDC");
        leg->SetBorderSize(1);
        leg->SetTextAlign(12);
        leg->AddEntry(hCopyTot,Form("%d simulated",(Int_t)hCopyTot->GetEntries()),"F");
        leg->AddEntry(hCopyRe,Form("%d reconstructed",(Int_t)hCopyRe->GetEntries()),"F");
        leg->SetBit(kCanDelete);
        leg->Draw();   

        // draw simulated phi vs reconstructed phi
        c->cd(2);
        name  = "PhiSim_vs_Rec"  + it->first;
        TH1* h3 = GetHisto(name+"_2D","","2D",0,0,0,0,0,0);
        h3->Draw("colz");
        h3->SetXTitle("#varphi_{sim} [deg]");
        h3->SetYTitle("#varphi_{rec} [deg]");
        
        // draw theta error
        c->cd(4);
        name = "PhiError" + it->first;
        TH1 *hT = GetHisto(name+"_1D","","1D",0,0,0,0,0,0);
        hT->Draw();
        hT->SetXTitle("#Delta #varphi [deg]");
        hT->SetYTitle("Entries");
        
        // draw abs theta error
        TVirtualPad *p5 = c->cd(5);
        p5->SetLogy();
        name = "AbsPhiError" + it->first;
        TH1 *hTabs = GetHisto(name+"_1D","","1D",0,0,0,0,0,0);
        hTabs->Draw();
        hTabs->SetXTitle("|#Delta #varphi| [deg]");
        hTabs->SetYTitle("Entries");
        name = "Phi68" + it->first;
        TH1 *h68 = GetHisto(name+"_1D","","1D",0,0,0,0,0,0);
        h68->SetFillColor(kRed);
        h68->Draw("same");
        
        // draw phi Error of vs Theta Simulated
        c->cd(3);
        name  = "PhiError_vs_Theta" + it->first;
        TH1* h4 = GetHisto(name+"_2D","","2D",0,0,0,0,0,0);
        h4->SetXTitle("#theta_{sim} [deg]");
        h4->SetYTitle("#Delta #varphi [deg]");
        h4->Draw("colz");
        
        // draw phi Error of vs Theta Simulated
        c->cd(6);
        name  = "PhiError_vs_Phi" + it->first;
        TH1* h7 = GetHisto(name+"_2D","","2D",0,0,0,0,0,0);
        h7->SetXTitle("#varphi_{sim} [deg]");
        h7->SetYTitle("#Delta #varphi [deg]");
        h7->Draw("colz");
        
        // draw Phi Error of vs S/N
        c->cd(7);
        name  = "PhiError_vs_SignalToNoise" + it->first;
        TH1* h5 = GetHisto(name+"_2D","","2D",0,0,0,0,0,0);
        h5->SetXTitle("S/N");
        h5->SetYTitle("#Delta #varphi [deg]");
        h5->Draw("colz");
        
        // draw Theta Error of vs Energy
        c->cd(8);
        name  = "PhiError_vs_EnergySim" + it->first;
        TH1* h6 = GetHisto(name+"_2D","","2D",0,0,0,0,0,0);
        h6->SetXTitle("Simulated log_{10}Energy");        h6->SetYTitle("#Delta #varphi [deg]");
        h6->Draw("colz");
        
        // draw Theta Error of vs Energy
        c->cd(9);
        name  = "PhiError_vs_Distance" + it->first;
        TH1* h8 = GetHisto(name+"_2D","","2D",0,0,0,0,0,0);
        h8->SetXTitle("Distance in FOV [km]");
        h8->SetYTitle("#Delta #theta [deg]");
        h8->Draw("colz");
    }
    cname = "Summary";
    TCanvas *c2 = fPhiViewer->FindCanvas(cname);
    if(!c2 )  c2 = fPhiViewer->AddTab(cname,cname);
    c2->Clear();
    c2->cd();
    c2->Divide(2,2);

    Int_t cuts_size = fCuts.size();
    Double_t theta[] = {0.,15.,30.,45.,60.,75.};
    Double_t phi68[(const Int_t)cuts_size];
    Double_t psi68[(const Int_t)cuts_size];
    Int_t i = 0;
    for (it=fCuts.begin(); it != fCuts.end(); it++) {
        TString hname = "Phi68" + it->first;
        TH1* histo = GetHisto(hname+"_1D","","1D",0,0,0,0,0,0);
        Int_t nbins = histo->GetNbinsX();
        phi68[i] = histo->GetBinCenter(nbins);
        hname = "Psi68" + it->first;
        histo = GetHisto(hname+"_1D","","1D",0,0,0,0,0,0);
        nbins = histo->GetNbinsX();
        psi68[i] = histo->GetBinCenter(nbins);
        i++;
    }
    
    c2->cd(1);
    TGraph *dirRes = new TGraph(6,theta,psi68);
    dirRes->SetName("DirectionResolution");
    dirRes->SetTitle("Direction Resolution");
    dirRes->GetXaxis()->SetTitle("#theta [deg]");
    dirRes->GetYaxis()->SetTitle("#psi_{68} [deg]");
    dirRes->SetMarkerStyle(22);
    dirRes->SetMarkerColor(kRed);
    dirRes->Draw("AP");

    c2->cd(2);    
    TGraph *phiRes = new TGraph(6,theta,phi68);
    phiRes->SetName("PhiResolution");
    phiRes->SetTitle("Phi Resolution");
    phiRes->GetXaxis()->SetTitle("#theta [deg]");
    phiRes->GetYaxis()->SetTitle("#varphi_{68} [deg]");
    phiRes->SetMarkerStyle(21);
    phiRes->SetMarkerColor(kBlue);
    phiRes->Draw("AP");

}

//_____________________________________________________________________________
void RecoTreePainter::DrawEnergy() {
    //
    //
    //
    fEnergyViewer = new EGViewer();
    fEnergyViewer->SetName("EnergyReconstuction");
    fEnergyViewer->SetTitle("Energy Reconstuction");
    fEnergyViewer->SetWindowName("Energy Reconstuction");

    map <string, Bool_t>::iterator it;
    for (it=fCuts.begin(); it != fCuts.end(); it++) {
        TCanvas *c = fEnergyViewer->FindCanvas((it->first).c_str());
        if(!c )  c = fEnergyViewer->AddTab((it->first).c_str(),(it->first).c_str());
        c->Clear();
        c->cd();
        c->Divide(3,3);
        TString name, title;

        //
        // draw simulated theta distribution: all events and superimposed in
        // the given bin
        //
        c->cd(1);
        TH1* h1 = GetHisto("ThetaSimulated_1D","Theta Simulated","1D",0,0,0,0,0,0);
        h1->DrawCopy();
        
        name = "ThetaSimulated"  + it->first;
        TH1* hTot = GetHisto(name+"_1D","","1D",0,0,0,0,0,0);        
        TH1* hCopyTot = hTot->DrawCopy("same");
        hCopyTot->SetFillColor(kBlue);
        hCopyTot->SetLineColor(kBlue);


        name  = "SimulatedTheta"+it->first;
        TH1* h2 = GetHisto(name,"","1D",0,0,0,0,0,0);
        TH1* hCopyRe = h2->DrawCopy("same");
        hCopyRe->SetXTitle("Simulated #{Theta}/deg");
        hCopyRe->SetFillColor(kRed);
        hCopyRe->SetLineColor(kRed);

        name  = "EnergySim_vs_Rec"  + it->first;
        TH1* h3 = GetHisto(name+"_2D","","2D",0,0,0,0,0,0);
        TLegend *legRec = new TLegend(0.65,0.75,0.98,0.98,"Entries","brNDC");
        legRec->SetBorderSize(1);
        legRec->SetTextAlign(12);
        legRec->AddEntry(hCopyTot,Form("%d simulated",(Int_t)hCopyTot->GetEntries()),"F");
        legRec->AddEntry(hCopyRe,Form("%d reconstructed",(Int_t)h3->GetEntries()),"F");
        legRec->SetBit(kCanDelete);
        legRec->Draw();

        // 
        // draw simulated energy vs reconstructed energy
        //
        c->cd(2);

        h3->Draw("colz");
        h3->SetXTitle("Simulated log_{10}Energy");
        h3->SetYTitle("Reconstructed log_{10}Energy");


        //
        // draw Relative Error of vs Energy Simulated
        //
        c->cd(5);
        name  = "EnergyErrorRelative_vs_Energy" + it->first;
        TH1* h4 = GetHisto(name+"_2D","","2D",0,0,0,0,0,0);
        h4->SetXTitle("Simulated log_{10}Energy");
        h4->SetYTitle("#Delta E/E");
        h4->Draw("colz");

        //
        // draw simulated energy superimposed by reconstructed energy
        // 
        c->cd(4);
        name = "EnergySimulated"  + it->first;
        TH1* h5 = GetHisto(name+"_1D","","1D",0,0,0,0,0,0);
        h5->Draw();
        h5->SetLineWidth(2);
        h5->SetMinimum(0);
        name  = "ReconstructedEnergy:"+it->first;
        TH1* h6 = GetHisto(name,"","1D",0,0,0,0,0,0);
        h6->SetMarkerStyle(29);
        h6->SetMarkerSize(0.3);
        h6->Draw("EPsame");
        h6->SetXTitle("log_{10}Energy");
        h6->SetLineColor(kRed);
        h5->Rebin(4);
        h6->Rebin(4);

        //
        // draw efficiency as a function of energy
        // 
        c->cd(3);
        name = "AcceptedEventsRatio:"+it->first;
        TH1* h71 = GetHisto(name,"","1D",0,0,0,0,0,0);
        h71->Rebin(4);
        h71->Divide(h5);
        h71->SetMarkerStyle(29);
        h71->SetMarkerColor(kBlue);
        h71->Draw("P");
        h71->SetXTitle("log_{10}Energy");
        h71->SetLineColor(kBlue);
        h71->SetMaximum(1.1);


        // 
        // draw Relative Error of vs Signal/Noise ratio
        //
        c->cd(6);
        name  = "EnergyErrorRelative_vs_SignalToNoise" + it->first;
        title = "EnergyError Relative vs S/N"+ it->first;
        TH2D* h8 = (TH2D*)GetHisto(name+"_2D",title,"2D",0,0,0,0,0,0);
        h8->Draw("colz");
        h8->SetXTitle("signal/noise");
        h8->SetYTitle("#Delta E/E");

        // 
        // draw Signal/Noise ratio as a function of Energy simulated
        //
        c->cd(7);
        name  = "SignalToNoise_vs_EnergySim" + it->first;
        title = "S/N vs Energy"+ it->first;
        TH2D* h9 = (TH2D*)GetHisto(name+"_2D",title,"2D",0,0,0,0,0,0);
        h9->Draw("colz");
        h9->SetXTitle("log_{10}Energy");
        h9->SetYTitle("signal/noise");

        // 
        // draw sigma and mean of the energy resolution as a function of
        // signal/noise
        //
        c->cd(8);
        name  = "EnergyErrorRelative_vs_SignalToNoise" + it->first;
        TH2D* h10 = (TH2D*)h8->Clone();
        h10->RebinX(4);
        h10->FitSlicesY();
        TH1D* h8_1 = (TH1D*)gDirectory->Get(name+"_2D_1");
        TH1D* h8_2 = (TH1D*)gDirectory->Get(name+"_2D_2");
        h8_1->SetMarkerColor(kBlue);
        h8_1->SetMarkerStyle(23);
        h8_1->SetMarkerSize(1.0);
        h8_1->SetXTitle("Signal/Noise");
        h8_1->SetYTitle("");
        h8_1->SetTitle("Energy resolution");
        h8_2->SetMarkerColor(kRed);
        h8_2->SetMarkerStyle(29);
        h8_2->SetMarkerSize(1.0);

        TLegend *leg = new TLegend(0.68,0.75,0.98,0.95);
        leg->SetBit(kCanDelete);
        leg->SetTextFont(72);

        if (h8_1) {
            h8_1->Draw("PE");
            h8_1->SetMaximum(2);
            h8_1->SetMinimum(-1);
            h8_1->SetStats(kFALSE);
            leg->AddEntry(h8_1,"#LTE#GT/E","p");
        }
        if (h8_2) {
            h8_2->Draw("PEsame");
            h8_2->SetStats(kFALSE);
            leg->AddEntry(h8_2,"#Delta E/E","p");
        }

        leg->SetBorderSize(1);
        leg->Draw();

        /*   name  = "StatisticalExpectation"+it->first;
             TH1* h11 = GetHisto(name,title,"1D",100,0,100,0,0,0);
             h11->Draw("same");
             h11->SetLineColor(kBlack);
         */

        // 
        // draw sigma and mean of the energy resolution as a function of energy
        // 
        c->cd(9);
        name  = "EnergyErrorRelative_vs_Energy" + it->first;
        TH2D* h12 = (TH2D*)GetHisto(name+"_2D",title,"2D",100,left_edge["Energy"],right_edge["Energy"], 100,-2,2);
        TH2D* h13 = (TH2D*)h12->Clone();
        h13->RebinX(10);
        h13->FitSlicesY();
        TH1D* h13_1 = (TH1D*)gDirectory->Get(name+"_2D_1");
        TH1D* h13_2 = (TH1D*)gDirectory->Get(name+"_2D_2");
        h13_1->SetMarkerColor(kBlue);
        h13_1->SetMarkerStyle(23);
        h13_1->SetXTitle("log_{10}Energy");
        h13_1->SetYTitle("");
        h13_1->SetTitle("Energy resolution");

        h13_2->SetMarkerColor(kRed);
        h13_2->SetMarkerStyle(29);

        leg = new TLegend(0.68,0.75,0.98,0.95);
        leg->SetBit(kCanDelete);
        leg->SetTextFont(72);

        if (h13_1) {
            h13_1->Draw("PE");
            h13_1->SetMaximum(2);
            h13_1->SetMinimum(-1);
            h13_1->SetStats(kFALSE);
            leg->AddEntry(h13_1,"#LTE#GT/E","p");
        }
        if (h13_2) {
            h13_2->Draw("PEsame");
            h13_2->SetStats(kFALSE);
            leg->AddEntry(h13_2,"#Delta E/E","p");
        }

        leg->SetBorderSize(1);
        leg->Draw();

    }
    TCanvas *c = fEnergyViewer->FindCanvas("c_summary");
    if(!c )  c = fEnergyViewer->AddTab("c_summary","Summary");
    c->Clear();
    c->cd();
    c->Divide(3,2);
    // draw sigma of energy resolution for all theta angles as a function of energy
    c->cd(1);
    TLegend *legend1 = new TLegend(0.7,0.6,0.95,0.98);
    legend1->SetTextFont(72);
    legend1->SetBit(kCanDelete);
    //    legend1->SetTextSize(0.02);
    Int_t cut(0);

    THStack* eRelErr = new THStack("EnergyErrorRelative_vs_Energy","#sigma of Energy Resolution");
    eRelErr->SetBit(kCanDelete);

    TH1* hCopy;
    for (it=fCuts.begin(); it != fCuts.end(); it++) {
        TString name  = "EnergyErrorRelative_vs_Energy" + it->first;
        TH1D* h13_2 = (TH1D*)gDirectory->Get(name+"_2D_2");
        cut++;
        hCopy = (TH1*)h13_2->Clone();
        hCopy->SetTitle("#sigma of Energy Resolution");
        hCopy->SetMarkerSize(0.5);
        hCopy->SetMarkerStyle(20+cut);
        hCopy->SetMarkerColor(cut+1);
        hCopy->SetStats(0);
        legend1->AddEntry(hCopy,TString(it->first).Data(),"p");
        eRelErr->Add(hCopy,"P");
    }
    eRelErr->Draw("nostack");
    eRelErr->GetXaxis()->SetTitle(hCopy->GetXaxis()->GetTitle());
    eRelErr->GetYaxis()->SetTitle(hCopy->GetYaxis()->GetTitle());
    legend1->SetBorderSize(1);
    legend1->Draw();

    // 
    // draw sigma of energy resolution for all theta angles as a function of signal/noise
    //
    c->cd(2);
    TLegend *legend2 = new TLegend(0.7,0.55,0.95,0.98);
    legend2->SetTextFont(72);
    legend2->SetBit(kCanDelete);
    //    legend1->SetTextSize(0.02);
    cut = 0;
    eRelErr = new THStack("EnergyErrorRelative_vs_SignalToNoise","#sigma of Energy Resolution");
    eRelErr->SetBit(kCanDelete);
    for (it=fCuts.begin(); it != fCuts.end(); it++) {
        TString name  = "EnergyErrorRelative_vs_SignalToNoise" + it->first;
        TH1D* h13_2 = (TH1D*)gDirectory->Get(name+"_2D_2");
        cut++;

        hCopy = (TH1*)h13_2->Clone();
        hCopy->SetTitle("#sigma of Energy Resolution");
        hCopy->SetMarkerSize(0.5);
        hCopy->SetStats(kFALSE);
        hCopy->SetMarkerStyle(20+cut);
        hCopy->SetMarkerColor(cut+1);
        legend2->AddEntry(hCopy,TString(it->first).Data(),"p");
        eRelErr->Add(hCopy);
    }
    eRelErr->Draw("nostack");
    eRelErr->GetXaxis()->SetTitle(hCopy->GetXaxis()->GetTitle());
    eRelErr->GetYaxis()->SetTitle(hCopy->GetYaxis()->GetTitle());
    legend2->SetBorderSize(1);
    legend2->Draw();

    //
    // draw efficiencies as a function of Energy for various bins in Theta
    //
    c->cd(3);
    TLegend *legend3 = new TLegend(0.7,0.6,0.95,0.98);
    legend3->SetTextFont(72);
    legend3->SetBit(kCanDelete);

    eRelErr = new THStack("AcceptedEventsRatio","Fraction of accepted events");
    eRelErr->SetBit(kCanDelete);

    cut = 0;
    for (it=fCuts.begin(); it != fCuts.end(); it++) {
        TString name = "AcceptedEventsRatio:"+it->first;
        TH1* h = GetHisto(name,"","1D",0,0,0,0,0,0);
        cut++;
        hCopy = (TH1*)h->Clone();
        hCopy->SetMarkerSize(0.5);
        hCopy->SetTitle("Fraction of accepted events");
        hCopy->SetStats(kFALSE);
        hCopy->SetMarkerStyle(20+cut);
        hCopy->SetMarkerColor(cut+1);
        legend3->AddEntry(hCopy,TString(it->first).Data(),"p");
        eRelErr->Add(hCopy,"P");
    }
    eRelErr->Draw("nostack");
    eRelErr->GetXaxis()->SetTitle(hCopy->GetXaxis()->GetTitle());
    eRelErr->GetYaxis()->SetTitle(hCopy->GetYaxis()->GetTitle());
    eRelErr->SetMaximum(eRelErr->GetMaximum("nostack"));
    eRelErr->SetMinimum(eRelErr->GetMinimum("nostack"));
    legend3->SetBorderSize(1);
    legend3->Draw();

    //
    // draw reconstructed altitude as a function of Theta
    //
    c->cd(4);
    TH2D* h;
    h = (TH2D*)GetHisto("Altitude_vs_Theta","Altitude vs Theta","2D",0,0,0,0,0,0);
    h->SetYTitle("Altitude [km]");
    h->SetXTitle("#Theta [deg]");
    TH1* hProf;
    hProf = h->ProfileX("_pfx",-1,-1,"s");
    hProf->SetYTitle("Altitude [km]");
    hProf->SetBit(kCanDelete);
    hProf->Draw();

    c->cd(5);
    h = (TH2D*)GetHisto("Quality_vs_Energy","","2D",0,0,0,0,0,0);
    h->SetYTitle("Quality");
    h->SetXTitle("#Theta [deg]");
    h->Rebin2D(5,5);
    h->Draw("box");

    //
    // summary: reconstructed events vs simulated events
    //
    TH1* h3 = GetHisto("EnergySimulated_1D","","1D",0,0,0,0,0,0);
    c->cd(6);
    TPaveText *summary = new TPaveText(0.0,0.0,1.,1.,"brNDC");
    summary->SetTextAlign(12);
    summary->SetFillStyle(0);
    summary->SetBorderSize(0);

    summary->AddText(Form("%d simulated events",(Int_t)h3->GetEntries()));
    summary->AddText(Form("%d reconstructed events",(Int_t)h->GetEntries()));
    summary->Draw();


}


//_____________________________________________________________________________
TH1* RecoTreePainter::GetHisto(const char *name, const char *title, Option_t *opt, Int_t xbins, Float_t xmin, Float_t xmax,
                                                                                   Int_t ybins, Float_t ymin, Float_t ymax) {
    //
    // 
    //
    TH1* th = (TH1*)fHistoList->FindObject(name);
    TString option(opt);
    if ( !title ) title = name;

    if (!th) {
        if      (option == "1D")    th = new TH1F( name, title, xbins, xmin, xmax);
        else if (option == "2D")    th = new TH2F( name, title, xbins, xmin, xmax, ybins, ymin, ymax);
        else if (option == "prof")  th = new TProfile( name, title, xbins, xmin, xmax);
        else if (option == "prof2") th = new TProfile2D( name, title, xbins, xmin, xmax,ybins, ymin, ymax,-2,2);
        else                              Error("GetHisto",Form("Option %s is not supported\n",opt));
        fHistoList->Add(th);
    }

    th->SetStats(kFALSE);
    return th;
}

//______________________________________________________________________________
void RecoTreePainter::SaveAsPs( Option_t* option, const char* prefix ) {
    TString opt(option);
    TString pfx = prefix;
    if ( !pfx.IsNull() ) pfx += "_";

    if ( opt.Contains("A") || opt.Contains("T") && fThetaViewer )
        fThetaViewer->SaveAsPS(pfx+"ThetaPlots.ps");
    if ( opt.Contains("A") || opt.Contains("P") && fPhiViewer )
        fPhiViewer->SaveAsPS(pfx+"PhiPlots.ps");
    if ( opt.Contains("A") || opt.Contains("E") && fEnergyViewer )
        fEnergyViewer->SaveAsPS(pfx+"EnergyPlots.ps");
}
