#ifndef __PARAMETRICSURFACE__
#define __PARAMETRICSURFACE__

#include "SSurface.h"
#include "FSIntersectedSegment.h"

#include <globals.hh>
#include <geomdefs.hh>
#include <G4ThreeVector.hh>
#include <float.h>
#include <vector>

namespace G4FresnelLens{
class ParametricSurface : public SSurface {
public:
    ParametricSurface(double z=0., double rho=0);
    virtual ~ParametricSurface();

    void ShiftSurfaceTo(int n, int tn, int* idx, double* rho, double* z, bool minimum);
    void ShiftSurfaceTo(const std::vector<SSurface*>& surfaces, bool minimum);
    virtual void Configure()=0;
    virtual double PointOnSurf(double x)=0;
    virtual double PointOnSurf2(double x);
    virtual double DistanceToSurf(const G4ThreeVector& p, const G4ThreeVector& v)=0;
    double DistanceToSurf(const G4ThreeVector& p, const G4ThreeVector& v,FSIntersectedSegment* is);
    virtual double DistanceToSurfSafe(const G4ThreeVector& p)=0;
    virtual bool Normal(FSIntersectedSegment* ints, const G4ThreeVector& p, double up, G4ThreeVector* norm)=0;
    virtual bool Point(int i, double step, double& rho, double& z);

    void SetR(double r) { fR=r; Configure(); }
    void SetZ(double z) { fZpos=z; Configure(); }
    void ShiftZ(double dz) { fZpos+=dz; Configure(); }

    virtual void print(){ printf("z=%g r=%g\n", fZpos, fR); }
    void selfCheck(int n=1000);

    // plugs
    virtual void FillSegment(double rho, FSIntersectedSegment *intS)=0;
    virtual EInside Inside(double rho, double z, FSIntersectedSegment *intS=0);
    virtual double FindNearestIntersection(const G4ThreeVector& p1,
                                           const G4ThreeVector& v,
                                           FSIntersectedSegment* intS=0);
    double TestEdgeTolerance(int dir, const G4ThreeVector& p, const G4ThreeVector& v);
    virtual double Derivative(double rho);

    double CheckDistance(const G4ThreeVector& p1, const G4ThreeVector& v);

    #ifdef USE_ROOT
    void draw(const char* rootfile, const char* canvasfile=0, double shift=0., const char* prefix="", const char* option="RECREATE");
    #endif
protected:
    virtual void GuessParameters(int n, int tn, int* idx, double* rho, double* z, bool minimum)=0;
    virtual void GuessParameters(const std::vector<SSurface*>& surfaces, bool minimum){};
    static double fSafetyGap;

private:

};

//__________________________________________________________________________________
inline double ParametricSurface::PointOnSurf2(double x2){
    return PointOnSurf(sqrt(x2));
}

//__________________________________________________________________________________
inline double ParametricSurface::FindNearestIntersection(const G4ThreeVector& p1, const G4ThreeVector& v, FSIntersectedSegment* intSeg){
    if ( intSeg ) FillSegment(-1., intSeg);
    #ifdef DEBUG
    return CheckDistance(p1,v);
    #endif
    return DistanceToSurf(p1,v);
}

//______________________________________________________________________________________________________________________
inline double ParametricSurface::TestEdgeTolerance(int dir, const G4ThreeVector& p, const G4ThreeVector& v){
    return DistanceToSurf(p,v);
}

//______________________________________________________________________________________________________________________
inline double ParametricSurface::Derivative(double rho){
    const double step=.5*um;
    double r1 = rho-step;
    double r2 = rho+step;
    double z1 = PointOnSurf(r1);
    double z2 = PointOnSurf(r2);
    double dz=z2-z1;
    double drho=r2-r1;
    return drho?dz/drho:DBL_MAX*dz;
}

//______________________________________________________________________________________________________________________
inline double ParametricSurface::DistanceToSurf(const G4ThreeVector& p, const G4ThreeVector& v,FSIntersectedSegment* intSeg){
    FillSegment(-1., intSeg);
    return DistanceToSurf(p, v);
}

////______________________________________________________________________________________________________________________
//inline void ParametricSurface::FillSegment(double rho, FSIntersectedSegment *intSeg){
    //const double step=.5*um;
    //intSeg->type = kfsCone;
    //intSeg->r1 = rho-step;
    //intSeg->r2 = rho+step;
    //intSeg->z1 = PointOnSurf(intSeg->r1);
    //intSeg->z2 = PointOnSurf(intSeg->r2);
    //intSeg->pars=0;
//
    //intSeg->surf=0;
    //intSeg->perp=rho;
//}

//______________________________________________________________________________________________________________________
inline EInside ParametricSurface::Inside(double rho, double z, FSIntersectedSegment *intSeg){
    double zs=PointOnSurf(rho);
    double tga=Derivative(rho);                // tg(a)
    double dz=kHalfTolerance*sqrt(1.+tga*tga); // kHTol/cos(a)

    if ( z<zs-dz ) {
        #ifdef DEBUG
        printf("%g, %g, diff=%g, inside\n", z, zs, z-zs);
        #endif
        return kInside;
    }
    if ( z>zs+dz ) {
        #ifdef DEBUG
        printf("%g, %g, diff=%g, outside\n", z, zs, z-zs);
        #endif
        return kOutside;
    }
    if ( intSeg ) FillSegment(rho, intSeg);
    #ifdef DEBUG
    printf("%g, %g, diff=%g, surf\n", z, zs, z-zs);
    #endif
    return kSurface;
}

//______________________________________________________________________________________________________________________
inline bool ParametricSurface::Point(int i, double step, double& rho, double& z){
    rho=fR1+step*i;
    if ( rho>fR ) return false;
    z=PointOnSurf(rho);
    return true;
}

};
#endif
