#ifndef __SECONORDERSURFACE__
#define __SECONORDERSURFACE__

#include "ParametricSurface.h"

namespace G4FresnelLens{

class SecondOrderSurface : public ParametricSurface
{
public:
    SecondOrderSurface(double k1=0., double k2=0., double k3=0., bool plusSol=false, double rhomin=0., double rhomax=0.);
   ~SecondOrderSurface();

    void Configure();
    double PointOnSurf(double x);
    double PointOnSurf2(double x);
    double EvalRho(double z);
    double EvalRho2(double z);
    double DistanceToSurf(const G4ThreeVector& p, const G4ThreeVector& v);
    double DistanceToSurfSafe(const G4ThreeVector& p);
    void   FillSegment(double rho, FSIntersectedSegment *intSeg);
    double Derivative(double rho);
    double Derivative(double rho, double z);
    bool Normal(FSIntersectedSegment* ints, const G4ThreeVector& p, double up, G4ThreeVector* norm);
    bool operator==(SSurface&);

    void print();

    double GuessParameters(int n, double* rho, double* z);
protected:
    void GuessParameters(int n, int tn, int* idx, double* rho, double* z, bool minimum);

    double fK0, fK1, fK2;
    double fK1_init, fK2_init;
    bool fPlusSolutionUsed;
};

//__________________________________________________________________________________
inline void SecondOrderSurface::FillSegment(double rho, FSIntersectedSegment *intSeg){
    intSeg->surf=this;
    intSeg->perp=rho;
}

//__________________________________________________________________________________
inline double SecondOrderSurface::EvalRho2(double z){
    double r2=z*z*fK0+z*fK1+fK2;
    if ( r2<0 ) {
        if ( (-r2)<kTolerance ) return 0.;
        else return DBL_MAX;
    }
    return r2;
}

//__________________________________________________________________________________
inline double SecondOrderSurface::EvalRho(double z){
    double r2=z*z*fK0+z*fK1+fK2;
    if ( r2<0 ) {
        if ( (-r2)<kTolerance ) return 0.;
        else return DBL_MAX;
    }
    return sqrt(r2);
}

//__________________________________________________________________________________
inline double SecondOrderSurface::PointOnSurf(double x){
    return PointOnSurf2(x*x);
}

//__________________________________________________________________________________
inline double SecondOrderSurface::Derivative(double perp){
    double denom=2.*fK0*PointOnSurf(perp)+fK1;
    double nom=2.*perp*(fPlusSolutionUsed?+1.:-1.);
    return denom?(nom/denom):DBL_MAX;
}

//__________________________________________________________________________________
inline double SecondOrderSurface::Derivative(double perp, double z){
    double denom=2.*fK0*z+fK1;
    double nom=2.*perp*(fPlusSolutionUsed?+1.:-1.);
    return denom?(nom/denom):DBL_MAX;
}

//__________________________________________________________________________________
inline void SecondOrderSurface::print(){
    printf("p2 surface\tr(%g,%g)\tz(%g, %g)\tmin/max=%g/%g\tk0=%g, k1=%g, k2=%g\n", fR1, fR, fZ1, fZ2, fZmin, fZmax, fK0, fK1, fK2);
}

//______________________________________________________________________________
inline bool SecondOrderSurface::Normal(FSIntersectedSegment* ints, const G4ThreeVector& p, double out, G4ThreeVector* norm){
    // calculate normal to the special surface, described by equation:
    // rho^2 = a*z^2 + b*z + c
    // or rho^2 = par0*z^2 + par1*z + par2
    double perp=p.perp();
    double invperp=1./perp;
    double tgtheta=Derivative(perp, p.z());
    double cos_theta=1./sqrt(1.+tgtheta*tgtheta);
    double psin_theta=tgtheta*cos_theta*invperp;
    double left=((fZ2-fZ1)<0?+1:-1)*out;
    norm->set(left*p.x()*psin_theta,left*p.y()*psin_theta,cos_theta*out);
    return true;
}

};

#endif
