#include "ArraySurface.h"
#include "FixedStack.h"

#ifdef USE_ROOT
#include <TROOT.h>
#endif

using namespace G4FresnelLens;

int ArraySurface::fStepsLimit=100;
#define NUM_TOLERANCE 1.e-15

//______________________________________________________________________________
ArraySurface::ArraySurface(int n, double* rho, double* z): SSurface(0., rho[n-1]), fN(n) {
    fZ=new double [fN];
    fRho=new double [fN];
    memcpy(fZ, z, fN*sizeof(double));
    memcpy(fRho, rho, fN*sizeof(double));

    /*Nonfixed step*/// GetCurIndexPointer=&ArraySurface::GetCurIndexConstStep;
    DetermineDistToSeg=&ArraySurface::DetermineDistToSegSmart;
    //DetermineDistToSeg=&ArraySurface::DetermineDistToSegDirect;

    //for (int i(0); i<fN; i++){
        //printf("%i %g %g\n", i, fRho[i], fZ[i]);
    //}

    Configure();
}

//______________________________________________________________________________
ArraySurface::~ArraySurface(){
    delete [] fZ;
    delete [] fRho;
}

//______________________________________________________________________________
void ArraySurface::Configure(){
    //TODO
    fStep=fRho[1]-fRho[0];
    fR1=fRho[0];

    fZmin=fZmax=fZ[0];
    for ( int i(1); i<fN; i++){
        if ( fZ[i]<fZmin ) fZmin=fZ[i];
        if ( fZ[i]>fZmax ) fZmax=fZ[i];
    }
    fZ1=fZ[0];
    fZ2=fZ[fN-1];

    double rho1=fRho[0];
    double rho2=fRho[fN-1];
    double z1=fZ[0];
    double z2=fZ[fN-1];
    int idxC=fN/2;
    double rhoC=fRho[idxC];
    double zC=z2-(rho2-rhoC)*(z2-z1)/(rho2-rho1);
    double zCa=fZ[idxC];
    fConvexity=zCa>zC?+1:-1;
}

//______________________________________________________________________________
EInside ArraySurface::Inside(int idx, double rho, double z, FSIntersectedSegment *intSeg){
    double zc1=fZ[idx];
    double zc2=fZ[idx+1];
    double rhoc1=fRho[idx];
    double rhoc2=fRho[idx+1];
    if ( intSeg ) FillSegment(idx, rho, intSeg);
    return InsideCone(rho, z, rhoc1, zc1, rhoc2, zc2);
}

//______________________________________________________________________________
double ArraySurface::DetermineDistToSegDirect(const G4ThreeVector& p, const G4ThreeVector& v, FSIntersectedSegment* intseg){
    double dist_cone=kInfinity;

    double perp=p.perp();
    int idx=this->GetCurIndex(perp);
    if( idx>=fN-1 ) idx=fN-2;
    int found=-1;
    double c=p.x()*v.x()+p.y()*v.y();
    if ( c>=0. ) {
        #ifdef DEBUG
        if (theDebugSwitch) {
            printf("go forward from %i to %i\n", idx, fN-1);
        }
        #endif
        dist_cone=FindCone(p, v, idx, fN-1, +1, found);

        #ifdef DEBUG
        if (theDebugSwitch) {
            printf("forward dist=%f and idx=%i\n", dist_cone, found);
        }
        #endif
    }
    else {
        double diff=p.perp2()-c*c/v.perp2();
        int min_i=this->GetCurIndex(diff>0.?sqrt(diff):0.);

        #ifdef DEBUG
        if (theDebugSwitch) {
            printf("current i1=%i min_i=%i perp=%f\n", idx, min_i, perp);
        }
        #endif

        if ( min_i<0) min_i=0;
        if ( min_i<idx) {
            #ifdef DEBUG
            if (theDebugSwitch) {
                printf("go backward from %i to %i\n",idx,0);
            }
            #endif
            dist_cone=FindCone(p, v, idx, min_i-1, -1, found);
            #ifdef DEBUG
            if (theDebugSwitch) {
                printf("backward dist=%f and idx=%i\n",dist_cone,found);
            }
            #endif
        }
        if ( dist_cone==kInfinity ) {
            #ifdef DEBUG
            if (theDebugSwitch) {
                printf("go forward from %i to %i\n",idx,fN-1);
            }
            #endif
            dist_cone=FindCone(p, v, idx, fN-1, +1, found);
            #ifdef DEBUG
            if (theDebugSwitch) {
                printf("forward dist=%f and idx=%i\n",dist_cone,found);
            }
            #endif
        }
    }

    if ( intseg && found>-1 ){
        FillSegment(found, -1, intseg);
    }
    return dist_cone;
}

//______________________________________________________________________________
double ArraySurface::DistanceToSurfSafe(const G4ThreeVector& p){
    double perp2=p.perp2();
    if ( perp2<=fR*fR && perp2>=fR1*fR1 ) {
        if ( p.z()>fZmax ) return p.z()-fZmax;
        if ( p.z()<fZmin ) return fZmin-p.z();
        return 0.;
    }
    double perp=sqrt(perp2);
    double rc=perp>fR?fR:fR1;
    if ( p.z()>fZmax ) return Hyp(perp, p.z(), rc, fZmax);
    if ( p.z()<fZmin ) return Hyp(perp, p.z(), rc, fZmin);
    return fabs(rc-perp);
}

//__________________________________________________________________________________
double ArraySurface::FindCone(const G4ThreeVector& p, const G4ThreeVector&  v, int point1, int point2, int dir, int& idx){
    double dist;
    #ifdef DEBUG
        if (theDebugSwitch) {
            printf("pos %e %e %e\n", p.x(), p.y(), p.z());
            printf("dir %e %e %e\n", v.x(), v.y(), v.z());
        }
    #endif
    if ( (point1-point2)*dir>0 ) Abort_("Bad iteration direction. Shouldn't happen.");
    for (int i(point1); i!=point2; i+=dir){
        dist=ConeIntersection(p, v, fRho[i], fZ[i], fRho[i+1], fZ[i+1]);
        if ( dist<kInfinity ){
            idx=i;
            #ifdef DEBUG
                if (theDebugSwitch) {
                    info_<<"Direct alg: yes "<<i-point1<<endline_;
                    info_<<"from/to/found "<<point1<<"  "<<point2<<"  "<<i<<endline_;
                }
            #endif
            return dist;
        }
    }

    #ifdef DEBUG
        if (theDebugSwitch) {     info_<<"Direct alg: noo "<<point2-point1<<endline_;
            info_<<"from/to "<<point1<<"  "<<point2<<endline_;
        }
    #endif
    idx=-1;
    return kInfinity;
}

//__________________________________________________________________________________
double ArraySurface::FindConeConcave(double aperp, const G4ThreeVector& p, const G4ThreeVector&  v, int aidx, FSIntersectedSegment* intseg){
    G4ThreeVector pos(p);
    double perp=aperp;
    int idx=-1;
    double dist(kInfinity);
    int currentI=aidx;
    bool out=VectorGoesOut(p,v);

    fixedISStack stackSteps;
    //stackSteps.clear();
    if ( out ){
        if ( currentI>0 ) stackSteps.push(0, currentI);   //left
        if ( (fN-1-currentI)>1 ) stackSteps.push(currentI+1, fN-1); //right
    }
    else {
        if ( (fN-1-currentI)>1 ) stackSteps.push(currentI+1, fN-1); //right
        if ( currentI>0 ) stackSteps.push(0, currentI);   //left
    }
    stackSteps.push(currentI, currentI+1);                            //center

    #ifdef DEBUG
    if (theDebugSwitch) {
        printf("Starting concave method for\n");
        info_<<"currentI "<<currentI<<endline_;
    }
    #endif

    fcIterationStep* itStep;
    for ( int i(0); i<fStepsLimit; i++ ){
        itStep = stackSteps.pop();
        if ( !itStep ) {
            #ifdef DEBUG
            if (theDebugSwitch) {
                info_<<"Concave alg: no "<<i<<endline_;
            }
            #endif
            return kInfinity;
        }

        #ifdef DEBUG
        if (theDebugSwitch) {
            info_<<i<<endline_;
            info_<<Form(" min %i max %i c %i", 0, fN-1, currentI)<<endline_;
            info_<<Form("tt l %f l1 %f ", fRho[0], fRho[fN-1])<<endline_;
            info_<<Form("coord %s (%e %e) (%e %e) idx(%i, %i) %e %e %e %e", (out?"out":"inn"),
                         pos.x(), pos.y(), pos.perp(), pos.z(),
                         itStep->leftI, itStep->rightI,
                         fRho[itStep->leftI], fZ[itStep->leftI], fRho[itStep->rightI], fZ[itStep->rightI])<<endline_;
        }
        #endif
        dist=ConeIntersection( p, v, fRho[itStep->leftI], fZ[itStep->leftI], fRho[itStep->rightI], fZ[itStep->rightI] );

        #ifdef DEBUG
        if ( theDebugSwitch ) {
            info_<<Form("dist=%g", dist)<<endline_;
            info_<<Form("dir %e %e %e", v.x(), v.y(), v.z())<<endline_;
        }
        #endif
        if ( !(dist<kInfinity) ) continue;

        //find intersection position
        pos=p+v*dist;
        perp=pos.perp();
        int backi=currentI;
        currentI=this->GetCurIndex(perp);

        if ( (itStep->rightI-itStep->leftI)==1 ){
            idx=currentI;
            #ifdef DEBUG
            if (theDebugSwitch) {
                info_<<"Concave alg: yes "<<i<<endline_;
            }
            #endif
            if (intseg) FillSegment(idx, perp, intseg);
            return dist;
        }

        // calculate new steps
        int li(itStep->leftI), ri(itStep->rightI);
        if ( currentI==backi ) {
            if ( currentI==li-1 ) {
                currentI=li;
                if (li) li--;
            }
            else if ( currentI==ri+1 ) {
                currentI=ri;
                if ( ri<fN-1 ) ri++;
            }
            #ifdef DEBUG
            else {
                printf("strange ci\n");
            }
            #endif
        }
        #ifdef DEBUG
        if ( theDebugSwitch )
          printf("add single seg at %i (prev %i)\n", currentI, backi);
        #endif

        if ( VectorGoesOut(pos,v) ){
          if ( (currentI-li)>0   )
            stackSteps.push(li, currentI);   //left
          if ( (ri-currentI)>1 )
            stackSteps.push(currentI+1, ri); //right
        }
        else {
          if ( (ri-currentI)>1 )
            stackSteps.push(currentI+1, ri); //right
          if ( (currentI-li)>0   )
            stackSteps.push(li, currentI);   //left
        }
        stackSteps.push(currentI, currentI+1);
    }
    error_<<"FindConeConcave returned no result after "<<fStepsLimit<<" steps"<<endline_;
    #ifdef DEBUG
        throw 1;
    #endif
    return kInfinity;
}

//__________________________________________________________________________________
double ArraySurface::FindConeConvex(double aperp, const G4ThreeVector& p, const G4ThreeVector&  v, int aidx, FSIntersectedSegment* intseg){
    G4ThreeVector pos(p);
    int idx=-1;
    int currentI=aidx;
    double perp2(p.perp2()), perp(aperp);
    double dist;
    double minRho=fRho[0];
    double maxRho=fRho[fN-1];
    double minZ=fZmin;
    double maxZ=fZmax;

    bool inlimits=false;

    #ifdef DEBUG
        if (theDebugSwitch) {
            printf("Starting convex method\n");
            printf("initins perp=%e z=%e r1=%e z1=%e r2=%e z2=%e %i\n",
            perp, pos.z(),
            fRho[currentI], fZ[currentI],
            fRho[currentI+1], fZ[currentI+1], currentI);
        }
    #endif

    double mindist=0.;
    for (int i(0); i<=fStepsLimit; i++){
        #ifdef DEBUG
        if (theDebugSwitch) {
            info_<<"Iteration "<<i<<endline_;
            info_<<"i cur/min/max "<<currentI<<"  "<<0<<"   "<<fN-1<<endline_;
            info_<<"rho cur/min/max "<<fRho[currentI]<<"  "<<fRho[0]<<"   "<<fRho[fN-1]<<endline_;
        }
        #endif

        dist=ConeIntersection1( p, v, fRho[currentI], fZ[currentI], fRho[currentI+1], fZ[currentI+1], inlimits, mindist);
        #ifdef DEBUG
        if (theDebugSwitch) {
            double dist1=ConeIntersection( p, v, fRho[currentI], fZ[currentI], fRho[currentI+1], fZ[currentI+1] );
            bool inlimits1;
            double dist2=ConeIntersection1( p, v, fRho[currentI], fZ[currentI], fRho[currentI+1], fZ[currentI+1], inlimits1, 0 );
            info_<<Form("dist_cone=%e (raw %e), (wo mindist %e) in limits=%i (%i)", dist, dist1, dist2, inlimits, inlimits1)<<endline_;
            printf("coord (%e %e) (%e %e) %e %e %e %e\n",
                    pos.x(), pos.y(), pos.perp(), pos.z(),
                    fRho[currentI], fZ[currentI], fRho[currentI+1], fZ[currentI+1]);
            printf("dir %g %g\n", v.theta()/deg, v.phi()/deg);
        }
        #endif

        if ( dist==kInfinity ) {
            #ifdef DEBUG
            if (theDebugSwitch) {
                info_<<"Conv alg: noo "<<i<<endline_;
            }
            #endif
            return kInfinity;
        }
        else if( inlimits ) {
            idx=currentI;
            #ifdef DEBUG
            if (theDebugSwitch) {
                info_<<"Conv alg: yes "<<i<<endline_;
            }
            #endif
            if (intseg) FillSegment(idx, -1., intseg);
            return dist;
        }
        mindist=dist*0.9; //FIXME: not optimized
        pos=p+v*dist;
        perp2=pos.perp2();

        #ifdef DEBUG
         if (theDebugSwitch) {
            printf("perp=%e z=%e r1=%e z1=%e r2=%e z2=%e i1=%i i2=%i\n", sqrt(perp2), pos.z(), minRho, minZ, maxRho, maxZ, 0, fN-1);
         }
        #endif
        if ( !InLimits2(perp2, pos.z(), minRho, minZ, maxRho, maxZ) ) {
            #ifdef DEBUG
            if (theDebugSwitch) {
                info_<<Form("Conv alg: noo not in limits %i: perp=%e z=%e r1=%e z1=%e r2=%e z2=%e", i, sqrt(perp2), pos.z(), minRho, minZ, maxRho, maxZ)<<endline_;
            }
            #endif
            return kInfinity;
        }
        perp=sqrt(perp2);
        currentI=this->GetCurIndex(perp);
        if ( currentI==fN-1 ){
            #ifdef DEBUG
            if (theDebugSwitch) {
                info_<<"Conv alg: noo "<<i<<endline_;
             }
            #endif
            return kInfinity;
        }

        #ifdef DEBUG
        if (theDebugSwitch) {
            debug_<<currentI<<endline_;
        }
        #endif
    }
    error_<<"FindConeConvex returned no result after "<<fStepsLimit<<" steps."<<endline_;
    #ifdef DEBUG
        throw 1;
    #endif
    return kInfinity;
}

//__________________________________________________________________________________
double ArraySurface::ConeIntersection(const G4ThreeVector& p, const G4ThreeVector& v, double r1, double z1, double r2, double z2){
    // adopted from Ltrace Lrt_bsearch
    // Int_t Lcheck_intersect_cone(Double_t r1, Double_t z1, Double_t r2, Double_t z2,
    //                Double_t phot[8], Double_t *dis)
//    printf("(%10.15f, %10.15f, %10.15f), (%10.15f, %10.15f, %10.15f)\n", p.x(), p.y(), p.z(), v.x(), v.y(), v.z());
//    printf("(%10.15f, %10.15f), (%10.15f, %10.15f)\n", r1, r2, z1, z2);

    double d1, d2;
    int nroot;
    if ( z1==z2 ) {
        // the quadratic equation for this case fails due to numeric precision
        double vz=v.z();
        if ( vz ) {
            d1=d2=(z1-p.z())/v.z();
            nroot=1;
        }
        else return kInfinity;
    }
    else {
        double a00 = (z2-z1)/(r2-r1);
        double z0  = z1-a00*r1;
        double a0  = a00*a00;

        double dz = p.z()-z0;
        double a = a0*v.perp2() - v.z()*v.z();
        double b = 2.0*(a0*v.x()*p.x() + a0*v.y()*p.y() - v.z()*dz);
        double c = a0*p.perp2() - dz*dz;
        nroot=SolveQuadratic(a, b, c, d1, d2);
    }

    //    if ( r1>=937.00 && r1<=938.0) printf("nroot=%i d1=%e d2=%e\n",nroot,d1,d2);
    //there can be 1 intersection
    if( nroot==0 || (nroot==1&&d1==0.0) ) return kInfinity;
    // d2 is more than d1 as defined in SolveQuadratic
    if(d2<0.0)  return kInfinity;

    G4ThreeVector p1;
    double z;
    if ( d1>=0.0 ){
        p1=p+v*d1;
        z=p1.z();
        /* check if the solution is in the specified range */
//        printf("dd %e (%10.15f, %10.15f, %10.15f)\n", d1, p1.x(), p1.y(), p1.z() );
        if( (z-z1)*(z-z2)<=NUM_TOLERANCE ) {
            double perp2=p1.perp2();
            if ( perp2>=r1*r1 && perp2<=r2*r2 ) return d1;
        }
    }

    if ( d1!=d2 ) {
        p1=p+v*d2;
        z=p1.z();
        /* check if the solution is in the specified range */
//        printf("dd %e (%10.15f, %10.15f, %10.15f)\n", d2, p1.x(), p1.y(), p1.z() );
        if( (z-z1)*(z-z2)<=NUM_TOLERANCE ) {
            double perp2=p1.perp2();
            if ( perp2>=r1*r1 && perp2<=r2*r2 ) return d2;
        }
    }
    return kInfinity;
}

//__________________________________________________________________________________
double ArraySurface::ConeIntersection1(const G4ThreeVector& p, const G4ThreeVector& v,
                                    double r1, double z1, double r2, double z2,
                                    bool& inlimits, double mindist){
    // adopted from Ltrace Lrt_bsearch
    // Int_t Lcheck_intersect_cone(Double_t r1, Double_t z1, Double_t r2, Double_t z2,
    //                Double_t phot[8], Double_t *dis)
//    printf("(%10.15f, %10.15f, %10.15f), (%10.15f, %10.15f, %10.15f)\n", p.x(), p.y(), p.z(), v.x(), v.y(), v.z());
//    printf("(%10.15f, %10.15f), (%10.15f, %10.15f)\n", r1, r2, z1, z2);
    inlimits=false;
    double d1, d2, z0;
    int nroot;
    double cone_higher; // help to chose right cone part and reject intersections with the reflected cone
    if ( z1==z2 ) {
        // the quadratic equation for this case fails due to numeric precision
        double vz=v.z();
        if ( vz ) {
            d1=d2=(z1-p.z())/v.z();
            nroot=1;
        }
        else return kInfinity;

        z0=z2;
        cone_higher=0.;
    }
    else {
        double a00 = (z2-z1)/(r2-r1);
        z0  = z1-a00*r1;
        cone_higher=(z2-z0);
        double a0  = a00*a00;

        double dz = p.z()-z0;
        double a = a0*v.perp2() - v.z()*v.z();
        double b = 2.0*(a0*v.x()*p.x() + a0*v.y()*p.y() - v.z()*dz);
        double c = a0*p.perp2() - dz*dz;

        nroot=SolveQuadratic(a, b, c, d1, d2);
    }

    //    if ( r1>=937.00 && r1<=938.0) printf("nroot=%i d1=%e d2=%e\n",nroot,d1,d2);
    //there can be 1 intersection
    if( nroot==0 || (nroot==1&&d1==0.0) ) return kInfinity;
    // d2 is more than d1 as defined in SolveQuadratic
    if(d2<0.0) return kInfinity;
    #ifdef CHECK_ConeIntersection1
    printf("dd000 %.10f %.10f\n", d1, d2);
    #endif

    G4ThreeVector p1;
    double z;
    double minpositive=kInfinity;
    if ( d1>=mindist ){
        p1=p+v*d1;
        z=p1.z();
        #ifdef CHECK_ConeIntersection1
        printf("dd %e (%10.15f, %10.15f), (pos %10.15f, %10.15f, %s), (cone %10.15f, %10.15f, %s)\n", d1, p1.perp(), p1.z(), z, z0, ((z>=z0)?"aga":"nee"), z1, z0, (cone_higher>=0?"aga":"nee") );
        #endif
        // check that solution is on the same halfspace
        if ( (z-z0)*cone_higher>=0. ) {
            minpositive=d1;
            if( (z-z1)*(z-z2)<NUM_TOLERANCE ) {
                double perp2=p1.perp2();
                if ( perp2>=r1*r1 && perp2<=r2*r2 ) {
                    inlimits=true;
                    return d1;
                }
            }
        }
        //else minpositive=d2;
    }

    if ( d1!=d2 && d2>=mindist ){
        p1=p+v*d2;
        z=p1.z();
        /* check if the solution is in the specified range */
        #ifdef CHECK_ConeIntersection1
        printf("dd %e (%10.15f, %10.15f), (pos %10.15f, %10.15f, %s), (cone %10.15f, %10.15f, %s)\n", d2, p1.perp(), p1.z(), z, z0, ((z>=z0)?"aga":"nee"), z1, z0, (cone_higher?"aga":"nee") );
        #endif
        if ( (z-z0)*cone_higher>=0. ) {
            if( (z-z1)*(z-z2)<NUM_TOLERANCE ) {
                double perp2=p1.perp2();
                if ( perp2>=r1*r1 && perp2<=r2*r2 ) {
                    inlimits=true;
                    return d2;
                }
            }

            if ( d2<minpositive ) minpositive=d2;
        }
        // else minpositive=kInfinity; //FIXME
    }

    inlimits=false;
    return minpositive;
}

//__________________________________________________________________________________
EInside ArraySurface::InsideCone(double rho, double z, double rho1, double z1, double rho2, double z2){
    double z0=(z2-z1)*(rho-rho1)/(rho2-rho1)+z1;

    // find a distance from the point to the cylinder surface using the triangle area
    double s=(rho-rho1)*(z2-z1)-(rho2-rho1)*(z-z1);
    double drho=rho2-rho1;
    double dz=z2-z1;
    double h2=s/sqrt(drho*drho+dz*dz);
    if ( h2>kHalfTolerance ){
        if ( z>z0 ) return kOutside;
        else return kInside;
    }
    return kSurface;
}

//__________________________________________________________________________________
bool ArraySurface::operator==(SSurface& s){
    ArraySurface* to1=static_cast<ArraySurface*>(&s);
    if (!to1) {
        return false;
    }
    ArraySurface& to=*to1;
    if( fN!=to.fN ||
        fStep!=to.fStep ||
        fConvexity!=to.fConvexity ||
        fZmin!=to.fZmin ||
        fZmax!=to.fZmax ){
        //print();
        //to.print();
        return false;
    }
    for (int i(0); i<fN; i++){
        if ( fZ[i]==to.fZ[i] &&
             fRho[i]==to.fRho[i] ) continue;
        //printf("%g!=%g, %g!=%g\n", fZ[i], to.fZ[i], fRho[i], to.fRho[i]);
        return false;
    }
    return true;
}
