#include "ParametricSurface.h"

#include <Randomize.hh>
#include <G4ThreeVector.hh>
#include <G4RandomDirection.hh>
#include <globals.hh>

using namespace std;

using namespace G4FresnelLens;

double ParametricSurface::fSafetyGap=1.e-9;
//______________________________________________________________________________
ParametricSurface::ParametricSurface(double z, double rho) : SSurface(z, rho) {

}

//______________________________________________________________________________
ParametricSurface::~ParametricSurface(){

}

//______________________________________________________________________________
void ParametricSurface::ShiftSurfaceTo(int n, int tn, int* idx, double* rho, double* z, bool minimum){
    GuessParameters(n, tn, idx, rho, z, minimum);
    fZpos=0;
    fZpos=z[0]-PointOnSurf(0.);
    for(int i(0); i<n; i++){
        printf("a %i %g %g\n", i, rho[i], z[i]);
        double zi=PointOnSurf(rho[i]);
        double dz=(minimum?min(zi, z[i]):max(zi, z[i]))-zi;
        fZpos+=dz;
    }
    fZpos+=minimum?-fSafetyGap:+fSafetyGap;
    Configure();
}

//______________________________________________________________________________
void ParametricSurface::ShiftSurfaceTo(const vector<SSurface*>& surfaces, bool minimum){
    GuessParameters(surfaces, minimum);
    fZpos=0;
    double step=5*um;
    vector<SSurface*>::const_iterator it=surfaces.begin();
    fZpos=(*it)->GetZ1()-PointOnSurf(0.);
    for ( ;it<surfaces.end(); it++ ) {
        double rho,z;
        int i=0;
        while( (*it)->Point(i, step, rho, z) ){
            double zi=PointOnSurf(rho);
            double dz=(minimum?min(zi, z):max(zi, z))-zi;
            fZpos+=dz;
            i++;
        }
        rho=(*it)->GetR2();
        z=(*it)->GetZ2();
        double zi=PointOnSurf(rho);
        double dz=(minimum?min(zi, z):max(zi, z))-zi;
        fZpos+=dz;
    }
    fZpos+=minimum?-fSafetyGap:+fSafetyGap;
    Configure();
}

//______________________________________________________________________________
void ParametricSurface::selfCheck(int n){
    double shiftmax=1000.*mm;
    printf("checking self\n");
    for( int i(0); i<n; i++ ){
        double r2=G4UniformRand()*fR*fR;
        double phi = twopi*G4UniformRand();
        double r=sqrt(r2);
        double x=r*cos(phi);
        double y=r*sin(phi);
        double z=PointOnSurf(r);

        double shift=G4UniformRand()*shiftmax;
        G4ThreeVector dir=G4RandomDirection();
        G4ThreeVector posint(x,y,z);
        G4ThreeVector pos=posint-shift*dir;

        double dist=DistanceToSurf(pos, dir);
        double dists=DistanceToSurfSafe(pos);

        double diff=shift-dist;
        double limit=.5e-9;
        char mes[1000];
        if ( fabs(diff)>limit ) {
            sprintf(mes, "%i (%g, %g) (%g, %g)-> dist!=shift %g %g,%g\n", i, r, z, dir.theta()/deg, dir.phi()/deg, dist, shift, diff);
            dist+=DistanceToSurf((pos+dir*(dist+1.e-8)), dir)+1.e-8;
            diff=shift-dist;
        }
        if ( fabs(diff)>limit ) {
            printf("%s",mes);
            printf("    dist1!=shift %g %g %g\n", dist, shift, dist-shift);
            continue;
        }
        if ( dists>dist ){
            printf("error safe dist>dist %g %g\n", dists, dist);
        }
    }
}


//______________________________________________________________________________
double ParametricSurface::CheckDistance(const G4ThreeVector& p1, const G4ThreeVector& v){
    double dist=DistanceToSurf(p1,v);
    if ( dist==kInfinity ) return kInfinity;
    G4ThreeVector p2=p1+v*dist;
    EInside ins=Inside(p2.perp(), p2.z());
    if( ins!=kSurface ) {
        printf("Dist from (%g, %g; [%g], %g) along (%g, %g) = %g, brings not to surface (%g, %g) (%s) dz%g\n",
               p1.x(), p1.y(), p1.perp(), p1.z(), v.theta()/deg, v.phi()/deg, dist,
               p2.perp(), p2.z(),
               ins==kOutside?"out":"in", p2.z()-PointOnSurf(p2.perp()));
        print();
    }

    return dist;
}

//__________________________________________________________________________________
#ifdef USE_ROOT
#include <TGraph.h>
#include <TCanvas.h>
#include <TString.h>
#include <TFile.h>
void ParametricSurface::draw(const char* file, const char* canvasfile, double shift, const char* prefix, const char* option){
    double step=1.e-2;
    int n=fR/step+1;

    TFile* f = new TFile(file, option);
    TGraph surf(n*2-2);
    for (int i(0); i<n; i++){
        double rho=fR-i*step;
        surf.SetPoint(surf.GetN(), -rho, PointOnSurf(rho)+shift);
    }
    for (int i(1); i<n; i++){
        double rho=i*step;
        surf.SetPoint(surf.GetN(), rho, PointOnSurf(rho)+shift);
    }
    surf.Write( TString(prefix)+"surf" );
    f->Close();

    if ( canvasfile ) {
        TCanvas c("surface", "surface");
        surf.Draw("AL");
        c.Print(canvasfile);
    }
}
#endif
