#include "SecondOrderSurface.h"
#include "LensUtils.h"

#include <CLHEP/Matrix/Matrix.h>

using CLHEP::HepMatrix;
using std::swap;
using namespace G4FresnelLens;

//______________________________________________________________________________
SecondOrderSurface::SecondOrderSurface(double k0, double k1, double k2, bool plusSol, double rhomin, double rhomax):
ParametricSurface(0., rhomax),
fK0(k0),
fK1(k1),
fK2(k2),
fK1_init(k1),
fK2_init(k2),
fPlusSolutionUsed(plusSol)
{
    fPlusSolutionUsed=true;
    Configure();
}

//______________________________________________________________________________
SecondOrderSurface::~SecondOrderSurface(){


}

//______________________________________________________________________________
void SecondOrderSurface::Configure(){
    // on shifting z0:
    // a -> a
    // b -> b-2az
    // c -> c-bz+azz
    if ( fZpos ){
        fK2=fK2_init-fK1_init*fZpos+fK0*fZpos*fZpos;
        fK1=fK1_init-2.*fK0*fZpos;
    }

    fZ1=fZmin=PointOnSurf(fR1);
    fZ2=fZmax=PointOnSurf(fR);
    if ( fZmin>fZmax ) swap(fZmin,fZmax);

    if ( fZmin==fZmax && fZmin<kInfinity ) {
        printf("abnormal parameters:");
        print();
        exit(1);
    }
}

//__________________________________________________________________________________
double SecondOrderSurface::GuessParameters(int n, double* rho, double* z){
    fZpos=0.;
    fR1=rho[0];
    fR=rho[n-1];

    const int size=3;
                 //row  col
    HepMatrix matA(size,size);
    HepMatrix matB(size,1);

    int i1=0;
    while ( rho[i1]<kTolerance && fabs(z[i1])<kTolerance ){
        i1++;
    }

    int idxs[3]={i1, n/2, n-1};
    for( int i(0); i<size; i++ ) {
        matA(i+1, 1)= z[idxs[i]]*z[idxs[i]];
        matA(i+1, 2)= z[idxs[i]];
        matA(i+1, 3)= 1.;
        matB(i+1, 1)= rho[idxs[i]]*rho[idxs[i]];
    }
    double detA=matA.determinant();

    double*   coeffs[3]={&fK0, &fK1, &fK2};
    double    detA_prim;
    HepMatrix matA_prim;
    for( int i(1); i<(size+1); i++ ) {
        matA_prim=matA;
        for( int j(1); j<(size+1); j++ ) matA_prim(j, i)=matB(j, 1);
        detA_prim=matA_prim.determinant();
        *coeffs[i-1]=detA_prim/detA;
    }
    fK1_init=fK1;
    fK2_init=fK2;

    //printf("new (%g, %g), (%g, %g), (%g, %g) -> (%g, %g, %g)\n",
            //rho[idxs[0]],z[idxs[0]],
            //rho[idxs[1]],z[idxs[1]],
            //rho[idxs[2]],z[idxs[2]],
            //fK0,fK1,fK2);

    int p=n/2;
    fPlusSolutionUsed=true;
    double ps1=PointOnSurf(rho[p]);
    fPlusSolutionUsed=false;
    double ps2=PointOnSurf(rho[p]);
    fPlusSolutionUsed=fabs(ps1-z[p])<fabs(ps2-z[p]);

    Configure();
    double dev(0.);
    for(int i(0); i<n; i++){
        double rhof=EvalRho(z[i]);
        if ( isnan(rhof) ) return kInfinity;
        double devc=fabs(rho[i]-rhof);
        if ( devc>dev ) dev=devc;
    }
    return dev;
}

//__________________________________________________________________________________
void SecondOrderSurface::GuessParameters(int n, int tn, int* idx, double* rho, double* z, bool minimum){
    AbstractMethod();
}

//__________________________________________________________________________________
double SecondOrderSurface::DistanceToSurf(const G4ThreeVector& p, const G4ThreeVector& v){
    double d1, d2;
    double a = v.perp2() - fK0*v.z()*v.z();
    double b = 2.0*(v.x()*p.x() + v.y()*p.y()) - v.z()*(2.*fK0*p.z()+fK1);
    double c = p.perp2() - fK0*p.z()*p.z()-fK1*p.z()-fK2;
    int nroot=SolveQuadratic(a, b, c, d1, d2);

    // there can be 1 intersection
    if( nroot==0 || (nroot==1&&d1==0.0) ) return kInfinity;
    if(d2<0.0) return kInfinity;

    G4ThreeVector p1;
    if ( d1>=0.0 ){
        p1=p+v*d1;
        double z=p1.z();
        if( (z-fZmin)*(z-fZmax)<0.) {
            double perp2=p1.perp2();
            if ( perp2>=fR1*fR1 && perp2<=fR*fR ) return d1;
        }
    }

    p1=p+v*d2;
    double z=p1.z();
    if( (z-fZmin)*(z-fZmax)<0.) {
        double perp2=p1.perp2();
        if ( perp2>=fR1*fR1 && perp2<=fR*fR ) return d2;
    }
    return kInfinity;
}

//__________________________________________________________________________________
double SecondOrderSurface::DistanceToSurfSafe(const G4ThreeVector& p){
    //
    // return undirectional distance to surface
    // can be underestimated
    // thus returning distance to the bounding tube
    //
    AbstractMethod();
    return 0.;
    //double perp2=p.perp2();
    //double z=p.z();
//
    //if ( perp2<=fR*fR ) {
        //if ( z>fZmax ) return z-fZmax;
        //if ( z<fZmin ) return fZmin-z;
        //return 0.;
    //}
    //double perp=sqrt(perp2);
    //if ( z>fZmax ) return Hyp(perp, z, fR, fZmax);
    //if ( z<fZmin ) return Hyp(perp, z, fR, fZmin);
    //return perp-fR;
}

//__________________________________________________________________________________
double SecondOrderSurface::PointOnSurf2(double rho2){
    double rz1, rz2;
    int roots=G4FresnelLens::SolveQuadratic(fK0, fK1, fK2-rho2, rz1, rz2);
    if ( roots==0 ) return kInfinity;
    return fPlusSolutionUsed?rz2:rz1;
}

//__________________________________________________________________________________
bool SecondOrderSurface::operator==(SSurface& s){
    SecondOrderSurface* to1=static_cast<SecondOrderSurface*>(&s);
    if (!to1) {
        return false;
    }
    SecondOrderSurface& to=*to1;
    if( fK0!=to.fK0 ||
        fK1!=to.fK1 ||
        fK2!=to.fK2 ||
        fPlusSolutionUsed!=to.fPlusSolutionUsed ){
        //print();
        //to.print();
        return false;
    }
    return true;
}
