#include "G4Detector.hh"

#include "Detector.hh"
#include "EusoMapping.hh"
#include "DetectorGeometry.hh"
#include "VirtualDetectorTransportManager.hh"
#include "ElectronicsFactory.hh"
#include "ERunParameters.hh"
#include "ERunParsPixelMapFiller.hh"
#include "ERunParsPmtsFiller.hh"
#include "EusoElectronics.hh"
#include "OpticsFactory.hh"
#include "Pupil.hh"


#include <TMath.h>

using namespace sou;

//______________________________________________________________________________
G4Detector::G4Detector() : fElectronics(0) {
    //
    // Constructor
    //
    // FIXME: call the factories to get the right associate child and parent
    // classes in config
    ElectronicsFactory::Get();
    OpticsFactory::Get();
    EusoMapping::Get();
    Build();
    SetSender("G4Detector");
    MsgForm(EsafMsg::Info,"Creating G4Detector.");
}

//______________________________________________________________________________
G4Detector::~G4Detector() {
    //
    // Destructor
    //
    if (fGeometry)
        delete fGeometry;
    if (fDetectorTransporter)
        delete fDetectorTransporter;
    ElectronicsFactory::Get()->DeleteEusoElectronics();
}

//______________________________________________________________________________
void G4Detector::Build() {
    //
    // Build or re-build the detector
    //
    fGeometry = new DetectorGeometry();
    fGeometry->SetFoV(Pi()/6.);

    Msg(EsafMsg::Info)  << "Creating G4DetectorTransporter." << MsgDispatch;
    fDetectorTransporter = OpticsFactory::Get()->MakeTransportManager();
    fElectronics = ElectronicsFactory::Get()->Build();

    Double_t scale = Conf()->GetNum("G4Detector.fScaleFactor");
    Double_t radius;
    if ( Conf()->GetBool("G4Detector.fFixMaxRadius") )
         radius = Conf()->GetNum("G4Detector.fMaxRadius");
    else radius = fDetectorTransporter->GetOuterRadius();

    fGeometry->SetRadius(scale*radius);
    fDetectorTransporter->SetScaleFactor(scale);

    // save runparameters
    if ( EEvent::GetCurrent() && EEvent::GetCurrent()->GetRunPars() ) {
        ERunParameters *runpars = EEvent::GetCurrent()->GetRunPars();

        fElectronics->SetDetectorScaleFactor(scale);
        fElectronics->BuildBackgroundChipDist();

        if ( runpars ) {
            // get from EusoMapping the config
            ERunParsPixelMapFiller pxfiller;
            runpars->Fill(pxfiller);

            ERunParsPmtsFiller pmtfiller( fElectronics );
            runpars->Fill(pmtfiller);
        }
    }
    fDetectorTransporter->GetFocalPlane()->SetElectronics(fElectronics);
}

void G4Detector::Reset() {
    //
    // Reset electronics
    //

    if ( fElectronics )
        fElectronics->Reset();
}

Telemetry* G4Detector::Get( PhotonsOnPupil* photons ) {
    //
    // Simulate one event and return a pointer to a Telemetry object
    //


    // wrapper to track the photons onto the real pupil
    Pupil pupil( photons, GetGeometry());

    if(pupil.GetNphotons()==0) Msg(EsafMsg::Warning)  << "NO PHOTONS ON PUPIL !!!"<< MsgDispatch;
    else Msg(EsafMsg::Info)  << "PHOTONS ON PUPIL: "<<pupil.GetNphotons()<< MsgDispatch;

    // transport photons from pupil to pmts
    fDetectorTransporter->Go( &pupil );

    Double_t tBegin = fDetectorTransporter->GetTimeFirstPhoton();
    Double_t tEnd = fDetectorTransporter->GetTimeLastPhoton();

    MsgForm(EsafMsg::Info,"Photon Time interval (ns):\t\tSTART=%ld  END=%ld",
            (long int)(tBegin), (long int)tEnd);

    // simulate electronics and trigger
    if ( !fElectronics->Simulate( tBegin, tEnd ) )
        return (Telemetry*)0;

    // return telemetry object
    return fElectronics->Data();

}


//______________________________________________________________________________
Bool_t G4Detector::ClearMemory() {
    //
    // Physically release the memory allocated by the arrays of this object
    //

    fElectronics->ClearMemory();

    return kTRUE;
}
