#include "MaterialParameters.hh"

#include <TGraph.h>
#include <TF1.h>
#include <TGraph2D.h>
#include <TEnv.h>

#include <G4MaterialPropertiesTable.hh>
#include <G4Material.hh>
#include <G4OpticalSurface.hh>
#include <globals.hh>

using namespace std;

//_______________________________________________________________________________
MaterialParameters::MaterialParameters(const char* re, const char* im) :
theTable(0), theReIndex(0), theImIndex(0), theAbs(0), theTransmittance(0), isTableOwn(true)
{
    init();
    if ( re ) setRealRI(re);
    if ( im ) setImaRI(im);
}

//_______________________________________________________________________________
MaterialParameters::MaterialParameters(const char* ri) :
theTable(0), theReIndex(0), theImIndex(0), theAbs(0), theTransmittance(0), isTableOwn(true)
{
    init();
    if ( ri ) setRI(ri);
}

//_______________________________________________________________________________
void MaterialParameters::init(){
    /// Refraction index  _n_ = n - i*k
    /// absorption coeff  alpha = 4pi*k/lambda
    /// absorption length L = 1/alpha                    dim: 1/mm for Geant4
    SetSender("MaterialParameters");
    theConvConst=hbarc*twopi/eV;

    theTable = new G4MaterialPropertiesTable();
}

//_______________________________________________________________________________
MaterialParameters::~MaterialParameters() {
    if (theReIndex) delete theReIndex;
    if (theImIndex) delete theImIndex;
    if (theAbs) delete theAbs;
    if (theTransmittance) delete theTransmittance;

    if ( isTableOwn ) delete theTable;
}

//double f(double a);
//_______________________________________________________________________________
void MaterialParameters::setRealRI(const char* name, double (*idx_fun)(double)){
    double theEnergyMin=theConvConst/(1000.);
    double theEnergyMax=theConvConst/(200.);
    double theEnergyStep=0.01*eV;

    const int n=int( (theEnergyMax-theEnergyMin)/theEnergyStep )+1;
    double en[n];
    double rindex[n];
    double wl, energy;
    for ( int i(0); i<n; i++ ){
        energy=theEnergyMin+theEnergyStep*i;
        wl=theConvConst/energy;
        en[i]=energy;
        rindex[i]=idx_fun(wl);
    }
    //cout<<Form("Generated RINDEX for %s with limits: %.0f nm - %.0f nm and step: %.2f eV", name, theWlMin, theWlMax, theEnergyStep/eV)<<endl;
    theTable->AddProperty("RINDEX", en, rindex, n);
}

//TF1 f();

//_______________________________________________________________________________
void MaterialParameters::setRealRI(TF1* f,double wmin,double wmax){
    double theEnergyMin=theConvConst/wmax;
    double theEnergyMax=theConvConst/wmin;
    double theEnergyStep=0.01*eV;

    const int n= int( (theEnergyMax-theEnergyMin)/theEnergyStep )+1;
    double en[n];
    double rindex[n];
    double wl, energy;
    if ( theReIndex ) delete theReIndex;
    theReIndex=new TGraph();
    for ( int i(0); i<n; i++ ){
        energy=theEnergyMin+theEnergyStep*i;
        wl=theConvConst/energy;
        en[i]=energy;
        rindex[i]=f->Eval(wl);
        theReIndex->SetPoint(theReIndex->GetN(),en[i],rindex[i]);
    }
    //cout<<Form("Generated RINDEX for %s with limits: %.0f nm - %.0f nm and step: %.2f eV", name, theWlMin, theWlMax, theEnergyStep/eV)<<endl;
    theTable->AddProperty("RINDEX", en, rindex, n);
}
//_______________________________________________________________________________
void MaterialParameters::loadRealRI(const char* re){
    if ( theReIndex ) delete theReIndex;
    theReIndex = new TGraph(re);
    this->ConvertNmToEv( theReIndex );
}

//_______________________________________________________________________________
void MaterialParameters::loadImaRI(const char* im){
    if (theImIndex) delete theImIndex;
    if (theAbs) delete theAbs;
    theImIndex = new TGraph(im);
    this->ConvertNmToEv( theImIndex );
}

//_______________________________________________________________________________
void MaterialParameters::setRI(const char* ri){
    loadRI(ri);
    setRealRI();
    setImaRI();
}

//_______________________________________________________________________________
void MaterialParameters::loadRI(const char* ri){
    if ( theReIndex ) delete theReIndex;
    if (theImIndex) delete theImIndex;
    if (theAbs) delete theAbs;
    TGraph2D* bothRI=new TGraph2D(ri);
    int n=bothRI->GetN();
    double* wl=bothRI->GetX();

    theReIndex = new TGraph(n, wl, bothRI->GetY());
    theImIndex = new TGraph(n, wl, bothRI->GetZ());
    this->ConvertNmToEv( theReIndex );
    this->ConvertNmToEv( theImIndex );

    delete bothRI;
}

//_______________________________________________________________________________
void MaterialParameters::setRealRI(const char* re){
    loadRealRI(re);
    setRealRI();
}

//_______________________________________________________________________________
void MaterialParameters::setRealRI(){
    theTable->AddProperty("RINDEX", theReIndex->GetX(), theReIndex->GetY(), theReIndex->GetN());

}

//_______________________________________________________________________________
void MaterialParameters::setImaRI(const char* im){
    loadImaRI(im);
    setImaRI();
}

//_______________________________________________________________________________
void MaterialParameters::setImaRI(){
    int n=theImIndex->GetN();
    theAbs = new TGraph(n);
    double en, idx, abs;

    for ( int i(0); i<n; i++ ){
        theImIndex->GetPoint(i,en,idx);
        abs=hbarc/en/idx*0.5;
        theAbs->SetPoint(i,en,abs);
        }

    theTable->AddProperty("ABSLENGTH", theAbs->GetX(), theAbs->GetY(), theAbs->GetN());
}

//_______________________________________________________________________________
void MaterialParameters::setAL(const char* abs){
    if (theAbs) delete theAbs;
    theAbs = new TGraph(abs);
    this->ConvertNmToEv( theAbs );
    theTable->AddProperty("ABSLENGTH", theAbs->GetX(), theAbs->GetY(), theAbs->GetN());
}

//_______________________________________________________________________________
void MaterialParameters::SetData(G4Material* mat){
    isTableOwn=false;
    mat->SetMaterialPropertiesTable(theTable);
}

//_______________________________________________________________________________
void MaterialParameters::SetData(G4OpticalSurface* surf){
    isTableOwn=false;
    surf->SetMaterialPropertiesTable(theTable);
}

//_______________________________________________________________________________
void MaterialParameters::ConvertNmToEv(TGraph* gr){
    int n = gr->GetN();
    double* wl=gr->GetX();
    double* idx_wl=gr->GetY();

    double* ev=new double[n];
    double* idx_ev=new double[n];

    double energy;
    int inv_i;
    for ( int i(0); i<n; i++ ){
        energy=theConvConst/wl[i];
        inv_i=n-i-1;
        ev[inv_i]=energy;
        idx_ev[inv_i]=idx_wl[i];
    }

    memcpy(wl,ev,n*sizeof(double));
    memcpy(idx_wl,idx_ev,n*sizeof(double));
    delete [] ev;
    delete [] idx_ev;
}

//_______________________________________________________________________________
void MaterialParameters::setTrans(const char* trans, double unitlength){
    //
    // T = (1-Pabs(unitlength))*(1-Prefl)^2 + ...
    // Pabs(unitlength) = 1 - exp(-unitlength/ABSLENGTH)
    // Pabs = 1 - T*(1+n)^4 /4/n^2
    // ABSLENGTH = -unitlength/ln(T*(1+n)^4 /4/n^2)
    //
    if (theTransmittance) delete theTransmittance;
    if (theAbs) delete theAbs;

    theTransmittance = new TGraph(trans);
    this->ConvertNmToEv( theTransmittance );

    int n=theTransmittance->GetN();
    theAbs = new TGraph(n);
    double en, tr, abs, ridx;
    double en_min, en_max, tr0;
    double en_min1, en_max1;
    getLimits(theReIndex, en_min, en_max);
    getLimits(theTransmittance, en_min1, en_max1);
    en_min=max(en_min, en_min1);
    en_max=min(en_max, en_max1);
    for ( int i(0); i<n; i++ ){
        theTransmittance->GetPoint(i,en,tr);
        if ( en<en_min ) continue;
        if ( en>en_max ) break;
        ridx=theReIndex->Eval(en, 0, "S");
//        printf("RealIndex  %g   Wavelength  %g\n",ridx,theConvConst/en);
        tr0=4.*ridx/((1.+ridx)*(1.+ridx));
        tr0*=tr0;
        if ( tr>tr0 ){
            double old=tr;
            tr=2.*tr0-tr;
           printf("Correcting transmittivity to be less then square of reflection probability (%.5f -> %.5f)", old, tr);
        }
        abs=unitlength/log(tr0/tr);
        theAbs->SetPoint(theAbs->GetN(),en,abs);

#ifdef DEBUG
            printf("output/trans_to_abslength.dat   w   %.5f\t%.10f\n", theConvConst/en, abs);
            printf("output/trans_to_r2.dat   w   %.5f\t%.10f\n", theConvConst/en, tr0);
#endif
    }
    printf(Form("Converted transmittivitu to ABSLENGTH with limits: %.0f nm - %.0f nm", theConvConst/en_max, theConvConst/en_min));

    theTable->AddProperty("ABSLENGTH", theAbs->GetX(), theAbs->GetY(), theAbs->GetN());
}

//______________________________________________________________________________
void MaterialParameters::getLimits(TGraph* gr, double& min, double& max){
    int n = gr->GetN();
    if ( !n ) return;
    double yy;
    gr->GetPoint(0, min, yy);
    gr->GetPoint(n-1, max, yy);
}