#include "OptSteppingAction.hh"
#include "OptUserStackingAction.hh"
#include "Photon.hh"

#include <G4SteppingManager.hh>
#include <G4Track.hh>
#include <G4Step.hh>
#include <G4StepPoint.hh>
#include <G4TrackStatus.hh>
#include <G4VPhysicalVolume.hh>
#include <G4ParticleDefinition.hh>
#include <G4ParticleTypes.hh>
#include <G4String.hh>
#include <G4ProcessManager.hh>
#include <G4ProcessVector.hh>
#include <G4OpBoundaryProcess.hh>
#include <G4TrackStatus.hh>
#include <G4EventManager.hh>
#include <G4Event.hh>
#include <G4MaterialPropertiesTable.hh>

#include <TTree.h>
#include <TVector3.h>
#include <TObject.h>
#include <TObjString.h>

#include "SimuApplication.hh"
#include "ConfigFileParser.hh"
#include "EsafConfigurable.hh"
#include "Config.hh"
#include <iostream>
using namespace std;

const char* oprocess[] ={  "Undefined",
                                  "FresnelRefraction", "FresnelReflection",
                                  "TotalInternalReflection",
                                  "LambertianReflection", "LobeReflection",
                                  "SpikeReflection", "BackScattering",
                                  "Absorption", "Detection", "NotAtBoundary",
                                  "SameMaterial", "StepTooSmall", "NoRINDEX" };

// Detector subsystem identifier
enum EDetectorSystem {
    keUndefined               = 0,
    kNotInside                = 1,
    kBaffle                   = 2,
    kOptics                   = 3,
    kFocalPlane               = 4,
    kWalls                    = 5,
    kEarth                    = 6,
    kG4FocalSurface           = 7
};

OptSteppingAction::OptSteppingAction(OptUserStackingAction* stack) : fStepCounter(0), fMaxIterations(1000)
{
    //theStatus = Undefined;
    fBoundary = NULL;
    fProcessManager = NULL;
    fProcessVector = NULL;
    fStatus = NULL;
    fStacking = stack;
    ConfigFileParser* conf = Config::Get()->GetCF("G4","G4Detector");
    string det=conf->GetStr("G4Detector.DetectorType");
    if (det.find("Tus")!=string::npos)fTus=true;
    else fTus=false;
    SetSender("OptSteppingAction");

}

OptSteppingAction::~OptSteppingAction()
{

}

void OptSteppingAction::UserSteppingAction(const G4Step * theStep)
{
    G4VPhysicalVolume* prevol=theStep->GetPreStepPoint()->GetPhysicalVolume();
    G4VPhysicalVolume* postvol=theStep->GetPostStepPoint()->GetPhysicalVolume();

    G4String PreStepVolume,PostStepVolume;
    if(prevol) PreStepVolume = prevol->GetName();
    if(postvol) PostStepVolume = postvol->GetName();

    fStepCounter++;
    if ( fStepCounter>fMaxIterations ){
        theStep->GetTrack()->SetTrackStatus(fStopAndKill);
        MsgForm(EsafMsg::Warning,"This photon exceeded the maximum number of iterations (%i), killed.",fMaxIterations);
    }

    //find the boundary process only once
    if( !fBoundary ){
        fProcessManager = theStep->GetTrack()->GetDefinition()->GetProcessManager();
        G4int nprocesses = fProcessManager->GetProcessListLength();
        fProcessVector = fProcessManager->GetProcessList();

        for(int i=0;i<nprocesses;i++){
            if( (*fProcessVector)[i]->GetProcessName()=="OpBoundary" ){
                fBoundary = (G4OpBoundaryProcess*)(*fProcessVector)[i];
                break;
            }
        }
    }

    theStatus = Undefined;
    if(fBoundary)theStatus = fBoundary->GetStatus();

    G4StepPoint *PostStep = theStep->GetPostStepPoint();
    G4StepPoint *PreStep = theStep->GetPreStepPoint();

    G4VPhysicalVolume* VolumePost = PostStep->GetPhysicalVolume();
    G4VPhysicalVolume* VolumePre = PreStep->GetPhysicalVolume();
    G4ThreeVector position =  PreStep->GetPosition();
    G4ThreeVector direction =  PreStep->GetMomentumDirection();
    G4ThreeVector pposition =  PostStep->GetPosition();
    G4ThreeVector pdirection =  PostStep->GetMomentumDirection();

    G4String Namepr = VolumePre->GetName();
    double rho = position.perp();
    double r_cut = 950.;
    double r_wall = 1325.;

    if((fabs(position.y())>r_cut || rho>r_wall) && !fTus){
        theStep->GetTrack()->SetTrackStatus(fStopAndKill);
        fStatus=-4;
    }

    double mult=+1.;
    if ( position.x()<0 )mult=-1.;

    if ( fStatus!=-4 && PostStep->GetStepStatus()==fPostStepDoItProc &&
         theStep->GetTrack()->GetTrackStatus()==fStopAndKill ) fStatus=-3;

#ifdef DEBUG
        int id = theStep->GetTrack()->GetTrackID();
        G4Material* mat=PreStep->GetMaterial();
        double l1(-1.),l2(-1.);
        double e=PreStep->GetTotalEnergy();
        G4String name1,name2;
        if (mat){
            name1=mat->GetName();
            G4MaterialPropertiesTable* mt=mat->GetMaterialPropertiesTable();
            if (mt){
                G4MaterialPropertyVector* mv=mt->GetProperty("ABSLENGTH");
                l1=mv?mv->GetProperty(e):-2.;
            }
        }
        mat=PostStep->GetMaterial();
        if (mat){
            name2=mat->GetName();
            G4MaterialPropertiesTable* mt=mat->GetMaterialPropertiesTable();
            if (mt){
                G4MaterialPropertyVector* mv=mt->GetProperty("ABSLENGTH");
                l2 =mv?mv->GetProperty(e):-2.;
            }
        }
//         printf("e=%.2f, wl=%.2f, l1=%g l2=%g (%s, %s) len=%g\n",
//             e/eV, hbarc*twopi/e/nm,
//             l1, l2,
//             name1.data(),name2.data(), theStep->GetStepLength());

        printf("    %i (%i) %s: %s(%s) -> %s(%s) (%i -> %i)\n", theStep->GetTrack()->GetCurrentStepNumber(), id,
                                             oprocess[theStatus],
                                             VolumePre?VolumePre->GetName().data():"xxx",name1.data(),
                                             VolumePost?VolumePost->GetName().data():"xxx",name2.data(),
                                             PreStep->GetStepStatus(),
                                             PostStep->GetStepStatus());
        printf("    p(%g, %g), d(%g, %g) -> %g -> p(%g, %g), d(%g, %g) %s (%i)\n",
               position.perp(), position.z(),
               direction.theta()/deg, direction.phi()/deg,
               theStep->GetStepLength(),
               pposition.perp(), pposition.z(),
               pdirection.theta()/deg, pdirection.phi()/deg,
               theStep->GetTrack()->GetTrackStatus()==fStopAndKill?"stop":"", fStatus);
#endif

    if(VolumePost){
        G4String NamePost = VolumePost->GetName();
        //if(NamePost!="absorber" && NamePost!="World"){
        if ( NamePost.contains("FresnelLens") || Namepr.contains("FresnelLens")||Namepr.contains("mirror_phys") ||NamePost.contains("mirror_phys") ){
            //||Namepr.contains("mirror_phys") ||NamePost.contains("mirror_phys") for tus
            if(fStatus==0) fStatus++;
            //printf("%i\n",fStatus);
        }

        if(NamePost=="absorber"){
            theStep->GetTrack()->SetTrackStatus(fStopAndKill);
            fStatus=(fStatus>0?-2:-1);
            //fStatus=-2;
        }
    }

    G4ThreeVector position2 =  PostStep->GetPosition();
    G4ThreeVector direction2 =  PostStep->GetMomentumDirection();

    if ( theStatus==Absorption || theStatus==Detection ){
        fStatus=-3;
    }
    G4TrackStatus TrackStatus = theStep->GetTrack()->GetTrackStatus();

    if(TrackStatus==fStopAndKill){
        fStepCounter=0;
         int history;
         if ( position2.z()<0 && direction2.z()<0  ) history=kEarth;
         else if ( fStatus==-2 ) history=kG4FocalSurface;
         else if ( fStatus==-3) history=kOptics;
         else if ( fStatus==-4) history=kWalls;
         else history=keUndefined;

         fStacking->SetTrackHistory(history);
         fStatus=0;
    }
}
