// $Id: SimpleDetector.hh 2845 2010-02-25 05:43:43Z fenu $
// Author: Sylvain Moreggia   2006/03/16

/*****************************************************************************
 * ESAF: Euso Simulation and Analysis Framework                              *
 *                                                                           *
 *  Id: SimpleDetector                                                           *
 *  Package: electronics                                                   *
 *  Coordinator: Marco.Pallavicini                                               *
 *                                                                           *
 *****************************************************************************/

#ifndef __SIMPLEDETECTOR_HH__
#define __SIMPLEDETECTOR_HH__

#include "euso.hh"
#include "Detector.hh"
#include "EarthVector.hh"

////////////////////////////////////////////////////////////////////////////////
//                                                                            //
// SimpleDetector                                                                 //
//                                                                            //  
// <brief class description>                                                  //  
//                                                                            //  
////////////////////////////////////////////////////////////////////////////////

class PhotoElectron;
class Photon;


class SimpleDetector :  public Detector,  public EsafConfigurable, public EsafMsgSource {
public:
    
    // ctor
    SimpleDetector();
    
    // dtor
    virtual ~SimpleDetector();

    // do the complete simulation of one event
    virtual Telemetry* Get( PhotonsOnPupil* );

    // get ready for next event
    virtual void Reset();

    // initialization
    void InitDetector();

    // apply optics effect
    Int_t ApplyOpticsFilter(const Photon& ph) const;
    
    // Create PhotoElectrons using Detection Quantum Efficiency (= Collection eff. * Quantum eff.)
    Int_t ApplyDQE(const Photon& ph) const;
    
    // interpolation functions according to wl
    Double_t InterpolateBG3(Double_t) const;
    Double_t InterpolateQE(Double_t) const;
    
    // orders pe- chronologically
    // then find the time of first GTU and ranges them in the map using GTU info
    // find starting position of shower point (taken at fFirstGTUTime)
    // set pixel ID as well
    void SpaceTimePEinfos(const vector<PhotoElectron*>&);
    
    // find pe- pixel
    // - using sqrt(x*x + y*y) value in the MES Z=0 plane
    // - i,j in [0..n-1]  (i<-->row    j<-->column)
    // - pixeldID = N --> i,j coordinates found back using N / n = i  and  N % n = j
    // Check if position is in the FoV, return kFALSE if not
    Bool_t FindPixelID(const EarthVector&,UInt_t&) const;
    
    // get pixel 2D position in the Mes Z=0 plane
    // return kTRUE if pixel is in the FoV, kFALSE otherwise
    Bool_t Pos2D_Z0plane(UInt_t,EarthVector&) const;
    
    // get lateral distance to the track of given pixel
    //Bool_t LateralDist2ToTrack(UInt_t, UInt_t&) const;   // disabled so far
        
    // add background around showerpoint pixels (no more around the whole showertrack !!)
    // For every pe-, set the spatial shift w.r.t. shower point
    // returns nb of bgnd pe-
    Int_t AddBackground();
    
    // add background around direct fluorescence pixels at a given GTU
    // needed for very inclined track, i.e. when total nb of bgnd pe- of previous method becomes huge
    // For every pe-, set the spatial shift w.r.t. shower point
    // returns nb of bgnd pe-
    //Int_t AddMinimizedBackground();  // disabled so far
    
    // kind of trigger simulation
    // Tag pe- according to the SNR per pixel per GTU : if (S + B) >= (<B> + 3*sigmaB), pixel/GTU considered as triggered
    // returns nb of triggered pe-
    Int_t ApplyTrigger();
    
    // calculate mean background per pixel per GTU (Nadir incidence)
    Double_t MeanBgndPerPixelPerGTU(Double_t);
    
    // getters
    UInt_t GetNbPhotoElectrons() const {return fPhotoElectrons.size();}
    
    // physically release the memory allocated by the arrays of this object
    virtual Bool_t ClearMemory() {return kTRUE;}


    EsafConfigClass(Electronics,SimpleDetector)
    
private:
    // configs
    Double_t fReducedSurf;                                      // factor of reduction of collecting surface
    Bool_t fTriggEff;                                           // true : fraction of light correctly focused (in 5mm bucket, optical aberration, red book)
    Double_t fBG3filter[400];                                   // tabulated value for BG3 filter
    Double_t fPMQE[51];                                         // tabulated values for PM Quantum Efficiency
    Double_t fGTUlength;                                        // GTU length
    Double_t fPixelSideLength;                                  // pixel side length projected on the MES plane Z=0
    Double_t fFirstPixelX;                                      // pixel high left corner X-position projected on the MES plane Z=0
    Double_t fFirstPixelY;                                      // pixel high left corner Y-position projected on the MES plane Z=0
    UInt_t fNbPixelOnSide;                                      // nb of pixels on a side of the square FS
    Double_t fMeanBgnd;                                         // value in ph-on-pupil/m2/ns/sr
    Double_t fSNRtrigg;                                         // nb of background sigma considered for SNR treatment
    UInt_t fNpersTrigg;                                         // persistency = nb of GTU with adjacents hitted pixels
    string fBgndWl;                                             // RedBook (within 300-400nm) or Extended (extrapolation within 200-600nm)
    string fBgndMode;                                           // standard (added all along track, all GTU), minimized (around fluo pixel only, at one GTU)
    string fBG3type;                                            // BG3 (long tail above 400nm), or BG3_coated (RedBook config, sharp cut above 400nm)
    UInt_t fNbMaxLateralPixels;                                 // define maximum area around track to be considered in the simulation
    UInt_t fMaxLateralDist2;                                    // = fNbMaxLateralPixels * fNbMaxLateralPixels
    Bool_t fCloudStatus;                                        // false : photon absorbed by cloud are kept for simu
    
    // event data
    map<UInt_t, vector<PhotoElectron*>* > fPhotoElectrons;      // list of photoelectrons, stored per GTU (index=0 <--> GTUnb=0)
    Bool_t fEvIsTriggered;                                      // tells if event is triggered
    UInt_t fNbGTUs;                                             // nb of GTUs containing pe- (GTU id within [0,fNbGTUs-1])
    Int_t fNbSignalPE;                                         // nb of pe- coming from signal photons
    Double_t fFirstGTUTime;                                     // time of first GTU (raising time)
    Double_t fFirstPETime;                                      // time of first pe-
    Double_t fLastPETime;                                       // time of last pe-
    EarthVector fStartPosShowPt;                                // MES 3D position of shower point at fFirstGTUTime
    EarthVector fEndPosShowPt;                                  // MES 3D position of shower point at end of track (earth impact or Fov bounds)
    EarthVector fDirShowPt;                                     // shower track 3D direction in MES
    UInt_t fStartPixelID;                                       // define first corner of covered rectangular area (bgnd generation)
    UInt_t fEndPixelID;                                         // define second corner of covered rectangular area (bgnd generation)

    ClassDef(SimpleDetector,0)
};

#endif  /* __SIMPLEDETECTOR_HH__ */

