// $Id$
// Author: fenu   2007/12/05

/*****************************************************************************
 * ESAF: Euso Simulation and Analysis Framework                              *
 *                                                                           *
 *  Id: LTTTrigger                                                           *
 *  Package: <packagename>                                                   *
 *  Coordinator: <coordinator>                                               *
 *                                                                           *
 *****************************************************************************/

//_____________________________________________________________________________
//
// LTTTrigger
//
// <extensive class description>
//
//   Config file parameters
//   ======================
//
//   <parameter name>: <parameter description>
//   -Valid options: <available options>

/*
Algorithm description
The LTT performs an integration over a predefined set of directions.
If the integration is above a certain value a trigger is issued.

*In the standard procedure red pixels are chosen as initial condition. 
    -> Or in the "special procedure the PTT trigger pixel is chosen
*The 16 directions are to be read in the direction file.
*A box of 2X2 pixels will be created around the red pixel.
*Starting from the red pixel and according to the 16 directions the box will be translated in the time frame -4 +4 GTUs 
 with respect to the triggering GTU.
* NEW the LTT can also use as starting condition the highest trigger of the PTT. One trigger per EC is used in this case.

*in Simulate the search for red pixels is performed.
*Integrate performs the integration and fills all the information in the .root file and
 in the ChipGtudata Object. In this last object all the informations are saved for the third level trigger
*Calculate_Directions makes a first transformation between the angle and focal surface reference system
*CorrectTransformation makes the correct translation of the box. REMEMBER: PIXEL ORDER INSIDE THE EC IS NOT AS EXPECTED.
 To translate the box a transformation  must be performed. In the new system the box can be translated 
 and then the system has to be transformed back again in the original system. 
*IntegrateCounts makes the incrementation of the integratin counter taking into account all the transformations and the 
 changes of EC
*Debub_tables make two .txt tables for visualising he movements of the box. Use just debugging and for 1 event.
*LTTTrigger.fChainPTT_LTT makes the trigger start from the PTT trigger. Remember in this case to set parameter in macrocell
 to activate also the PTT.
*/
//
// PLEASE READ:
// If using LTTTrigger.EC_PDM=1
// remember to put EusoElectronics.fNightGlowCode = -2
// to work at the level of PDM you will have to simulate the whole PDM!
// Otherwise 
// LTTTrigger.EC_PDM=0
//
// If you change something on the layout structure please make a test to 
// check if the box  is properly translated
// Especially if you change the definition of the geometry 
// within one PMT this will for sure affect the LTT algorithm
// put LTTTrigger.fDebug=1 and run just one event.
// Look to Debug_Table1 and Debug_Table2 (end of this file) for understanding the meaning 
// of the debug table. Names used there are self explaining

//
// FOR TRIGGER DEFINITION: if Side cut is used put a limit of +-170km on the y axis. 
//      Otherwise in the efficiency plot a plateau of 80% will be reached nut never 100%
//      This is because part of the FOV is not covered for sidecut.
//      in Y direction FOV +-22deg while in X is +-30
//

//To read the .root file use 
//
//   TFile  *f = new TFile("nameOffile.root");
//   TTree *etree = (TTree*)f->Get("etree");
//   EEvent *evs = new EEvent();
//   evs->SetBranches(etree);
//   etree->GetEntry(ii);//ii number of event to read

// after that 
// evs->GetLTTTrigger()->GetLTTTrackNum()
// will give the number of PDMs having at least one triggering track
// The event will have triggered if evs->GetLTTTrigger()->GetLTTTrackNum()>0
// For other useful informations on the trigger
// look in the file packages/common/root/include/ELTTTriggerSegment.hh
//
//Trick: if the integration track is cut add so many counts as necessary to complete the track.
//This is in order to avoid that the algorithm integrates on short tracks with high threshold.
#include "LTTTrigger.hh"
#include "MacroCellData.hh"
#include "MacroCell.hh"
#include "ERunParameters.hh"
#include "NumbersFileParser.hh"
#include "ELTTTriggerDataAdder.hh"
#include "LTTTriggerSegment.hh"
#include "EEvent.hh"
#include "EHeader.hh"
#include "EsafRandom.hh"

ClassImp (LTTTrigger)
//_____________________________________________________________________________
LTTTrigger::LTTTrigger ():TriggerEngine (string ("LTTTrigger"), kLTTTrigger)
{
  //
  // Constructor
  //

  fRedThreshold = (Int_t) Conf ()->GetNum ("LTTTrigger.fRedThreshold");
  fYellowThreshold = (Int_t) Conf ()->GetNum ("LTTTrigger.fYellowThreshold");
  fIntegrateSum = (Int_t) Conf ()->GetNum ("LTTTrigger.fIntegrateSum");
  flag_EC = (Int_t) Conf ()->GetNum ("LTTTrigger.fEC_PDM");
  fPixel_size = (Float_t) Conf ()->GetNum ("LTTTrigger.fPixel_size");
  fGtuLight = (Float_t) Conf ()->GetNum ("LTTTrigger.fGtu_light");
  fNum_Dir =  (Int_t) Conf ()->GetNum ("LTTTrigger.fNum_Dir");

  fNum_Steps =  (Int_t) Conf ()->GetNum ("LTTTrigger.fNum_Steps");

  fChainPTT_LTT= (Int_t) Conf ()->GetNum ("LTTTrigger.fChainPTT_LTT");

  //read and calculate path of the box on the chip
  // starting from the directions given as input
  fDataDirectory = Conf ()->GetStr ("LTTTrigger.fDataDirectory");
  fDirectionFileName = Conf ()->GetStr ("LTTTrigger.fDirectionFileName");
  fDirectionFileName = fDataDirectory + "/" + fDirectionFileName;
  fDirectionFileName =
    Conf ()->GetCfgDir () + "/" + Conf ()->GetType () + "/" +
    fDirectionFileName;
  Flag_debug_table = (Int_t) Conf ()->GetNum ("LTTTrigger.fDebug");

  ConfigFileParser* pConf = Config::Get()->GetCF("Electronics","MacroCell");
  fGTU = pConf->GetNum("MacroCell.fGtuTimeLength");
  fGTU/=1000.;

  pConf = Config::Get()->GetCF("Electronics","EusoElectronics");
  fBack = pConf->GetNum("EusoElectronics.fNightGlowRateOnAxis");

  NumbersFileParser *
    dummy2 = new NumbersFileParser (fDirectionFileName, 2);


  for (Int_t j (0); j < fNum_Dir; j++)
    {

      vector < Double_t > dummy2_line = dummy2->GetRow (j);

      for (size_t k (0); k < dummy2_line.size (); k++)
	{

	  angle.push_back (dummy2_line[k]);
	}
      dummy2_line.clear ();
    }
  Calculate_Directions ();

  angle.clear ();
}

//_____________________________________________________________________________
LTTTrigger::~LTTTrigger ()
{
  //
  // Destructor
  //

  cout << "Destroy " << endl;
  Clear ();
}

//______________________________________________________________________________
void
LTTTrigger::Simulate (MacroCellData * pData)
{
  //
  // Simulate LTT algorythm
  //
// finds the red pixels and starting from these conditions calls the integration method

  Int_t matsize = pData->Cell ()->GetEC (0)->FrontEnd ()->Channels ();
  Int_t side_EC = sqrt (matsize);
  Int_t side_PMT = side_EC / 2;

  // get map of ChipGtuData elements
  map < Int_t, map < Int_t, ChipGtuData * >*>&m1 = pData->fChipData2;
  map < Int_t, map < Int_t, ChipGtuData * >*>::const_iterator it1;

  Int_t ec_number = 1;
// start from red pixel
if(fChainPTT_LTT==0){
  Msg (EsafMsg::Info) << "Simulating LTT algorythm (Bertaina) starting from RED pixel" << MsgDispatch;
  // iterate on CHIPS in this macrocell
  for (it1 = m1.begin (); it1 != m1.end (); it1++)
    {
      Int_t chip_id = it1->first;
      map < Int_t, ChipGtuData * >&m2 = *(it1->second);
      map < Int_t, ChipGtuData * >::const_iterator it2;

      // iterate on GTU for the current chip
     UInt_t GTU_NUM=0;
     for (it2 = m2.begin (); it2 != m2.end (); it2++)
	{
	  if (GTU_NUM>15 && GTU_NUM<m2.size()-15)//analyze just within borders Check also in Standard Macrocell
	    //change of analysis range!
	  for (Int_t iteratore = 0; iteratore < matsize; iteratore++)
	    {

	      //writes first debug table
	      Int_t nn_PMT =
		it2->second->FrontEnd ()->PmtChannel (iteratore) / side_PMT;
	      Int_t mm_PMT =
		it2->second->FrontEnd ()->PmtChannel (iteratore) % side_PMT;
	      if (Flag_debug_table)
		Debug_Table1 (it2->second->FrontEnd ()->Pmt (iteratore)->
			      Geometry ()->Position (nn_PMT, mm_PMT).X (),
			      it2->second->FrontEnd ()->Pmt (iteratore)->
			      Geometry ()->Position (nn_PMT, mm_PMT).Y (),
			      it2->second->FrontEnd ()->Pmt (iteratore)->
			      Geometry ()->Position (nn_PMT, mm_PMT).Z (),
			      it2->second->FrontEnd ()->Row (iteratore),
			      it2->second->FrontEnd ()->Column (iteratore),
			      it2->first, chip_id);
//is the pixel above red threshold?
	      if (it2->second->GetCounter (iteratore) >= fRedThreshold)
		{

		  //Write on the event array
		  //This array helps to find the first position of each event
		  //refere to this when you're analyzing the events

		  pos_PDM_map.push_back (PDM_Map.size ());

		  PDM_Map.push_back (chip_id);
		  PDM_Map.push_back (it2->second->FrontEnd ()->
				     Row (iteratore));
		  PDM_Map.push_back (it2->second->FrontEnd ()->
				     Column (iteratore));
		  PDM_Map.push_back (it2->first);	// In wich gtu this event is

		}
	    }			//for(Int_t iteratore=0;iteratore<matsize ;iteratore++) {
	  GTU_NUM++;
	}			//for(it2=m2.begin(); it2!=m2.end(); it2++) {



      ec_number++;

    }				//for(it1=m1.begin(); it1!=m1.end(); it1++) {

  // call integration method//////////////////////////////////////////////////////////////////////////
  integrate (PDM_Map, pos_PDM_map, pData,side_PMT,side_EC);

}else if(pData->GetPTTTriggerMaxCounts ()>-1) {
//reading from PTT Trigger output
  Msg (EsafMsg::Info) << "Simulating LTT algorythm (Bertaina) starting from PTT trigger" << MsgDispatch;

                  pos_PDM_map.push_back (PDM_Map.size ());
                  PDM_Map.push_back (pData->GetPTTTriggerChipID ());
	          PDM_Map.push_back (pData->GetPTTTriggerMaxRow ());
                  PDM_Map.push_back (pData->GetPTTTriggerMaxCol ());
                  PDM_Map.push_back (pData->GetPTTTriggerMaxGtu ());

  // call integration method//////////////////////////////////////////////////////////////////////////
		  integrate (PDM_Map, pos_PDM_map, pData,side_PMT,side_EC);

	}

  //Clear vectors. 
  Clear ();
}

//______________________________________________________________________________
void
LTTTrigger::Clear ()
{
/////////////////////////////////////////////////////////////////////////////////////
//clear/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////
  PDM_Map.clear ();
  pos_PDM_map.clear ();

}

//_____________________________________________________________________________________
void
LTTTrigger::integrate (vector < int >PDM_map_arr,
		       vector < int >PDM_map_arr_pos, MacroCellData * pData,Int_t side_PMT,Int_t side_EC)
{
  /////////////////////////////////////////////////////////////////////////////////////
  //performs the integration///////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////////////

  Int_t up_counts = 0;
  Int_t dir_up;
  Int_t chip_ID_up;
  Int_t red_ID_up;
  Int_t flag_up = 1;
  Int_t GTU_up;
  Int_t row_up;
  Int_t col_up;
  Int_t offset_row,offset_col;
  // Cycle over all the red pixels
  for (UInt_t i = 0; i < PDM_map_arr_pos.size (); i++)
    {
      //position of the ec in the pdm
      Double_t divide = (PDM_map_arr[PDM_map_arr_pos[i]] - 1) / 9.;
      Double_t num_PDM_DOUB;
      Double_t rest = modf (divide, &num_PDM_DOUB);
      rest *= 9;
      Double_t integ;

      Double_t funz = round (1.e10 * rest / 3.) * 1.e-10;
      //round at the 10th digit. Otherwise rounding problems arise
      // for example: modf (1.9999999999999998,&int) is different from 
      //  modf (1.9999999999999999,&int) 
      // why? 

      rest = funz;
      Double_t rest2 = modf (rest, &integ);


      Int_t row_PDM = integ;
      Int_t col_PDM = rest2 * 3.;

// CYCLE OVER DIRECTIONS
      for (Int_t dir = 0; dir < fNum_Dir; dir++)
	{


//Cycle over +-4 GTUS
	  Int_t counts = 0;
	  for (Int_t f = -fNum_Steps; f <= fNum_Steps; f++)
	    {
 Int_t analyzed_gtu = PDM_map_arr[PDM_map_arr_pos[i] + 3] + f;
//CALCULATING OFFSETS//////////////////////////////////////////////////////////
	      Float_t ffl = f;
	      Float_t offset_row_fl = ffl * kx_ky[2 * dir];
	      Float_t offset_col_fl = ffl * kx_ky[(2 * dir) + 1];

	      offset_row_fl = offset_row_fl + 0.5;
	      offset_col_fl = offset_col_fl + 0.5;

	      Int_t offset_row_int;
	      Int_t offset_col_int;

	      Double_t offset_row_dob;
	      Double_t offset_col_dob;

	      Float_t diff_off_row = modf (offset_row_fl, &offset_row_dob);
	      Float_t diff_off_col = modf (offset_col_fl, &offset_col_dob);

	      offset_row_int = offset_row_dob;
	      offset_col_int = offset_col_dob;

	      diff_off_row = fabs (diff_off_row);
	      diff_off_col = fabs (diff_off_col);



	      if (offset_row_fl >= 0)
		{
		  if (diff_off_row < 0.5)
		    offset_row = offset_row_int - 1;
		  else
		    offset_row = offset_row_int;
		}
	      else
		{
		  if (diff_off_row < 0.5)
		    offset_row = offset_row_int - 1;
		  else
		    offset_row = offset_row_int - 2;
		}


	      if (offset_col_fl >= 0)
		{
		  if (diff_off_col < 0.5)
		    offset_col = offset_col_int - 1;
		  else
		    offset_col = offset_col_int;
		}
	      else
		{
		  if (diff_off_col < 0.5)
		    offset_col = offset_col_int - 1;
		  else
		    offset_col = offset_col_int - 2;
		}

	      //CALCULATING OFFSETS (END)//////////////////////////////////////////////////////////
	      //if gtu analyzed_gtu has data (if it exists on the map m2) start calculation of integral

	      Int_t posit_row;
	      Int_t posit_col;
	      Int_t flag_inside;

//makes the integration:
// first calculate the new coordinate.
// then increment the counter
           CorrectTransformation (PDM_map_arr[PDM_map_arr_pos[i] + 1],
                                   PDM_map_arr[PDM_map_arr_pos[i] + 2],
                                   offset_row, offset_col, &posit_row,
				  &posit_col, &flag_inside,side_PMT,side_EC);

	      IntegrateCounts (&counts, analyzed_gtu, posit_row, posit_col,
			       flag_inside, pData,
			       PDM_map_arr, PDM_map_arr_pos, i, row_PDM,
			       col_PDM, dir, 1,side_PMT,side_EC);
	      //up/////////////////////////////////////////////////////
CorrectTransformation (PDM_map_arr[PDM_map_arr_pos[i] + 1],
                                   PDM_map_arr[PDM_map_arr_pos[i] + 2],
                                   offset_row - 1, offset_col, &posit_row,
		       &posit_col, &flag_inside,side_PMT,side_EC);


	      IntegrateCounts (&counts, analyzed_gtu, posit_row, posit_col,
			       flag_inside, pData,
			       PDM_map_arr, PDM_map_arr_pos, i, row_PDM,
			       col_PDM, dir, 2,side_PMT,side_EC);
	      //right//////////////////////////////////////////////////
              CorrectTransformation (PDM_map_arr[PDM_map_arr_pos[i] + 1],
                                     PDM_map_arr[PDM_map_arr_pos[i] + 2],
                                     offset_row, offset_col + 1, &posit_row,
                                     &posit_col, &flag_inside,side_PMT,side_EC);


	      IntegrateCounts (&counts, analyzed_gtu, posit_row, posit_col,
			       flag_inside, pData,
			       PDM_map_arr, PDM_map_arr_pos, i, row_PDM,
			       col_PDM, dir, 3,side_PMT,side_EC);

	      //up-right///////////////////////////////////////////////
              CorrectTransformation (PDM_map_arr[PDM_map_arr_pos[i] + 1],
                                     PDM_map_arr[PDM_map_arr_pos[i] + 2],
                                     offset_row - 1, offset_col + 1,
                                     &posit_row, &posit_col, &flag_inside,side_PMT,side_EC);


	      IntegrateCounts (&counts, analyzed_gtu, posit_row, posit_col,
			       flag_inside, pData,
			       PDM_map_arr, PDM_map_arr_pos, i, row_PDM,
			       col_PDM, dir, 4,side_PMT,side_EC);


	    }			// closing   for(Int_t f=-4;f<5;f++){

	  if (counts >= fIntegrateSum && up_counts < counts)	//if triggering assign to _up variables
	    {			// the informations regarding this red pixel
	      up_counts = counts;
	      dir_up = dir;
	chip_ID_up = PDM_map_arr[PDM_map_arr_pos[i]];
	      red_ID_up = i;
	      flag_up = 0;
            GTU_up = PDM_map_arr[PDM_map_arr_pos[i] + 3];
            row_up = PDM_map_arr[PDM_map_arr_pos[i] + 1];
            col_up = PDM_map_arr[PDM_map_arr_pos[i] + 2];

	    }

	}			//closing for(Int_t dir=1;dir<16;dir++){
    }				// closing   for(UInt_t i=0;i<PDM_map_arr_pos.size();i++){

  if (flag_up == 0)
    {
      Msg (EsafMsg::
	   Info) << "TRIGGERING... NO MORE THAN 1 TRACK X 1 Macro Cell " <<
	MsgDispatch;
      Msg (EsafMsg::
	   Info) << "Integrated " << up_counts << " in direction " << dir_up
	<< " ( Theta= " << angle[2 * dir_up] << "  Phi= " << angle[2 *
								   dir_up +
								   1] <<
	" )    Chip ID " << chip_ID_up << " Red ID " << red_ID_up <<
	MsgDispatch;

      //FILLING the INFORMATION VECTOR              

      FillInfoVector(up_counts,GTU_up,row_up,col_up,chip_ID_up,dir_up,EEvent::GetCurrent ()->GetHeader ()->GetNum ());

 FillChipData(EEvent::GetCurrent()->GetHeader()->GetNum(),up_counts,GTU_up,row_up,col_up,chip_ID_up,dir_up,pData);
    }
  else
    {
    FillChipData(-1,-1,-1,-1,-1,-1,-1,pData);
  
    }
}

void
LTTTrigger::FillInfoVector(Int_t up_counts,Int_t  GTU_up,Int_t row_up,Int_t col_up,Int_t chip_ID_up,Int_t dir_up,Int_t Triggering){
 
    fTracks.push_back (LTTTriggerSegment ());
      LTTTriggerSegment & trkseg = fTracks.back ();
      trkseg.SetNumEvt (EEvent::GetCurrent ()->GetHeader ()->GetNum ());
      trkseg.SetMaxCount (up_counts);
      trkseg.SetMaxGtu (GTU_up);
      trkseg.SetMaxRow (row_up);
      trkseg.SetMaxCol (col_up);
      trkseg.SetMaxChipID (chip_ID_up);
      trkseg.SetMaxDir (dir_up);

  vector < LTTTriggerSegment >::iterator it;

  for (it = fTracks.begin (); it != fTracks.end (); it++)
    {
      if (EEvent::GetCurrent ())
	{
	  ELTTTriggerDataAdder a (&(*it), Triggering, 1);
	  EEvent::GetCurrent ()->Fill (a);
	}
    }
  fTracks.clear ();
}

void
LTTTrigger::FillChipData(Int_t evtNum,Int_t up_counts,Int_t GTU_up,Int_t row_up,Int_t col_up,Int_t chip_ID_up,Int_t dir_up,MacroCellData* pData){
     pData->SetLTTTrigger (evtNum);
      pData->SetLTTTriggerMaxCounts (up_counts);
      pData->SetLTTTriggerMaxGtu (GTU_up);
      pData->SetLTTTriggerMaxRow (row_up);
      pData->SetLTTTriggerMaxCol (col_up);
      pData->SetLTTTriggerChipID (chip_ID_up);
      pData->SetLTTTriggerMaxDir (dir_up);

}

//______________________________________________________________________________
void
LTTTrigger::Calculate_Directions ()
{
/////////////////////////////////////////////////////////////////////////////////////
//Calculate directions///////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////


  for (UInt_t i = 0; i < (angle.size () / 2); i++)
    {

      Float_t tg_th2 = tan (TMath::DegToRad () * (angle[2 * i] / 2.));
      Float_t tg_fi = tan (TMath::DegToRad () * angle[(2 * i) + 1]);


      Float_t kx =
	sqrt (pow (tg_th2, 2) /
	      (pow (fPixel_size / fGtuLight, 2) * (1 + pow (tg_fi, 2))));

      if (angle[(2 * i) + 1] > 90. && angle[(2 * i) + 1] <= 270)
	kx = -kx;


      Float_t ky = kx * tg_fi;

      kx_ky.push_back (kx);
      kx_ky.push_back (ky);

    }


}

//_____________________________________________________________________________________
void
LTTTrigger::CorrectTransformation (Int_t pos_row, Int_t pos_col,
				   Int_t offset_row, Int_t offset_col,
				   Int_t * final_row, Int_t * final_col,
				   Int_t * flag_inside,Int_t side_PMT,Int_t side_EC)
{
  //
  //corrects the coordinates of the EC
  // to move inside the EC you need to correct the coordinate system
  //
  //the transformation is made of 2 subtrasformations

  *flag_inside = 0;

  //trasformation
  //first part
  if (pos_row < side_PMT && pos_col < side_PMT)
    {
      *final_col = pos_row;
      *final_row = pos_col;
    }

  if (pos_row > side_PMT - 1 && pos_col < side_PMT)
    {
      *final_col = pos_row - side_PMT;
      *final_row = pos_col + side_PMT;
    }

  if (pos_row < side_PMT && pos_col > side_PMT - 1)
    {
      *final_col = pos_row + side_PMT;
      *final_row = pos_col - side_PMT;
    }

  if (pos_row > side_PMT - 1 && pos_col > side_PMT - 1)
    {
      *final_col = pos_row;
      *final_row = pos_col;
    }

  //second part
  if (*final_row < side_PMT)
    *final_row += side_PMT;
  else
    *final_row -= side_PMT;

  // Translate by offset
  //if out of EC move in nearby place
  *final_row += offset_row;
  *final_col += offset_col;
  //check wether the box is inside the EC or not.
  // if not calculate  the new position and issue a flag to 
  // determine in which EC is the box going
  if (*final_row < side_EC && *final_row > -1 && *final_col < side_EC && *final_col > -1)	//move within EC
    {
      *flag_inside = 1;
    }

  if (*final_row > side_EC - 1)	//to lower EC
    {
      *flag_inside = 2;
      *final_row = *final_row - side_EC;
    }

  if (*final_row < 0)		//to upper EC
    {
      *flag_inside = 3;
      *final_row = *final_row + side_EC;
    }

  if (*final_col < 0)		//to left EC
    {
      *flag_inside = 4;
      *final_col = side_EC + *final_col;
    }

  if (*final_col > side_EC - 1)	//to right EC
    {
      *flag_inside = 5;
      *final_col = *final_col - side_EC;
    }

  if (*final_col > side_EC - 1 && *final_row > side_EC - 1)
    {
      *flag_inside = 6;
      *final_row = *final_row - side_EC;
      *final_col = *final_col - side_EC;
    }

  if (*final_col < 0 && *final_row > side_EC - 1)
    {
      *flag_inside = 7;
      *final_row = *final_row - side_EC;
      *final_col = side_EC + *final_col;
    }

  if (*final_col < 0 && *final_row < 0)
    {
      *flag_inside = 8;
      *final_row = *final_row + side_EC;
      *final_col = side_EC + *final_col;
    }

  if (*final_col > side_EC - 1 && *final_row < 0)
    {
      *flag_inside = 9;
      *final_row = *final_row + side_EC;
      *final_col = *final_col - side_EC;

    }

  if (flag_EC == 1 && *flag_inside != 1)	//if flag_EC ==1 you are integrating over each single EC if box is out of EC
    {				// put flag_inside=0 this will avoid  incrementing of counts in IntegrateCounts
      *flag_inside = 0;
    }


  //second
  if (*final_row < side_PMT)
    *final_row += side_PMT;
  else
    *final_row -= side_PMT;

 
  //Inverts transformation functions
  //first

  //first
  if (*final_row < side_PMT && *final_col < side_PMT)
    {
      Int_t aux = *final_col;
      *final_col = *final_row;
      *final_row = aux;
    }

  if (*final_row > side_PMT - 1 && *final_col < side_PMT)
    {
      Int_t aux = *final_col;
      *final_col = *final_row - side_PMT;
      *final_row = aux + side_PMT;
    }

  if (*final_row < side_PMT && *final_col > side_PMT - 1)
    {
      Int_t aux = *final_col;
      *final_col = *final_row + side_PMT;
      *final_row = aux - side_PMT;
    }

  if (*final_row > side_PMT - 1 && *final_col > side_PMT - 1)
    {
      Int_t aux = *final_col;
      *final_col = *final_row;
      *final_row = aux;
    }  
 
}

//_____________________________________________________________________________________
void
LTTTrigger::IntegrateCounts (Int_t * counts, Int_t analyzed_gtu,
			     Int_t posit_row, Int_t posit_col,
			     Int_t flag_inside, MacroCellData * pData,
			     vector < int >PDM_map_arr,
			     vector < int >PDM_map_arr_pos, Int_t i,
			     Int_t row_PDM, Int_t col_PDM, Int_t dir,
			     Int_t box,Int_t side_PMT,Int_t side_EC)
{
// once the coordinate and the chip identifier is calculated make the incrementation of the counter

  map < Int_t, map < Int_t, ChipGtuData * >*>&m1 = pData->fChipData2;
  map < Int_t, map < Int_t, ChipGtuData * >*>::const_iterator it1;
 
  TRandom *rndm = EsafRandom::Get ();
  Int_t nbckhits = rndm->Poisson (fBack*fGTU);

  if (flag_inside == 1)
    {

      map < Int_t, ChipGtuData * >&m2 =
 *(m1[PDM_map_arr[PDM_map_arr_pos[i]]]);
      if (m2.begin ()->first <= analyzed_gtu
	  && m2.end ()->first > analyzed_gtu)
	{

	  Int_t pix =
          (m2[PDM_map_arr[PDM_map_arr_pos[i] + 3]])->FrontEnd ()->
	    ChanRowCol (posit_row, posit_col);

	  //writes second debug table
	  Int_t nn_PMT =
          m2[PDM_map_arr[PDM_map_arr_pos[i] + 3]]->FrontEnd ()->
	    PmtChannel (pix) / side_PMT;
	  Int_t mm_PMT =
          m2[PDM_map_arr[PDM_map_arr_pos[i] + 3]]->FrontEnd ()->
	    PmtChannel (pix) % side_PMT;
	  if (Flag_debug_table)
          Debug_Table2 (m2[PDM_map_arr[PDM_map_arr_pos[i] + 3]]->
			  FrontEnd ()->Pmt (pix)->Geometry ()->
			  Position (nn_PMT, mm_PMT).X (),
m2[PDM_map_arr[PDM_map_arr_pos[i] + 3]]->
			  FrontEnd ()->Pmt (pix)->Geometry ()->
			  Position (nn_PMT, mm_PMT).Y (),
                         m2[PDM_map_arr[PDM_map_arr_pos[i] + 3]]->
			  FrontEnd ()->Pmt (pix)->Geometry ()->
			  Position (nn_PMT, mm_PMT).Z (), i, dir,
                        analyzed_gtu, PDM_map_arr[PDM_map_arr_pos[i]],
			  box);


	  if ((m2[analyzed_gtu])->GetCounter (pix) >= fYellowThreshold)
	      *counts += (m2[analyzed_gtu])->GetCounter (pix);  
	}
    }
  else if (flag_inside == 2)
    {
      if (row_PDM > 0)
	{
	  map < Int_t, ChipGtuData * >&m2 =
         *(m1[PDM_map_arr[PDM_map_arr_pos[i]] - 3]);
	  if (m2.begin ()->first <= analyzed_gtu
	      && m2.end ()->first > analyzed_gtu)
	    {
	      Int_t pix =
              (m2[PDM_map_arr[PDM_map_arr_pos[i] + 3]])->FrontEnd ()->
		ChanRowCol (posit_row, posit_col);
	      //writes second debug table
	      Int_t nn_PMT =
              m2[PDM_map_arr[PDM_map_arr_pos[i] + 3]]->FrontEnd ()->
		PmtChannel (pix) / side_PMT;
	      Int_t mm_PMT =
              m2[PDM_map_arr[PDM_map_arr_pos[i] + 3]]->FrontEnd ()->
		PmtChannel (pix) % side_PMT;

	      if (Flag_debug_table)
              Debug_Table2 (m2[PDM_map_arr[PDM_map_arr_pos[i] + 3]]->
			      FrontEnd ()->Pmt (pix)->Geometry ()->
			      Position (nn_PMT, mm_PMT).X (),
                            m2[PDM_map_arr[PDM_map_arr_pos[i] + 3]]->
			      FrontEnd ()->Pmt (pix)->Geometry ()->
			      Position (nn_PMT, mm_PMT).Y (),
                              m2[PDM_map_arr[PDM_map_arr_pos[i] + 3]]->
			      FrontEnd ()->Pmt (pix)->Geometry ()->
			      Position (nn_PMT, mm_PMT).Z (), i, dir,
			      analyzed_gtu,
                            PDM_map_arr[PDM_map_arr_pos[i]] - 3, box);

	      if ((m2[analyzed_gtu])->GetCounter (pix) >= fYellowThreshold)
		*counts += (m2[analyzed_gtu])->GetCounter (pix);
	    }
	}
    }
  else if (flag_inside == 3)
    {

      if (row_PDM < 2)
	{
	  map < Int_t, ChipGtuData * >&m2 =
 *(m1[PDM_map_arr[PDM_map_arr_pos[i]] + 3]);
	  if (m2.begin ()->first <= analyzed_gtu
	      && m2.end ()->first > analyzed_gtu)
	    {


	      Int_t pix =
              (m2[PDM_map_arr[PDM_map_arr_pos[i] + 3]])->FrontEnd ()->
		ChanRowCol (posit_row, posit_col);


	      //writes second debug table
	      Int_t nn_PMT =
              m2[PDM_map_arr[PDM_map_arr_pos[i] + 3]]->FrontEnd ()->
		PmtChannel (pix) / side_PMT;
	      Int_t mm_PMT =
                m2[PDM_map_arr[PDM_map_arr_pos[i] + 3]]->FrontEnd ()->
		PmtChannel (pix) % side_PMT;
	      if (Flag_debug_table)
              Debug_Table2 (m2[PDM_map_arr[PDM_map_arr_pos[i] + 3]]->
			      FrontEnd ()->Pmt (pix)->Geometry ()->
			      Position (nn_PMT, mm_PMT).X (),
                            m2[PDM_map_arr[PDM_map_arr_pos[i] + 3]]->
			      FrontEnd ()->Pmt (pix)->Geometry ()->
			      Position (nn_PMT, mm_PMT).Y (),
                              m2[PDM_map_arr[PDM_map_arr_pos[i] + 3]]->
			      FrontEnd ()->Pmt (pix)->Geometry ()->
			      Position (nn_PMT, mm_PMT).Z (), i, dir,
			      analyzed_gtu,
                            PDM_map_arr[PDM_map_arr_pos[i]] + 3, box);



	      if ((m2[analyzed_gtu])->GetCounter (pix) >= fYellowThreshold)
		*counts += (m2[analyzed_gtu])->GetCounter (pix);
	    }
	}
    }
  else if (flag_inside == 4)
    {


      if (col_PDM > 0)
	{
	  map < Int_t, ChipGtuData * >&m2 =
          *(m1[PDM_map_arr[PDM_map_arr_pos[i]] - 1]);
	  if (m2.begin ()->first <= analyzed_gtu
	      && m2.end ()->first > analyzed_gtu)
	    {



	      Int_t pix =
                (m2[PDM_map_arr[PDM_map_arr_pos[i] + 3]])->FrontEnd ()->
		ChanRowCol (posit_row, posit_col);
//writes second debug table
	      Int_t nn_PMT =
                m2[PDM_map_arr[PDM_map_arr_pos[i] + 3]]->FrontEnd ()->
		PmtChannel (pix) / side_PMT;
	      Int_t mm_PMT =
              m2[PDM_map_arr[PDM_map_arr_pos[i] + 3]]->FrontEnd ()->
		PmtChannel (pix) % side_PMT;
	      if (Flag_debug_table)
                Debug_Table2 (m2[PDM_map_arr[PDM_map_arr_pos[i] + 3]]->
			      FrontEnd ()->Pmt (pix)->Geometry ()->
			      Position (nn_PMT, mm_PMT).X (),
                            m2[PDM_map_arr[PDM_map_arr_pos[i] + 3]]->
			      FrontEnd ()->Pmt (pix)->Geometry ()->
			      Position (nn_PMT, mm_PMT).Y (),
                              m2[PDM_map_arr[PDM_map_arr_pos[i] + 3]]->
			      FrontEnd ()->Pmt (pix)->Geometry ()->
			      Position (nn_PMT, mm_PMT).Z (), i, dir,
			      analyzed_gtu,
                            PDM_map_arr[PDM_map_arr_pos[i]] - 1, box);



	      if ((m2[analyzed_gtu])->GetCounter (pix) >= fYellowThreshold)
		*counts += (m2[analyzed_gtu])->GetCounter (pix);
	    }
	}

    }
  else if (flag_inside == 5)
    {

      if (col_PDM < 2)
	{

	  map < Int_t, ChipGtuData * >&m2 =
          *(m1[PDM_map_arr[PDM_map_arr_pos[i]] + 1]);
	  if (m2.begin ()->first <= analyzed_gtu
	      && m2.end ()->first > analyzed_gtu)
	    {


	      Int_t pix =
              (m2[PDM_map_arr[PDM_map_arr_pos[i] + 3]])->FrontEnd ()->
		ChanRowCol (posit_row, posit_col);

//writes second debug table
	      Int_t nn_PMT =
              m2[PDM_map_arr[PDM_map_arr_pos[i] + 3]]->FrontEnd ()->
		PmtChannel (pix) / side_PMT;
	      Int_t mm_PMT =
              m2[PDM_map_arr[PDM_map_arr_pos[i] + 3]]->FrontEnd ()->
		PmtChannel (pix) % side_PMT;
	      if (Flag_debug_table)
              Debug_Table2 (m2[PDM_map_arr[PDM_map_arr_pos[i] + 3]]->
			      FrontEnd ()->Pmt (pix)->Geometry ()->
			      Position (nn_PMT, mm_PMT).X (),
                            m2[PDM_map_arr[PDM_map_arr_pos[i] + 3]]->
			      FrontEnd ()->Pmt (pix)->Geometry ()->
			      Position (nn_PMT, mm_PMT).Y (),
                            m2[PDM_map_arr[PDM_map_arr_pos[i] + 3]]->
			      FrontEnd ()->Pmt (pix)->Geometry ()->
			      Position (nn_PMT, mm_PMT).Z (), i, dir,
			      analyzed_gtu,
                            PDM_map_arr[PDM_map_arr_pos[i]] + 1, box);


	      if ((m2[analyzed_gtu])->GetCounter (pix) >= fYellowThreshold)
		*counts += (m2[analyzed_gtu])->GetCounter (pix);
	    }
	}

    }
  else if (flag_inside == 0)
    {
      // JUST INTEGRATION ON EC 
      //do nohing if the box is out of the EC
    }
  else if (flag_inside > 5)
    {
      Msg (EsafMsg::Info) << "FLAG OUT CORNER!!!!!!!" << MsgDispatch;
    }


}

//_____________________________________________________________________________________
//PIXEL POSITION
void
LTTTrigger::Debug_Table1 (Float_t x, Float_t y, Float_t z, Int_t row,
			  Int_t col, Int_t analyzed_gtu, Int_t chip_id)
{
  FILE *ff = fopen ("Debug_Table1.txt", "a");

  fprintf (ff, "%f  %f  %f  %d  %d  %d  %d\n", x, y, z, row, col,
	   analyzed_gtu, chip_id);

  fclose (ff);
}

//_____________________________________________________________________________________
//INTEGRATION BOX POSITION
void
LTTTrigger::Debug_Table2 (Float_t x, Float_t y, Float_t z, Int_t i, Int_t dir,
			  Int_t analyzed_gtu, Int_t chip_id, Int_t box)
{
  FILE *ff = fopen ("Debug_Table2.txt", "a");

  fprintf (ff, "%f  %f  %f  %d  %d  %d  %d  %d\n", x, y, z, i, dir,
	   analyzed_gtu, chip_id, box);

  fclose (ff);
}
