// $Id$
// Author: fenu   2007/12/05

/*****************************************************************************
 * ESAF: Euso Simulation and Analysis Framework                              *
 *                                                                           *
 *  Id: BertainaTrigger                                                           *
 *  Package: <packagename>                                                   *
 *  Coordinator: <coordinator>                                               *
 *                                                                           *
 *****************************************************************************/

//_____________________________________________________________________________
//
// PTTTrigger
//
// <extensive class description>
//
//   Config file parameters
//   ======================
//
//   <parameter name>: <parameter description>
//   -Valid options: <available options>
/*
Algorithm description:
Looks for persistency of signal over time in some EC.

1) Checks all the pixels above a certain threshold. (PTTTrigger.PixelThr )
2) Calculates the excess on a 3X3 box around the pixel with respect to this threshold 
3) Counts for how many consecutive GTUs the condition counts>PTTTrigger.PixelThr is 
   verified within one EC.
4) If the number of consecutive GTUs satisfying this condition is above a threshold 
    (PTTTrigger.Persistency) integration is started.
5) The integration has to be performed on the excesses from GTU(i-PTTTrigger.Persistency) 
   to GTU(i) on the entire EC for each pixel which was above PTTTrigger.PixelThr.
6) If the integration for at least one pixel is above a certain threshold 
   (PTTTrigger.Integration) a trigger 
   is issued.



   Methods description
*  PTTTrigger: initializer
*  ~PTTTrigger destructor
*  Simulate the entire algorithm logic  is here
*  Transform performs all the necessary transformations 
   for properly building and translating the box
*  Debug table make a table for visualising the box 
   translation and the position of all the boxes

*/
// If you change something on the layout structure please make a test to
// check if the box  is properly translated
// Especially if you change the definition of the geometry
// within one PMT this will for sure affect the PTT algorithm
// put PTTTrigger.fDebug=1 and run just one event.
// Look to Debug_Table1 and Debug_Table2 (end of this file) for understanding the meaning
// of the debug table. Names used there are self explaining

//
// FOR TRIGGER DEFINITION: if Side cut is used put a limit of +-170km on the y axis.
//      Otherwise in the efficiency plot a plateau of 80% will be reached nut never 100%
//      This is because part of the FOV is not covered for sidecut.
//      in Y direction FOV +-22deg while in X is +-30
//

//To read the .root file use
//
//   TFile  *f = new TFile("nameOffile.root");
//   TTree *etree = (TTree*)f->Get("etree");
//   EEvent *evs = new EEvent();
//   evs->SetBranches(etree);
//   etree->GetEntry(ii);//ii number of event to read

// after that
// evs->GetPTTTrigger()->GetPTTTrackNum()
// will give the number of PDMs having at least one triggering track
// The event will have triggered if evs->GetPTTTrigger()->GetPTTTrackNum()>0
// For other useful informations on the trigger
// look in the file packages/common/root/include/ELTTTriggerSegment.hh
//



//
#include "PTTTrigger.hh"
#include "MacroCellData.hh"
#include "MacroCell.hh"
#include "ERunParameters.hh"
#include "NumbersFileParser.hh"
#include "EPTTTriggerDataAdder.hh"
#include "PTTTriggerSegment.hh"
#include "EEvent.hh"
#include "EHeader.hh"

ClassImp (PTTTrigger)
//_____________________________________________________________________________
PTTTrigger::PTTTrigger ():TriggerEngine (string ("PTTTrigger"), kPTTTrigger)
{
  //
  // Constructor
  //
  fPixelThr = (Int_t) Conf ()->GetNum ("PTTTrigger.PixelThr");
  fPersistency = (Int_t) Conf ()->GetNum ("PTTTrigger.Persistency");
  fIntegration = (Int_t) Conf ()->GetNum ("PTTTrigger.Integration");
  Flag_debug_table = (Int_t) Conf ()->GetNum ("PTTTrigger.fDebug");
}

//_____________________________________________________________________________
PTTTrigger::~PTTTrigger ()
{
  //
  // Destructor
  //

  cout << "Destroy " << endl;
  Clear ();
}

//______________________________________________________________________________
void
PTTTrigger::Simulate (MacroCellData * pData)
{
  // Simulate PTT algorythm
  Msg (EsafMsg::Info) << "Simulating PTT algorythm (Catalano) " << MsgDispatch;
  map < Int_t, map < Int_t, ChipGtuData * >*>&m1 = pData->fChipData2;
  map < Int_t, map < Int_t, ChipGtuData * >*>::const_iterator it1;
  Int_t matsize = pData->Cell ()->GetEC (0)->FrontEnd ()->Channels ();
  side_EC = sqrt (matsize);
  side_PMT = side_EC / 2;
  Int_t maximum_int = 0;

  Int_t flag2 = 0;
  // iterate on CHIPS in this macrocell
  for (it1 = m1.begin (); it1 != m1.end (); it1++)
    {

      map < Int_t, ChipGtuData * >&m2 = *(it1->second);
      map < Int_t, ChipGtuData * >::const_iterator it2;
      // iterate on GTU for the current chip
      Int_t pers = 0;
      Int_t integr[side_EC][side_EC][fPersistency];
      Int_t sum_int[side_EC][side_EC];

      //initialize counters
      for (Int_t i = 0; i < side_EC; i++)
	for (Int_t ii = 0; ii < side_EC; ii++)
	  {
	    sum_int[i][ii] = 0;
	    for (Int_t iii = 0; iii < fPersistency; iii++)
	      integr[i][ii][iii] = 0;

	  }
	Int_t gtu_count=0;
      // iterate on GTUs in this chip
      for (it2 = m2.begin (); it2 != m2.end (); it2++)
	{
	if(gtu_count>15 && gtu_count<m2.end ()->first-15){//GTU border
	  Int_t integer_int, rest;
	  double integer;
	  rest = modf (it2->first / fPersistency, &integer);
	  integer_int = integer;
	  Int_t gtu = it2->first - integer_int * fPersistency;
	  for (Int_t i = 0; i < side_EC; i++)
	    for (Int_t ii = 0; ii < side_EC; ii++)
	      integr[i][ii][gtu] = 0;

	  //run over all the pixels in one GTU
	  Int_t flag = 0;
	  for (Int_t iteratore = 0; iteratore < matsize; iteratore++)
	    {
	      Int_t nn_PMT =
		it2->second->FrontEnd ()->PmtChannel (iteratore) / (side_PMT);
	      Int_t mm_PMT =
		it2->second->FrontEnd ()->PmtChannel (iteratore) % (side_PMT);




	      if (Flag_debug_table)
		Debug_Table1 (it2->second->FrontEnd ()->Pmt (iteratore)->
			      Geometry ()->Position (nn_PMT, mm_PMT).X (),
			      it2->second->FrontEnd ()->Pmt (iteratore)->
			      Geometry ()->Position (nn_PMT, mm_PMT).Y (),
			      it2->second->FrontEnd ()->Pmt (iteratore)->
			      Geometry ()->Position (nn_PMT, mm_PMT).Z (),
			      it2->first, it1->first,
			      it2->second->FrontEnd ()->Row (iteratore),
			      it2->second->FrontEnd ()->Column (iteratore));



	      Int_t nn_EC = it2->second->FrontEnd ()->Row (iteratore);
	      Int_t mm_EC = it2->second->FrontEnd ()->Column (iteratore);

	      Int_t center_row, center_col, center_row_aux, center_col_aux;
	      center_row = nn_EC;
	      center_col = mm_EC;
//transform
	      Transform (&center_row, &center_col);
	      center_row_aux = center_row;
	      center_col_aux = center_col;

//inverts
	      Transform (&center_row_aux, &center_col_aux);

	      Int_t new_row, new_col;
	      // since a 3X3 box is build you need to stay within 1 pixel border from the edge of the EC
	      if (center_row > 0 && center_row < side_EC - 1 && center_col > 0
		  && center_col < side_EC - 1)
		{
		  //build the 3X3 box
		  for (Int_t x = -1; x < 2; x++)
		    {
		      for (Int_t y = -1; y < 2; y++)
			{
			  new_row = center_row + x;
			  new_col = center_col + y;

			  //Inverts transformation functions
			  Transform (&new_row, &new_col);
			  Int_t channel =
			    it2->second->FrontEnd ()->ChanRowCol (new_row,
								  new_col);

			  nn_PMT =
			    it2->second->FrontEnd ()->PmtChannel (channel) /
			    (side_PMT);
			  mm_PMT =
			    it2->second->FrontEnd ()->PmtChannel (channel) %
			    (side_PMT);


			  if (Flag_debug_table)
			    Debug_Table2 (it2->second->FrontEnd ()->
					  Pmt (channel)->Geometry ()->
					  Position (nn_PMT, mm_PMT).X (),
					  it2->second->FrontEnd ()->
					  Pmt (channel)->Geometry ()->
					  Position (nn_PMT, mm_PMT).Y (),
					  it2->second->FrontEnd ()->
					  Pmt (channel)->Geometry ()->
					  Position (nn_PMT, mm_PMT).Z (), x,
					  y, it2->first, it1->first, channel);




			  //excess is calculated
			  Int_t eccesso =
			    it2->second->GetCounter (channel) - fPixelThr + 1;

			  //if above threshold 
			  if (it2->second->GetCounter (channel) >= fPixelThr)
			    {
			      // calculate the excess of the 3X3 box
			      //assigning it to the central pixel
			      integr[center_row_aux][center_col_aux][gtu] +=
				eccesso;
			      // in this GTU something is above threshold
			      flag = 1;
			    }

			}	//for(Int_t y=-1;y<2;y++){
		    }		//     for(Int_t x=-1;x<2;x++)

		}		// if(center_row>0 && center_row<11 && center_col>0 && center_col<11)
	    }			// for(Int_t iteratore=0;iteratore<matsize ;iteratore++) {

	  //if within this EC and GTU something is above threshold....
	  if (flag == 1)
	    {
	      // update the persistency count
	      pers += 1;
	      // if now the persistency  is above a certain value...
	      if (pers >= fPersistency)
		{
		  for (Int_t i = 0; i < side_EC; i++)
		    {
		      for (Int_t ii = 0; ii < side_EC; ii++)
			{
			  // calculates the sum of the excesses for each pixel in the 
			  // fPersistency GTUs previous to the one under analysis
			  sum_int[i][ii] = 0;
			  for (Int_t iii = 0; iii < fPersistency; iii++)
			    sum_int[i][ii] += integr[i][ii][iii];

			  // is there one pixel with integration above the threshold?
			  if (sum_int[i][ii] >= fIntegration)
			    {
			      flag2 = 1;
			      //TRIGGER!!!!
			      if (maximum_int < sum_int[i][ii])
				{
				  // is the trigger the maximum?
				  maximum_int = sum_int[i][ii];
				  row_up = i;
				  col_up = ii;
				  GTU_up = it2->first;
				  chip_ID_up = it1->first;
				}
			    }
			}
		    }
		}
	    }
	  else
	    {			//if in this GTU no pixel above threshold...
	      //reinitialize all the counters
	      pers = 0;
	      for (Int_t i = 0; i < side_EC; i++)
		for (Int_t ii = 0; ii < side_EC; ii++)
		  for (Int_t iii = 0; iii < fPersistency; iii++)
		    integr[i][ii][iii] = 0;
	    }
	}
	gtu_count++;
	}			//  for(it2=m2.begin(); it2!=m2.end(); it2++) {

    }				// for(it1=m1.begin(); it1!=m1.end(); it1++) {

  if (flag2 == 1)
    {
      //FILLING the INFORMATION VECTOR
      fTracks.push_back (PTTTriggerSegment ());
      PTTTriggerSegment & trkseg = fTracks.back ();

      trkseg.SetNumEvt (EEvent::GetCurrent ()->GetHeader ()->GetNum ());
      trkseg.SetMaxCount (maximum_int);
      trkseg.SetMaxGtu (GTU_up);
      trkseg.SetMaxRow (row_up);
      trkseg.SetMaxCol (col_up);
      trkseg.SetMaxChipID (chip_ID_up);

      pData->SetPTTTrigger (EEvent::GetCurrent ()->GetHeader ()->GetNum ());
      pData->SetPTTTriggerMaxCounts (maximum_int);
      pData->SetPTTTriggerMaxGtu (GTU_up);
      pData->SetPTTTriggerMaxRow (row_up);
      pData->SetPTTTriggerMaxCol (col_up);
      pData->SetPTTTriggerChipID (chip_ID_up);

      FillEEvent (EEvent::GetCurrent ()->GetHeader ()->GetNum ());

      Msg (EsafMsg::
	   Info) << "TRIGGERING... NO MORE THAN 1 TRIGGER X 1 Macro Cell " <<
	MsgDispatch;
      Msg (EsafMsg::
	   Info) << "Integrated " << maximum_int << " Chip ID " << chip_ID_up
	<< " GTU ID " << GTU_up << MsgDispatch;


    }
  else
    {
      pData->SetPTTTrigger (-1);
      pData->SetPTTTriggerMaxCounts (-1);
      pData->SetPTTTriggerMaxGtu (-1);
      pData->SetPTTTriggerMaxRow (-1);
      pData->SetPTTTriggerMaxCol (-1);
      pData->SetPTTTriggerChipID (-1);
    }


  Clear ();
}

//______________________________________________________________________________
void
PTTTrigger::Clear ()
{
/////////////////////////////////////////////////////////////////////////////////////
//clear/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////

}

//______________________________________________________________________________
void
PTTTrigger::Transform (Int_t * row, Int_t * col)
{
  //trasformation

  //transformation
  //first part
  if (*row < side_PMT && *col < side_PMT)
    {
      Int_t aux = *col;
      *col = *row;
      *row = aux;
    }

  if (*row > side_PMT - 1 && *col < side_PMT)
    {
      Int_t aux = *col;
      *col = *row - side_PMT;
      *row = aux + side_PMT;
    }

  if (*row < side_PMT && *col > side_PMT - 1)
    {
      Int_t aux = *col;
      *col = *row + side_PMT;
      *row = aux - side_PMT;
    }

  if (*row > side_PMT - 1 && *col > side_PMT - 1)
    {
      Int_t aux = *col;
      *col = *row;
      *row = aux;
    }

  //second part
  if (*row < side_PMT)
    *row += side_PMT;
  else
    *row -= side_PMT;

}


void
PTTTrigger::FillEEvent (Int_t Triggering)
{
  vector < PTTTriggerSegment >::iterator it;

  for (it = fTracks.begin (); it != fTracks.end (); it++)
    {

      if (EEvent::GetCurrent ())
	{

	  EPTTTriggerDataAdder a (&(*it), Triggering, 1);
	  EEvent::GetCurrent ()->Fill (a);
	}

    }
  fTracks.clear ();

}

//PIXEL POSITION
void
PTTTrigger::Debug_Table1 (Float_t x, Float_t y, Float_t z, Int_t analyzed_gtu,
			  Int_t chip_id, Int_t row, Int_t col)
{
  FILE *ff = fopen ("Debug_Table1.txt", "a");

  fprintf (ff, "%f  %f  %f  %d  %d  %d  %d\n", x, y, z, analyzed_gtu, chip_id,
	   row, col);

  fclose (ff);
}

//INTEGRATION BOX POSITION
void
PTTTrigger::Debug_Table2 (Float_t x, Float_t y, Float_t z, Int_t i, Int_t dir,
			  Int_t analyzed_gtu, Int_t chip_id, Int_t box)
{
  FILE *ff = fopen ("Debug_Table2.txt", "a");

  fprintf (ff, "%f  %f  %f  %d  %d  %d  %d  %d\n", x, y, z, i, dir,
	   analyzed_gtu, chip_id, box);

  fclose (ff);
}
