// $Id: SimpleDetector.cc 2910 2011-05-16 11:03:36Z biktem $
// Author: Sylvain Moreggia   2006/03/16

/*****************************************************************************
 * ESAF: Euso Simulation and Analysis Framework                              *
 *                                                                           *
 *  Id: SimpleDetector                                                           *
 *  Package: <packagename>                                                   *
 *  Coordinator: <coordinator>                                               *
 *                                                                           *
 *****************************************************************************/

//_____________________________________________________________________________
//
// SimpleDetector
//
// <extensive class description>
//
//
// ******* WARNING ********
// Data members of Photon and ParentPhoton have not the std meaning
//
// - ParentPhoton.fType      = type of photon
//                           1: fluo
//                           2: ckov
//
// - Photon.history          = their history IN ATMOSPHERE
//                          -1: none
//                           0: direct
//                           1: reflected
//                           2: rayleigh scat
//                           3: clouds scat
//                           4: aerosols scat
//
// - Photon.fate             = their history IN DETECTOR
//                          -1: outFoV
//                           0: alive
//                           1: absorbed by optics
//                           2: absorbed by BG3
//                           3: absorbed by DQE
//
// - Photon.iterations       = NB of INTERACTIONS in atmosphere (not in detector)
//
//
// NB : within detector, distance on "focal surface" is in (nb pixels)� unit
//

#include "SimpleDetector.hh"
#include "ESimpleDetector.hh"
#include "DetectorGeometry.hh"
#include "ESimpleDetectorPhotonAdder.hh"
#include "ESimpleDetectorPhotoElecAdder.hh"
#include "EEvent.hh"
#include "ETruth.hh"
#include "Photon.hh"
#include "ParentPhoton.hh"
#include "PhotonsOnPupil.hh"
#include "PhotoElectron.hh"
#include "NumbersFileParser.hh"
#include "EConst.hh"
#include "Pupil.hh"
#include "EsafRandom.hh"
#include "SimpleTelemetry.hh"
#include "Config.hh"
ClassImp(SimpleDetector)

using namespace TMath;
using namespace sou;
using namespace EConst;


//_____________________________________________________________________________
SimpleDetector::SimpleDetector() : Detector(), EsafConfigurable(), EsafMsgSource() {
    //
    // Constructor
    //
    // NB : position is set in SimuApplication.cc by reading General/Euso.cfg file
    //

    fGeometry = new DetectorGeometry;
    Reset();
    InitDetector();
}

//_____________________________________________________________________________
SimpleDetector::~SimpleDetector() {
    //
    // Destructor
    //

    Reset();
    SafeDelete(fGeometry);
}

//_____________________________________________________________________________
void SimpleDetector::Reset() {
    //
    // get ready for next event
    //


    // remove pe- of previous event
    map<UInt_t, vector<PhotoElectron*>* >::const_iterator it;
    for(it = fPhotoElectrons.begin(); it != fPhotoElectrons.end(); it++) {
        vector<PhotoElectron*>* pelist = it->second;
        if(pelist) {
            for(size_t j=0; j < pelist->size(); j++) {
	        PhotoElectron* pe = (*pelist)[j];
	        SafeDelete(pe);
            }
            pelist->clear();
            SafeDelete(pelist);
        }
    }
    fPhotoElectrons.clear();

    // reset other event data
    fEvIsTriggered = kFALSE;
    fNbGTUs        = 0;
    fNbSignalPE    = 0;
    fFirstGTUTime  = 0.;
    fFirstPETime   = 0.;
    fLastPETime    = 0.;
    fStartPixelID  = 0;
    fEndPixelID    = 0;
    fStartPosShowPt.SetXYZ(0,0,HUGE);
    fEndPosShowPt.SetXYZ(0,0,HUGE);
    fDirShowPt.SetXYZ(0,0,HUGE);

}

//_____________________________________________________________________________
void SimpleDetector::InitDetector() {
    //
    // initialization
    //

    // read configs
    fGeometry->SetRadius(Conf()->GetNum("SimpleDetector.fRadius")*sou::m);
    fGeometry->SetFoV(Conf()->GetNum("SimpleDetector.fFoV")*sou::deg);
    fReducedSurf         = Conf()->GetNum("SimpleDetector.fReducedSurf");
    fTriggEff            = Conf()->GetBool("SimpleDetector.fTriggEff");
    fGTUlength           = Conf()->GetNum("SimpleDetector.fGTUlength")*sou::nanosecond;
    fPixelSideLength     = Conf()->GetNum("SimpleDetector.fPixelSideLength")*sou::m;
    fBgndWl              = Conf()->GetStr("SimpleDetector.fBgndWl");
    fBgndMode            = Conf()->GetStr("SimpleDetector.fBgndMode");
    fBG3type             = Conf()->GetStr("SimpleDetector.fBG3type");
    fSNRtrigg            = Conf()->GetNum("SimpleDetector.fSNRtrigg");
    fNpersTrigg          = (UInt_t) Conf()->GetNum("SimpleDetector.fNpersTrigg");
    fNbMaxLateralPixels  = (UInt_t) Conf()->GetNum("SimpleDetector.fNbMaxLateralPixels");
    fMaxLateralDist2     = fNbMaxLateralPixels * fNbMaxLateralPixels;
    fCloudStatus         = Conf()->GetBool("SimpleDetector.fCloudStatus");

    // check if GTU*Clight < PixelLength
    // otherwise, treatment using shower point is not wholly valid
    if(fGTUlength * Clight() > fPixelSideLength) Msg(EsafMsg::Panic)<< "<InitDetector> GTU*Clight MUST BE > PixelSideLength" << MsgDispatch;
    if(fBgndMode != "standard") Msg(EsafMsg::Panic)<< "<InitDetector> fBgndMode = \"standard\" is the only one possible so far" << MsgDispatch;
    if(fReducedSurf < 1.) Msg(EsafMsg::Panic)<< "<InitDetector> fReducedSurf MUST be >= 1" << MsgDispatch;
    if(fTriggEff) Msg(EsafMsg::Warning)<< "<InitDetector> OK for trigger simple study : Reduction of photons due to optical aberration is taken into account" << MsgDispatch;

    // build BG3 filter
    if(fBG3type == "BG3") {
        NumbersFileParser nf("config/Optics/OpticalAdaptor/BG3.dat",3);
        vector<Double_t> wl    = nf.GetCol(0);
        vector<Double_t> trans = nf.GetCol(1);
        vector<Double_t> fake  = nf.GetCol(2);
        for(Int_t i=0; i<400; i++) fBG3filter[i] = trans[i];
    }
    else if(fBG3type == "BG3_coated") {
        NumbersFileParser nf("config/Optics/OpticalAdaptor/BG3_coated.dat",3);
        vector<Double_t> wl    = nf.GetCol(0);
        vector<Double_t> trans = nf.GetCol(1);
        vector<Double_t> fake  = nf.GetCol(2);
        for(Int_t i=0; i<400; i++) fBG3filter[i] = trans[i];
    }
    else Msg(EsafMsg::Panic)<< "<InitDetector> Wrong argument for BG3 type = "<<fBG3type << MsgDispatch;

    // build BG3 filter
    NumbersFileParser nf2("config/Electronics/R8900M36Photomultiplier/AverageQuantumEfficiency.dat",2);
    vector<Double_t> wl2    = nf2.GetCol(0);
    vector<Double_t> trans2 = nf2.GetCol(1);
    for(Int_t i=0; i<51; i++) fPMQE[i] = trans2[i];

}

//______________________________________________________________________________
Telemetry* SimpleDetector::Get( PhotonsOnPupil* photons ) {
    //
    // Simulate one event
    //


    // if no photons on pupil
    if ( !photons ) {
        MsgForm(EsafMsg::Warning,"PhotonsOnPupil is NULL. Photon tracing disabled.");
        return NULL;
    }

    // set usefull quantities
    fNbPixelOnSide = UInt_t(ceil(fGeometry->GetPos().Z() * Tan(fGeometry->GetFoV()) / fPixelSideLength)) + fNbMaxLateralPixels + 2;
    fFirstPixelX = - fPixelSideLength * fNbPixelOnSide;
    fFirstPixelY = fPixelSideLength * fNbPixelOnSide;
    fNbPixelOnSide *= 2;  // "fNbMaxLateralPixels + 2" to handle FoV bounds (+2, cause X,Y <0 AND X,Y >0)
    Double_t configbgnd = Conf()->GetNum("SimpleDetector.fMeanBgnd");  // in ph/m2/ns/sr
    fMeanBgnd = MeanBgndPerPixelPerGTU(configbgnd);
    Msg(EsafMsg::Info)<< "Mean Background (Nadir) = "<<fMeanBgnd <<" per pixel per GTU" << MsgDispatch;

    // reset previous event
    Reset();

    // wrapper to track the photons onto the real pupil
    Pupil pupil( photons, GetGeometry());

    // init
    Photon* ph = 0;
    Int_t fate = 0;
    vector<PhotoElectron*> templist;
    EEvent* ev = EEvent::GetCurrent();
    Int_t nTracked=0, nTotal(photons->GetNphotons());
    Msg(EsafMsg::Info) << "Convert photons into PhotoElectrons: " << MsgFlush;





    // loop on photons on pupil
    // 1. Apply optics + electronics filters : photons still alive give a pe-
    while(kTRUE) {
        ph = pupil.GetPhoton();
	if(!ph) break; // end of list

	// update display
        if ( nTracked*50 % nTotal == 0  || (nTracked-1)*50 % nTotal > nTracked*50 % nTotal )
            if ( nTracked*10 % nTotal == 0 || (nTracked-1)*10 % nTotal > nTracked*10 % nTotal ){
                Msg(EsafMsg::Info) << Form("%d",nTracked*100/nTotal) << '%'<< MsgFlush;
            }
            else Msg(EsafMsg::Info) << "." << MsgFlush;
        nTracked++;


	// 0.0. apply factor of reduction of EUSO detector surface + fraction of photons in 5mm bucket (opt. aberr.)
        TRandom* rndm = EsafRandom::Get();
	if(fReducedSurf > 1.)
	    if(rndm->Rndm() > (1./fReducedSurf)) continue;

	// 0.1 if trigger efficiency taken into account, the uncorrectly focused photons are trashed
	// Red book values with linear interpolation
	if(fTriggEff) {
	    Double_t triggeff = 1.;
	    if((ph->dir).Theta() < 10.*DegToRad())       triggeff = 0.74 + (ph->dir).Theta() / (10.*DegToRad()) * (0.54 - 0.74);
	    else if((ph->dir).Theta() < 20.*DegToRad())  triggeff = 0.54 + ((ph->dir).Theta() - 10.*DegToRad()) / (10.*DegToRad()) * (0.61 - 0.54);
	    else if((ph->dir).Theta() <= 30.*DegToRad()) triggeff = 0.61 + ((ph->dir).Theta() - 20.*DegToRad()) / (10.*DegToRad()) * (0.74 - 0.61);
	    if(rndm->Rndm() > triggeff) continue;
	}

	// 0. transfer atmosphere infos from parentphoton to photon
        ParentPhoton* parentph = ph->parent;
	if(parentph->Type() == 3) parentph->SetType(CerenkovReflected);
	ph->history  = parentph->AtmoHistory();
	ph->iterations = parentph->AtmoNbinter();

	if(fCloudStatus && parentph->IsCloudAbsorbed()) continue;


	// 1. FoV handling : if outFoV, tagged accordingly
	if(!fGeometry->IsInFoV(ph->dir)) {
	    ph->fate = -1;
            if( ev ) {
                ESimpleDetectorPhotonAdder a(ph,kTRUE);
                ev->Fill(a);
            }
	    continue;
	}


	// 2. optics effects
	fate = ApplyOpticsFilter(*ph);
	if(fate != 0) {
	    ph->fate = fate;
            if( ev ) {
                ESimpleDetectorPhotonAdder a(ph,kTRUE);
                ev->Fill(a);
            }
	    continue;
	}


	// 3. Detection quantum efficiency (DQE = QE * CE)
	fate = ApplyDQE(*ph);
	if(fate != 0) {
	    ph->fate = fate;
            if( ev ) {
                ESimpleDetectorPhotonAdder a(ph,kTRUE);
                ev->Fill(a);
            }
	    continue;
	}

        // 4. if photon still alive --> gives a pe-
	if(ph->fate != 0) Msg(EsafMsg::Panic)<< "<Get> Photon fate SHOULD be zero here : "<<ph->fate << MsgDispatch;
	if( ev ) {
            ESimpleDetectorPhotonAdder a(ph,kTRUE);
            ev->Fill(a);
        }
	templist.push_back(new PhotoElectron(*ph));
    }
    fNbSignalPE = templist.size();
    Msg(EsafMsg::Info) << " -done" << MsgDispatch;
    Msg(EsafMsg::Info)<< "NB of PhotoElectrons given by signal photons = "<<fNbSignalPE << MsgDispatch;
    if(fNbSignalPE == 0) return (Telemetry*)0;


    // 2. orders pe- chronologically
    //    then find the time of first GTU and ranges pe- in the map using GTU info
    //    Set pixel ID + shift w.r.t. shower point
    SpaceTimePEinfos(templist);


    // 3. Add background around shower point pixel (for each GTU)
    //    For every pe-, set the spatial shift w.r.t. shower point
    Int_t nbpebgnd = 0;
    if(fBgndMode == "standard") nbpebgnd = AddBackground();
    //else if(fBgndMode == "minimized") nbpebgnd = AddMinimizedBackground();  // disabled so far
    else Msg(EsafMsg::Panic)<< "Wrong argument for fBgndMode = "<<fBgndMode << MsgDispatch;


    // 4. kind of trigger simulation
    //    Tag pe- according to the SNR per pixel per GTU : if (S + B) >= (<B> + fSNRtrigg*sigmaB), pe- is considered triggered
    //    returns nb of triggered pe-
    Int_t nbpetriggered = ApplyTrigger();
    Msg(EsafMsg::Info)<< "NB of Bgnd PhotoElectrons = "<<nbpebgnd << MsgDispatch;
    Msg(EsafMsg::Info)<< "NB of PhotoElectrons with SNR > SNRtrigg = "<<nbpetriggered << MsgDispatch;
    if(fEvIsTriggered) Msg(EsafMsg::Info)<< "EVENT IS TRIGGERED !" << MsgDispatch;
    else Msg(EsafMsg::Info)<< "Event is NOT TRIGGERED.." << MsgDispatch;


    // 5. fill ESimpleDetector (root output for present detector simu)
    ESimpleDetector* Edet = 0;
    if(ev) Edet = ev->GetSimpleDetector();
    if(Edet) {
        Edet->SetEventIsTriggered(fEvIsTriggered);
        Edet->SetMeanBgnd(fMeanBgnd);
        Edet->SetTriggerLevel(fSNRtrigg);
        Edet->SetTriggPersistency(fNpersTrigg);
        Edet->SetNbPixelOnSide(fNbPixelOnSide);
        Edet->SetPixelSideLength(fPixelSideLength);
        Edet->SetFirstPixelXY(fFirstPixelX,fFirstPixelY);
        Edet->SetFirstTime(fFirstPETime);
        Edet->SetLastTime(fLastPETime);
        Edet->SetFirstGTUTime(fFirstGTUTime);
        Edet->SetNbGTUs(fNbGTUs);
        Edet->SetGTUlength(fGTUlength);
    }
    // fill EPhotoElectron list
    map<UInt_t, vector<PhotoElectron*>* >::const_iterator it;
    for(it = fPhotoElectrons.begin(); it != fPhotoElectrons.end(); it++) {
	vector<PhotoElectron*>* pelist = it->second;

	// if pe- exist in this GTU
	if(pelist) {
	    for(size_t i=0; i < pelist->size(); i++) {
	        ESimpleDetectorPhotoElecAdder a((*pelist)[i],kTRUE);
		ev->Fill(a);
	    }
	}
    }



    return (Telemetry*)0;
}

//______________________________________________________________________________
Int_t SimpleDetector::ApplyOpticsFilter(const Photon& ph) const {
    //
    // Total optical efficiency + BG3
    // return photon's fate value
    //

    // inits
    TRandom* rndm = EsafRandom::Get();

    // 1. total optical throughput efficiency (taken from OM-DESIGN-001-2)
    Double_t eff = 0.53;
    if(eff < rndm->Rndm()) return 1;

    // 2. BG3 filter
    eff = InterpolateBG3(ph.wl);
    if(eff < rndm->Rndm()) return 2;

    // if photon is still alive
    return 0;
}

//______________________________________________________________________________
Int_t SimpleDetector::ApplyDQE(const Photon& ph) const {
    //
    // Detection Quantum Efficiency rules the creation of photoelectrons from incident photons
    // return photon's fate value
    //

    // inits
    TRandom* rndm = EsafRandom::Get();

    // DQE = CE * QE
    // mean collection efficiency = 0.67025 drawn from config/Electronics/R8900M36Photomultiplier/AverageCollectionEfficiency.dat
    //    Double_t NORM = 0.67025;
    ConfigFileParser * eConf = Config::Get ()->GetCF ("Electronics", "ElectronicsFactory");
    ConfigFileParser * pConf = Config::Get ()->GetCF ("Electronics", eConf->GetStr("ElectronicsFactory.PmtType"));

    Double_t NORM = pConf->GetNum ("R8900M64Photomultiplier.fMeanCollectionEfficiency"); // ICRC2011 Assumption

    // apply DQE
    Double_t eff = InterpolateQE(ph.wl);
    eff *= NORM;
    if(eff < rndm->Rndm()) return 3;

    // if photon is still alive
    return 0;
}

//______________________________________________________________________________
Double_t SimpleDetector::InterpolateQE(Double_t wl) const {
    //
    // interpolate using tabulated values
    //

    Double_t rtn = 0.;
    Int_t index = Int_t(floor(wl/(10.*nm)));
    index -= 20;
    if(wl <= 200*nm || wl >= 710*nm) rtn = 0.;
    else if(index < 0) Msg(EsafMsg::Warning)<< "<ApplyDQE> index SHOULD NOT be negative, (wl, index) = ("<<wl/nm<<", "<<index<<")" << MsgDispatch;
    else rtn = fPMQE[index] + (wl/(10.*nm) - floor(wl/(10.*nm))) * (fPMQE[index+1] - fPMQE[index]);

    return rtn;
}

//______________________________________________________________________________
Double_t SimpleDetector::InterpolateBG3(Double_t wl) const {
    //
    // interpolate using tabulated values
    //

    Double_t rtn = 0.;
    Int_t index = Int_t(floor(wl/nm));
    index -= 200;
    if(wl <= 200*nm || wl >= 600*nm) rtn = 0.;
    else if(index < 0) Msg(EsafMsg::Warning)<< "<ApplyOpticsFilter> index SHOULD NOT be negative, (wl, index) = ("<<wl/nm<<", "<<index<<")" << MsgDispatch;
    else rtn = fBG3filter[index] + (wl/nm - floor(wl/nm)) * (fBG3filter[index+1] - fBG3filter[index]);

    return rtn;
}

//______________________________________________________________________________
void SimpleDetector::SpaceTimePEinfos(const vector<PhotoElectron*>& templist) {
    //
    // - orders pe- chronologically
    //   then find the time of first GTU and ranges them in the map using GTU info
    // - set pixel ID
    // - find pixels rectangular area involved in the event
    //



    // 1. find time of first and last created pe-
    fFirstPETime = HUGE;
    fLastPETime = 0.;
    TArrayD tofarray(fNbSignalPE);
    for(Int_t i=0; i < fNbSignalPE; i++) {
	tofarray[i] = templist[i]->Time();
        if(tofarray[i] < fFirstPETime) fFirstPETime = tofarray[i];
        if(tofarray[i] > fLastPETime)  fLastPETime  = tofarray[i];
    }



    // 2. find time of first GTU (first pe- sampled randomly in the first GTU)
    fFirstGTUTime = fFirstPETime - fGTUlength * EsafRandom::Get()->Rndm();
    fNbGTUs = Int_t(ceil( (fLastPETime - fFirstGTUTime) / fGTUlength ));
    for(size_t i=0; i < fNbGTUs; i++) fPhotoElectrons[i] = new vector<PhotoElectron*>;


    // 3. sort tofarray elements using time info
    Int_t sortedindex[fNbSignalPE];
    Sort(fNbSignalPE,tofarray.GetArray(),sortedindex,false); // false means in increasing order



    // 4. find pixelID for each pe-
    //    and fill fPhotoElectrons using GTU info
    Double_t tof = 0.;
    UInt_t pixelID = 0;
    PhotoElectron* pe = 0;
    Bool_t IsInFoV = kTRUE;
    UInt_t GTUnb(0), oldGTUnb(100000);
    for(Int_t i=0; i < fNbSignalPE; i++) {
        pe = templist[sortedindex[i]];
	IsInFoV = FindPixelID(pe->AtmoPos(),pixelID);
	if(!IsInFoV) Msg(EsafMsg::Panic)<< "<SpaceTimePEinfos> PE- is OUT OF FOV, it SHOULD NOT" << MsgDispatch;
	pe->SetPixelID(pixelID);
        tof = pe->Time();
	GTUnb = UInt_t(floor( (tof - fFirstGTUTime) / fGTUlength ));
	//if(GTUnb == 0 && oldGTUnb == 100000)   fPhotoElectrons[0] = new vector<PhotoElectron*>;
	//else if(oldGTUnb < GTUnb)              fPhotoElectrons[GTUnb] = new vector<PhotoElectron*>;
	if(GTUnb > 0 && oldGTUnb > GTUnb) Msg(EsafMsg::Panic)<< "<SpaceTimePEinfos> It seems pe- are not chronologically sorted "<< MsgDispatch;
	fPhotoElectrons[GTUnb]->push_back(pe);
	oldGTUnb = GTUnb;
    }




    // 5. Find the shower point position at fFirstGTUTime
    //    Find the end of the observable track (earth impact or FoV bounds)
    //    Set track direction (3D and 2D)
    //

	// 5.0. init
    UInt_t startpixID = 0;
    UInt_t endpixID = 0;
    Bool_t kIsInFoV = kFALSE;
    Double_t tof_alongtrack = 0.;
    const Double_t increm = 1.*m;
    EEvent* ev       = EEvent::GetCurrent();
    ETruth* truth    = ev->GetTruth();
    EarthVector EUSO = fGeometry->GetPos();
    EarthVector FirstInteraction = truth->GetTrueInitPos();   // time info is taken w.r.t. first interaction
    EarthVector TOAimpact = truth->GetTrueTOAImpact();
    EarthVector TrackEnd = truth->GetTrueEarthImpact();

	// 5.1. Track direction
    fDirShowPt = (FirstInteraction - TOAimpact).Unit();

        // 5.2. track position at fFirstGTUTime
    tof = 0.;
    fStartPosShowPt = FirstInteraction;
    Int_t infloop = 0;
    while(tof < fFirstGTUTime) {
        fStartPosShowPt += increm*fDirShowPt;
	tof_alongtrack += increm / Clight();
	tof = tof_alongtrack + (EUSO - fStartPosShowPt).Mag() / Clight();
	if(infloop++ > 1000000) {
	    Msg(EsafMsg::Warning)<< "<SpaceTimePEinfos> Infinite loop broken by hand"<< MsgDispatch;
	    Msg(EsafMsg::Warning)<< "<SpaceTimePEinfos> Thus, starting Shower Point position taken at first photon position"<< MsgDispatch;
	    fStartPosShowPt = templist[sortedindex[0]]->AtmoPos();
	    break;
	}
    }
    kIsInFoV = FindPixelID(fStartPosShowPt,startpixID);
    if(!kIsInFoV) {
	Msg(EsafMsg::Warning)<< "<SpaceTimePEinfos> Starting Shower Point position is OUT OF FOV (may occur)"<< MsgDispatch;
	Msg(EsafMsg::Warning)<< "<SpaceTimePEinfos> Thus, StartPixelID is recalculated"<< MsgDispatch;
	kIsInFoV = FindPixelID(templist[sortedindex[0]]->AtmoPos(),startpixID);
	if(!kIsInFoV) Msg(EsafMsg::Panic)<< "<SpaceTimePEinfos> PROBLEM IS NOT SOLVED, StartPixelID still out of FoV. IT SHOULD NOT."<< MsgDispatch;
    }

	// 5.3. end of track position (earth impact or FoV bounds) : DOES NOT SET endpix, see below
    kIsInFoV = FindPixelID(TrackEnd,endpixID);
    if(!kIsInFoV) {
        // handle cases with end of track out of FoV
	TrackEnd = fStartPosShowPt;
        infloop = 0;
	kIsInFoV = kTRUE;
	while(kIsInFoV) {
	    TrackEnd += increm*fDirShowPt;
	    kIsInFoV = FindPixelID(TrackEnd,endpixID);
	    if(infloop++ > 10000000) Msg(EsafMsg::Panic)<< "<SpaceTimePEinfos> inf loop broken by hand in searching for end of track, out FOV"<< MsgDispatch;
	}
	TrackEnd -= increm*fDirShowPt; // one step back to be in the Fov
    }
    fEndPosShowPt = TrackEnd;



/*
// code below :
// ==========
// only used for ancient version of AddBackground() method : define a rectangular area around the showertrack
// in order to add bgnd around it at every GTU --> TOO MUCH PE�, particularly for inclined events


	// 5.4. end of track position (in case track goes out of FoV) : SET endpix here
    TrackEnd = truth->GetTrueEarthImpact();
    kIsInFoV = FindPixelID(TrackEnd,endpixID);
    if(!kIsInFoV) {
        // handle cases with end of track out of FoV
	TrackEnd = fStartPosShowPt;
	EarthVector dirloc;
        EarthVector Nadirloc(0,0,-1);
        Double_t magloc(0);
	Double_t thetaloc(0.);
	EarthVector posZ0loc;
	Int_t jloc = 10;
	Int_t iloc = 10;
	infloop = 0;
	Bool_t kBreak = kFALSE;
	while(kTRUE) {
	    TrackEnd += increm*fDirShowPt;
	    if( (TrackEnd - fStartPosShowPt).Mag() > (truth->GetTrueEarthImpact() - fStartPosShowPt).Mag() ) break;
            // cpy of FindPixelID, but forgets out of FoV pb
            dirloc = (TrackEnd - fGeometry->GetPos()).Unit();
            thetaloc = dirloc.Angle(Nadirloc);
            magloc = -TrackEnd.Z() / dirloc.Z();
            posZ0loc = TrackEnd + magloc*dirloc;
	    jloc = Int_t(floor((posZ0loc.X() - fFirstPixelX) / fPixelSideLength));
            iloc = Int_t(floor((fFirstPixelY - posZ0loc.Y()) / fPixelSideLength));
            if((jloc < 0)) {
	        jloc = 0;
		kBreak = kTRUE;
	    }
	    else if(jloc > Int_t(fNbPixelOnSide) - 1) {
	        jloc = Int_t(fNbPixelOnSide) - 1;
		kBreak = kTRUE;
	    }
            if((iloc < 0)) {
	        iloc = 0;
		kBreak = kTRUE;
	    }
	    else if(iloc > Int_t(fNbPixelOnSide) - 1) {
	        iloc = Int_t(fNbPixelOnSide) - 1;
		kBreak = kTRUE;
	    }
	    if(kBreak) break;
	    if(infloop++ > 1000000) {
	        Msg(EsafMsg::Warning)<< "<SpaceTimePEinfos, out of FoV track end serach> Infinite loop broken by hand"<< MsgDispatch;
	        iloc = endpixID / fNbPixelOnSide;
	        jloc = endpixID % fNbPixelOnSide;
	        break;
	    }
	}
        endpixID = UInt_t(iloc)*fNbPixelOnSide + UInt_t(jloc);
    }



    // 6. Set start pixel and end pixel IDs, shifting determined values of row/col by fNbMaxLateralPixels
    //    Does not care about FoV bounds, it will be handled later in bgnd generation
    Bool_t rowstartflag = kFALSE;
    Bool_t colstartflag = kFALSE;
    Bool_t rowendflag = kFALSE;
    Bool_t colendflag = kFALSE;
    Int_t StartRowIsUp = 1;
    Int_t StartColIsUp = 1;
    if( (startpixID / fNbPixelOnSide) < (endpixID / fNbPixelOnSide) ) StartRowIsUp = -1;
    if( (startpixID % fNbPixelOnSide) < (endpixID % fNbPixelOnSide) ) StartColIsUp = -1;
        // check pixels are in required range
    if(StartRowIsUp > 0) {
        if( (startpixID / fNbPixelOnSide ) > (fNbPixelOnSide - 1 - fNbMaxLateralPixels) ) {
            Msg(EsafMsg::Warning)<< "<SpaceTimePEinfos> startpixID ROW is too large = "
	                         <<startpixID / fNbPixelOnSide + fNbMaxLateralPixels<<", instead < "<<fNbPixelOnSide - 1 << MsgDispatch;
	    rowstartflag = kTRUE;
	}
        if( (endpixID / fNbPixelOnSide) < fNbMaxLateralPixels ) {
            Msg(EsafMsg::Warning)<< "<SpaceTimePEinfos> endpixID ROW is too small = "
	                         <<endpixID / fNbPixelOnSide - fNbMaxLateralPixels<<", instead being positive" << MsgDispatch;
	    rowendflag = kTRUE;
	}
    }
    else {
        if( (endpixID / fNbPixelOnSide ) > (fNbPixelOnSide - 1 - fNbMaxLateralPixels) ) {
            Msg(EsafMsg::Warning)<< "<SpaceTimePEinfos> endpixID ROW is too large = "
	                         <<endpixID / fNbPixelOnSide + fNbMaxLateralPixels<<", instead < "<<fNbPixelOnSide - 1 << MsgDispatch;
	    rowendflag = kTRUE;
	}
        if( (startpixID / fNbPixelOnSide) < fNbMaxLateralPixels ) {
            Msg(EsafMsg::Warning)<< "<SpaceTimePEinfos> startpixID ROW is too small = "
	                         <<startpixID / fNbPixelOnSide - fNbMaxLateralPixels<<", instead being positive" << MsgDispatch;
	    rowstartflag = kTRUE;
	}
    }
    if(StartColIsUp > 0) {
        if( (startpixID % fNbPixelOnSide ) > (fNbPixelOnSide - 1 - fNbMaxLateralPixels) ) {
            Msg(EsafMsg::Warning)<< "<SpaceTimePEinfos> startpixID COL is too large = "
	                         <<startpixID % fNbPixelOnSide + fNbMaxLateralPixels<<", instead < "<<fNbPixelOnSide - 1 << MsgDispatch;
	    colstartflag = kTRUE;
	}
        if( (endpixID % fNbPixelOnSide) < fNbMaxLateralPixels ) {
            Msg(EsafMsg::Warning)<< "<SpaceTimePEinfos> endpixID COL is too small = "
	                         <<endpixID % fNbPixelOnSide - fNbMaxLateralPixels<<", instead being positive" << MsgDispatch;
	    colendflag = kTRUE;
	}
    }
    else {
        if( (endpixID % fNbPixelOnSide ) > (fNbPixelOnSide - 1 - fNbMaxLateralPixels) ) {
            Msg(EsafMsg::Warning)<< "<SpaceTimePEinfos> endpixID COL is too large = "
	                         <<endpixID % fNbPixelOnSide + fNbMaxLateralPixels<<", instead < "<<fNbPixelOnSide - 1 << MsgDispatch;
	    colendflag = kTRUE;
	}
        if( (startpixID % fNbPixelOnSide) < fNbMaxLateralPixels ) {
            Msg(EsafMsg::Warning)<< "<SpaceTimePEinfos> startpixID COL is too small = "
	                         <<startpixID % fNbPixelOnSide - fNbMaxLateralPixels<<", instead being positive" << MsgDispatch;
	    colstartflag = kTRUE;
	}
    }
        // set values
    fStartPixelID = startpixID;
    fEndPixelID   = endpixID;
    if(!rowstartflag) fStartPixelID += Sign(1,StartRowIsUp)*fNbMaxLateralPixels*fNbPixelOnSide;
    if(!colstartflag) fStartPixelID += Sign(1,StartColIsUp)*fNbMaxLateralPixels;
    if(!rowendflag)   fEndPixelID   -= Sign(1,StartRowIsUp)*fNbMaxLateralPixels*fNbPixelOnSide;
    if(!colendflag)   fEndPixelID   -= Sign(1,StartColIsUp)*fNbMaxLateralPixels;
*/
}

//______________________________________________________________________________
Bool_t SimpleDetector::FindPixelID(const EarthVector& atmopos, UInt_t& rtn) const {
    //
    // find pixel according to position in atmosphere
    // - pixel nb 0 is at left-up corner (X<0 and Y>0)
    // - using sqrt(x*x + y*y) value in the MES Z=0 plane
    // - i,j in [0..n-1]  (i<-->row    j<-->column)
    // - pixelID = N --> i,j coordinates found back using N / n = i  and  N % n = j
    //
    // return kFALSE if position is out of FoV
    //

    // 0. photon entering direction in detector
    EarthVector dir = (atmopos - fGeometry->GetPos()).Unit();

    // 1. check if photon is in the FoV
    EarthVector Nadir(0,0,-1);
    Double_t theta = dir.Angle(Nadir);
    if(theta > fGeometry->GetFoV() + 2e-3) return kFALSE;   // 2e-3 <=> around 2.5�s at speed of light (used to handle fFirstTimeGTU shift at FoV bound)

    // 2. find parent photon's horizontal position in the MES plane Z=0
    //    intersection between Z=0 plane and track defined by (photon atmopos and (EUSO - atmopos).Unit() direction)
    Double_t mag(0);
    mag = -atmopos.Z() / dir.Z();

    EarthVector posZ0 = atmopos + mag*dir;
    if(fabs(posZ0.Z()) > kAltitudeTolerance) Msg(EsafMsg::Warning)<< "<FindPixelID> search pos Z coo. NOT EQUAL to zero = "<<posZ0 << MsgDispatch;

    // get pixel ID
    UInt_t j = UInt_t(floor((posZ0.X() - fFirstPixelX) / fPixelSideLength));
    UInt_t i = UInt_t(floor((fFirstPixelY - posZ0.Y()) / fPixelSideLength));

    rtn = i*fNbPixelOnSide + j;

    return kTRUE;
}

//______________________________________________________________________________
Bool_t SimpleDetector::Pos2D_Z0plane(UInt_t pixID, EarthVector& rtn) const {
    //
    // get pixel 2D position in the Mes Z=0 plane
    // Invert method of FindPixelID()
    //
    // return kTRUE if pixel is in the FoV, kFALSE otherwise
    //

    // init
    UInt_t i = pixID / fNbPixelOnSide;
    UInt_t j = pixID % fNbPixelOnSide;

    // 1. take position at the middle of pixel
    rtn.SetXYZ(fFirstPixelX + (Double_t(j) + 0.5)*fPixelSideLength,fFirstPixelY - (Double_t(i) + 0.5)*fPixelSideLength,0.);

    // 2. tell if pixel is out of FoV
    if(rtn.Mag() > (fGeometry->GetPos().Z() * Tan(fGeometry->GetFoV())) ) return kFALSE;

    return kTRUE;
}

//______________________________________________________________________________
Double_t SimpleDetector::MeanBgndPerPixelPerGTU(Double_t bckgnd_photon) {
    //
    // input : bckgnd_photon is in ph / m2 / ns / sr
    // Calculate mean (over wl) background in pe/pixel/GTU (at Nadir)
    // Background wl spectrum is assumed FLAT
    //
    // NB : for a detector of 2.5m diameter
    //      500 ph/ns/sr/m2 within 300-400nm           --> ~1.7 pe/pixel/GTU
    //      500 ph/ns/sr/m2 EXTENDED within 200-600nm  --> ~2.8 pe/pixel/GTU (BG3)  ~2.35 (coated BG3)
    //


    // 0. Set bounds according to background mode used
    Double_t wlmin(0.), wlmax(0.);
    Double_t WlRangeCorrection(1.);
    if(fBgndWl == "RedBook") {
        wlmin = 300*nm;
	wlmax = 400*nm;
    }
    else if(fBgndWl == "Extended") {
        wlmin = 200*nm;
	wlmax = 600*nm;
	WlRangeCorrection = 4.;  // rough extrapolation
    }


    // 1. find wl-averaged efficiencies of QE and BG3 filters (convolution)
    Double_t wl = wlmin;
    Double_t wlincrement = 0.1*nm;
    Double_t meaneff(0.), nbsteps(0.);
    while(wl < wlmax) {
         meaneff += InterpolateQE(wl)*InterpolateBG3(wl);
	 wl += wlincrement;
	 nbsteps += 1.;
    }
    meaneff /= nbsteps;
    // effects of optics absorption + PM collection efficiency
    meaneff *= 0.67025 * 0.53;

    // 2. Detector surface + pixel solid angle effects + GTUlength + wl range used
    Double_t pixomega = fPixelSideLength*fPixelSideLength / fGeometry->GetPos().Z() / fGeometry->GetPos().Z();
    Double_t eusosurf = Pi() * fGeometry->GetRadius()/m * fGeometry->GetRadius()/m;
    eusosurf /= fReducedSurf;
    Double_t integ_factor = pixomega * eusosurf * fGTUlength/nanosecond * WlRangeCorrection;


    return bckgnd_photon * integ_factor * meaneff;
}


//______________________________________________________________________________
Int_t SimpleDetector::AddBackground() {
    //
    // At every GTU, add background around showerPOINT  (not all around shower track)
    // For every pe-, set the spatial shift w.r.t. shower point of all pe- (signal and bgnd)
    //
    // NB : looking at shower point at GTU start and GTU end could miss an intermediate pixel
    //      not so important if the intermediate pixel is considered in simu analysis or reconstruction
    //

    Int_t rtn(0);
    UInt_t GTUnb = 0;
    Bool_t IsInFoV = kFALSE;
    Int_t GTUshiftstop = 15;  // will be used to stop bgnd addition after endoftrack has been reached

    // loop on GTUs
    map<UInt_t, vector<PhotoElectron*>* >::const_iterator it;
    for(it = fPhotoElectrons.begin(); it != fPhotoElectrons.end(); it++) {

	vector<PhotoElectron*>* pelist = it->second;
	GTUnb = it->first;


	// 1. find shower point pixels for this GTU
	//    to add bgnd around + to set spatial shift afterwards
	//   NB : WHEN END OF TRACK IS REACHED --> goes on with the same angular speed (as if track goes on under ground or out of FoV)
	UInt_t LowEdgeRefPixel = 0;  // at GTU low edge
	UInt_t UpEdgeRefPixel  = 0;  // at GTU up edge
	UInt_t LowEdge_i = 0;
	UInt_t UpEdge_i  = 0;
	UInt_t LowEdge_j = 0;
	UInt_t UpEdge_j  = 0;
	Bool_t IsSamePixel = kFALSE;
	Bool_t reach_endtrack = kFALSE;
	Double_t tof = 0.;
	Double_t tof_alongtrack = 0.;
	Double_t current_tof = 0.;
        Int_t infloop = 0;
	const Double_t increm = 1.*sou::m;
	EarthVector ShowPtPos;
	EarthVector EUSO = fGeometry->GetPos();
	    // low GTU edge
	current_tof = Double_t(GTUnb) * fGTUlength;
        tof = 0.;
        ShowPtPos = fStartPosShowPt;
        while(tof <= current_tof) {
            ShowPtPos += increm*fDirShowPt;
	    tof_alongtrack += increm / Clight();
	    tof = tof_alongtrack + (EUSO - ShowPtPos).Mag() / Clight() - (EUSO - fStartPosShowPt).Mag() / Clight();
	    if(infloop++ > 1000000) Msg(EsafMsg::Panic)<< "<AddBackground, low edge> Infinite loop broken by hand"<< MsgDispatch;
	}
	    // check if end of track is reached
	if( (ShowPtPos - fStartPosShowPt).Mag() > (fEndPosShowPt - fStartPosShowPt).Mag() ) reach_endtrack = kTRUE;

	// if end of track is reached, go on until GTU = GTUendtrack + GTUshiftstop
	// (to include a part of scattered photons not too far from usefull signal)
	if(reach_endtrack) GTUshiftstop--;
	if(GTUshiftstop > 0) {
	    IsInFoV = FindPixelID(ShowPtPos,LowEdgeRefPixel);
	    if(!IsInFoV) {
		ShowPtPos = fEndPosShowPt;
		IsInFoV = FindPixelID(ShowPtPos,LowEdgeRefPixel);
	    }
	    LowEdge_i = LowEdgeRefPixel / fNbPixelOnSide;
	    LowEdge_j = LowEdgeRefPixel % fNbPixelOnSide;
		// up GTU edge
	    if(!reach_endtrack) {
		current_tof += fGTUlength;
        	infloop = 0;
        	while(tof <= current_tof) {
        	    ShowPtPos += increm*fDirShowPt;
		    tof_alongtrack += increm / Clight();
		    tof = tof_alongtrack + (EUSO - ShowPtPos).Mag() / Clight() - (EUSO - fStartPosShowPt).Mag() / Clight();
		    if(infloop++ > 1000000) Msg(EsafMsg::Panic)<< "<AddBackground, up edge> Infinite loop broken by hand"<< MsgDispatch;
		}
		IsInFoV = FindPixelID(ShowPtPos,UpEdgeRefPixel);
		if(!IsInFoV) Msg(EsafMsg::Warning)<< "<AddBackground, up edge> Track pos is OUT OF FOV, it SHOULD NOT" << MsgDispatch;
		UpEdge_i = UpEdgeRefPixel / fNbPixelOnSide;
		UpEdge_j = UpEdgeRefPixel % fNbPixelOnSide;
		if(UpEdgeRefPixel == LowEdgeRefPixel) IsSamePixel = kTRUE;
	    }
	    else {  // not a significant effect when impact on ground is reached, but simpler when FoV bound is reached
		UpEdge_i = LowEdge_i;
		UpEdge_j = LowEdge_j;
		IsSamePixel = kTRUE;
	    }


	    // 2. init + find region around shower point where bgnd will be added
	    Int_t prev_pixID = -1;
	    Int_t pixID = 0;
            UInt_t low_i = LowEdge_i;
	    UInt_t up_i  = UpEdge_i;
	    if(low_i > up_i) {
        	low_i    = UpEdge_i;
		up_i     = LowEdge_i;
	    }
            UInt_t low_j = LowEdge_j;
	    UInt_t up_j  = UpEdge_j;
	    if(low_j > up_j) {
        	low_j    = UpEdge_j;
		up_j     = LowEdge_j;
	    }
	    if(low_i >= fNbMaxLateralPixels) low_i -= fNbMaxLateralPixels;
	    else low_i = 0;
	    if(low_j >= fNbMaxLateralPixels) low_j -= fNbMaxLateralPixels;
	    else low_j = 0;
	    if((up_i + fNbMaxLateralPixels) < fNbPixelOnSide) up_i  += fNbMaxLateralPixels;
	    else up_i = fNbPixelOnSide - 1;
	    if((up_j + fNbMaxLateralPixels) < fNbPixelOnSide) up_j  += fNbMaxLateralPixels;
	    else up_j = fNbPixelOnSide - 1;


	    size_t NBPIXELSid = (up_i - low_i + 1) * (up_j - low_j + 1);
            TArrayI pixelIDs_temp(NBPIXELSid);



	    // 3. store IDs of involved pixels (shower point, fMaxLateralDist2 and Fov bounds)
	    size_t ind = 0;
	    UInt_t LatDist2 = 0;
	    UInt_t LatDist2_upedge = 0;
	    EarthVector dummy;
	    for(UInt_t i=low_i; i <= up_i; i++) {
		for(UInt_t j=low_j; j <= up_j; j++) {
		    // check if searched pixel is in FoV + if its lateral distance� to the track < fMaxLateralDist2
		    LatDist2 = (i - LowEdge_i)*(i - LowEdge_i) + (j - LowEdge_j)*(j - LowEdge_j);
		    LatDist2_upedge = (i - UpEdge_i)*(i - UpEdge_i)  + (j - UpEdge_j)*(j - UpEdge_j);
		    if(LatDist2_upedge < LatDist2) LatDist2 = LatDist2_upedge;
                    IsInFoV = Pos2D_Z0plane(i*fNbPixelOnSide + j,dummy);
		    if(!IsInFoV || (LatDist2 > fMaxLateralDist2)) continue;
		    else pixelIDs_temp[ind++] = i*fNbPixelOnSide + j;
		}
	    }



	    // 4. copy usefull pixels ID
	    if(ind < NBPIXELSid) NBPIXELSid = ind;
            TArrayI pixelIDs(NBPIXELSid);
	    for(size_t n=0; n < NBPIXELSid; n++) pixelIDs[n] = pixelIDs_temp[n];



	    // 5. sort pe- according to pixelID
            Int_t sortedindex[NBPIXELSid];
            Sort((Int_t)NBPIXELSid,pixelIDs.GetArray(),sortedindex,false); // false means in increasing order



	    // 6. then add background ONCE per pixel
	    prev_pixID = -1;
	    pixID = 0;
	    Int_t nbPEbgnd = 0;
	    vector<PhotoElectron*> buffer;
	    EarthVector pixel2Dpos;
	    for(size_t m=0; m < NBPIXELSid; m++) {
		nbPEbgnd = 0;
		pixID = pixelIDs[sortedindex[m]];
		if(pixID > prev_pixID) {
		    // find pixel corresponding incident angle
		    Pos2D_Z0plane(pixID, pixel2Dpos);
		    Double_t th = (fGeometry->GetPos() - pixel2Dpos).Theta();

		    // create background pe-
		    // Bgnd source assumed _lambertian_ :
		    //    - one Cos(th) stems from euso surface projection
		    //    - the other one because pixel surface projection on ground is constant in SimpleDetector
		    nbPEbgnd = EsafRandom::Get()->Poisson(Cos(th) * Cos(th) * fMeanBgnd);
		    PhotoElectron* pe = 0;
		    for(Int_t p=0; p < nbPEbgnd; p++) {
			pe = new PhotoElectron();
			rtn++;
			pe->SetTime(it->first * fGTUlength + fFirstGTUTime + 0.5*fGTUlength);
			pe->SetPixelID(UInt_t(pixID));
			buffer.push_back(pe);
		    }
		}
		prev_pixID = pixID;
	    }


	    for(size_t m=0; m < buffer.size(); m++) pelist->push_back(buffer[m]);
	    buffer.clear();



	    // 7. Set pe- spatial shift w.r.t. shower point pixels
	    //    If shower point travel several pixels within a GTU, only the extrem ones are considered as reference
	    PhotoElectron* pe = 0;
	    UInt_t lowDist2Shift = 0;
	    UInt_t upDist2Shift = 0;
	    for(size_t m=0; m < pelist->size(); m++) {
		pe = (*pelist)[m];
		pixID = pe->PixelID();
		UInt_t pixi = pixID / fNbPixelOnSide;
		UInt_t pixj = pixID % fNbPixelOnSide;
		lowDist2Shift = (pixi - LowEdge_i)*(pixi - LowEdge_i) + (pixj - LowEdge_j)*(pixj - LowEdge_j);
		if(!IsSamePixel) {
	            upDist2Shift  = (pixi - UpEdge_i) *(pixi - UpEdge_i)  + (pixj - UpEdge_j) *(pixj - UpEdge_j);
	            if(lowDist2Shift < upDist2Shift) pe->SeDist2ShiftToShowPt(lowDist2Shift);
	            else pe->SeDist2ShiftToShowPt(upDist2Shift);
		}
		else pe->SeDist2ShiftToShowPt(lowDist2Shift);
	    }
	}
	// if pe- arrives "too late" (after GTUendtrack + GTUshiftstop), their dist� is set to 1000000
        else {
	    for(size_t m=0; m < pelist->size(); m++) (*pelist)[m]->SeDist2ShiftToShowPt(1000000);
        }

	// pe- hold in pixels without bgnd are set to SNR = -1000
	PhotoElectron* pe = 0;
	for(size_t m=0; m < pelist->size(); m++) {
	    pe = (*pelist)[m];
	    if(pe->Dist2ShiftToShowPt() > fMaxLateralDist2) pe->SetSNR(-1000.);
	}
    }

    return rtn;
}

//______________________________________________________________________________
Int_t SimpleDetector::ApplyTrigger() {
    //
    // trigger simulation :
    // ==================
    // 1. pixels with SNR > fSNRtrig are considered hitted, SNR =  (S + B - <B>) / sqrt(<B>)
    // 2. persistency = from one GTU to the next one, if adjacent hitted pixels exist, Npers is incremented
    //    when Npers = fNpersTrigg --> event is triggered
    //
    // NB : pe- hold in pixels without bgnd have been set to SNR = -1000, thus not considered for trigger simu
    // NB : pe- in pixels with SNR > fSNRtrig are tagged "triggered"
    //

    Msg(EsafMsg::Info)<< "<ApplyTrigger> SNRtrig = " <<fSNRtrigg<<", persistency = "<<fNpersTrigg<<" GTU" << MsgDispatch;


    // inits
    Int_t rtn(0);
    UInt_t GTUid = 0;
    vector< vector<Int_t> > list_hitpixID; // one list of hitted pixels per GTU, the list at GTU=i is modified at GTU=i+1,i+2,i+3 etc..
    vector<UInt_t> Npers;                  // one nb per GTU
    vector<Int_t> init;
    for(UInt_t i=0; i < fNbGTUs; i++) {
        Npers.push_back(0);
	list_hitpixID.push_back(init);
    }

    // loop on GTUs
    map<UInt_t, vector<PhotoElectron*>* >::const_iterator it;
    for(it = fPhotoElectrons.begin(); it != fPhotoElectrons.end(); it++) {
	GTUid = it->first;
	vector<PhotoElectron*>* pelist = it->second;
	vector<PhotoElectron*> usefullpelist;

	// if photons exist in this GTU
	if(pelist) {

	    // 0. init
	    Int_t prev_pixID = -1;
	    Int_t pixID = 0;
	    Int_t nbpe_inpixel = 0;
	    PhotoElectron* pe = 0;


	    // 1. get rid of useless pe- --> then use usefullpelist  (see AddBackground() method, SNR = -1000.)
	    for(size_t n=0; n < pelist->size(); n++) {
	        pe = (*pelist)[n];
		if(pe->SNR() > -1000.) usefullpelist.push_back(pe);
	    }
	    size_t pelistsize = usefullpelist.size();


	    // 2. sort pe- according to pixelID
            TArrayI pixelIDs(pelistsize);
	    for(size_t n=0; n < pelistsize; n++) pixelIDs[n] = Int_t(usefullpelist[n]->PixelID());
            Int_t sortedindex[pelistsize];
            Sort((Int_t)pelistsize,pixelIDs.GetArray(),sortedindex,false); // false means in increasing order



	    // 3. loop on pe- hold in this GTU
	    //    treatment pixel per pixel
	    vector<PhotoElectron*> buffer;
	    for(size_t m=0; m < pelistsize; m++) {
		pe = usefullpelist[sortedindex[m]];
		pixID = Int_t(pe->PixelID());
		if(pixID > prev_pixID) {
		    // another pixel is considered, work out SNR treatment for previous pixel
		    Double_t SNR = 0.;
		    if(fMeanBgnd > 0) SNR = (Float_t(nbpe_inpixel) - fMeanBgnd) / Sqrt(fMeanBgnd);
		    else SNR = -100.;
		    // previous pixel is hitted --> stored in list of hitted pixel
		    if((SNR > fSNRtrigg) && buffer.size()>0) list_hitpixID[GTUid].push_back(prev_pixID);
                    for(size_t t=0; t < buffer.size(); t++) {
		        buffer[t]->SetSNR(SNR);
		        rtn++;
                        if(SNR > fSNRtrigg) buffer[t]->Triggered();
		        else rtn--;
		    }
		    // reset for next pixel
		    buffer.clear();
		    nbpe_inpixel = 1;
		}
		else if(pixID == prev_pixID) {
		    nbpe_inpixel++;
		}
		else Msg(EsafMsg::Panic)<< "<ApplyTrigger> pixelID are not increasingly sorted : SHOULD NOT"<< MsgDispatch;
		buffer.push_back(pe);
		prev_pixID = pixID;
	    }
	    // SNR treatment for last pixel
	    Double_t SNR = 0.;
            if(fMeanBgnd > 0) SNR = (Float_t(nbpe_inpixel) - fMeanBgnd) / Sqrt(fMeanBgnd);
	    else SNR = -100.;
	    // last pixel is hitted --> stored in list of hitted pixel
	    if((SNR > fSNRtrigg) && buffer.size()>0) list_hitpixID[GTUid].push_back(pixID);
            for(size_t t=0; t < buffer.size(); t++) {
                buffer[t]->SetSNR(SNR);
                rtn++;
                if(SNR > fSNRtrigg) buffer[t]->Triggered();
                else rtn--;
            }
	    buffer.clear();



	    // 4. look for signal persistency
	    // 4.1 for the current GTU
	    if(!fEvIsTriggered) {
		if(list_hitpixID[GTUid].size() > 0) Npers[GTUid] = 1;
		else Npers[GTUid] = 0;

		// 4.2 for previous GTU : remove pixels in prev GTU containing no adjacent pixels with the current GTU
		// init
		Int_t prevGTUid = 0;
		Int_t iprev = -1;
		Int_t jprev = -1;
		Int_t icurr = -1;
		Int_t jcurr = -1;
		Int_t pixIDprevGTU = 0;
		Int_t pixIDcurrGTU = 0;
		vector<Int_t> buf; // used to update pixels of prev GTU
		// loop on previous GTUs
		for(UInt_t j=1; j < fNpersTrigg; j++) {
	            // go back in time until GTUid - (fNpersTrigg-1)
		    prevGTUid = GTUid - j;
	            if(prevGTUid < 0) break;

		    // if current GTU has no hitted pixels : remove all pixels in previous lists
		    if(Npers[GTUid] == 0) {
			list_hitpixID[prevGTUid].clear();
			continue;
		    }

		    // else : loops on both pixel list (prev and current GTU) to look for adjacent pixels
		    size_t prevsize = list_hitpixID[prevGTUid].size();
		    size_t currsize = list_hitpixID[GTUid].size();
		    for(size_t m=0; m < currsize; m++) {
			pixIDcurrGTU = list_hitpixID[GTUid][m];
			icurr = pixIDcurrGTU / fNbPixelOnSide;
			jcurr = pixIDcurrGTU % fNbPixelOnSide;
			for(size_t n=0; n < prevsize; n++) {
		            pixIDprevGTU = list_hitpixID[prevGTUid][n];
		            iprev = pixIDprevGTU / fNbPixelOnSide;
		            jprev = pixIDprevGTU % fNbPixelOnSide;
			    if( abs(icurr - iprev) <= 1 || abs(jcurr - jprev) <= 1 ) {
				buf.push_back(pixIDcurrGTU);
				break;
			    }
			}
		    }


		    // clear list of previous GTU, and replace it by "buf" content
		    // if this content is not null, increment Npers of previous GTU
		    // if Npers > fNpersTrigg, event is triggered and simu is stopped
		    list_hitpixID[prevGTUid].clear();
		    for(size_t i=0; i < buf.size(); i++) list_hitpixID[prevGTUid].push_back(buf[i]);
		    if(buf.size() > 0 ) Npers[prevGTUid]++;
		    if(Npers[prevGTUid] == fNpersTrigg) {
			fEvIsTriggered = kTRUE;
			break;  // stop trigger simu
		    }
		    buf.clear();
		}
	    }
	}


	// if no pe- at this GTU
	else {
	    Npers[GTUid] = 0;
	    Int_t prevGTUid = 0;
	    for(UInt_t j=1; j <= fNpersTrigg; j++) {
	        // go back in time until (GTUid - fNpersTrigg)
		prevGTUid = GTUid - j;
	        if(prevGTUid < 0) break;
		// current GTU has no hitted pixels : remove all pixels in previous lists
		list_hitpixID[prevGTUid].clear();
	    }
	}
    }

    return rtn;
}

/* used by old version of SimpleDetector::AddBackground()  (see below)
//______________________________________________________________________________
Bool_t SimpleDetector::LateralDist2ToTrack(UInt_t pixID, UInt_t& Distance2) const {
    //
    // get lateral distance� to the track of given pixel
    //
    // return kTRUE if pixel is in the FoV, kFALSE otherwise
    //

    // find pixel position in the Mes Z=0 plane
    EarthVector Pos2D;
    Bool_t rtn = Pos2D_Z0plane(pixID,Pos2D);

    // get projected position of Pos2D on the track
    EarthVector Dir2DShowPt = fDirShowPt;
    Dir2DShowPt.SetZ(0.);
    Dir2DShowPt = Dir2DShowPt.Unit();
    EarthVector StartPos2D = fStartPosShowPt;
    StartPos2D.SetZ(0.);
    EarthVector pos_startpos2D = Pos2D - StartPos2D;
    EarthVector projectedpos2D = StartPos2D + pos_startpos2D.Dot(Dir2DShowPt)*Dir2DShowPt;

    // get pixel ID of projected position
    UInt_t pixIDprojected = 0;
    Bool_t IsinFoV = FindPixelID(projectedpos2D,pixIDprojected);

    // if projected pos is in foV, calculate lateral distance� of pixID w.r.t. pixIDprojected, in  (nb pixels)� unit
    if(IsinFoV) {
        UInt_t ipos = pixID / fNbPixelOnSide;
        UInt_t jpos = pixID % fNbPixelOnSide;
        UInt_t iposprojected = pixIDprojected / fNbPixelOnSide;
        UInt_t jposprojected = pixIDprojected % fNbPixelOnSide;
        Distance2 = (ipos - iposprojected)*(ipos - iposprojected) + (jpos - jposprojected)*(jpos - jposprojected);
    }
    // else, uses a trick : typically underestimate the distance�
    else {
        Double_t dist2  = pos_startpos2D.Mag2() - pos_startpos2D.Dot(Dir2DShowPt)*pos_startpos2D.Dot(Dir2DShowPt);
        dist2 /= fPixelSideLength * fPixelSideLength; // distance� is in unit : (nb pixels)�
        Distance2 = UInt_t(floor(dist2 + 0.5));
    }

    return rtn;
}
*/


/*

// add bgnd at all GTU all along the track ---> TOO MUCH PE� !!, particularly for inclined showers
// if reactivated, to reactivate SpaceTimePEinfos() method too !!

//______________________________________________________________________________
Int_t SimpleDetector::AddBackground() {
    //
    // Add background around showertrack (for each GTU)
    // For every pe-, set the spatial shift w.r.t. shower point
    //
    // NB : looking at shower point at GTU start and GTU end could miss an intermediate pixel
    //      not so important if the intermediate pixel is considered in simu analysis or reconstruction
    //

    Int_t rtn(0);
    UInt_t GTUnb = 0;
    Bool_t IsInFoV = kFALSE;

    // loop on GTUs
    map<UInt_t, vector<PhotoElectron*>* >::const_iterator it;
    for(it = fPhotoElectrons.begin(); it != fPhotoElectrons.end(); it++) {

	vector<PhotoElectron*>* pelist = it->second;
	GTUnb = it->first;



	// 1. init + find rectangular region coordinates, where bgnd will be added
	Int_t prev_pixID = -1;
	Int_t pixID = 0;
        UInt_t low_i = fStartPixelID / fNbPixelOnSide;
	UInt_t up_i  = fEndPixelID   / fNbPixelOnSide;
	if(low_i > up_i) {
            low_i    = fEndPixelID   / fNbPixelOnSide;
	    up_i     = fStartPixelID / fNbPixelOnSide;
	}
        UInt_t low_j = fStartPixelID % fNbPixelOnSide;
	UInt_t up_j  = fEndPixelID   % fNbPixelOnSide;
	if(low_j > up_j) {
            low_j    = fEndPixelID   % fNbPixelOnSide;
	    up_j     = fStartPixelID % fNbPixelOnSide;
	}
	size_t NBPIXELSid = (up_i - low_i + 1) * (up_j - low_j + 1);
        TArrayI pixelIDs_temp(NBPIXELSid);



	// 2. store IDs of involved pixels (area defined by fStartPixelID, fEndPixelID, fMaxLateralDist2 and Fov bounds)
	size_t ind = 0;
	UInt_t LatDist2 = 0;
	for(UInt_t i=low_i; i <= up_i; i++) {
	    for(UInt_t j=low_j; j <= up_j; j++) {
		// check if searched pixel is in FoV + if its lateral distance� to the track < fMaxLateralDist2
		IsInFoV = LateralDist2ToTrack(i*fNbPixelOnSide + j,LatDist2);
		if(!IsInFoV || (LatDist2 > fMaxLateralDist2)) continue;
		else pixelIDs_temp[ind++] = i*fNbPixelOnSide + j;
	    }
	}



	// 3. copy usefull pixels ID
	if(ind < NBPIXELSid) NBPIXELSid = ind;
        TArrayI pixelIDs(NBPIXELSid);
	for(size_t n=0; n < NBPIXELSid; n++) pixelIDs[n] = pixelIDs_temp[n];



	// 4. sort pe- according to pixelID
        Int_t sortedindex[NBPIXELSid];
        Sort(NBPIXELSid,pixelIDs.GetArray(),sortedindex,false); // false means in increasing order



	// 5. then add background ONCE per pixel
	prev_pixID = -1;
	pixID = 0;
	Int_t nbPEbgnd = 0;
	vector<PhotoElectron*> buffer;
	EarthVector pixel2Dpos;
	for(size_t m=0; m < NBPIXELSid; m++) {
	    nbPEbgnd = 0;
	    pixID = pixelIDs[sortedindex[m]];
	    if(pixID > prev_pixID) {
		// find pixel corresponding incident angle
		Pos2D_Z0plane(pixID, pixel2Dpos);
		Double_t th = (fGeometry->GetPos() - pixel2Dpos).Theta();

		// create background pe-
		// Bgnd source assumed _lambertian_ :
		//    - one Cos(th) stems from euso surface projection
		//    - the other one because pixel surface projection on ground is constant in SimpleDetector
		nbPEbgnd = EsafRandom::Get()->Poisson(Cos(th) * Cos(th) * fMeanBgnd);
		PhotoElectron* pe = 0;
		for(Int_t p=0; p < nbPEbgnd; p++) {
		    pe = new PhotoElectron();
		    rtn++;
		    pe->SetTime(it->first * fGTUlength + fFirstGTUTime + 0.5*fGTUlength);
		    pe->SetPixelID(UInt_t(pixID));
		    buffer.push_back(pe);
		}
	    }
	    prev_pixID = pixID;
	}
	for(size_t m=0; m < buffer.size(); m++) pelist->push_back(buffer[m]);
	buffer.clear();




	// 6. find shower point pixels for this GTU, to set spatial shift (part 7. below)
	// NB : WHEN END OF TRACK IS REACHED --> goes on with the same angular speed (as if track goes on under ground or out of FoV)
	UInt_t LowEdgeRefPixel = 0;  // at GTU low edge
	UInt_t UpEdgeRefPixel  = 0;  // at GTU up edge
	UInt_t LowEdge_i = 0;
	UInt_t UpEdge_i  = 0;
	UInt_t LowEdge_j = 0;
	UInt_t UpEdge_j  = 0;
	Bool_t IsSamePixel = kFALSE;
	Bool_t reach_endtrack = kFALSE;
	Double_t tof = 0.;
	Double_t tof_alongtrack = 0.;
	Double_t current_tof = 0.;
        Int_t infloop = 0;
	const Double_t increm = 1.*sou::m;
	EarthVector ShowPtPos;
	EarthVector EUSO = fGeometry->GetPos();
	    // low GTU edge
	current_tof = Double_t(GTUnb) * fGTUlength;
        tof = 0.;
        ShowPtPos = fStartPosShowPt;
        while(tof <= current_tof) {
            ShowPtPos += increm*fDirShowPt;
	    tof_alongtrack += increm / Clight();
	    tof = tof_alongtrack + (EUSO - ShowPtPos).Mag() / Clight() - (EUSO - fStartPosShowPt).Mag() / Clight();
	    if(infloop++ > 1000000) Msg(EsafMsg::Panic)<< "<AddBackground, low edge> Infinite loop broken by hand"<< MsgDispatch;
	}
	    // check if end of track is reached
	if( (ShowPtPos - fStartPosShowPt).Mag() > (fEndPosShowPt - fStartPosShowPt).Mag() ) reach_endtrack = kTRUE;
	IsInFoV = FindPixelID(ShowPtPos,LowEdgeRefPixel);
	if(!IsInFoV) {
	    Msg(EsafMsg::Warning)<< "<AddBackground, low edge> Track pos is OUT OF FOV, may occur if event is close to FoV bounds" << MsgDispatch;
	    ShowPtPos = fEndPosShowPt;
	    IsInFoV = FindPixelID(ShowPtPos,LowEdgeRefPixel);
	}
	LowEdge_i = LowEdgeRefPixel / fNbPixelOnSide;
	LowEdge_j = LowEdgeRefPixel % fNbPixelOnSide;
	    // up GTU edge
	if(!reach_endtrack) {
	    current_tof += fGTUlength;
            infloop = 0;
            while(tof <= current_tof) {
        	ShowPtPos += increm*fDirShowPt;
		tof_alongtrack += increm / Clight();
		tof = tof_alongtrack + (EUSO - ShowPtPos).Mag() / Clight() - (EUSO - fStartPosShowPt).Mag() / Clight();
		if(infloop++ > 1000000) Msg(EsafMsg::Panic)<< "<AddBackground, up edge> Infinite loop broken by hand"<< MsgDispatch;
	    }
	    IsInFoV = FindPixelID(ShowPtPos,UpEdgeRefPixel);
	    if(!IsInFoV) Msg(EsafMsg::Warning)<< "<AddBackground, up edge> Track pos is OUT OF FOV, it SHOULD NOT" << MsgDispatch;
	    UpEdge_i = UpEdgeRefPixel / fNbPixelOnSide;
	    UpEdge_j = UpEdgeRefPixel % fNbPixelOnSide;
	    if(UpEdgeRefPixel == LowEdgeRefPixel) IsSamePixel = kTRUE;
	}
	else IsSamePixel = kTRUE;  // not a significant effect when impact on ground is reached, but simpler when FoV bound is reached



	// 7. Set pe- spatial shift w.r.t. shower point pixels
	//    If shower point travel several pixels within a GTU, only the extrem ones are considered as reference
	PhotoElectron* pe = 0;
	UInt_t lowDist2Shift = 0;
	UInt_t upDist2Shift = 0;
	for(size_t m=0; m < pelist->size(); m++) {
	    pe = (*pelist)[m];
	    pixID = pe->PixelID();
	    UInt_t pixi = pixID / fNbPixelOnSide;
	    UInt_t pixj = pixID % fNbPixelOnSide;
	    lowDist2Shift = (pixi - LowEdge_i)*(pixi - LowEdge_i) + (pixj - LowEdge_j)*(pixj - LowEdge_j);
	    if(!IsSamePixel) {
	        upDist2Shift  = (pixi - UpEdge_i) *(pixi - UpEdge_i)  + (pixj - UpEdge_j) *(pixj - UpEdge_j);
	        if(lowDist2Shift < upDist2Shift) pe->SeDist2ShiftToShowPt(lowDist2Shift);
	        else pe->SeDist2ShiftToShowPt(upDist2Shift);
	    }
	    else pe->SeDist2ShiftToShowPt(lowDist2Shift);
	}
    }

    return rtn;
}
*/


/* disabled : usefull only around fluo track, the rest of signal is not treated correctly
//______________________________________________________________________________
Int_t SimpleDetector::AddMinimizedBackground() {
    //
    // Add background around direct fluorescence pixels at a given GTU (Squared surface 7x7 pixels)
    // Needed for very inclined track, i.e. when total nb of bgnd pe- of previous method becomes huge
    // For every pe-, set the spatial shift w.r.t. shower point
    //
    // NB : looking at shower point at GTU start and GTU end could miss an intermediate pixel
    //      not so important if the intermediate pixel is considered in simu analysis or reconstruction
    //

    Int_t rtn(0);
    UInt_t GTUnb = 0;
    Bool_t IsInFoV = kFALSE;

    // loop on GTUs
    map<UInt_t, vector<PhotoElectron*>* >::const_iterator it;
    for(it = fPhotoElectrons.begin(); it != fPhotoElectrons.end(); it++) {

	vector<PhotoElectron*>* pelist = it->second;
	GTUnb = it->first;

	// if signal pe- exist in this GTU
	if(pelist) {


	    // 1. init + find rectangular region coordinates, where bgnd will be added
	    Int_t prev_pixID = -1;
	    Int_t pixID = 0;
	    size_t NBPIXELSid = pelist->size()*49;
            TArrayI pixelIDs_temp(NBPIXELSid);

	    // 2. store IDs of direct fluo hit pixels + surrouding pixels (49 pixels)
	    size_t ind = 0;
	    for(size_t n=0; n < pelist->size(); n++) {
		if((*pelist)[n]->AtmoHistory() == 0) {
		    pixID = Int_t((*pelist)[n]->PixelID());
		    UInt_t ipe = pixID / fNbPixelOnSide;
		    UInt_t jpe = pixID % fNbPixelOnSide;
	            UInt_t low_i = ipe - 3;
	            UInt_t up_i  = ipe + 3;
                    UInt_t low_j = jpe - 3;
	            UInt_t up_j  = jpe + 3;
		    UInt_t dummy = 0;
		    for(UInt_t i=low_i; i <= up_i; i++) {
			for(UInt_t j=low_j; j <= up_j; j++) {
			    // check if searched pixel is in FoV
			    IsInFoV = LateralDist2ToTrack(i*fNbPixelOnSide + j,dummy);
			    if(!IsInFoV) continue;
			    else pixelIDs_temp[ind++] = i*fNbPixelOnSide + j;
			}
		    }
		}
	    }


	    // 3. copy usefull pixels ID
	    if(ind < NBPIXELSid) NBPIXELSid = ind;
            TArrayI pixelIDs(NBPIXELSid);
	    for(size_t n=0; n < NBPIXELSid; n++) pixelIDs[n] = pixelIDs_temp[n];



	    // 4. sort pe- according to pixelID
            Int_t sortedindex[NBPIXELSid];
            Sort(NBPIXELSid,pixelIDs.GetArray(),sortedindex,false); // false means in increasing order



	    // 5. then add background ONCE per pixel
	    prev_pixID = -1;
	    pixID = 0;
	    Int_t nbPEbgnd = 0;
	    vector<PhotoElectron*> buffer;
	    EarthVector pixel2Dpos;
	    for(size_t m=0; m < NBPIXELSid; m++) {
		nbPEbgnd = 0;
		pixID = pixelIDs[sortedindex[m]];
		if(pixID > prev_pixID) {
		    // find pixel corresponding incident angle
		    Pos2D_Z0plane(pixID, pixel2Dpos);
		    Double_t th = (fGeometry->GetPos() - pixel2Dpos).Theta();

		    // create background pe-
		    // Bgnd source assumed _lambertian_ :
		    //    - one Cos(th) stems from euso surface projection
		    //    - the other one because pixel surface projection on ground is constant in SimpleDetector
		    nbPEbgnd = EsafRandom::Get()->Poisson(Cos(th) * Cos(th) * fMeanBgnd);
		    PhotoElectron* pe = 0;
		    for(Int_t p=0; p < nbPEbgnd; p++) {
			pe = new PhotoElectron();
			rtn++;
			pe->SetTime(it->first * fGTUlength + fFirstGTUTime + 0.5*fGTUlength);
			pe->SetPixelID(UInt_t(pixID));
			buffer.push_back(pe);
		    }
		}
		prev_pixID = pixID;
	    }
	    for(size_t m=0; m < buffer.size(); m++) pelist->push_back(buffer[m]);
	    buffer.clear();




	    // 6. find shower point pixels
	    // NB : WHEN END OF TRACK IS REACHED --> goes on with the same angular speed (as if track goes on under ground or out of FoV)
	    UInt_t LowEdgeRefPixel = 0;  // at GTU low edge
	    UInt_t UpEdgeRefPixel  = 0;  // at GTU up edge
	    UInt_t LowEdge_i = 0;
	    UInt_t UpEdge_i  = 0;
	    UInt_t LowEdge_j = 0;
	    UInt_t UpEdge_j  = 0;
	    Bool_t IsSamePixel = kFALSE;
	    Bool_t reach_endtrack = kFALSE;
	    Double_t tof = 0.;
	    Double_t tof_alongtrack = 0.;
	    Double_t current_tof = 0.;
            Int_t infloop = 0;
	    const Double_t increm = 1.*sou::m;
	    EarthVector ShowPtPos;
	    EarthVector EUSO = fGeometry->GetPos();
		// low GTU edge
	    current_tof = Double_t(GTUnb) * fGTUlength;
            tof = 0.;
            ShowPtPos = fStartPosShowPt;
            while(tof <= current_tof) {
        	ShowPtPos += increm*fDirShowPt;
		tof_alongtrack += increm / Clight();
		tof = tof_alongtrack + (EUSO - ShowPtPos).Mag() / Clight() - (EUSO - fStartPosShowPt).Mag() / Clight();
		if(infloop++ > 1000000) Msg(EsafMsg::Panic)<< "<AddBackground, low edge> Infinite loop broken by hand"<< MsgDispatch;
	    }
		// check if end of track is reached
	    if( (ShowPtPos - fStartPosShowPt).Mag() > (fEndPosShowPt - fStartPosShowPt).Mag() ) reach_endtrack = kTRUE;
	    IsInFoV = FindPixelID(ShowPtPos,LowEdgeRefPixel);
	    if(!IsInFoV) {
		Msg(EsafMsg::Warning)<< "<AddBackground, low edge> Track pos is OUT OF FOV, may occur if event is close to FoV bounds" << MsgDispatch;
		ShowPtPos = fEndPosShowPt;
		IsInFoV = FindPixelID(ShowPtPos,LowEdgeRefPixel);
	    }
	    LowEdge_i = LowEdgeRefPixel / fNbPixelOnSide;
	    LowEdge_j = LowEdgeRefPixel % fNbPixelOnSide;
		// up GTU edge
	    if(!reach_endtrack) {
		current_tof += fGTUlength;
        	infloop = 0;
        	while(tof <= current_tof) {
        	    ShowPtPos += increm*fDirShowPt;
		    tof_alongtrack += increm / Clight();
		    tof = tof_alongtrack + (EUSO - ShowPtPos).Mag() / Clight() - (EUSO - fStartPosShowPt).Mag() / Clight();
		    if(infloop++ > 1000000) Msg(EsafMsg::Panic)<< "<AddBackground, up edge> Infinite loop broken by hand"<< MsgDispatch;
		}
		IsInFoV = FindPixelID(ShowPtPos,UpEdgeRefPixel);
		if(!IsInFoV) Msg(EsafMsg::Warning)<< "<AddBackground, up edge> Track pos is OUT OF FOV, it SHOULD NOT" << MsgDispatch;
		UpEdge_i = UpEdgeRefPixel / fNbPixelOnSide;
		UpEdge_j = UpEdgeRefPixel % fNbPixelOnSide;
		if(UpEdgeRefPixel == LowEdgeRefPixel) IsSamePixel = kTRUE;
	    }
	    else IsSamePixel = kTRUE;



	    // 7. Set pe- spatial shift w.r.t. shower point pixels
	    //    If shower point travel several pixels within a GTU, only the extrem ones are considered as reference
	    PhotoElectron* pe = 0;
	    UInt_t lowDist2Shift = 0;
	    UInt_t upDist2Shift = 0;
	    for(size_t m=0; m < pelist->size(); m++) {
		pe = (*pelist)[m];
		pixID = pe->PixelID();
		UInt_t pixi = pixID / fNbPixelOnSide;
		UInt_t pixj = pixID % fNbPixelOnSide;
		lowDist2Shift = (pixi - LowEdge_i)*(pixi - LowEdge_i) + (pixj - LowEdge_j)*(pixj - LowEdge_j);
		if(!IsSamePixel) {
	            upDist2Shift  = (pixi - UpEdge_i) *(pixi - UpEdge_i)  + (pixj - UpEdge_j) *(pixj - UpEdge_j);
	            if(lowDist2Shift < upDist2Shift) pe->SeDist2ShiftToShowPt(lowDist2Shift);
	            else pe->SeDist2ShiftToShowPt(upDist2Shift);
		}
		else pe->SeDist2ShiftToShowPt(lowDist2Shift);
	    }
	}
    }

    return rtn;
}
*/

