// class SimpleTelemetry
// very simple implementation of Telemetry
// it just writes data onto a ASCII file
// M. Pallavicini - created
//

#include "SimpleTelemetry.hh"
#include "MacroCellData.hh"
#include "MacroCell.hh"
#include "MacroCellHit.hh"
#include "MCTruth.hh"

ClassImp(SimpleTelemetry)

//______________________________________________________________________________
SimpleTelemetry::SimpleTelemetry() : 
   Telemetry(), fTextDump("Empty"), fp(NULL)
{
    // 
    // constructor
    //
    SetState( kFALSE );
}

//______________________________________________________________________________
SimpleTelemetry::~SimpleTelemetry() {
    // 
    // delete all MacroCellData objects in the vector
    //
    if ( IsOpen() )
        Close();
    Clear();
} 

//______________________________________________________________________________
void SimpleTelemetry::Clear() {
    //
    // Delete current event
    //
    
    for( unsigned int i=0; i < fData.size(); i++) {
        MacroCellData* p = fData[i];
        if ( p )
            delete p;
    }
    
    fData.clear();
}

//______________________________________________________________________________
void SimpleTelemetry::Add( MacroCellData* pData ) {
    //
    // add data coming from a MacroCell
    //
    fData.push_back( pData );
}

//______________________________________________________________________________
void SimpleTelemetry::SimulateTCU() {
    // 
    // Simulate TCU: trigger and on-board data handling
    // does nothing in this implementation
    //
    Msg(EsafMsg::Info) << fData.size() << " MacroCellData collected." << MsgDispatch;
    Msg(EsafMsg::Info) << "TCU simulation done" << MsgDispatch;
    return;
}

//______________________________________________________________________________
Bool_t SimpleTelemetry::Open( const char* fname ) {
    //
    // Open the output file
    //
    if ( IsOpen() )
        return kFALSE;

    fFileName = fname;

    // check if a file with this name already exists
    // in this case, change file name
    if (fFileName.find(".dat.gz") != (fFileName.size()-7))
    fFileName += ".dat.gz";

    fp = gzopen(fFileName.c_str(),"rt");
    if ( fp ) {
        fFileName.resize(fFileName.size()-7);
        gzclose( fp );
        char new_file_name[500];
        char temp_s[100];
        time_t tt = time(NULL);
        strftime(temp_s,99,"%d-%b-%G:%H:%M:%S",localtime(&tt));
        sprintf(new_file_name,"%s.%s",fFileName.c_str(),temp_s);
        fFileName = new_file_name;
    }
    fFileName += ".dat.gz";
    fp = gzopen(fFileName.c_str(),"wt+");

    if ( fp == NULL ) {
        FatalError("Error opening Telemetry file");
    }
    Msg(EsafMsg::Info) << "SimpleTelemetry: " <<  fFileName << " telemetry file opened" << MsgDispatch;

    SetState( kTRUE );

    return kTRUE;
}

//______________________________________________________________________________
Bool_t SimpleTelemetry::Close() {
    // 
    // Close the file if open
    //
    
    Msg(EsafMsg::Info)<< "Closing telemetry file" << MsgDispatch;

    if ( !IsOpen() )
        return kFALSE;
    
    gzclose(fp);
  
    SetState( kFALSE );
  
    return kTRUE;
}

//______________________________________________________________________________
Bool_t SimpleTelemetry::Write() {
    // 
    // Write data on file
    //
    
    // if file is not open, open it
    if ( !IsOpen() ) 
        return kFALSE;
   
 
    // write header data
    WriteHeader();
    
    // write data
    WriteData();
    
    // clean memory
    Clear();
    
    return kTRUE;
}

//______________________________________________________________________________
string& SimpleTelemetry::Dump() {
    //
    // Dump on a string a returns it the same string is used at each call
    // 
    
    fTextDump = "This is a dummy event\n";
    return fTextDump;
}

//______________________________________________________________________________
void SimpleTelemetry::WriteHeader() {
    // 
    // Write header
    //

    // run number
    //gzprintf(fp,"fRunNumber = %d\n",gSimuApplication->GetRunNumber());
    // run date
    //gzprintf(fp,"RunDate = %s\n",EusoTimeOrbit::Get()->GetRunDate().c_str());

    // general data
    // FIXME: mostly dummy or hard wired
    // FIXME: need to recover infos from right places; to be done
//      if ( GetTruth()) {
//          gzprintf(fp,"EventGenerator = %s\n",
//  			GetTruth()->GetLightToEuso()->GetName().c_str());
    if ( !IsOpen() ) {
      gzprintf(fp,"GenerationTree = 000000\n");
      gzprintf(fp,"ShowerDumpLevel = 0\n");
      gzprintf(fp,"EnergySlope = 0\n");
      gzprintf(fp,"EnergyRangeLow = 0\n");
      gzprintf(fp,"EnergyRangeHigh = 0\n");
      gzprintf(fp,"ThetaRangeLow = 0\n");
      gzprintf(fp,"ThetaRangeHigh = 0\n");
      gzprintf(fp,"PhiRangeLow = 0\n");
      gzprintf(fp,"PhiRangeHigh = 6.28\n");
      gzprintf(fp,"ElectromagneticFlag = 1\n");
      gzprintf(fp,"HadronEnergyCutOff = 0\n");
      gzprintf(fp,"MuonEnergyCutOff = 0\n");
      gzprintf(fp,"ElectronEnergyCutOff = 0\n");
      gzprintf(fp,"PhotonEnergyCutOff = 0\n");
      gzprintf(fp,"ThinHybridLevel = 1\n");
      gzprintf(fp,"EusoAltitude = 430000\n");
      gzprintf(fp,"EntrancePupilDiameter = 2\n");
      gzprintf(fp,"FocalNumber = 0\n");
      gzprintf(fp,"EusoPixelSize = 2\n");
      gzprintf(fp,"PixelDigitalThreshold = 6\n");
      gzprintf(fp,"TriggerThreshold = 2\n");
    }
}

//______________________________________________________________________________
void SimpleTelemetry::WriteData() {
    // 
    // Write data
    // 

    if ( !IsOpen() ) return;

    // number of macrocells that gave signal
    gzprintf(fp,"NumMacrocells = %d\n",NumCell());
    // total number of macrocell hits
    int hits=0;
    for( int iCell=0; iCell < NumCell(); iCell++) {
        MacroCellData *pData = fData[iCell];
        if ( pData ) {
            hits += pData->NumMacroCellHits();   
        }
    }
    gzprintf(fp,"NumMacroCellHits = %d\n",hits);

    // macrocell hits
    // write a table with the follwing data:
    //    hit,cell,gtu,row,col,counts,theta,phi 
    int hit_counter=0;
    for( int iCell=0; iCell < NumCell(); iCell++) {
        MacroCellData *pD = fData[iCell];
        if ( pD ) {
            for(int iGtu = pD->GtuStart(); iGtu < pD->GtuEnd(); iGtu++) {
                if ( pD->Hits( iGtu ) ) {
                    for( unsigned int iH=0; iH < pD->Hits(iGtu)->size(); iH++) {
                        MacroCellHit *pH = pD->Hit( iGtu, iH );
                        if ( pH ) {
                            hit_counter++;
                            if ( iGtu != pH->Gtu()) {
                                cerr << "\n\n******** GTU Mapping Error !!!\n\n";
                            }
                            gzprintf(fp,"%d\t%d\t%d\t%d\t%d\t%d\t%f\t%f\n",
                                    hit_counter, pD->Cell()->Id(), iGtu,  
                                    pH->Row(), pH->Col(), pH->Hits(),
                                    pH->GetThetaFOV(), pH->GetPhiFOV());	
                        }
                    }
                }
            }
        }
    }
}
