/* -------------------------------------------------------------------------
 *   trace_optF1v4.c
 *
 *   --- main program of raytracing in the Fresnel lens system
 *
 * Copyright (c) 2000-2007 N.Sakaki, Y.Takizawa, Y.Kawasaki
 * All rights reserved
 * $Id: Ltrace_optF1v4.cc,v 1.0 2007/11/13 Dmitry V.Naumov adapted for ESAF
 * -------------------------------------------------------------------------
 */
#include <stdio.h>
#include <stdlib.h>
#include "Lspline.hh"
#include "Lutils.hh"
#include "TMath.h"
#include "TFile.h"
#include "EsafRandom.hh"
#include "EConst.hh"
#include "Ltrace_optF1v4.hh"

#define RT_SEARCH_ITER_MAX         100
#define RT_INTERACTION_MAX         20
#define S_DATA 130000
/* alias */
#define X      0
#define Y      1
#define Z      2
#define PX     3
#define PY     4
#define PZ     5
#define L      6
#define T      7

//#define DEBUG
using namespace sou;
/* lens surface data */
static Int_t n_s1=0,n_s2=0,n_s4=0,n_s6=0,n_s7=0;
static Double_t sr1[S_DATA],sz1[S_DATA];
static Double_t sr2[S_DATA],sz2[S_DATA];
static Double_t sr4[S_DATA],sz4[S_DATA];
static Double_t sr6[S_DATA],sz6[S_DATA];
static Double_t sr7[S_DATA],sz7[S_DATA];

/* lens material data */
static Int_t n_acryl=0, n_cytop=0, n_bg3=0;
static Double_t ac_la[LAcryl_N],ac_n[LAcryl_N],ac_k[LAcryl_N];
static Double_t cy_la[LAcryl_N],cy_n[LAcryl_N],cy_k[LAcryl_N];
static Double_t bg3_la[LBg3_N],bg3_n[LBg3_N],bg3_k[LBg3_N];



/* diffractive structure data */
static Int_t n_sd1=0;
#define N_DIFFR_DATA   130000
static Double_t SD1_R[N_DIFFR_DATA];
static Double_t SD1_D[N_DIFFR_DATA];

static Double_t Lref(Double_t n1, Double_t n2, Double_t ci, Double_t ct);
static Int_t    Lcheck_intersect_cylinder(Double_t r1, Double_t z1, Double_t r2, Double_t z2,
				    Double_t phot[8], Double_t *dis);
static Int_t    Lcheck_intersect_cone    (Double_t r1, Double_t z1, Double_t r2, Double_t z2,
				    Double_t phot[8], Double_t *dis);
static Int_t    LCheckIris(tel_param *p, Double_t phot[6], Int_t flag);

/*
 * ========================================================================
 * functions for diffractive optics start
 * ========================================================================
 */
/*
 * ------------------------------------------------------------------------
 *  Calculate normal vector to the sphere approximating the lens surface
 *  Input:
 *    aX, aY, aZ: photon position on the lens surface
 *    aCZ       : center of the approximated sphere
 *
 *  Output:
 *    nnnDIFF[3]: normal unit vector
 * ------------------------------------------------------------------------
 */
Int_t LCalSpheVec(Double_t aX, Double_t aY, Double_t aZ, Double_t aCZ, Double_t nnnDIFF[3])
{

  Double_t N_;

  nnnDIFF[0] = aX;
  nnnDIFF[1] = aY;
  nnnDIFF[2] = aZ - aCZ;

  N_ = TMath::Sqrt(Lvdot(nnnDIFF,nnnDIFF));

  nnnDIFF[0] = nnnDIFF[0]/N_;
  nnnDIFF[1] = nnnDIFF[1]/N_;
  nnnDIFF[2] = nnnDIFF[2]/N_;

  return 0;
}


/*
 * ------------------------------------------------------------------------
 *  read parameters of diffractive optics
 *    Input:
 *        dirname    the name of the directory name
 *                        where data files are stored
 *    Output:
 *        (return value) status
 * ------------------------------------------------------------------------
 */
Int_t LReadDiffractData(char *dirname, FILE *fplog)
{
  FILE *fr;
  char Infname[128];

#ifdef DEBUG
  if(fplog)
    fprintf(fplog,"Reading Diffractive Surface Data....  ");
#endif

  /* read diffractive optics data */
  sprintf(Infname,"%.100s/DOE_F1v4.dat",dirname);

  if(( fr = fopen(Infname,"r" )) == NULL)
    {
      if(fplog)
	fprintf(fplog,"%s not OPEN\n",Infname );
      return (-1);
    }
  while(n_sd1<N_DIFFR_DATA
	&& fscanf(fr,"%lf %lf",&SD1_R[n_sd1],&SD1_D[n_sd1])==2)
    n_sd1++;
  fclose(fr);
  if(n_sd1==N_DIFFR_DATA)
    return (-2);

#ifdef DEBUG
  if(fplog)
    fprintf(fplog,"Done.\n");
#endif

  return (0);
}

/*
 * ------------------------------------------------------------------------
 *  calculate the diffractive structure
 *   Input:
 *      r   the distance in mm to the optical axis at the intersection point
 *      s   surface ID (i.e. S1,S2,...)
 *   Output:
 *     (return value)
 * ------------------------------------------------------------------------
 */
Double_t LCal_D(Double_t r)
{
  Int_t i;

  i=Lrt_bsearch(SD1_R,r,0,n_sd1-1);
  return (SD1_D[i]);
}

/*
 * ------------------------------------------------------------------------
 *  calculate the diffracted ray
 *  Input:
 *        in_ph[3]      array x,y,z of photon
 *        nnn[3]        normal vector to the diffractive surface
 *                         at the intersection
 *        n1            refractive index of xxx
 *        n2            refractive index of xxx
 *        d
 *        wl            wavelength in mm
 *        m             index
 *  Output:
 *        in_ph[3]
 * ------------------------------------------------------------------------
 */
Int_t Ldiffract(Int_t ID, Double_t in_phdir[3],
	     Double_t x3, Double_t y3, Double_t z3, Double_t Zc,
	     Double_t n1, Double_t n2, Double_t wl, Double_t wl0, Int_t m)
{
  Int_t    i;
  Double_t CosIN, dir, nnnR[3], r3, ppp[3], nnn[3], d, norm;

  r3=TMath::Sqrt(x3*x3+y3*y3);
  nnnR[0]= x3/r3;
  nnnR[1]= y3/r3;
  nnnR[2]=0.0;

  nnn[0]= 0.0;
  nnn[1]= 0.0;
  nnn[2]= 1.0;

  d=LCal_D(r3);

  if(in_phdir[Z]>0.)
    dir=1.0;
  else
    dir=-1.0;

  CosIN=in_phdir[Z];

  for(i=0;i<3;i++)
    ppp[i]=(n1*(in_phdir[i]-CosIN*nnn[i])
	    +(Double_t)m*wl/wl0*d*nnnR[i]*dir)/n2;

  norm=Lvdot(ppp,ppp);
  if(norm>1.0)
    return -1;

  norm=sqrt(1.0-norm);

#ifdef DEBUG
  fprintf(stderr,"Diffractive: (%.3f , %.3f , %.3f)=>",
	  in_phdir[0],in_phdir[1],in_phdir[2]);
#endif

  for(i=0;i<3;i++)
    in_phdir[i]=ppp[i]+norm*nnn[i]*(CosIN>0.0?(1.0):(-1.0));

#ifdef DEBUG
  fprintf(stderr,"(%.3f , %.3f , %.3f)\n",in_phdir[0],in_phdir[1],in_phdir[2]);
#endif

  return (0);
}

/*
 * ========================================================================
 * functions for diffractive optics END
 * ========================================================================
 */

/*
 * ------------------------------------------------------------------------
 * calculate reflection probability assuming no polariation
 *   n1,n2: refractive index of the media of this side and the other side
 *   ci,ct: cosine of the incident light and reflected light to the surface
 *          normal vector
 * ------------------------------------------------------------------------
 */
static Double_t Lref(Double_t n1, Double_t n2, Double_t ci, Double_t ct)
{
  Double_t rp,rs;

  rp=(n2*ci-n1*ct)/(n2*ci+n1*ct);
  rs=(n1*ci-n2*ct)/(n1*ci+n2*ct);

  return((rp*rp+rs*rs)/2.0);
}

/*
 * ------------------------------------------------------------------------
 *  read lens surface shapes, and optical indices
 * ------------------------------------------------------------------------
 */
Int_t LReadLensData(tel_param *param, FILE *fplog)
{
  FILE *fp;
  char filename[128];

#ifdef DEBUG
  if(fplog)
    fprintf(fplog,"Reading Surface Data....  ");
#endif

  /* Read S1 surface */
  sprintf(filename,"%.100s/S1_F1v4.dat",param->Llens_dir);
  if((fp=fopen(filename,"r"))==NULL)
    {
      if(fplog)
	fprintf(fplog,"%s not found.\n",filename);
      return -1;
    }
    while(n_s1<S_DATA &&
        fscanf(fp,"%lf %lf",&sr1[n_s1], &sz1[n_s1])!=EOF)
    {
        sz1[n_s1]=param->LSz1-sz1[n_s1];
        n_s1++;
    }
    fclose(fp);

  /* Read S2 surface */
  sprintf(filename,"%.100s/S2_F1v4.dat",param->Llens_dir);
  if((fp=fopen(filename,"r"))==NULL)
    {
      if(fplog)
	fprintf(fplog,"%s not found.\n",filename);
      return -2;
    }
  while(n_s2<S_DATA &&
        fscanf(fp,"%lf %lf",&sr2[n_s2], &sz2[n_s2])!=EOF)
    {
      sz2[n_s2]=param->LSz2-sz2[n_s2];
      n_s2++;
    }
  fclose(fp);

  /* Read S4 surface */
  sprintf(filename,"%.100s/S4_F1v4.dat",param->Llens_dir);
  if((fp=fopen(filename,"r"))==NULL)
    {
      if(fplog)
	fprintf(fplog,"%s not found.\n",filename);
      return -4;
    }
  while(n_s4<S_DATA &&
        fscanf(fp,"%lf %lf",&sr4[n_s4], &sz4[n_s4])!=EOF)
    {
      sz4[n_s4]=param->LSz4-sz4[n_s4];
      n_s4++;
    }
  fclose(fp);

  /* Read S6 surface */
  sprintf(filename,"%.100s/S6_F1v4.dat",param->Llens_dir);
  if((fp=fopen(filename,"r"))==NULL)
    {
      if(fplog)
	fprintf(fplog,"%s not found.\n",filename);
      return -6;
    }
  while(n_s6<S_DATA &&
        fscanf(fp,"%lf %lf",&sr6[n_s6], &sz6[n_s6])!=EOF)
    {
      sz6[n_s6]=param->LSz6-sz6[n_s6];
      n_s6++;
    }
  fclose(fp);

  /* Read S7 surface */
  sprintf(filename,"%.100s/S7_F1v4.dat",param->Llens_dir);
  if((fp=fopen(filename,"r"))==NULL)
    {
      if(fplog)
	fprintf(fplog,"%s not found.\n",filename);
      return -7;
    }
  while(n_s7<S_DATA &&
        fscanf(fp,"%lf %lf",&sr7[n_s7], &sz7[n_s7])!=EOF)
    {
      sz7[n_s7]=param->LSz7-sz7[n_s7];
      n_s7++;
    }
  fclose(fp);

#ifdef DEBUG
  if(fplog)
    {
      fprintf(fplog,"Done.\n");
      fprintf(fplog,"Reading Optical Data....   ");
    }
#endif

  /* Read PMMA data */
  sprintf(filename,"%.100s/Mitsubishi_000.dat",param->Llens_dir);
  if((fp=fopen(filename,"r"))==NULL)
    return -10;
  while(n_acryl<LAcryl_N &&
	fscanf(fp,"%lf %lf %lf",&ac_la[n_acryl],
	       &ac_n[n_acryl], &ac_k[n_acryl])!=EOF)
    {
      n_acryl++;
    }
  fclose(fp);

  /* Read CYOP data */
  sprintf(filename,"%.100s/CYTOP.dat",param->Llens_dir);
  if((fp=fopen(filename,"r"))==NULL)
    return -10;
  while(n_cytop<LCytop_N &&
	fscanf(fp,"%lf %lf %lf",&cy_la[n_cytop],
	       &cy_n[n_cytop], &cy_k[n_cytop])!=EOF)
    {
      n_cytop++;
    }
  fclose(fp);

  /* Read BG3 data */
  sprintf(filename,"%.100s/bg3.dat",param->Llens_dir);
  if((fp=fopen(filename,"r"))==NULL)
    return -11;
  while(n_bg3<LBg3_N &&
	fscanf(fp,"%lf %lf %lf",&bg3_la[n_bg3],
	       &bg3_n[n_bg3], &bg3_k[n_bg3])!=EOF)
    {
      n_bg3++;
    }
  fclose(fp);

#ifdef DEBUG
  if(fplog)
    fprintf(fplog,"Done.\n");
#endif
  return 0;
}

/*
 * ------------------------------------------------------------------------
 *  get refractive index of the acryl at wavelength lambda
 *    using cubic spline interpolation
 * ------------------------------------------------------------------------
 */
Double_t LGet_n_Acryl(Double_t lambda)
{
  static Int_t init=0;
  static Double_t *z;

  if(!init)
    {
      init=1;
      if((z=(Double_t *)malloc(sizeof(Double_t)*n_acryl))==NULL)
	return -1000.0;
      Lcsp_maketable(ac_la, ac_n, z, n_acryl);
    }

  return Lcspline(lambda, ac_la, ac_n, z, n_acryl);
}

/*
 * ------------------------------------------------------------------------
 *  get extinction coefficient of the acryl at wavelength lambda
 *    using cubic spline interpolation
 * ------------------------------------------------------------------------
 */
Double_t LGet_k_Acryl(Double_t lambda)
{
  static Int_t init=0;
  static Double_t *z;

  if(!init)
    {
      init=1;
      if((z=(Double_t *)malloc(sizeof(Double_t)*n_acryl))==NULL)
	return -1000.0;
      Lcsp_maketable(ac_la, ac_k, z, n_acryl);
    }

  return Lcspline(lambda, ac_la, ac_k, z, n_acryl);
}

/*
 * ------------------------------------------------------------------------
 *  get refractive index of CYTOP at wavelength lambda
 *    using cubic spline interpolation
 * ------------------------------------------------------------------------
 */
Double_t LGet_n_Cytop(Double_t lambda)
{
  static Int_t init=0;
  static Double_t *z;

  if(!init)
    {
      init=1;
      if((z=(Double_t *)malloc(sizeof(Double_t)*n_cytop))==NULL)
	return -1000.0;
      Lcsp_maketable(cy_la, cy_n, z, n_cytop);
    }

  return Lcspline(lambda, cy_la, cy_n, z, n_cytop);
}

/*
 * ------------------------------------------------------------------------
 *  get extinction coefficient of CYTOP at wavelength lambda
 *    using cubic spline interpolation
 * ------------------------------------------------------------------------
 */
Double_t LGet_k_Cytop(Double_t lambda)
{
  static Int_t init=0;
  static Double_t *z;

  if(!init)
    {
      init=1;
      if((z=(Double_t *)malloc(sizeof(Double_t)*n_cytop))==NULL)
	return -1000.0;
      Lcsp_maketable(cy_la, cy_k, z, n_cytop);
    }

  return Lcspline(lambda, cy_la, cy_k, z, n_cytop);
}

/*
 * ------------------------------------------------------------------------
 *  get the refracted ray
 *    if reflected it is thrown away.
 *
 *   in_ph[3]: incident photon vector (normalized)
 *   nnn[3]  : surface normal vector (direct to this side)
 *   n1,n2   : refractive indices of the media of this side and the other side
 * ------------------------------------------------------------------------
 */
Int_t Lrefract(Double_t in_phdir[3],Double_t nnn[3],Double_t n1,Double_t n2)
{
  Int_t i, TotalReflection=0;
  Double_t CosIN, CosOUT=1.0, ppp[3], norm, out_phdir[3];

  CosIN=Lvdot(in_phdir,nnn);
  for(i=0;i<3;i++)
    ppp[i]=(in_phdir[i]-CosIN*nnn[i])*n1/n2;

  norm=Lvdot(ppp,ppp);
#ifdef NO_REFLECTION
  if(norm>1.0)
    return(RT_NO_REFRACTED_LIGHT);  /* total reflection */
#else
  if(norm>1.0)
    TotalReflection=1;
  else
    {
#endif

      norm=sqrt(1.0-norm);
      for(i=0;i<3;i++)
	out_phdir[i]=ppp[i]+norm*nnn[i]*(CosIN>0.0?(1.0):(-1.0));

#ifndef NO_REFLECTION
      /* check reflection */
      CosOUT=Lvdot(out_phdir,nnn);
    }

  if(TotalReflection==1 || EsafRandom::Get()->Rndm() < Lref(n1,n2,CosIN,CosOUT))
    {
#ifdef DEBUG
      fprintf(stderr,"Reflection: (%.3f,%.3f,%.3f)=>",
             in_phdir[0],in_phdir[1],in_phdir[2]);
#endif
      for(i=0;i<3;i++)
	in_phdir[i]-=2.0*CosIN*nnn[i];
#ifdef DEBUG
      fprintf(stderr,"(%.3f,%.3f,%.3f)\n",
             in_phdir[0],in_phdir[1],in_phdir[2]);
#endif
      return(RT_REFLECTED);
    }
#endif

  /* in_phdir[]: direction of incident ray -> refracted ray */
#ifdef DEBUG
      fprintf(stderr,"Refraction: (%.3f,%.3f,%.3f)=>",
             in_phdir[0],in_phdir[1],in_phdir[2]);
#endif
  for(i=0;i<3;i++)
    in_phdir[i]=out_phdir[i];
#ifdef DEBUG
  fprintf(stderr,"(%.3f,%.3f,%.3f)\n",
	  in_phdir[0],in_phdir[1],in_phdir[2]);
#endif

  return(RT_REFRACTED);
}

/*
 * ------------------------------------------------------------------------
 *  calculate the differential coefficient at lens_r[i]
 *   Input:
 *      i                : the radius index of lens_r[i]
 *      lens_r[],lens_z[]: lens surface data sets of the radius and the height
 *      n_lens_elem      : number of data elements of lens_r[]
 *   Output:
 *     (return value)    : differential coefficeint at lens_r[i]
 * ------------------------------------------------------------------------
 */
Double_t Ldifferential_r(Int_t i, Double_t *lens_r,Double_t *lens_z, Int_t n_lens_elem)
{
  Double_t dr1,dr2;
  Double_t dz1,dz2;
  Int_t j1,j2;

  dz1=dz2=0.;
  j1=i+1;
  j2=i-1;

  if(i>=n_lens_elem-1)
    return 0.0;
  else if(i==n_lens_elem-2)
    {
      if(lens_r[i+1]-lens_r[i]<0.001)
        return 0.0;
      else
        return (lens_z[i+1]-lens_z[i])/(lens_r[i+1]-lens_r[i]);
    }

  dr1=lens_r[j1]-lens_r[i];
  if(dr1<0.001)
    {
      dz1+=lens_z[j1]-lens_z[i];
      j1++;
    }

  if(j2<0)
    {
      j2=-j2;
      dr2=lens_r[i]+lens_r[j2];
    }
  else
    {
      dr2=lens_r[i]-lens_r[j2];
    }
  if(dr2<0.001)
    {
      dz2+=lens_z[j2]-lens_z[i];
      j2--;
    }

  return (lens_z[j1]-dz1-lens_z[j2]+dz2)/(dr1+dr2);
}

/*
 *
 * check if there are any intersections with a cylinder
 *
 */
static Int_t Lcheck_intersect_cylinder(Double_t r1, Double_t z1, Double_t r2, Double_t z2,
				    Double_t phot[8], Double_t *dis)
{
  Double_t aa[3], xx[2], tmp, z;
  Int_t    nroot, i;

  aa[0] = phot[PX]*phot[PX] + phot[PY]*phot[PY];
  aa[1] = 2.0*(phot[X]*phot[PX] + phot[Y]*phot[PY]);
  aa[2] = (phot[X]*phot[X] + phot[Y]*phot[Y]) - r1*r1;

  nroot=LSolveQuadratic(aa,xx);
  if(nroot<2)
    return(RT_NO_INTERSECTION1);

  /* solution is the smaller positive value */
  /* swap */
  if(xx[0]>xx[1])
    {
      tmp=xx[0];
      xx[0]=xx[1];
      xx[1]=tmp;
    }
  if(xx[1]<0)
    return(RT_NO_INTERSECTION1);

  for(i=0;i<2;i++)
    {
      if(xx[i]>=0.0)
	{
	  z=phot[Z]+phot[PZ]*xx[i];
	  /* check if the solution is in the specified range */
	  if((z-z1)*(z-z2)<0)
	    {
	      *dis=xx[i];
	      return 0;
	    }
	}
    }

  *dis=(xx[0]>0 ? xx[0]:xx[1]);
  return(RT_NO_INTERSECTION2);
}

/*
 *
 * check if there are any intersections with a cone
 *
 */
static Int_t Lcheck_intersect_cone(Double_t r1, Double_t z1, Double_t r2, Double_t z2,
				Double_t phot[8], Double_t *dis)
{
  Double_t a00, a0, aa[3], xx[2], z0, dbuf1, tmp, x, y, r, zz[2];
  Int_t    nroot, i;

  a00 = (z2-z1)/(r2-r1);
  z0  = z1-a00*r1;
  a0  = a00*a00;

  dbuf1 = phot[Z]-z0;
  aa[0] = a0*phot[PX]*phot[PX] + a0*phot[PY]*phot[PY] - phot[PZ]*phot[PZ];
  aa[1] = 2.0*(a0*phot[PX]*phot[X] + a0*phot[PY]*phot[Y] - phot[PZ]*dbuf1);
  aa[2] = a0*phot[X]*phot[X] + a0*phot[Y]*phot[Y] - dbuf1*dbuf1;

  nroot=LSolveQuadratic(aa,xx);
  if(nroot<2)
    return(RT_NO_INTERSECTION1);

  /* solution is the smaller positive value */
  /* swap */
  if(xx[0]>xx[1])
    {
      tmp=xx[0];
      xx[0]=xx[1];
      xx[1]=tmp;
    }
  if(xx[1]<0.0)
    return(RT_NO_INTERSECTION1);

  /* We usually have a solution on a virtual cone with -a00 */
  /* check this condition */
  for(i=0;i<2;i++)
    if(xx[i]>=0.0)
      {
	x=phot[PX]*xx[i]+phot[X];
	y=phot[PY]*xx[i]+phot[Y];
	zz[i]=phot[PZ]*xx[i]+phot[Z];
	r=sqrt(x*x+y*y);
	if((r-r1)*(r-r2)<0.0)
	  {
	    *dis=xx[i];
	    return 0;
	  }
      }

  if(xx[0]<0.0)
    *dis=xx[1];
  else
    {
      /* the nearer one to the surface will be what we want... */
      if(fabs(zz[0]-z1)<fabs(zz[1]-z1))
	*dis=xx[0];
      else
	*dis=xx[1];
    }

  return(RT_NO_INTERSECTION2);
}

/*
 * ------------------------------------------------------------------------
 *
 *  ID:            Lens Surface ID
 *  phot[8]: input and output photon data
 *    phot[0,1,2]: x,y,z position in mm
 *    phot[3,4,5]: l,m,n direction
 *    phot[6]    : wavelength in nm
 *    phot[7]    : time in ns
 *  offset_z     : the offset z-position of sphare center approximating
 *                 lens surface
 *  radius       : the radius of sphare approximated lens surface
 *  lens_r[],lens_z[]: lens surface data sets of the radius and the height
 *  n_lens_elem  : number of data elements of lens_r[]
 *  medium1_n, medium2_n: refractive indices of the media in this side
 *                         and the other side
 * ------------------------------------------------------------------------
 */
#ifdef CHECK_TIMING
void CreateHisto(){

    cout<<"histo status"<<endl;
    thehisto=new TH1D("st","st",500,18.,4.);
}
void DeleteHisto(){

    TFile* f = new TFile("output/histo_st.root","RECREATE");
    thehisto->Write("st");
    delete f;
}

#endif

Int_t Llens_interaction(Int_t ID, Double_t phot[8],
		     Double_t Zc, Double_t radius, Double_t lens_r_max,
		     Double_t *lens_r, Double_t *lens_z, Int_t n_lens_elem,
		     Double_t medium1_n, Double_t medium2_n, Double_t *dis)
{




  Int_t r_idx,iter,flag,Found,nroot,dir,status;
//  Int_t idx_dir;
//  Double_t tmp_dis, tmp_x[3];
  Double_t dbuf1, aa[3], xx[2], x3,y3,z3,r3, r1,r2, z1,z2;
  Double_t nnn[3];
  Double_t a00;
  Double_t tmp, v[3];
#ifdef DEBUG
  Double_t approx_x[3];
#endif

  if(phot[PZ]==0){


    return(RT_HIT_SIDE_WALL);
  }

  if(ID==4 || ID==5) /* plain surface */
    {
      dbuf1 = (Zc - phot[Z])/phot[PZ];
      x3=phot[PX]*dbuf1+phot[X];
      y3=phot[PY]*dbuf1+phot[Y];
      z3=phot[PZ]*dbuf1+phot[Z];
      r3=sqrt(x3*x3+y3*y3);
      if(r3>lens_r_max){

        return(RT_HIT_SIDE_WALL);
      }

      *dis=dbuf1;
    }
  else
    {
  /* as an initial guess, intersection with an approximated sphere
     is calculated */

      dbuf1 = phot[Z]-Zc;
      aa[0] = phot[PX]*phot[PX] + phot[PY]*phot[PY] + phot[PZ]*phot[PZ];
      aa[1] = 2.0*(phot[PX]*phot[X] + phot[PY]*phot[Y] + phot[PZ]*dbuf1);
      aa[2] = phot[X]*phot[X] + phot[Y]*phot[Y] + dbuf1*dbuf1-radius*radius;

      nroot=LSolveQuadratic(aa,xx);
      if(nroot<2){

        return(RT_NO_INTERSECTION1);
      }

      /* swap */
      if(xx[0]>xx[1])
	{
	  tmp=xx[0];
	  xx[0]=xx[1];
	  xx[1]=tmp;
	}

      /* We need a positive, smaller solution. */
      if(xx[0]>0.0)
	dbuf1=xx[0];
      else
	{
	  if(xx[1]>0.0)
	    dbuf1=xx[1];
	  else{

	    return(RT_NO_INTERSECTION1);
      }
	}

      x3=phot[PX]*dbuf1+phot[X];
      y3=phot[PY]*dbuf1+phot[Y];
      z3=phot[PZ]*dbuf1+phot[Z];
      /* distance to the optical axis at the initially guessed intersection */
      r3=sqrt(x3*x3+y3*y3);

      if(r3>lens_r_max && xx[0]>-100.0) /* -100 is not optimized */
	{
	  /* The other solution may be acceptable just for an initial guess. */
	  dbuf1=xx[0];
	  x3=phot[PX]*dbuf1+phot[X];
	  y3=phot[PY]*dbuf1+phot[Y];
	  z3=phot[PZ]*dbuf1+phot[Z];

	  r3=sqrt(x3*x3+y3*y3);
	  if(r3>lens_r_max){

	    return(RT_HIT_SIDE_WALL);
      }
	}
      *dis=dbuf1;

#ifdef DEBUG
      approx_x[0]=x3;approx_x[1]=y3;approx_x[2]=z3;
#endif
    }

  if(ID==5) /* not Fresnel */
    {
      nnn[0]=nnn[1]=0.;
      nnn[2]=1.0;
    }
  else
    {
      /* search the index near r3 */
      r_idx=Lrt_bsearch(lens_r,r3,0,n_lens_elem-1);
      if(r_idx>n_lens_elem-2) r_idx=n_lens_elem-2;

      /* search intersection point with real fresnel lens */
      iter=0; /* number of iterations */
      do{
	Found=0;

	r1=lens_r[r_idx];
	z1=lens_z[r_idx];
	r2=lens_r[r_idx+1];
	z2=lens_z[r_idx+1];

	if(r1==r2)
	  {
	    /* intersection with a part of the cylinder */
	    status=Lcheck_intersect_cylinder(r1, z1, r2, z2, phot, dis);

	    x3=phot[PX]*(*dis)+phot[X];
	    y3=phot[PY]*(*dis)+phot[Y];
	    z3=phot[PZ]*(*dis)+phot[Z];

	    if(!status)     /* intersection found */
	      {
		Found=1;
#ifdef DEBUG
		fprintf(stderr,"Hit Backcut\n");
		/* return (RT_ABSORBED);*/
#endif
		/* normal vector */
		if(z2>z1)
		  dir=1;
		else
		  dir=-1;
		nnn[X]=dir*x3/r1;
		nnn[Y]=dir*y3/r1;
		nnn[Z]=0.0;
	      }
	    else /* intersection with cylinder not found */
	      {
		if(status==RT_NO_INTERSECTION2 && fabs(z3-z1)<100.0)
		  {
		    v[X]=x3;
		    v[Y]=y3;
		    v[Z]=0.0;

		    if((z3-z1)*Lvdot(v,&phot[PX])*phot[PZ]>0)
		      r_idx--;
		    else
		      r_idx++;
		  }
		else
		  {
		    *dis=(z1-phot[Z])/phot[PZ];
		    z3=z1;
		    x3=phot[PX]*(*dis)+phot[X];
		    y3=phot[PY]*(*dis)+phot[Y];
		    r3=sqrt(x3*x3+y3*y3);
		    r_idx=Lrt_bsearch(lens_r,r3,0,n_lens_elem-1);
		  }
	      }
#ifdef DEBUG
	    fprintf(stderr,"ridx0(%d): %d (%.1f: %.3f %.3f %.3f (%.1f %.1f %.1f)(%.3f %.3f %.3f))\n",
		    status,r_idx,z1,r1,r2,r3,phot[X],phot[Y],phot[Z],phot[PX],phot[PY],phot[PZ]);
#endif
	  }
	else  /* r1 != r2 */
	  {
	    /* intersection with a part of the cone */
	    status=Lcheck_intersect_cone(r1, z1, r2, z2, phot, dis);

	    x3=phot[PX]*(*dis)+phot[X];
	    y3=phot[PY]*(*dis)+phot[Y];
	    z3=phot[PZ]*(*dis)+phot[Z];

	    r3=sqrt(x3*x3+y3*y3);

	    if(!status)   /* intersection found */
	      {
		Found=1;
		/* normal vector */
		if(r_idx>n_lens_elem-2)
		  r_idx=n_lens_elem-2;
		a00=(Ldifferential_r(r_idx,lens_r,lens_z,n_lens_elem)
		     *(r3-lens_r[r_idx])-
		     Ldifferential_r(r_idx+1,lens_r,lens_z,n_lens_elem)
		     *(r3-lens_r[r_idx+1]))/(lens_r[r_idx+1]-lens_r[r_idx]);
		nnn[X]=a00*x3;
		nnn[Y]=a00*y3;
		nnn[Z]=-r3;
		dbuf1=TMath::Sqrt(Lvdot(nnn,nnn));
		nnn[X]/=dbuf1;
		nnn[Y]/=dbuf1;
		nnn[Z]/=dbuf1;
	      }
	    else /* not found */
	      {
		if(status==RT_NO_INTERSECTION2 &&
		   fabs(z3-z1)<10. && fabs(r3-r1)<5.)
		  {
		    if(r3<r1)
		      r_idx--;
		    if(r3>r2)
		      r_idx++;
		  }
		else
		  {
		    *dis=(z1-phot[Z])/phot[PZ];
		    z3=z1;
		    x3=phot[PX]*(*dis)+phot[X];
		    y3=phot[PY]*(*dis)+phot[Y];
		    r3=sqrt(x3*x3+y3*y3);
		    r_idx=Lrt_bsearch(lens_r,r3,0,n_lens_elem-1);
		  }
	      }
#ifdef DEBUG
	    fprintf(stderr,"ridx1(%d): %d (%.1f: %.3f %.3f %.3f (%.1f %.1f %.1f)(%.3f %.3f %.3f))\n",
		    status,r_idx,z1,r1,r2,r3,phot[X],phot[Y],phot[Z],phot[PX],phot[PY],phot[PZ]);
#endif
	  }

	if(r_idx>=S_DATA){

	  return(RT_OUT_OF_SURFACE2);
    }
	if(r_idx<0 && r3>lens_r_max && fabs(z3-z1)<10.0){


	  return(RT_OUT_OF_SURFACE2);
    }
	if(r_idx<0)
	  {
	    r_idx=1;
	    /* return(RT_OUT_OF_SURFACE1); */
	  }

	/* iteration doesn't converge */
	if(iter>RT_SEARCH_ITER_MAX)
	  {
#ifdef DEBUG
	    fprintf(stderr,"app.(%.1f,%.1f,%.1f)=>end(%.1f,%.1f,%.1f)\n",
		    approx_x[0],approx_x[1],approx_x[2],x3,y3,z3);
#endif



	    return(RT_EXCEED_ITER_MAX);
	  }

	iter++;

      } while(!Found);

#if 0
      /*
       * check no intersections at a nearer point
       */
      v[X]=x3;
      v[Y]=y3;
      v[Z]=0.0;
      if(vdot(v,&phot[PX])<0)
	idx_dir=1;
      else
	idx_dir=-1;

      do {
	r_idx+=idx_dir;

	r1=lens_r[r_idx];
	z1=lens_z[r_idx];
	r2=lens_r[r_idx+1];
	z2=lens_z[r_idx+1];

	if(r1==r2)
	  status=check_intersect_cylinder(r1, z1, r2, z2, phot, &tmp_dis);
	else
	  status=check_intersect_cone(r1, z1, r2, z2, phot, &tmp_dis);

	if(!status)  /* intersection found */
	  {
#ifdef DEBUG
	    if(r1==r2)
	      {
		fprintf(stderr,"Hit Backcut2\n");
		/*return (RT_ABSORBED);*/
	      }
#endif
	    tmp_x[X]=phot[PX]*tmp_dis+phot[X];
	    tmp_x[Y]=phot[PY]*tmp_dis+phot[Y];
	    tmp_x[Z]=phot[PZ]*tmp_dis+phot[Z];
	    r3=sqrt(tmp_x[X]*tmp_x[X]+tmp_x[Y]*tmp_x[Y]);

	    if(r1==r2)
	      {
		if(z2>z1)
		  dir=1;
		else
		  dir=-1;
		nnn[X]=dir*tmp_x[X]/r3;
		nnn[Y]=dir*tmp_x[Y]/r3;
		nnn[Z]=0.0;
	      }
	    else
	      {
		/* normal vector */
		if(r_idx>n_lens_elem-2)
		  r_idx=n_lens_elem-2;
		a00=(differential_r(r_idx,lens_r,lens_z,n_lens_elem)
		     *(r3-lens_r[r_idx])-
		     differential_r(r_idx+1,lens_r,lens_z,n_lens_elem)
		     *(r3-lens_r[r_idx+1]))/(lens_r[r_idx+1]-lens_r[r_idx]);
		nnn[X]=a00*tmp_x[X];
		nnn[Y]=a00*tmp_x[Y];
		nnn[Z]=-r3;
		dbuf1=sqrt(vdot(nnn,nnn));
		nnn[X]/=dbuf1;
		nnn[Y]/=dbuf1;
		nnn[Z]/=dbuf1;
	      }

	    if(vdot(nnn,&phot[PX])>=0) /* wrong intersection; back side */
	      status=1;
	    else                      /* update intersection point */
	      {
		x3=tmp_x[X];
		y3=tmp_x[Y];
		z3=tmp_x[Z];
		*dis=tmp_dis;
	      }
	  }
      }while(!status);
#endif
    }

  /* refraction */
  if(phot[PZ]<0.0)
    {
      tmp=medium2_n;
      medium2_n=medium1_n;
      medium1_n=tmp;
    }
  flag=Lrefract(&phot[PX], nnn, medium1_n, medium2_n);
#if 0
  if(flag==RT_REFLECTED)
    return(RT_REFLECTION_DISCARDED); /* reflected photons are thrown away */
#endif

  /* distance between injection point and the Surface */
  phot[T]+=(*dis)* medium1_n/EConst::Clight();

  phot[X]=x3;
  phot[Y]=y3;
  phot[Z]=z3;

  return flag;
}

static Int_t LCheckIris(tel_param *param, Double_t in_ph[6], Int_t flag)
{
  Double_t x3, y3, z3=0.0, r3=0.0, dist, radius;

  if(flag==SURFACE3)
    {
      z3=param->LZc3;
      r3=param->LRc3;
    }
  else if(flag==SURFACED)
    {
      z3=param->LZcD;
      r3=param->LRcD;
    }

  dist=(z3 - in_ph[Z])/in_ph[PZ];
  if(dist>=0.0)
    {
      x3=in_ph[PX]*dist+in_ph[X];
      y3=in_ph[PY]*dist+in_ph[Y];

      in_ph[X]=x3;
      in_ph[Y]=y3;
      in_ph[Z]=z3;
      in_ph[T]+=dist/EConst::Clight();

      radius=TMath::Sqrt(x3*x3+y3*y3);
      if(radius > param->Lr_wall)
	return RT_HIT_SIDE_WALL;
      else if(radius > r3)
	{
	  if(flag==SURFACE3)
	    return RT_OUT_OF_IRIS;
	  else if(flag==SURFACED)
	    return RT_OUT_OF_DOES;
	}
    }

    return 0;
}

/* -------------------------------------------------------------------------
 * int trace_lens(tel_param *param, double ph_in[8], double ph_out[8],
 *            FILE *fp)
 *
 *   --- trace photons in the EUSO lens system (the first lens to the last one)
 *
 * Input:
 *        param        tel_param* to store the optics parameters
 *                       (see also in tracemain_optF1v2.h)
 *        ph_in[]      info on injected photon
 *        ph_out[]     info on output photon
 *        ph_xx[0,1,2] x,y,z position in mm
 *        ph_xx[3,4,5] x,y,z direction (unit vector)
 *        ph_xx[6]     wavelength in nm
 *        ph_xx[7]     time in ns
 *        fp           FILE* for (error) message logging
 *
 * Output:
 *  return value
 *        0             normal end
 *       <0             error
 *                      (see trace_optF1v2.h for more detail
 *                        from RT_NO_SPINTERSECTION to RT_ESCAPE_FROM_ENTRANCE)
 * -------------------------------------------------------------------------
 */
Int_t Ltrace_lens(tel_param *param, Double_t ph_in[], Double_t ph_out[], FILE *fplog)
{


  Int_t    status=0, n_int, WithinOptics;
  Int_t    PrevPoint, CurPoint, NextPoint;
  Double_t nn_cytop,kk_cytop,nn_pmma,kk_pmma;
  Double_t AbsorbCoeff, TransProb;
  Double_t dis;
//   Int_t    s3=0;

  ph_out[L]=ph_in[L];


  /* Opt. index  for this lambda */
  nn_cytop=LGet_n_Cytop(ph_in[L]);
  kk_cytop=LGet_k_Cytop(ph_in[L]);
  nn_pmma=LGet_n_Acryl(ph_in[L]);
  kk_pmma=LGet_k_Acryl(ph_in[L]);

  if(ph_in[Z]>=param->LSz7 && ph_in[PZ]<0.0)
    {
      PrevPoint=SURFACE_FS;
      CurPoint=SURFACE_FS;
      NextPoint=SURFACE7;
    }
  else
    {
      PrevPoint=SURFACE0;
      CurPoint=SURFACE0;
      NextPoint=SURFACE1;
    }
  n_int=0;
  WithinOptics=1;

  while(n_int++ < RT_INTERACTION_MAX && WithinOptics)
    {
      PrevPoint=CurPoint;
      CurPoint=NextPoint;

      switch(NextPoint)
	{
	case SURFACE0:
      break;
	case SURFACE_FS:
	  WithinOptics=0;
	  break;

	case SURFACE1:
	  status=Llens_interaction(1, ph_in, param->LZc1, param->LRc1,
				  param->Lr_lens, sr1, sz1, n_s1,
				  LREFRACTIVE_IDX_MEDIA,nn_cytop,&dis);
      //printf("1 %f %f %f stat=%i\n",ph_in[X],ph_in[Y],ph_in[Z], status);
#ifdef CHECK_TIMING
      thehisto->Fill(status,1);
#endif
#ifdef DEBUG
	  printf("1 %f %f %f\n",ph_in[X],ph_in[Y],ph_in[Z]);
#endif
        if(status==RT_REFLECTED)
            {
              NextPoint=PrevPoint;
              status=0;
            }
          else if(PrevPoint==SURFACE0)
            NextPoint=SURFACE2;
          else
            NextPoint=SURFACE0;
	  break;

	case SURFACE2:
	  status=Llens_interaction(2, ph_in, param->LZc2, param->LRc2,
				  param->Lr_lens, sr2, sz2, n_s2,
				  nn_cytop,LREFRACTIVE_IDX_MEDIA, &dis);
      #ifdef CHECK_TIMING
      thehisto->Fill(status,1);
#endif
     // printf("2 %f %f %f stat=%i\n",ph_in[X],ph_in[Y],ph_in[Z], status);
#ifdef DEBUG
	  printf("2 %f %f %f\n",ph_in[X],ph_in[Y],ph_in[Z]);
#endif
          if(status==RT_REFLECTED)
            {
              NextPoint=PrevPoint;
              status=0;
            }
          else if(PrevPoint==SURFACE1)
            NextPoint=SURFACE3;
          else
            NextPoint=SURFACE1;
	  break;

	case SURFACE3:
	  if((status=LCheckIris(param,ph_in,SURFACE3))<0){

	    return (status+SURFACE3);
      }
	  if(ph_in[PZ]>0.0)
	    NextPoint=SURFACE4;
	  else
	    NextPoint=SURFACE2;
	  break;

	case SURFACE4:
	  status=Llens_interaction(4, ph_in, param->LZc4, param->LRc4,
				  param->Lr_lens, sr4, sz4, n_s4,
				  LREFRACTIVE_IDX_MEDIA, nn_pmma, &dis);
      #ifdef CHECK_TIMING
       thehisto->Fill(status,1);
#endif
      //printf("4 %f %f %f stat=%i\n",ph_in[X],ph_in[Y],ph_in[Z], status);
          if(status==RT_REFLECTED)
            {
              NextPoint=PrevPoint;
              status=0;
            }
	  else if(ph_in[PZ]>0.0)
	    NextPoint=SURFACED;
	  else
	    NextPoint=SURFACE3;
	  break;

	case SURFACED:
#ifndef NO_DIFFRACTIVE
	  if(ph_in[PZ]<0)
	    Ldiffract(1, &ph_in[PX], ph_in[X], ph_in[Y], ph_in[Z],
		     param->LZcD, 1.0003, 1.0003, ph_in[L]*1e-6,
		     param->Llambda0*1e-6, 1);
#endif
	  status=Llens_interaction(5, ph_in, param->LZcD, 0.0,
				  param->LRcD, NULL, NULL, 0,
				  nn_pmma, LREFRACTIVE_IDX_MEDIA, &dis);
      #ifdef CHECK_TIMING
       thehisto->Fill(status,1);
#endif
      //printf("d %f %f %f stat=%i\n",ph_in[X],ph_in[Y],ph_in[Z],status);
          if(status==RT_REFLECTED)
            {
              NextPoint=PrevPoint;
              status=0;
            }
	  else
	    {
#ifndef NO_DIFFRACTIVE
	      if(ph_in[PZ]>0)
		Ldiffract(1, &ph_in[PX], ph_in[X], ph_in[Y], ph_in[Z],
			 param->LZcD, 1.0003, 1.0003, ph_in[L]*1e-6,
			 param->Llambda0*1e-6, 1);
#endif
	    }
	  if(ph_in[PZ]>0.0)
	    NextPoint=SURFACE6;
	  else
	    NextPoint=SURFACE4;
	  break;

	case SURFACE6:
	  status=Llens_interaction(6, ph_in, param->LZc6, param->LRc6,
				  param->Lr_lens, sr6, sz6, n_s6,
				  LREFRACTIVE_IDX_MEDIA, nn_cytop, &dis);

      #ifdef CHECK_TIMING
       thehisto->Fill(status,1);
#endif
      //printf("6 %f %f %f\n",ph_in[X],ph_in[Y],ph_in[Z]);
#ifdef DEBUG
	  printf("6 %f %f %f\n",ph_in[X],ph_in[Y],ph_in[Z]);
#endif
          if(status==RT_REFLECTED)
            {
              NextPoint=PrevPoint;
              status=0;
            }
          else if(PrevPoint==SURFACED)
            NextPoint=SURFACE7;
          else
            NextPoint=SURFACED;
	  break;

	case SURFACE7:
	  status=Llens_interaction(7, ph_in, param->LZc7, param->LRc7,
				  param->Lr_lens, sr7, sz7, n_s7,
				  nn_cytop, LREFRACTIVE_IDX_MEDIA, &dis);
      #ifdef CHECK_TIMING
       thehisto->Fill(status,1);
#endif
      //printf("7 %f %f %f\n",ph_in[X],ph_in[Y],ph_in[Z]);
#ifdef DEBUG
	  printf("7 %f %f %f\n",ph_in[X],ph_in[Y],ph_in[Z]);
#endif
          if(status==RT_REFLECTED)
            {
              NextPoint=PrevPoint;
              status=0;
            }
          else if(PrevPoint==SURFACE6)
            NextPoint=SURFACE_FS;
          else
            NextPoint=SURFACE6;
	  break;
	}
#ifdef DEBUG
      fprintf(stderr,"%d %.1f %.1f %.1f  %.3f %.3f %.3f\n",
	      -CurPoint,
	      ph_in[X],ph_in[Y],ph_in[Z],ph_in[PX],ph_in[PY],ph_in[PZ]);
#endif
      if(status<0){

	return (status+CurPoint);
      }

#ifndef NO_ABSORPTION
      /* absorption between S1 and S2 or between S6 and S7 */
      if((PrevPoint==SURFACE1 && CurPoint==SURFACE2) ||
	 (PrevPoint==SURFACE2 && CurPoint==SURFACE1) ||
	 (PrevPoint==SURFACE6 && CurPoint==SURFACE7) ||
	 (PrevPoint==SURFACE7 && CurPoint==SURFACE6))
	{
	  AbsorbCoeff=kk_cytop*TMath::Pi()*4.0/(ph_in[L]*1.e-6);
	  TransProb=exp(-AbsorbCoeff*dis);

	  if(EsafRandom::Get()->Rndm()>TransProb){
#ifdef CHECK_TIMING
          thehisto->Fill(-11,1);
#endif
	    return(RT_ABSORBED+CurPoint);
      }
	}
      if((PrevPoint==SURFACE4 && CurPoint==SURFACED) ||
	 (PrevPoint==SURFACED && CurPoint==SURFACE4))
      {
	AbsorbCoeff=kk_pmma*TMath::Pi()*4.0/(ph_in[L]*1.e-6);
	TransProb=exp(-AbsorbCoeff*dis);

	if(EsafRandom::Get()->Rndm()>TransProb){
        #ifdef CHECK_TIMING
        thehisto->Fill(-11,1);
#endif
	  return(RT_ABSORBED+CurPoint);
    }
      }

#endif
    }

  if(n_int>=RT_INTERACTION_MAX){
      #ifdef CHECK_TIMING
    thehisto->Fill(-13,1);
#endif
    return (RT_STRAY_LIGHT);
  }

  ph_out[X]  = ph_in[X];
  ph_out[Y]  = ph_in[Y];
  ph_out[Z]  = ph_in[Z];
  ph_out[PX] = ph_in[PX];
  ph_out[PY] = ph_in[PY];
  ph_out[PZ] = ph_in[PZ];
  ph_out[T]  = ph_in[T];

  return 0;
}
