/* -------------------------------------------------------------------------
 *   tracemain_optF1v4.c
 *
 *   ---  main program for raytracing
 *
 * Copyright (c) 2000-2007 N.Sakaki, Y.Takizawa, Y.Kawasaki
 * All rights reserved.
 * $Id: Ltracemain_optF1v4.cc 2836 2009-07-16 08:22:02Z biktem $
 * $Id: Ltracemain_optF1v4.cc,v 1.0 2007/11/13 Dmitry V.Naumov adapted for ESAF
 * -------------------------------------------------------------------------
 */
#include <stdio.h>
#include <stdlib.h>
#include "TMath.h"
#include "EConst.hh"
#include "Ltrace_optF1v4.hh"
#define X      0
#define Y      1
#define Z      2
#define PX     3
#define PY     4
#define PZ     5
#define L      6
#define T      7

Int_t Lread_tel_param(char *filename, tel_param *param, FILE *fplog);
Int_t Lprint_param(tel_param *p);
Int_t Ltrace_main(tel_param *param,
               Double_t ph_in[8], Double_t ph_out[8], FILE *fplog);
Double_t LFocalSurface(Double_t r, tel_param *p);

/* -------------------------------------------------------------------------
 * int trace_main(tel_param *param, double ph_in[8], double ph_out[8],
 *            FILE *fp)
 *
 *   --- trace photons in the EUSO optics (Entrance to FS)
 *
 * [input value]
 * param  : tel_param* to store the optics parameters
 *            (see also in tracemain_optF1v2.h)
 * ph_in[]: info on injected photon
 * ph_out[]: info on output photon
 * ph_xx[0,1,2] : x,y,z position in mm
 * ph_xx[3,4,5] : x,y,z direction (unit vector)
 * ph_xx[6] : wavelength in nm
 * ph_xx[7] : time in ns
 * fp       : FILE* for (error) message logging
 *
 * [return value]
 * 0: normal end
 * <0: error (see trace_optF1v2.h for more detail)
 * -------------------------------------------------------------------------
 */
int Ltrace_main(tel_param *param,
	       double ph_in[8], double ph_out[8], FILE *fplog)
{
  static double *z0;
  static int N_FOCAL_SURFACE;
  static int fs_init=0;
  int j;
  int flg, fsloop;
  double r,r2,dist,xx,yy,zz;

  if(!fs_init)
    {
      fs_init=1;
      N_FOCAL_SURFACE=(int)(param->Lr_fs/LFS_DEF_STEP);
      z0=(double *)malloc(sizeof(double)*N_FOCAL_SURFACE);
      if(!z0)
	{
	  fprintf(fplog,"Cannot allocate memory for array for focalsurface\n");
	  exit(-10);
	}
      for(j=0;j<N_FOCAL_SURFACE;j++)
	{
	  r=(double)j*LFS_DEF_STEP;
	  z0[j]=LFocalSurface(r,param);
	}
    }

  /* the contents of ph_in[] will be overwritten in trace_lens() */
  flg=Ltrace_lens(param,ph_in,ph_out,fplog);
  if(flg!=0)
    {
#ifdef DEBUG
      if(fplog)
	fprintf(fplog,"flg=%d in trace_main\n",flg);
#endif
      return flg;
    }
  else if(ph_out[PZ]<=0.0)
    {
      return (RT_ESCAPE_FROM_ENTRANCE+SURFACE1);
    }
  else /* ph_out[PZ]>0.0 */
    {
      j=0;
      zz=z0[j];
      r=1.0e30;
      fsloop=0;
      do {
	fsloop++;
	r2=r;

	dist=(zz-ph_out[Z])/ph_out[PZ];
	xx=ph_out[PX]*dist+ph_out[X];
	yy=ph_out[PY]*dist+ph_out[Y];
	r=TMath::Sqrt(xx*xx+yy*yy);         /* distance to optical axis */

	j=(Int_t)(r/LFS_DEF_STEP);
	if(j>=N_FOCAL_SURFACE)
	  j=N_FOCAL_SURFACE-1; /* should be error? */

	if(fabs(z0[j]-zz)<LFS_DEF_STEP)
	  {
	    if(r > param->Lr_fs)
	      flg=1;       /* out of focal surface */
	    zz=LFocalSurface(r,param);
	  }
	else
	  zz=z0[j];
      }while(fabs(r-r2)>LFS_TOLERANCE && !flg && fsloop<LFSLOOP_MAX);

      if(flg)
	{
#ifdef DEBUG
	  fprintf(stderr,"curpos(%.1f,%.1f,%.1f) vec(%.2f,%.2f,%.2f) "
		  "est pos on FS(%.1f,%.1f,%.1f) r=%.1f\n",
		  ph_out[X],ph_out[Y],ph_out[Z],
		  ph_out[PX],ph_out[PY],ph_out[PZ],
		  xx,yy,zz,r
		  );
#endif
	  return (SURFACE_FS+RT_OUT_OF_SURFACE1);
	}
      else if(fsloop>=LFSLOOP_MAX)
	{
	  return (SURFACE_FS+RT_STRAY_LIGHT);
	}
      else
	{
	  ph_out[X]=xx;
	  ph_out[Y]=yy;
	  ph_out[Z]=zz;
	  ph_out[T]+=dist/EConst::Clight();
	}
    }

  return 0;
}

/* -------------------------------------------------------------------------
 * int read_tel_param(char *filename, tel_param *param, FILE *fplog)
 *
 *  --- read optics parameters from a file
 *
 *  [input value]
 *  filename: the name of the file for optics parameters
 *  param   : pointer to tel_param structure (see also tracemain_optF1v2.h)
 *  fplog   : FILE* for (error) message logging
 *
 *  [return value]
 *  -1 : "(char *)filename" not found
 *   0 : normal end
 * -------------------------------------------------------------------------
 */
int Lread_tel_param(char *filename, tel_param *param, FILE *fplog)
{
  FILE *fr;
  char buf[256],item[256];
  double val;

  if(( fr=fopen(filename,"r") )==NULL)
    {
      if(fplog)
	fprintf(fplog,"Telescope parameter file %s cannot be opened.\n",filename);
      return -1;
    }
  while(fgets(buf,255,fr))
    {
      if(buf[0]!=0 || buf[0]!='#')
	{
	  sscanf(buf,"%255s",item);
	  switch(item[0])
	    {
	    case 'R':
	      sscanf(buf,"%*s %lf",&val);
	      switch(item[1])
		{
		case '_':
		  switch(item[2])
		    {
		    case 'L':
		      param->Lr_lens=val;
		      break;
		    case 'W':
		      param->Lr_wall=val;
		      break;
		    case 'F':
		      param->Lr_fs=val;
		      break;
		    }
		  break;
		case 'c':
		  switch(item[2])
		    {
		    case '1':
		      param->LRc1=val;
		      break;
		    case '2':
		      param->LRc2=val;
		      break;
		    case '3':
		      param->LRc3=val;
		      break;
		    case '6':
		      param->LRc6=val;
		      break;
		    case '7':
		      param->LRc7=val;
		      break;
		    case 'D':
		      param->LRcD=val;
		      break;
		    }
		  break;
		}
	      break;

	    case 'S':
	      sscanf(buf,"%*s %lf",&val);
	      if(item[1]=='z')
		{
		  switch(item[2])
		    {
		    case '0':
		      param->LSz0=val;
		      break;
		    case '1':
		      param->LSz1=val;
		      break;
		    case '2':
		      param->LSz2=val;
		      break;
		    case '3':
		      param->LZc3=val;
		      break;
		    case '4':
		      param->LSz4=val;
		      break;
		    case '6':
		      param->LSz6=val;
		      break;
		    case '7':
		      param->LSz7=val;
		      break;
		    case 'D':
		      param->LZcD=val;
		      break;
		    }
		}
	      break;

	    case 'F':
	      sscanf(buf,"%*s %lf",&val);
	      switch(item[2])
		{
		case '_':
		  if(item[3]=='C')
		    param->LFS_C=val;
		  else if(item[3]=='K')
		    param->LFS_K=val;
		  else if(item[3]=='O')
		    param->LFS_OFFSET=val;
		  break;

		case 'A':
		  param->LFSA=val;
		  break;
		case 'B':
		  param->LFSB=val;
		  break;
		case 'C':
		  param->LFSC=val;
		  break;
		case 'D':
		  param->LFSD=val;
		  break;
		}
	      break;

	    case 'l':
	      switch(item[1])
		{
		case 'e':
		  sscanf(buf,"%*s %255s",param->Llens_dir);
		  break;
		case 'a':
		  sscanf(buf,"%*s %lf",&param->Llambda0);
		  break;
		}
	      break;
	    }
	}
    }
  fclose(fr);

  param->LSz1 += param->LSz0;
  param->LSz2 += param->LSz0;
  param->LZc3 += param->LSz0;
  param->LSz4 += param->LSz0;
  param->LSz6 += param->LSz0;
  param->LSz7 += param->LSz0;
  param->LZcD += param->LSz0;

  param->LZc1 =  param->LRc1+param->LSz1;
  param->LZc2 =  param->LRc2+param->LSz2;
  param->LZc4 =             param->LSz4;
  param->LZc6 = -param->LRc6+param->LSz6;
  param->LZc7 = -param->LRc7+param->LSz7;

  param->LFS_OFFSET += param->LSz0;

  return 0;
}

/* -------------------------------------------------------------------------
 * int print_param(tel_param *p)
 *
 *  --- print optics parameters for debbuging purpose
 *
 *  [input value]
 *  p: pointer to tel_param structure
 *
 *  [return value]
 *  always 0
 *
 * -------------------------------------------------------------------------
 */
int Lprint_param(tel_param *p)
{
  printf("r_lens=%.3f ",p->Lr_lens);
  printf("r_wall=%.3f ",p->Lr_wall);
  printf("r_fs=%.3f\n",p->Lr_fs);
  printf("Rc1=%.5f ",p->LRc1);
  printf("Sz1=%.5f ",p->LSz1);
  printf("Zc1=%.5f\n",p->LZc1);

  printf("Rc2=%.5f ",p->LRc2);
  printf("Sz2=%.5f ",p->LSz2);
  printf("Zc2=%.5f\n",p->LZc2);

  printf("Sz4=%.5f ",p->LSz4);

  printf("Rc6=%.5f ",p->LRc6);
  printf("Sz6=%.5f ",p->LSz6);
  printf("Zc6=%.5f\n",p->LZc6);

  printf("Rc7=%.5f ",p->LRc7);
  printf("Sz7=%.5f ",p->LSz7);
  printf("Zc7=%.5f\n",p->LZc7);

  printf("Rc3=%.5f ",p->LRc3);
  printf("Sz1=%.5f ",p->LSz1);
  printf("Zc3=%.5f\n",p->LZc3);

  printf("dir=%s\n",p->Llens_dir);

  printf("FS_C=%g ",p->LFS_C);
  printf("FS_K=%.2f\n",p->LFS_K);
  printf("FSA=%g ",p->LFSA);
  printf("FSB=%g ",p->LFSB);
  printf("FSC=%g ",p->LFSC);
  printf("FSD=%g ",p->LFSD);
  printf("FS_offset=%.2f\n",p->LFS_OFFSET);
  return 0;
}

/* -------------------------------------------------------------------------
 * double FocalSurface(double r, tel_param *p)
 *
 *    --- calculate the Focal surface height as a function of the radius
 *        from the center
 *        As photons move toward z>0, FS_C should always be negative.
 *
 *  [input value]
 *  r: radius in mm
 *  p: pointer to tel_param structure (optics parameter)
 *
 *  [return value]
 *  relative height of FS to that at r=0 in mm
 *
 * -------------------------------------------------------------------------
 */
/* photons move to the direction z>0 */
double LFocalSurface(double r /* mm */, tel_param *p)  /* mm */
{
  double r2;

  r2=r*r;
  return r2*(p->LFS_C /(1.0+TMath::Sqrt(1.0-(1.0 + p->LFS_K)* p->LFS_C * p->LFS_C *r2))
	     +r2*(p->LFSA +r2*(p->LFSB+r2*(p->LFSC + r2* p->LFSD))))+ p->LFS_OFFSET;
}
