// $Id: LIdealFocalSurface.cc 2824 2009-03-16 14:36:44Z naumov $
// Author: Dmitry V.Naumov  2007/11/16

/*****************************************************************************
 * ESAF: Euso Simulation and Analysis Framework                              *
 *                                                                           *
 *  Id: MIdealFocalSurface                                                   *
 *  Package: Optics                                                          *
 *  Coordinator: Alessandro.Thea                                             *
 *                                                                           *
 *****************************************************************************/

//_____________________________________________________________________________
//
// MIdealFocalSurface
//
//   MIdealFocalSurface is the optimized focal surface for the MOpticalSystem
//   Its shape is described by the radial profile:
//
//                       r^2
//   z(r) =  --------------------------- + A*r^4 + B*r^6 + C*r^8 + D*r^10
//          (1+(sqrt( (1+k)*(r/rho)^2) ))
//
//   Config file parameters
//   ======================
//
//   fR [mm] : Maximum extension of the ideal focal surface from the optical
//   axis.
//
//   fPos [mm] : z coodinate of the tip of the surface in detector coordinate
//   system.
//
//   fRho, fK, fA, fB, fC, fD : focal surface parameters.
//
//   fPrec [mm] : precision required to claim that the photon has hit the FS.
//

#include "MIdealFocalSurface.hh"
#include "Mtracemain_optF1v4.hh"
#include <TVector3.h>
#include "Photon.hh"
#include "TMath.h"

ClassImp(MIdealFocalSurface)

const Int_t kMaxIter=50;

using namespace sou;
using namespace TMath;
Mtel_param  fMIdealFocalSurfaceParam;

//______________________________________________________________________________
MIdealFocalSurface::MIdealFocalSurface() {
    //
    // Constructor
    //
    FILE* fLogFile=0;
    if(Mread_tel_param("config/Optics/MOpticalSystem/telparm_F1v6.dat",&fMIdealFocalSurfaceParam,fLogFile)!=0)
       Msg(EsafMsg::Error) << "  I do not find telparm_F1v6.dat file! " << MsgDispatch;


    fR   = Conf()->GetNum("MIdealFocalSurface.fR")*mm;
    Double_t FS_Offset = fMIdealFocalSurfaceParam.MFS_OFFSET;
    fPos = TVector3(0,0,FS_Offset*mm);
    fDZdown = Zfs(0)-Zfs(fR)-1.*mm;
}

//______________________________________________________________________________
MIdealFocalSurface::~MIdealFocalSurface() {
    //
    // Destructor
    //

}

//______________________________________________________________________________
Bool_t MIdealFocalSurface::HitPosition(Photon *p) {
    //
    // Find whether the photons hits the ideal surface, and in such a case
    // calculated the final position
    //


    // reject photons going downward
    if ( p->dir.Z() < 0 )
        return kFALSE;

    TVector3 in_pos = p->pos;
    TVector3 ax_pos(0,0,-fDZdown), ax_dir(0,0,1);

    TVector3 bottomHit = in_pos+p->dir*((fPos[Z]-fDZdown-in_pos[Z])/p->dir[Z]);

    if ( bottomHit.Perp()>fR )
        return kFALSE;
    //the minimum distance between the photon and the Zfs axis
    TVector3 dummy=ax_dir.Cross(p->dir);
    //double min_dist=Abs((bottomHit-ax_pos).Dot(dummy))/dummy.Mag();

    //interaction point on top of the cylinder
    //limit inclination of the photon

    Double_t  dt = NextInt( bottomHit, p->dir);

    TVector3 topHit = bottomHit+p->dir*dt;
/*
#ifdef DEBUG
    cout << "p->dir " << p->dir << endl;
    cout << "topHit= " << topHit << " bottomHit=" << bottomHit << endl;
#endif
*/
    TVector3 step = (topHit-bottomHit)*.5;
    dummy=bottomHit+step;

    for(int i=0; i < kMaxIter ; i++) {
/*
#ifdef DEBUG
        cout << "i=" << i << " dummy[Z]=" << dummy[Z] << " Zfs(dummy.Perp())=" << Zfs(dummy.Perp()) << endl;
#endif
*/
        if (Abs(Zfs(dummy.Perp())-dummy[Z]) < fPrec) {
            break;
        }
        step*=.5;
        if (Zfs(dummy.Perp()) > dummy[Z]) dummy+=step;
        else dummy-=step;
    }

    if ( dummy.Perp() > fR )
        return kFALSE;

    p->posOnIfs=dummy;
    p->hitIfs=kTRUE;

    return kTRUE;
}


//______________________________________________________________________________
Double_t MIdealFocalSurface::Zfs(Double_t r) {
    //
    // Zfs returns the z value of the focal surface as a function of radius in
    // local coordinates
    //

    return MFocalSurface(r,&fMIdealFocalSurfaceParam);
}

//______________________________________________________________________________
Double_t MIdealFocalSurface::Profile(Double_t r) {
    //
    // Profile returns the z value of the focal surface as a function of radius
    // in detector coordinates
    //

    return Zfs(r)+fPos.Z();
}
