/* -------------------------------------------------------------------------
 *   tracemain_optF1v6.c
 *
 *   ---  main program for raytracing
 *
 * Copyright (c) 2000-2007 N.Sakaki, Y.Takizawa, Y.Kawasaki
 * All rights reserved.
 * $Id$
 * -------------------------------------------------------------------------
 */
#include <stdio.h>
#include <stdlib.h>
//#include <time.h>
#include <math.h>
#include <string.h>
#include "EConst.hh"
#include "EsafRandom.hh"
#include "Mtrace_optF1v4.hh"
#include "Mtracemain_optF1v4.hh"
#define X      0
#define Y      1
#define Z      2
#define PX     3
#define PY     4
#define PZ     5
#define L      6
#define T      7

/* -------------------------------------------------------------------------
 * int trace_main(tel_param *param, double ph_in[8], double ph_out[8],
 *            FILE *fp)
 *
 *   --- trace photons in the EUSO optics (Entrance to FS)
 *
 * [input value]
 * param  : tel_param* to store the optics parameters
 *            (see also in tracemain_optF1v2.h)
 * ph_in[]: info on injected photon
 * ph_out[]: info on output photon
 * ph_xx[0,1,2] : x,y,z position in mm
 * ph_xx[3,4,5] : x,y,z direction (unit vector)
 * ph_xx[6] : wavelength in nm
 * ph_xx[7] : time in ns
 * fp       : FILE* for (error) message logging
 *
 * [return value]
 * 0: normal end
 * <0: error (see trace_optF1v2.h for more detail)
 * -------------------------------------------------------------------------
 */
int Mtrace_main(Mtel_param *param,
           double ph_in[8], double ph_out[8], FILE *fplog)
{
  static double *z0;
  static int N_FOCAL_SURFACE;
  static int fs_init=0;
  int j;
  int flg, fsloop;
  double r,r2,dist,xx,yy,zz;

  if(!fs_init)
    {
      fs_init=1;
      N_FOCAL_SURFACE=(int)(param->Mr_fs/MFS_DEF_STEP);
      z0=(double *)malloc(sizeof(double)*N_FOCAL_SURFACE);
      if(!z0)
    {
      fprintf(fplog,"Cannot allocate memory for array for focalsurface\n");
      exit(-10);
    }
      for(j=0;j<N_FOCAL_SURFACE;j++)
    {
      r=(double)j*MFS_DEF_STEP;
      z0[j]=MFocalSurface(r,param);
    }
    }

  /* the contents of ph_in[] will be overwritten in trace_lens() */
  flg=trace_lens(param,ph_in,ph_out,fplog);
  if(flg!=0)
    {
#ifdef DEBUG
      if(fplog)
    fprintf(fplog,"flg=%d in trace_main\n",flg);
#endif
      return flg;
    }
  else if(ph_out[PZ]<=0.0)
    {
      return (RT_ESCAPE_FROM_ENTRANCE+SURFACE1);
    }
  else /* ph_out[PZ]>0.0 */
    {
      j=0;
      zz=z0[j];
      r=1.0e30;
      fsloop=0;
      do {
    fsloop++;
    r2=r;

    dist=(zz-ph_out[Z])/ph_out[PZ];
    xx=ph_out[PX]*dist+ph_out[X];
    yy=ph_out[PY]*dist+ph_out[Y];
    r=sqrt(xx*xx+yy*yy);         /* distance to optical axis */

    j=int(r/MFS_DEF_STEP);

    if(j>=N_FOCAL_SURFACE || j < 0)/*changed by francesco 15-01-2010*/
      j=N_FOCAL_SURFACE-1; /* should be error? */

    if(fabs(z0[j]-zz)<MFS_DEF_STEP)
      {
        if(r > param->Mr_fs)
          flg=1;       /* out of focal surface */
	else/*changed by francesco 15-01-2010*/
        zz=MFocalSurface(r,param);
      }
    else
      zz=z0[j];
      }while(fabs(r-r2)>MFS_TOLERANCE && !flg && fsloop<MFSLOOP_MAX);

      if(flg)
    {
#ifdef DEBUG
      fprintf(stderr,"curpos(%.1f,%.1f,%.1f) vec(%.2f,%.2f,%.2f) "
          "est pos on FS(%.1f,%.1f,%.1f) r=%.1f\n",
          ph_out[X],ph_out[Y],ph_out[Z],
          ph_out[PX],ph_out[PY],ph_out[PZ],
          xx,yy,zz,r
          );
#endif
      return (SURFACE_FS+RT_OUT_OF_SURFACE1);
    }
      else if(fsloop>=MFSLOOP_MAX)
    {
      return (SURFACE_FS+RT_STRAY_LIGHT);
    }
      else
    {
      ph_out[X]=xx;
      ph_out[Y]=yy;
      ph_out[Z]=zz;
      ph_out[T]+=dist/EConst::Clight();
          if(CheckRcut(param,ph_out,SURFACE_FS))
            return (SURFACE_FS+RT_OUT_OF_SURFACE1);
    }
    }

  return 0;
}

/* -------------------------------------------------------------------------
 * int read_tel_param(char *filename, tel_param *param, FILE *fplog)
 *
 *  --- read optics parameters from a file
 *
 *  [input value]
 *  filename: the name of the file for optics parameters
 *  param   : pointer to tel_param structure (see also tracemain_optF1v2.h)
 *  fplog   : FILE* for (error) message logging
 *
 *  [return value]
 *  -1 : "(char *)filename" not found
 *   0 : normal end
 * -------------------------------------------------------------------------
 */
int Mread_tel_param(const char *filename, Mtel_param *param, FILE *fplog)
{
  FILE *fr;
  char buf[256],item[256];
  double val;

  if(( fr=fopen(filename,"r") )==NULL)
    {
      if(fplog)
    fprintf(fplog,"Telescope parameter file %s cannot be opened.\n",filename);
      return -1;
    }
  while(fgets(buf,255,fr))
    {
      if(buf[0]!=0 || buf[0]!='#')
    {
      sscanf(buf,"%255s",item);
      switch(item[0])
        {
        case 'R':
          sscanf(buf,"%*s %lf",&val);
          switch(item[1])
        {
        case '_':
          switch(item[2])
            {
            case 'L':
              param->Mr_lens=val;
              break;
            case 'W':
              param->Mr_wall=val;
              break;
            case 'F':
              param->Mr_fs=val;
              break;
                    case 'c':
                      param->Mr_cut[item[5]-'0']=val;
                      break;
            }
          break;
        case 'c':
          switch(item[2])
            {
            case '1':
              param->MRc1=val;
              break;
            case '2':
              param->MRc2=val;
              break;
            case '3':
              param->MRc3=val;
              break;
            case '6':
              param->MRc6=val;
              break;
            case '7':
              param->MRc7=val;
              break;
            case 'D':
              param->MRcD=val;
              break;
            }
          break;
        }
          break;

        case 'S':
          sscanf(buf,"%*s %lf",&val);
          if(item[1]=='z')
        {
          switch(item[2])
            {
            case '0':
              param->MSz0=val;
              break;
            case '1':
              param->MSz1=val;
              break;
            case '2':
              param->MSz2=val;
              break;
            case '3':
              param->MZc3=val;
              break;
            case '4':
              param->MSz4=val;
              break;
            case '6':
              param->MSz6=val;
              break;
            case '7':
              param->MSz7=val;
              break;
            case 'D':
              param->MZcD=val;
              break;
            }
        }
          break;

        case 'F':
          sscanf(buf,"%*s %lf",&val);
          switch(item[2])
        {
        case '_':
          if(item[3]=='C')
            param->MFS_C=val;
          else if(item[3]=='K')
            param->MFS_K=val;
          else if(item[3]=='O')
            param->MFS_OFFSET=val;
          break;

        case 'A':
          param->MFSA=val;
          break;
        case 'B':
          param->MFSB=val;
          break;
        case 'C':
          param->MFSC=val;
          break;
        case 'D':
          param->MFSD=val;
          break;
        }
          break;

        case 'l':
          switch(item[1])
        {
        case 'e':
          sscanf(buf,"%*s %255s",param->Mlens_dir);
          break;
        case 'a':
          sscanf(buf,"%*s %lf",&param->Mlambda0);
          break;
        }
          break;
        }
    }
    }
  fclose(fr);

  param->MSz1 += param->MSz0;
  param->MSz2 += param->MSz0;
  param->MZc3 += param->MSz0;
  param->MSz4 += param->MSz0;
  param->MSz6 += param->MSz0;
  param->MSz7 += param->MSz0;
  param->MZcD += param->MSz0;

  param->MZc1 =  param->MRc1+param->MSz1;
  param->MZc2 =  param->MRc2+param->MSz2;
  param->MZc4 =             param->MSz4;
  param->MZc6 = -param->MRc6+param->MSz6;
  param->MZc7 = -param->MRc7+param->MSz7;

  param->MFS_OFFSET += param->MSz0;

  param->Mr_cut[0]=3000.0;
  param->Mr_cut[5]=param->Mr_cut[4];
  param->Mr_cut[2]=param->Mr_cut[1];
  param->Mr_cut[7]=param->Mr_cut[6];

  return 0;
}

/* -------------------------------------------------------------------------
 * int print_param(tel_param *p)
 *
 *  --- print optics parameters for debbuging purpose
 *
 *  [input value]
 *  p: pointer to tel_param structure
 *
 *  [return value]
 *  always 0
 *
 * -------------------------------------------------------------------------
 */
int Mprint_param(Mtel_param *p)
{
  int i;

  printf("Mr_lens=%.3f ",p->Mr_lens);
  printf("Mr_wall=%.3f ",p->Mr_wall);
  printf("Mr_fs=%.3f\n",p->Mr_fs);
  printf("MRc1=%.5f ",p->MRc1);
  printf("MSz1=%.5f ",p->MSz1);
  printf("MZc1=%.5f\n",p->MZc1);

  printf("MRc2=%.5f ",p->MRc2);
  printf("MSz2=%.5f ",p->MSz2);
  printf("MZc2=%.5f\n",p->MZc2);

  printf("MSz4=%.5f ",p->MSz4);

  printf("MRc6=%.5f ",p->MRc6);
  printf("MSz6=%.5f ",p->MSz6);
  printf("MZc6=%.5f\n",p->MZc6);

  printf("MRc7=%.5f ",p->MRc7);
  printf("MSz7=%.5f ",p->MSz7);
  printf("MZc7=%.5f\n",p->MZc7);

  printf("Rc3=%.5f ",p->MRc3);
  printf("MSz1=%.5f ",p->MSz1);
  printf("MZc3=%.5f\n",p->MZc3);

  printf("dir=%s\n",p->Mlens_dir);

  printf("FS_C=%g ",p->MFS_C);
  printf("FS_K=%.2f\n",p->MFS_K);
  printf("FSA=%g ",p->MFSA);
  printf("FSB=%g ",p->MFSB);
  printf("FSC=%g ",p->MFSC);
  printf("FSD=%g ",p->MFSD);
  printf("FS_offset=%.2f\n",p->MFS_OFFSET);
  for(i=1;i<9;i++)
    printf("Rcut%d=%.2f\n",i,p->Mr_cut[i]);
  return 0;
}

/* -------------------------------------------------------------------------
 * double FocalSurface(double r, tel_param *p)
 *
 *    --- calculate the Focal surface height as a function of the radius
 *        from the center
 *        As photons move toward z>0, FS_C should always be negative.
 *
 *  [input value]
 *  r: radius in mm
 *  p: pointer to tel_param structure (optics parameter)
 *
 *  [return value]
 *  relative height of FS to that at r=0 in mm
 *
 * -------------------------------------------------------------------------
 */
/* photons move to the direction z>0 */
double MFocalSurface(double r /* mm */, Mtel_param *p)  /* mm */
{
  double r2;

  r2=r*r;
  return r2*(p->MFS_C /(1.0+sqrt(1.0-(1.0 + p->MFS_K)* p->MFS_C * p->MFS_C *r2))
         +r2*(p->MFSA +r2*(p->MFSB+r2*(p->MFSC + r2* p->MFSD))))+ p->MFS_OFFSET;
}
