// $Id: NOpticalSystem.cc 2828 2009-04-01 14:29:16Z biktem $
// Author: Dmitry V.Naumov   2007/11/13

/*****************************************************************************
 * ESAF: Euso Simulation and Analysis Framework                              *
 *                                                                           *
 *  Id: NOpticalSystem                                                       *
 *  Package: Optics                                                          *
 *  Coordinator: Alessandro.Thea                                             *
 *                                                                           *
 *****************************************************************************/

//_____________________________________________________________________________

#include <TMath.h>
#include "EEvent.hh"

#include "Config.hh"
#include "EsafRandom.hh"

#include "NOpticalSystem.hh"

#include "Ntrace_lenses.hh"
#include "Nphoton.hh"
#include <fstream>
//#define DEBUG
using namespace sou;
using namespace NTraceLens;

ClassImp(NOpticalSystem)
//_____________________________________________________________________________
NOpticalSystem::NOpticalSystem() {
    //
    // Constructor
    //
    int gen = Conf()->GetNum("NOpticalSystem.EsafRandom");
    SetRunGen(gen);

    double eff = Conf()->GetNum("NOpticalSystem.fEffectiveEfficiencyFactor");
    SetEffFactor(eff);

    const char* lens_dir = Conf()->GetStr("NOpticalSystem.lens_dir").data();
    const char* tel_par = Conf()->GetStr("NOpticalSystem.tel_par").data();
    ReadTelParm(lens_dir,tel_par);

    fDZdown=1*mm;
    fDZup=fOpticalSystemParam->surface[7].Sz+1.*mm;//-fOpticalSystemParam->surface[0].Sz

    // take the size of the system from Ltrace
    fR=fOpticalSystemParam->r_wall;
    fPos=EVector(0,0,Conf()->GetNum("NOpticalSystem.fPos.Z")*mm);


    float s7(0.),s1(0.);
    char c[50];
    char* par = fOpticalSystemParam->name;
    fstream f1(Form("%sS1_%s.dat",lens_dir,par), fstream::in );
    f1.seekg(-15, ios_base::end);
    f1.read(c,15);

    sscanf(c, "%*s %f", &s1);


    fstream f2(Form("%sS7_%s.dat",lens_dir,par), fstream::in );
    f2.seekg(-15, ios_base::end);
    f2.read(c,15);

    sscanf(c, "%*s %f",&s7);
    fFirstLensDZ=(fabs(s1)+0.5)*mm;
    fSecondLensDZ=(fabs(s7)+0.5)*mm;

#ifdef DEBUG
    printf("fFirstLensDZ %s",c);
    printf("fSecondLensDZ %s",c);
    printf("fSecondLensDZ %f \n",fSecondLensDZ);
    printf("fFirstLensDZ %f \n",fFirstLensDZ);
    printf("fDZup %f \n",fDZup);
#endif


}

//_____________________________________________________________________________
NOpticalSystem::~NOpticalSystem() {
    //
    // Destructor
    //
}

//______________________________________________________________________________
Photon* NOpticalSystem::Transport(Photon *p) const {
    NPhoton ph_in;
    EVector dummy=p->pos-fPos;
    ph_in.x[0]=dummy[X]/mm;
    ph_in.x[1]=dummy[Y]/mm;
    ph_in.x[2]=dummy[Z]/mm;
    ph_in.px[0]=p->dir[X];
    ph_in.px[1]=p->dir[Y];
    ph_in.px[2]=p->dir[Z];
    ph_in.lambda=p->wl/nm;
    ph_in.t=p->time/ns;

    ph_in.history=0;

#ifdef DEBUG
    cout<<"NOpticalSystem"<<endl;
    cout<<"position = "<<p->pos<<endl;
    cout<<"direction = "<<p->dir<<endl;
    double theta=p->dir.Angle(EVector(0.,0.,1.));
    cout<<"theta = "<<theta/deg<<endl;
    cout<<"wl = "<<p->wl/nm<<endl;
    cout<<"time = "<<p->time/ns<<endl;
#endif /* DEBUG */
    int flag = trace_main(fOpticalSystemParam,&ph_in,fLogFile);

#ifdef DEBUG
#endif /* DEBUG */

    // FIXME: add to mtrace possibility to save latest photon information, for cases when it's absorbed
    if ( !flag ) {
        dummy[X]=ph_in.x[0]*mm;
        dummy[Y]=ph_in.x[1]*mm;
        dummy[Z]=ph_in.x[2]*mm;
        p->pos=dummy+fPos;

        p->dir[X]=ph_in.px[0];
        p->dir[Y]=ph_in.px[1];
        p->dir[Z]=ph_in.px[2];

        p->wl   = ph_in.lambda*nm;
        p->time = ph_in.t*ns;
    }
    p->fate = flag;
#ifdef DEBUG
    cout<<"NOpticalSystem OUT"<<endl;
    cout<<"position = "<<p->pos<<endl;
    cout<<"direction = "<<p->dir<<endl;
    theta=p->dir.Angle(EVector(0.,0.,1.));
    cout<<"theta = "<<theta/deg<<endl;
    cout<<"wl = "<<p->wl/nm<<endl;
    cout<<"time = "<<p->time/ns<<endl;
#endif

    if ( flag != 0 ) return 0;

    Double_t DL;
    // sometime Ntrace returns photons out of the boundaries
    if (p->pos.Perp() > fOpticalSystemParam->r_wall || p->pos[Z] > Top() || p->pos[Z] < Bottom() ){
        // the photon is outside the cyl, trace it back to the borders
        EVector savedir = p->dir;
        p->dir = -p->dir;
        if ((DL = CylinderIntersection(p,kTRUE)) > 0){
            p->pos+=p->dir.Unit()*DL;
            p->time-=DL/TMath::C();
        }
        p->dir=savedir;
    } else if ((DL = CylinderIntersection(p,kTRUE)) > 0){
    // all the photon must be returned on the boundaries of the optics

        p->pos+=p->dir.Unit()*DL;
        p->time+=DL/TMath::C();
    }

#ifdef DEBUG
        cout<<"new position = "<<p->pos<<endl;
        cout<<"new direction = "<<p->dir<<endl;
        cout<<"new wl = "<<p->wl/nm<<endl;
        cout<<"new time = "<<p->time/ns<<endl;
#endif /* DEBUG */

    if (IsGoingOut(p)){
        return p;
    }
    else
        return 0;
}

//______________________________________________________________________________
Bool_t NOpticalSystem::IsGoingOut( Photon *p ) const {
    //
    // helper method to select outward going photons
    //

    Bool_t side, face;
    // check if photon is on the last surface
    side = TMath::Abs((p->pos-fPos).Perp()-fR) < kTolerance  &&
        (p->pos-fPos).XYvector()*p->dir.XYvector() > 0 &&
        ((p->pos[Z] < FirstLensTop() && p->pos[Z] > FirstLensBottom()) ||
         (p->pos[Z] < SecondLensTop() && p->pos[Z] > SecondLensBottom()) );
    // or if it's on the top/bottom face
    face = (TMath::Abs(p->pos[Z]-FirstLensBottom()) < kTolerance && p->dir[Z]<0) ||
        (TMath::Abs(p->pos[Z]-SecondLensTop()) < kTolerance && p->dir[Z]>0);
    return side || face;
}

//______________________________________________________________________________
Bool_t NOpticalSystem::HitSide(Photon *p) const {
    //
    // Check if p hist the side of the Cylinder.
    //

    Double_t dt = CylinderIntersection( p, fR, SecondLensBottom(), FirstLensTop() );
    if (dt < 0 || (p->pos+p->dir*dt)[Z]-SecondLensBottom() < kTolerance
               || (p->pos+p->dir*dt)[Z]-FirstLensTop() < kTolerance )
        return kFALSE;
    return kTRUE;
}
