/* -------------------------------------------------------------------------
 *   trace_lenses.c
 *
 *   --- main program of raytracing in the Fresnel lens system
 *
 * Copyright (c) 2000-2010 N.Sakaki, Y.Takizawa, Y.Kawasaki
 * All rights reserved
 * $Id$
 * -------------------------------------------------------------------------
 */
#define NTRACE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "EConst.hh"
#include "EsafRandom.hh"
#include "Nrand_gen.hh"
#include "Nspline.hh"
#include "Nutils.hh"
#include "Ntrace_lenses.hh"
#include "Nphoton.hh"
//#include "EConst.hh"

#define RT_SEARCH_ITER_MAX        100
#define RT_INTERACTION_MAX         20
//#define DEBUG
using namespace NTraceLens;

static double ref(double n1, double n2, double ci, double ct);
static int check_intersect_cylinder(double r1, double z1, double r2, double z2,
				    NPhoton *photon, double *dis);
static int check_intersect_cone    (double r1, double z1, double r2, double z2,
				    NPhoton *photon, double *dis);
static int CheckIris(tel_param *para, NPhoton *ph, int flag);

static int fGen;
static double fEffectiveEfficiencyFactor;

void  NTraceLens::SetRunGen(int i){

    fGen=i;

}

void NTraceLens::SetEffFactor(double ef){

fEffectiveEfficiencyFactor=ef;

}

#define LARGECURV                1.0e10
/*
 * ========================================================================
 * functions for diffractive optics start
 * ========================================================================
 */
/*
 * ------------------------------------------------------------------------
 *  Calculate normal vector to the sphere approximating the lens surface
 *  Input:
 *    aX, aY, aZ: photon position on the lens surface
 *    aCZ       : center of the approximated sphere
 *
 *  Output:
 *    nnnDIFF[3]: normal unit vector
 * ------------------------------------------------------------------------
 */
int NTraceLens::CalSpheVec(double aX, double aY, double aZ, double aCZ, double nnnDIFF[3])
{

  double N_;

  nnnDIFF[0] = aX;
  nnnDIFF[1] = aY;
  nnnDIFF[2] = aZ - aCZ;

  N_ = sqrt(vdot(nnnDIFF,nnnDIFF));

  nnnDIFF[0] = nnnDIFF[0]/N_;
  nnnDIFF[1] = nnnDIFF[1]/N_;
  nnnDIFF[2] = nnnDIFF[2]/N_;

  return 0;
}


/*
 * ------------------------------------------------------------------------
 *  read parameters of diffractive optics
 *    Input:
 *        dirname    the name of the directory name
 *                        where data files are stored
 *    Output:
 *        (return value) status
 * ------------------------------------------------------------------------
 */
int NTraceLens::ReadDiffractData(tel_param *param, FILE *fplog)
{
  FILE *fr;
  char Infname[256];
  double SD1_R,SD1_D,*tptr;

#ifndef QUIET
  if(fplog)
    fprintf(fplog,"Reading Diffractive Surface Data (%s) ....  ",
	    param->surface[SURFDOE].filename);
#endif

  /* read diffractive optics data */
  sprintf(Infname,"%.128s/%.127s",param->lens_dir,param->surface[SURFDOE].filename);

  if(( fr = fopen(Infname,"r" )) == NULL)
    {
      if(fplog)
	fprintf(fplog,"Cannot not OPEN %s\n",Infname );
      return (-1);
    }
  param->surface[SURFDOE].size=NSURFPTS;
  param->surface[SURFDOE].r=(double *)malloc(sizeof(double)*NSURFPTS);
  param->surface[SURFDOE].z=(double *)malloc(sizeof(double)*NSURFPTS);
  if(param->surface[SURFDOE].r==NULL || param->surface[SURFDOE].z==NULL)
    {
      if(fplog)
	fprintf(fplog,"cannot allocate memory (DOE)\n");
      fclose(fr);
      return (-1);
    }
  param->surface[SURFDOE].ndata=0;

  while(fscanf(fr,"%lf %lf",&SD1_R,&SD1_D)==2)
    {
      if(param->surface[SURFDOE].ndata >= param->surface[SURFDOE].size)
	{
	  param->surface[SURFDOE].size+=NSURFPTS;
	  tptr=(double *)realloc(param->surface[SURFDOE].r,
				 sizeof(double)*param->surface[SURFDOE].size);
	  if(tptr==NULL)
	    {
	      if(fplog)
		fprintf(fplog,"cannot allocate memory (DOE,r)\n");
	      fclose(fr);
	      free(param->surface[SURFDOE].r);
	      free(param->surface[SURFDOE].z);
	      param->surface[SURFDOE].r=NULL;
	      param->surface[SURFDOE].z=NULL;
	      param->surface[SURFDOE].size=0;
	      param->surface[SURFDOE].ndata=0;
	      return (-1);
	    }
	  else
	    {
	      param->surface[SURFDOE].r=tptr;
	      tptr=(double *)realloc(param->surface[SURFDOE].z,
				     sizeof(double)*param->surface[SURFDOE].size);
	      if(tptr==NULL)
		{
		  if(fplog)
		    fprintf(fplog,"cannot allocate memory (DOE,z)\n");
		  fclose(fr);
		  free(param->surface[SURFDOE].r);
		  free(param->surface[SURFDOE].z);
		  param->surface[SURFDOE].r=NULL;
		  param->surface[SURFDOE].z=NULL;
		  param->surface[SURFDOE].size=0;
		  param->surface[SURFDOE].ndata=0;
		  return (-1);
		}
	      else
		param->surface[SURFDOE].z=tptr;
	    }
	}
      param->surface[SURFDOE].r[param->surface[SURFDOE].ndata]=SD1_R;
      param->surface[SURFDOE].z[param->surface[SURFDOE].ndata]=SD1_D;
      param->surface[SURFDOE].ndata++;
    }
  fclose(fr);

#ifndef QUIET
  if(fplog)
    fprintf(fplog,"Done.\n");
#endif

  return (0);
}

/*
 * ------------------------------------------------------------------------
 *  calculate the diffractive structure
 *   Input:
 *      r   the distance in mm to the optical axis at the intersection point
 *      s   surface ID (i.e. S1,S2,...)
 *   Output:
 *     (return value)
 * ------------------------------------------------------------------------
 */
double NTraceLens::Cal_D(double r, tel_param *param)
{
  int i;

  i=rt_bsearch(param->surface[SURFDOE].r,r,0,param->surface[SURFDOE].ndata-1);
  return (param->surface[SURFDOE].z[i]);
}

/*
 * ------------------------------------------------------------------------
 *  calculate the diffracted ray
 *  Input:
 *        *photon       pointer to photon
 *        nnn[3]        normal vector to the diffractive surface
 *                         at the intersection
 *        n1            refractive index of xxx
 *        n2            refractive index of xxx
 *        d
 *        wl            wavelength in mm
 *        m             index
 *  Output:
 *        *photon
 * ------------------------------------------------------------------------
 */
int NTraceLens::diffract(int ID, tel_param *param, NPhoton *photon, double Zc,
	     double n1, double n2, double wl, double wl0, int m)
{
  int    i;
  double CosIN, CosD, dir, nnnR[3], r3, ppp[3], nnn[3], d, norm;
  double x3, y3, z3;

  x3=photon->x[X];
  y3=photon->x[Y];
  z3=photon->x[Z];

  if(fabs(param->surface[SURFDOE].Rc)>LARGECURV) /* plane */
    {
      nnn[0]= 0.0;
      nnn[1]= 0.0;
      nnn[2]= 1.0;
    }
  else /* sphere */
    {
      CalSpheVec(x3, y3, z3, Zc, nnn);
    }

  r3=sqrt(x3*x3+y3*y3);
  nnnR[0]= x3/r3;
  nnnR[1]= y3/r3;
  nnnR[2]=0.0;

  d=Cal_D(r3,param);

  if(photon->px[Z]>0.)
    dir=1.0;
  else
    dir=-1.0;

  CosIN=vdot(photon->px,nnn);
  CosD=vdot(nnn,nnnR);

  for(i=0;i<3;i++)
    ppp[i]=(n1*(photon->px[i]-CosIN*nnn[i])
	    +(double)m*wl/wl0*d*(nnnR[i]-CosD*nnn[i])*dir)/n2;

  norm=vdot(ppp,ppp);
  if(norm>1.0)
    return -1;

  norm=sqrt(1.0-norm);

#ifdef DEBUG
  fprintf(stderr,"Diffractive: (%.3f , %.3f , %.3f)=>",
	  photon->px[X],photon->px[Y],photon->px[Z]);
#endif

  for(i=0;i<3;i++)
    photon->px[i]=ppp[i]+norm*nnn[i]*(CosIN>0.0?(1.0):(-1.0));

#ifdef DEBUG
  fprintf(stderr,"(%.3f , %.3f , %.3f)\n",
	  photon->px[X],photon->px[Y],photon->px[Z]);
#endif

  return (0);
}

/*
 * ========================================================================
 * functions for diffractive optics END
 * ========================================================================
 */

/*
 * ------------------------------------------------------------------------
 * calculate reflection probability assuming no polariation
 *   n1,n2: refractive index of the media of this side and the other side
 *   ci,ct: cosine of the incident light and reflected light to the surface
 *          normal vector
 * ------------------------------------------------------------------------
 */
static double ref(double n1, double n2, double ci, double ct)
{
  double rp,rs;

  rp=(n2*ci-n1*ct)/(n2*ci+n1*ct);
  rs=(n1*ci-n2*ct)/(n1*ci+n2*ct);

  return((rp*rp+rs*rs)/2.0);
}

/*
 * ------------------------------------------------------------------------
 *  read lens surface shapes, and optical indices
 * ------------------------------------------------------------------------
 */
int NTraceLens::ReadLensData(tel_param *param, FILE *fplog)
{
  FILE *fp;
  char filename[256];
  int i;
  double tr,tsz,mat_wl,mat_n,mat_k,*tptr;

  for(i=0;i<NSURFACES;i++)
    {
      param->surface[i].ndata=0;
      param->surface[i].size=0;
      if(!(i==1 || i==2 || i==4 || i==5 || i==6 || i==7)) continue;

#ifndef QUIET
  if(fplog)
    fprintf(fplog,"Reading Surface(%d) Data (%s) ....  ",
	   i,param->surface[i].filename);
#endif

      sprintf(filename,"%.128s/%.128s",
	      param->lens_dir,param->surface[i].filename);
      if((fp=fopen(filename,"r"))==NULL)
	{
	  if(fplog)
	    fprintf(fplog,"%s not found.\n",filename);
        return -1;
	}

      param->surface[i].size=NSURFPTS;
      param->surface[i].r=(double *)malloc(sizeof(double)*NSURFPTS);
      param->surface[i].z=(double *)malloc(sizeof(double)*NSURFPTS);
      if(param->surface[i].r==NULL || param->surface[i].z==NULL)
	{
	  if(fplog)
	    fprintf(fplog,"cannot allocate memory (S%d)\n",i);
	  fclose(fp);
	}

      while(fscanf(fp,"%lf %lf",&tr, &tsz)!=EOF)
	{
	  if(param->surface[i].ndata >= param->surface[i].size)
	    {
	      param->surface[i].size+=NSURFPTS;
	      tptr=(double *)realloc(param->surface[i].r,
				     sizeof(double)*param->surface[i].size);
	      if(tptr==NULL)
		{
		  free(param->surface[i].r);
		  free(param->surface[i].z);
		  param->surface[i].r=NULL;
		  param->surface[i].z=NULL;
		  param->surface[i].size=0;
		  param->surface[i].ndata=0;
		  return (-1);
		}
	      else
		{
		  param->surface[i].r=tptr;
		  tptr=(double *)realloc(param->surface[i].z,
					 sizeof(double)*param->surface[i].size);
		  if(tptr==NULL)
		    {
		      if(fplog)
			fprintf(fplog,"cannot allocate memory (S%d,z)\n",i);
		      free(param->surface[i].r);
		      free(param->surface[i].z);
		      param->surface[i].r=NULL;
		      param->surface[i].z=NULL;
		      param->surface[i].ndata=0;
		      fclose(fp);
		      return -1;
		    }
		  else
		    {
		      param->surface[i].z=tptr;
		    }
		}
	    }
	  param->surface[i].r[param->surface[i].ndata]=tr;
	  param->surface[i].z[param->surface[i].ndata]=tsz;

	  param->surface[i].z[param->surface[i].ndata]=param->surface[i].Sz-param->surface[i].z[param->surface[i].ndata];
	  param->surface[i].ndata++;
	}
      fclose(fp);

      if(param->surface[i].r[param->surface[i].ndata-1] < param->surface[i].r_lens)
    param->surface[i].r_lens=param->surface[i].r[param->surface[i].ndata-1];

#ifndef QUIET
  if(fplog)
    fprintf(fplog,"Done.\n");
#endif
    }

  /* Read Material data */
  for(i=2;i<param->nmat;i++)
    {
#ifndef QUIET
  if(fplog)
    fprintf(fplog,"Reading Optical Data, %s (%s) ....   ",
	    param->material[i].name, param->material[i].filename);
#endif

      sprintf(filename,"%.128s/%.128s",param->lens_dir,param->material[i].filename);
      if((fp=fopen(filename,"r"))==NULL)
	{
	  if(fplog)
	    fprintf(fplog,"Cannot find %s\n",filename);
	  return -10;
	}

      param->material[i].ndata=0;
      param->material[i].size=NMATPTS;
      param->material[i].wl=(double *)malloc(sizeof(double)*NMATPTS);
      param->material[i].n=(double *)malloc(sizeof(double)*NMATPTS);
      param->material[i].k=(double *)malloc(sizeof(double)*NMATPTS);
      if(param->material[i].wl==NULL || param->material[i].n==NULL ||
	 param->material[i].k==NULL)
	{
	  if(fplog)
	    fprintf(fplog,"cannot allocate memory (%s)\n",param->material[i].name);
	  fclose(fp);
	  return (-1);
	}

      while(fscanf(fp,"%lf %lf %lf",&mat_wl,&mat_n,&mat_k)!=EOF)
	{
	  if(param->material[i].ndata >= param->material[i].size)
	    {
	      param->material[i].size+=NMATPTS;
	      tptr=(double *)realloc(param->material[i].wl,
				     sizeof(double)*param->material[i].size);
	      if(tptr==NULL)
		{
		  if(fplog)
		    fprintf(fplog,"cannot allocate memory (%s,wl)\n",param->material[i].name);
		  fclose(fp);
		  free(param->material[i].wl);
		  free(param->material[i].n);
		  free(param->material[i].k);
		  param->material[i].wl=NULL;
		  param->material[i].n=NULL;
		  param->material[i].k=NULL;
		  param->material[i].size=0;
		  param->material[i].ndata=0;
		  return (-1);
		}
	      else
		{
		  param->material[i].wl=tptr;
		  tptr=(double *)realloc(param->material[i].n,
					 sizeof(double)*param->material[i].size);
		  if(tptr==NULL)
		    {
		      if(fplog)
			fprintf(fplog,"cannot allocate memory (%s,n)\n",param->material[i].name);
		      fclose(fp);
		      free(param->material[i].wl);
		      free(param->material[i].n);
		      free(param->material[i].k);
		      param->material[i].wl=NULL;
		      param->material[i].n=NULL;
		      param->material[i].k=NULL;
		      param->material[i].size=0;
		      param->material[i].ndata=0;
		      return (-1);
		    }
		  else
		    {
		      param->material[i].n=tptr;

		      tptr=(double *)realloc(param->material[i].k,
					     sizeof(double)*param->material[i].size);
		      if(tptr==NULL)
			{
			  fclose(fp);
			  free(param->material[i].wl);
			  free(param->material[i].n);
			  free(param->material[i].k);
			  param->material[i].wl=NULL;
			  param->material[i].n=NULL;
			  param->material[i].k=NULL;
			  param->material[i].size=0;
			  param->material[i].ndata=0;
			  return (-1);
			}
		      else
			param->material[i].k=tptr;
		    }
		}
	    }
	  param->material[i].wl[param->material[i].ndata]=mat_wl;
	  param->material[i].n[param->material[i].ndata]=mat_n;
	  param->material[i].k[param->material[i].ndata]=mat_k;
	  param->material[i].ndata++;
	}
      fclose(fp);

      param->material[i].zn=(double *)malloc(sizeof(double)*param->material[i].ndata);
      param->material[i].zk=(double *)malloc(sizeof(double)*param->material[i].ndata);
      if(param->material[i].zn==NULL || param->material[i].zk==NULL)
	{
	  if(fplog)
	    fprintf(fplog,"Canoot allocate memory for optical data (%s)\n",
		    param->material[i].name);
	  return (-10);
	}
      csp_maketable(param->material[i].wl, param->material[i].n,
		    param->material[i].zn, param->material[i].ndata);
      csp_maketable(param->material[i].wl, param->material[i].k,
		    param->material[i].zk, param->material[i].ndata);
#ifndef QUIET
  if(fplog)
    fprintf(fplog,"Done.\n");
#endif
    }

  return 0;
}

/*
 * ------------------------------------------------------------------------
 *  get refractive index of material at wavelength lambda
 *    using cubic spline interpolation
 * ------------------------------------------------------------------------
 */
double NTraceLens::Get_n(int matid, tel_param *param, double lambda)
{
  if(matid==0 || matid==1)
    return (*(param->material[matid].n));
  else if(matid<param->nmat)
    return cspline(lambda, param->material[matid].wl,
		   param->material[matid].n, param->material[matid].zn,
		   param->material[matid].ndata);
  else
    return 0.;
}

/*
 * ------------------------------------------------------------------------
 *  get extinction coefficient of material at wavelength lambda
 *    using cubic spline interpolation
 * ------------------------------------------------------------------------
 */
double NTraceLens::Get_k(int matid, tel_param *param, double lambda)
{
  if(matid==0 || matid==1)
    return (*(param->material[0].k));
  else if(matid<param->nmat)
    return cspline(lambda, param->material[matid].wl,
		   param->material[matid].k, param->material[matid].zk,
		   param->material[matid].ndata);
  else
    return 0.;
}
/*
 * ------------------------------------------------------------------------
 *  get the refracted ray
 *    if reflected it is thrown away.
 *
 *  *photon: info on photon (only incident photon direction is used)
 *   nnn[3]  : surface normal vector (direct to this side)
 *   n1,n2   : refractive indices of the media of this side and the other side
 * ------------------------------------------------------------------------
 */
int NTraceLens::refract(NPhoton *photon, double nnn[3],double n1,double n2)
{
  int i, TotalReflection=0;
  double CosIN, CosOUT=1.0, ppp[3], norm, out_phdir[3];

  CosIN=vdot(photon->px,nnn);
  for(i=0;i<3;i++)
    ppp[i]=(photon->px[i]-CosIN*nnn[i])*n1/n2;

  norm=vdot(ppp,ppp);
#ifdef NO_REFLECTION
  if(norm>1.0)
    return(RT_NO_REFRACTED_LIGHT);  /* total reflection */
#else
  if(norm>1.0)
    TotalReflection=1;
  else
    {
#endif

      norm=sqrt(1.0-norm);
      for(i=0;i<3;i++)
	out_phdir[i]=ppp[i]+norm*nnn[i]*(CosIN>0.0?(1.0):(-1.0));

#ifndef NO_REFLECTION
      /* check reflection */
      CosOUT=vdot(out_phdir,nnn);
    }

  if(TotalReflection==1 || fGen==1?EsafRandom::Get()->Rndm():RANDOM() < ref(n1,n2,CosIN,CosOUT))
    {
#ifdef DEBUG
      fprintf(stderr,"Reflection: (%.3f,%.3f,%.3f)=>",
             photon->px[X],photon->px[Y],photon->px[Z]);
#endif
      for(i=0;i<3;i++)
	photon->px[i]-=2.0*CosIN*nnn[i];
#ifdef DEBUG
      fprintf(stderr,"(%.3f,%.3f,%.3f)\n",
             photon->px[X],photon->px[Y],photon->px[Z]);
#endif
      return(RT_REFLECTED);
    }
#endif

  /* photon->px[]: direction of incident ray -> refracted ray */
#ifdef DEBUG
      fprintf(stderr,"Refraction: (%.3f,%.3f,%.3f)=>",
             photon->px[X],photon->px[Y],photon->px[Z]);
#endif

  for(i=0;i<3;i++)
    photon->px[i]=out_phdir[i];

#ifdef DEBUG
  fprintf(stderr,"(%.3f,%.3f,%.3f)\n",
	  photon->px[X],photon->px[Y],photon->px[Z]);
  fprintf(stderr,"normal vect=(%.3f,%.3f,%.3f)\n",nnn[0],nnn[1],nnn[2]);
#endif

  return(RT_REFRACTED);
}

/*
 * ------------------------------------------------------------------------
 *  calculate the differential coefficient at lens_r[i]
 *   Input:
 *      i                : the radius index of lens_r[i]
 *      lens_r[],lens_z[]: lens surface data sets of the radius and the height
 *      n_lens_elem      : number of data elements of lens_r[]
 *   Output:
 *     (return value)    : differential coefficeint at lens_r[i]
 * ------------------------------------------------------------------------
 */
double NTraceLens::differential_r(int i, double *lens_r,double *lens_z, int n_lens_elem)
{
  double dr1,dr2;
  double dz1,dz2;
  int j1,j2;

  dz1=dz2=0.;
  j1=i+1;
  j2=i-1;

  if(i>=n_lens_elem-1)
    return 0.0;
  else if(i==n_lens_elem-2)
    {
      if(lens_r[i+1]-lens_r[i]<0.001)
        return 0.0;
      else
        return (lens_z[i+1]-lens_z[i])/(lens_r[i+1]-lens_r[i]);
    }

  dr1=lens_r[j1]-lens_r[i];
  if(dr1<0.001)
    {
      dz1+=lens_z[j1]-lens_z[i];
      j1++;
    }

  if(j2<0)
    {
      j2=-j2;
      dr2=lens_r[i]+lens_r[j2];
    }
  else
    {
      dr2=lens_r[i]-lens_r[j2];
    }
  if(dr2<0.001)
    {
      dz2+=lens_z[j2]-lens_z[i];
      j2--;
    }

  return (lens_z[j1]-dz1-lens_z[j2]+dz2)/(dr1+dr2);
}

/*
 *
 * check if there are any intersections with a cylinder
 *
 */
static int check_intersect_cylinder(double r1, double z1, double r2, double z2,
				    NPhoton *photon, double *dis)
{
  double aa[3], xx[2], tmp, z;
  int    nroot, i;

  aa[0] = (photon->px[X])*(photon->px[X]) + (photon->px[Y])*(photon->px[Y]);
  aa[1] = 2.0*((photon->x[X])*(photon->px[X]) + (photon->x[Y])*(photon->px[Y]));
  aa[2] = ((photon->x[X])*(photon->x[X]) + (photon->x[Y])*(photon->x[Y])) - r1*r1;

  nroot=SolveQuadratic(aa,xx);
  if(nroot<2)
    return(RT_NO_INTERSECTION1);

  /* solution is the smaller positive value */
  /* swap */
  if(xx[0]>xx[1])
    {
      tmp=xx[0];
      xx[0]=xx[1];
      xx[1]=tmp;
    }
  if(xx[1]<0)
    return(RT_NO_INTERSECTION1);

  for(i=0;i<2;i++)
    {
      if(xx[i]>=0.0)
	{
	  z=photon->x[Z]+photon->px[Z]*xx[i];
	  /* check if the solution is in the specified range */
	  if((z-z1)*(z-z2)<0)
	    {
	      *dis=xx[i];
	      return 0;
	    }
	}
    }

  *dis=(xx[0]>0 ? xx[0]:xx[1]);
  return(RT_NO_INTERSECTION2);
}

/*
 *
 * check if there are any intersections with a cone
 *
 */
static int check_intersect_cone(double r1, double z1, double r2, double z2,
				NPhoton *photon, double *dis)
{
  double a00, a0, aa[3], xx[2], z0, dbuf1, tmp, x, y, r, zz[2];
  int    nroot, i;

  a00 = (z2-z1)/(r2-r1);
  z0  = z1-a00*r1;
  a0  = a00*a00;

  dbuf1 = (photon->x[Z])-z0;
  aa[0] = a0*(photon->px[X])*(photon->px[X]) + a0*(photon->px[Y])*(photon->px[Y] )- (photon->px[Z])*(photon->px[Z]);
  aa[1] = 2.0*(a0*(photon->px[X])*(photon->x[X]) + a0*(photon->px[Y])*(photon->x[Y]) - (photon->px[Z])*dbuf1);
  aa[2] = a0*(photon->x[X])*(photon->x[X]) + a0*(photon->x[Y])*(photon->x[Y]) - dbuf1*dbuf1;

  nroot=SolveQuadratic(aa,xx);
  if(nroot<2)
    return(RT_NO_INTERSECTION1);

  /* solution is the smaller positive value */
  /* swap */
  if(xx[0]>xx[1])
    {
      tmp=xx[0];
      xx[0]=xx[1];
      xx[1]=tmp;
    }
  if(xx[1]<0.0)
    return(RT_NO_INTERSECTION1);

  /* We usually have a solution on a virtual cone with -a00 */
  /* check this condition */
  for(i=0;i<2;i++)
    if(xx[i]>=0.0)
      {
	x=photon->px[X]*xx[i]+photon->x[X];
	y=photon->px[Y]*xx[i]+photon->x[Y];
	zz[i]=photon->px[Z]*xx[i]+photon->x[Z];
	r=sqrt(x*x+y*y);
	if((r-r1)*(r-r2)<0.0)
	  {
	    *dis=xx[i];
	    return 0;
	  }
      }

  if(xx[0]<0.0)
    *dis=xx[1];
  else
    {
      /* the nearer one to the surface will be what we want... */
      if(fabs(zz[0]-z1)<fabs(zz[1]-z1))
	*dis=xx[0];
      else
	*dis=xx[1];
    }

  return(RT_NO_INTERSECTION2);
}

/*
 * ------------------------------------------------------------------------
 *
 *  ID:            Lens Surface ID
 *  phot[8]: input and output photon data
 *    phot[0,1,2]: x,y,z position in mm
 *    phot[3,4,5]: l,m,n direction
 *    phot[6]    : wavelength in nm
 *    phot[7]    : time in ns
 *  offset_z     : the offset z-position of sphare center approximating
 *                 lens surface
 *  radius       : the radius of sphare approximated lens surface
 *  lens_r[],lens_z[]: lens surface data sets of the radius and the height
 *  n_lens_elem  : number of data elements of lens_r[]
 *  medium1_n, medium2_n: refractive indices of the media in this side
 *                         and the other side
 * ------------------------------------------------------------------------
 */
int lens_interaction(int ID, NPhoton *photon, tel_param *param,
		     double medium1_n, double medium2_n, double *dis)
{
  int r_idx,iter=0,flag,Found,nroot,dir,status,idx_dir;
  double dbuf1, aa[3], xx[2], x3,y3,z3,r3, r1,r2, z1,z2;
  double nnn[3];
  double a00;
  double tmp, tmp_dis, v[3], tmp_x[3];
  double Zc, Rc, lens_r_max,  *lens_r, *lens_z;
  int n_lens_elem;
#ifdef DEBUG
  double approx_x[3];
#endif

  Zc=param->surface[ID].Zc;
  Rc=param->surface[ID].Rc;
  lens_r_max=param->surface[ID].r_lens;
  lens_r=param->surface[ID].r;
  lens_z=param->surface[ID].z;
  n_lens_elem=param->surface[ID].ndata;

  if(photon->px[Z]==0)
    return(RT_HIT_SIDE_WALL);

  if(fabs(Rc)>LARGECURV)
    {
      dbuf1 = (param->surface[ID].Sz-photon->x[Z])/photon->px[Z];
      x3=photon->px[X]*dbuf1+photon->x[X];
      y3=photon->px[Y]*dbuf1+photon->x[Y];
      z3=photon->px[Z]*dbuf1+photon->x[Z];
      r3=sqrt(x3*x3+y3*y3);
      nnn[0]=0.0;
      nnn[1]=0.0;
      nnn[2]=1.0;
      if(r3>lens_r_max)
	return(RT_HIT_SIDE_WALL);
      *dis=dbuf1;
    }
  else
    {
  /* as an initial guess, intersection with an approximated sphere
     is calculated */

      dbuf1 = photon->x[Z]-Zc;
      aa[0] = photon->px[X]*photon->px[X] + photon->px[Y]*photon->px[Y] + photon->px[Z]*photon->px[Z];
      aa[1] = 2.0*(photon->px[X]*photon->x[X] + photon->px[Y]*photon->x[Y] + photon->px[Z]*dbuf1);
      aa[2] = photon->x[X]*photon->x[X] + photon->x[Y]*photon->x[Y] + dbuf1*dbuf1-Rc*Rc;

      nroot=SolveQuadratic(aa,xx);
      if(nroot<2)
	return(RT_NO_INTERSECTION1);

      /* swap */
      if(xx[0]>xx[1])
	{
	  tmp=xx[0];
	  xx[0]=xx[1];
	  xx[1]=tmp;
	}

      /* We need a positive, smaller solution. */
      if(xx[0]>0.0)
	dbuf1=xx[0];
      else
	{
	  if(xx[1]>0.0)
	    dbuf1=xx[1];
	  else
	    return(RT_NO_INTERSECTION1);
	}

      x3=photon->px[X]*dbuf1+photon->x[X];
      y3=photon->px[Y]*dbuf1+photon->x[Y];
      z3=photon->px[Z]*dbuf1+photon->x[Z];
      /* distance to the optical axis at the initially guessed intersection */
      r3=sqrt(x3*x3+y3*y3);

      if(r3>lens_r_max && xx[0]>-100.0) /* -100 is not optimized */
	{
	  /* The other solution may be acceptable just for an initial guess. */
	  dbuf1=xx[0];
	  x3=photon->px[X]*dbuf1+photon->x[X];
	  y3=photon->px[Y]*dbuf1+photon->x[Y];
	  z3=photon->px[Z]*dbuf1+photon->x[Z];

	  r3=sqrt(x3*x3+y3*y3);
	  if(r3>lens_r_max)
	    return(RT_HIT_SIDE_WALL);
	}
      *dis=dbuf1;

#ifdef DEBUG
      fprintf(stderr,"1st est.: %.3f ",r3);
      approx_x[0]=x3;approx_x[1]=y3;approx_x[2]=z3;
#endif
    }
  {
      /* search taking account of Fresnel structure */
      /* search the index near r3 */
      r_idx=rt_bsearch(lens_r,r3,0,n_lens_elem-1);

#ifdef DEBUG
      fprintf(stderr,"%.3f\n",r3);
#endif
      if(r_idx>n_lens_elem-2)
	r_idx=n_lens_elem-2;

      /* search intersection point with real fresnel lens */
      iter=0; /* number of iterations */
      do{
	Found=0;

	r1=lens_r[r_idx];
	z1=lens_z[r_idx];
	r2=lens_r[r_idx+1];
	z2=lens_z[r_idx+1];

	if(r1==r2)
	  {
	    /* intersection with a part of the cylinder */
	    status=check_intersect_cylinder(r1, z1, r2, z2, photon, dis);

	    x3=photon->px[X]*(*dis)+photon->x[X];
	    y3=photon->px[Y]*(*dis)+photon->x[Y];
	    z3=photon->px[Z]*(*dis)+photon->x[Z];

	    if(!status)     /* intersection found */
	      {
		Found=1;
#ifdef DEBUG
		fprintf(stderr,"Hit Backcut\n");
		/* return (RT_ABSORBED);*/
#endif
		/* normal vector */
		if(z2>z1)
		  dir=1;
		else
		  dir=-1;
		nnn[X]=dir*x3/r1;
		nnn[Y]=dir*y3/r1;
		nnn[Z]=0.0;
	      }
	    else /* intersection with cylinder not found */
	      {
		if(status==RT_NO_INTERSECTION2 && fabs(z3-z1)<100.0)
		  {
		    v[X]=x3;
		    v[Y]=y3;
		    v[Z]=0.0;

		    if((z3-z1)*vdot(v,&photon->px[X])*photon->px[Z]>0)
		      r_idx--;
		    else
		      r_idx++;
		  }
		else
		  {
		    *dis=(z1-photon->x[Z])/photon->px[Z];
		    z3=z1;
		    x3=photon->px[X]*(*dis)+photon->x[X];
		    y3=photon->px[Y]*(*dis)+photon->x[Y];
		    r3=sqrt(x3*x3+y3*y3);
		    r_idx=rt_bsearch(lens_r,r3,0,n_lens_elem-1);
		  }
	      }
#ifdef DEBUG
	    fprintf(stderr,"ridx0(%d): %d (%.1f: %.3f %.3f %.3f (%.1f %.1f %.1f)(%.3f %.3f %.3f))\n",
		    status,r_idx,z1,r1,r2,r3,photon->x[X],photon->x[Y],photon->x[Z],photon->px[X],photon->px[Y],photon->px[Z]);
#endif
	  }
	else  /* r1 != r2 */
	  {
	    /* intersection with a part of the cone */
	    if(z1==z2)
	      {
		*dis = (z1-photon->x[Z])/photon->px[Z];
		status=0;
	      }
	    else
	      status=check_intersect_cone(r1, z1, r2, z2, photon, dis);

	    x3=photon->px[X]*(*dis)+photon->x[X];
	    y3=photon->px[Y]*(*dis)+photon->x[Y];
	    z3=photon->px[Z]*(*dis)+photon->x[Z];

	    r3=sqrt(x3*x3+y3*y3);

	    if(!status)   /* intersection found */
	      {
		Found=1;
		/* normal vector */
		if(r_idx>n_lens_elem-2)
		  r_idx=n_lens_elem-2;
		a00=(differential_r(r_idx,lens_r,lens_z,n_lens_elem)
		     *(r3-lens_r[r_idx])-
		     differential_r(r_idx+1,lens_r,lens_z,n_lens_elem)
		     *(r3-lens_r[r_idx+1]))/(lens_r[r_idx+1]-lens_r[r_idx]);
		nnn[X]=a00*x3;
		nnn[Y]=a00*y3;
		nnn[Z]=-r3;
		dbuf1=sqrt(vdot(nnn,nnn));
		nnn[X]/=dbuf1;
		nnn[Y]/=dbuf1;
		nnn[Z]/=dbuf1;
	      }
	    else /* not found */
	      {
		if(status==RT_NO_INTERSECTION2 &&
		   fabs(z3-z1)<10. && fabs(r3-r1)<5.)
		  {
		    if(r3<r1)
		      r_idx--;
		    if(r3>r2)
		      r_idx++;
		  }
		else
		  {
		    *dis=(z1-photon->x[Z])/photon->px[Z];
		    z3=z1;
		    x3=photon->px[X]*(*dis)+photon->x[X];
		    y3=photon->px[Y]*(*dis)+photon->x[Y];
		    r3=sqrt(x3*x3+y3*y3);
		    r_idx=rt_bsearch(lens_r,r3,0,n_lens_elem-1);
		  }
	      }
#ifdef DEBUG
	    fprintf(stderr,"ridx1(%d): %d (%.1f: %.3f %.3f %.3f (%.1f %.1f %.1f)(%.3f %.3f %.3f))\n",
		    status,r_idx,z1,r1,r2,r3,photon->x[X],photon->x[Y],photon->x[Z],photon->px[X],photon->px[Y],photon->px[Z]);
#endif
	  }

	if(r_idx>=n_lens_elem)
	  return(RT_OUT_OF_SURFACE2);
	if(r_idx<0 && r3>lens_r_max && fabs(z3-z1)<10.0)
	  return(RT_OUT_OF_SURFACE2);
	if(r_idx<0)
	  {
	    r_idx=1;
	    /* return(RT_OUT_OF_SURFACE1); */
	  }

	/* iteration doesn't converge */
	if(iter>RT_SEARCH_ITER_MAX)
	  {
#ifdef DEBUG
	    fprintf(stderr,"app.(%.1f,%.1f,%.1f)=>end(%.1f,%.1f,%.1f)\n",
		    approx_x[0],approx_x[1],approx_x[2],x3,y3,z3);
#endif
	    return(RT_EXCEED_ITER_MAX);
	  }

	iter++;

      } while(!Found);
    }

  /* refraction */
  if(photon->px[Z]<0.0)
    {
      tmp=medium2_n;
      medium2_n=medium1_n;
      medium1_n=tmp;
    }
#ifdef DEBUG
  fprintf(stderr,"ridx=%d n_med1=%.5f n_med2=%.5f\n",r_idx,medium1_n,medium2_n);
#endif
  flag=refract(photon, nnn, medium1_n, medium2_n);
#if 0
  if(flag==RT_REFLECTED)
    return(RT_REFLECTION_DISCARDED); /* reflected photons are thrown away */
#endif

  /* distance between injection point and the Surface */
  photon->t+=(*dis)* medium1_n/EConst::Clight();

  photon->x[X]=x3;
  photon->x[Y]=y3;
  photon->x[Z]=z3;

  return flag;
}

int NTraceLens::CheckRcut(tel_param *param, NPhoton *photon, int flag)
{
  int surfid;
  surfid=(-flag)/1000;
  if(surfid>8)
    surfid=8;

  if(fabs(photon->x[Y])>param->surface[surfid].r_cut)
    return RT_OUT_OF_SURFACE1;
  else
    return 0;
}

static int CheckIris(tel_param *param, NPhoton *photon, int flag)
{
  double x3, y3, z3=0.0, r3=0.0, dist, radius;
  int surfid;

  surfid=-flag/1000;

  z3=param->surface[surfid].Sz;
  r3=param->surface[surfid].r_lens;

  dist=(z3 - photon->x[Z])/photon->px[Z];
  if(dist>=0.0)
    {
      x3=photon->px[X]*dist+photon->x[X];
      y3=photon->px[Y]*dist+photon->x[Y];

      photon->x[X]=x3;
      photon->x[Y]=y3;
      photon->x[Z]=z3;
      photon->t+=dist*param->material[param->surface[3].matid].n[0]/EConst::Clight();

      radius=sqrt(x3*x3+y3*y3);
      if(radius > param->r_wall)
	return RT_HIT_SIDE_WALL;
      else if(radius > r3)
	{
	  return RT_OUT_OF_IRIS;
	}
    }

    return 0;
}

/* -------------------------------------------------------------------------
 * int trace_lens(tel_param *param, NPhoton *photon, FILE *fp)
 *
 *   --- trace photons in the EUSO lens system (the first lens to the last one)
 *
 * Input:
 *        param        tel_param* to store the optics parameters
 *                       (see also in tracemain_optF1v2.h)
 *        photon       info on photon (position[mm], direction, time[ns],wavelength[m]
 *        fp           FILE* for (error) message logging
 *
 * Output:
 *  return value
 *        0             normal end
 *       <0             error
 *                      (see trace_optF1v2.h for more detail
 *                        from RT_NO_SPINTERSECTION to RT_ESCAPE_FROM_ENTRANCE)
 * -------------------------------------------------------------------------
 */



//#define DEBUG
int NTraceLens::trace_lens(tel_param *param, NPhoton *photon, FILE *fplog)
{
  int    status=0, n_int, WithinOptics, i, matid;
  int    PrevPoint, CurPoint, NextPoint;
  double nn[8],kk[8];
  double AbsorbCoeff, TransProb;
  double dis;
  int    s3=0;
  /* Opt. index  for this lambda */
  for(i=0;i<param->nmat;i++)
    {
      nn[i]=Get_n(i,param,photon->lambda);
      kk[i]=Get_k(i,param,photon->lambda);
    }

  if(photon->x[Z]>=param->surface[7].Sz && photon->px[Z]<0.0)
    {
      PrevPoint=SURFACE_FS;
      CurPoint=SURFACE_FS;
      NextPoint=SURFACE7;
    }
  else
    {
      PrevPoint=SURFACE0;
      CurPoint=SURFACE0;
      NextPoint=SURFACE1;
    }
  n_int=0;
  WithinOptics=1;

#ifdef DEBUG
  fprintf(stderr,"0 %.1f %.1f %.1f  %.3f %.3f %.3f\n",
	  photon->x[X],photon->x[Y],photon->x[Z],
	  photon->px[X],photon->px[Y],photon->px[Z]);
#else


  if(param->flag_printray)
    printf("0 %.1f %.1f %.1f\n",photon->x[X],photon->x[Y],photon->x[Z]);
#endif

  while(n_int++ < RT_INTERACTION_MAX && WithinOptics)
    {
      PrevPoint=CurPoint;
      CurPoint=NextPoint;

      switch(CurPoint)
	{
	case SURFACE0:
	case SURFACE_FS:
	  WithinOptics=0;
	  break;

	case SURFACE1:
	  status=lens_interaction(1, photon, param,
				  nn[param->surface[1].matid],
				  nn[param->surface[2].matid],
				  &dis);
#ifdef DEBUG
      printf("1 %f %f %f stat=%i\n",photon->x[X],photon->x[Y],photon->x[Z], status);

#endif
          if(status==RT_REFLECTED)
            {
              NextPoint=PrevPoint;
              status=0;
            }
          else if(PrevPoint==SURFACE0)
            NextPoint=SURFACE2;
          else
            NextPoint=SURFACE0;
	  break;

	case SURFACE2:
	  status=lens_interaction(2, photon, param,
				  nn[param->surface[2].matid],
				  nn[param->surface[3].matid],
				  &dis);
#ifdef DEBUG
      printf("2 %f %f %f stat=%i\n",photon->x[X],photon->x[Y],photon->x[Z], status);

#endif
          if(status==RT_REFLECTED)
            {
              NextPoint=PrevPoint;
              status=0;
            }
          else if(PrevPoint==SURFACE1)
            NextPoint=SURFACE3;
          else
            NextPoint=SURFACE1;
	  break;

	case SURFACE3:
	  if((status=CheckIris(param,photon,SURFACE3))<0){
	    return (status+SURFACE3);}
	  if(photon->px[Z]>0.0)
	    NextPoint=SURFACE4;
	  else
	    NextPoint=SURFACE2;
	  break;

	case SURFACE4:
	  status=lens_interaction(4, photon, param,
				  nn[param->surface[4].matid],
				  nn[param->surface[5].matid],
				  &dis);
#ifdef DEBUG
      printf("4 %f %f %f stat=%i\n",photon->x[X],photon->x[Y],photon->x[Z], status);

#endif
          if(status==RT_REFLECTED)
            {
              NextPoint=PrevPoint;
              status=0;
            }
	  else if(photon->px[Z]>0.0)
	    NextPoint=SURFACE5;
	  else
	    NextPoint=SURFACE3;
	  break;

	case SURFACE5:
#ifndef NO_DIFFRACTIVE
	  if(photon->px[Z]<0.0)
	    {
	      diffract(5, param, photon,
		       param->surface[5].Zc, 1.0003, 1.0003,
		       photon->lambda*1e-6, param->lambda0*1e-6, 1);
	    }
#endif
	  status=lens_interaction(5, photon, param,
				  nn[param->surface[5].matid],
				  nn[param->surface[6].matid],
				  &dis);
#ifdef DEBUG
      printf("5 %f %f %f stat=%i\n",photon->x[X],photon->x[Y],photon->x[Z], status);

#endif
#ifndef NO_DIFFRACTIVE
	  if(photon->px[Z]>0.0)
	    {
	      diffract(5, param, photon,
		        param->surface[5].Zc, 1.0003, 1.0003,
		       photon->lambda*1e-6, param->lambda0*1e-6, 1);
	    }
#endif

          if(status==RT_REFLECTED)
            {
              NextPoint=PrevPoint;
              status=0;
            }
	  else if(photon->px[Z]>0.0)
	    NextPoint=SURFACE6;
	  else
	    NextPoint=SURFACE4;
	  break;

	case SURFACE6:
	  status=lens_interaction(6, photon, param,
				  nn[param->surface[6].matid],
				  nn[param->surface[7].matid],
				  &dis);
#ifdef DEBUG
      printf("6 %f %f %f stat=%i\n",photon->x[X],photon->x[Y],photon->x[Z], status);

#endif
          if(status==RT_REFLECTED)
            {
              NextPoint=PrevPoint;
              status=0;
            }
          else if(PrevPoint==SURFACE5)
            NextPoint=SURFACE7;
          else
            NextPoint=SURFACE5;
	  break;

	case SURFACE7:
	  status=lens_interaction(7, photon, param,
				  nn[param->surface[7].matid],
				  nn[param->surface[8].matid],
				  &dis);
#ifdef DEBUG
      printf("7 %f %f %f stat=%i\n",photon->x[X],photon->x[Y],photon->x[Z], status);

#endif
          if(status==RT_REFLECTED)
            {
              NextPoint=PrevPoint;
              status=0;
            }
          else if(PrevPoint==SURFACE6)
            NextPoint=SURFACE_FS;
          else
            NextPoint=SURFACE6;
	  break;
	}
//      if(WithinOptics)
//	StorePhotonHistory(photon);
#ifndef DEBUG
      if(param->flag_printray && CurPoint>SURFACE_FS)
	printf("%d %.1f %.1f %.1f\n",
	       -CurPoint/1000,photon->x[X],photon->x[Y],photon->x[Z]);
#else
      fprintf(stderr,"%d %.1f %.1f %.1f  %.3f %.3f %.3f\n",
	      -CurPoint/1000,
	      photon->x[X],photon->x[Y],photon->x[Z],
	      photon->px[X],photon->px[Y],photon->px[Z]);
#endif
      if(status<0){
	return (status+CurPoint);}
      status=CheckRcut(param,photon,CurPoint);
      if(status<0){
        return (status+CurPoint);}

#ifndef NO_ABSORPTION
      /* absorption between S1 and S2, S4 and S5, or between S6 and S7 */

      if(PrevPoint>CurPoint)
	matid=param->surface[-CurPoint/1000].matid;
      else
	matid=param->surface[-CurPoint/1000+1].matid;

      if(matid>1)
	{
	  AbsorbCoeff=kk[matid]*M_PI*4.0/(photon->lambda*1.e-6);
	  TransProb=exp(-AbsorbCoeff*dis);

	  if(fGen==1?EsafRandom::Get()->Rndm():RANDOM()>TransProb){
          return(RT_ABSORBED+CurPoint);}
	}
#endif
    }

  if(n_int>=RT_INTERACTION_MAX){
      return (RT_STRAY_LIGHT);}


  return 0;
}



