/* -------------------------------------------------------------------------
 *   trace_optics.c
 *
 *   ---  main program for raytracing
 *
 * Copyright (c) 2000-2010 N.Sakaki, Y.Takizawa, Y.Kawasaki
 * All rights reserved.
 * $Id$
 * -------------------------------------------------------------------------
 */
#define NTRACE

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>
#include <math.h>
#include <string.h>
#include "EConst.hh"
#include "Nrand_gen.hh"
#include "Ntrace_lenses.hh"
#include "Ntrace_optics.hh"
#include "Nphoton.hh"

//#define DEBUG

using namespace NTraceLens;


/* -------------------------------------------------------------------------
 * int trace_main(tel_param *param, NPhoton *photon, FILE *fp)
 *
 *   --- trace photons in the EUSO optics (Entrance to FS)
 *
 * [input value]
 * param  : tel_param* to store the optics parameters
 *            (see also in tracemain_optF1v2.h)
 * photon : info of photon (position, vector, time, wavelength,...)
 * fp       : FILE* for (error) message logging
 *
 * [return value]
 * 0: normal end
 * <0: error (see trace_optF1v2.h for more detail)
 * -------------------------------------------------------------------------
 */
int NTraceLens::trace_main(tel_param *param, NPhoton *photon, FILE *fplog)
{
  static double *z0;
  static int N_FOCAL_SURFACE;
  static int fs_init=0;
  int j;
  int flg, fsloop;
  double r,r2,dist,xx,yy,zz;

  if(!fs_init)
    {
      fs_init=1;
      N_FOCAL_SURFACE=(int)(param->r_fs/FS_DEF_STEP);
      z0=(double *)malloc(sizeof(double)*N_FOCAL_SURFACE);
      if(!z0)
	{
	  fprintf(fplog,"Cannot allocate memory for array for focalsurface\n");
	  exit(-10);
	}
      for(j=0;j<N_FOCAL_SURFACE;j++)
	{
	  r=(double)j*FS_DEF_STEP;
	  z0[j]=FocalSurface(r,param);
	}
    }

  /* the contents of photon will be overwritten in trace_lens() */
  flg=trace_lens(param,photon,fplog);
  if(flg!=0)
    {
#ifdef DEBUG
      if(fplog)
	fprintf(fplog,"flg=%d in trace_main\n",flg);
#endif
      return flg;
    }
  else if(photon->px[Z]<=0.0)
    {
      return (RT_ESCAPE_FROM_ENTRANCE+SURFACE1);
    }
  else /* photon->px[Z]>0.0 */
    {
      j=0;
      zz=z0[j];
      r=1.0e30;
      fsloop=0;
      do {
	fsloop++;
	r2=r;

	dist=(zz-photon->x[Z])/photon->px[Z];
	xx=photon->px[X]*dist+photon->x[X];
	yy=photon->px[Y]*dist+photon->x[Y];
	r=sqrt(xx*xx+yy*yy);         /* distance to optical axis */

	j=int(r/FS_DEF_STEP);
	if(j>=N_FOCAL_SURFACE || j < 0) /*changed by francesco 15-01-2010*/
	  j=N_FOCAL_SURFACE-1; /* should be error? */

	if(fabs(z0[j]-zz)<FS_DEF_STEP)
	  {
	    if(r > param->r_fs)
	      flg=1;       /* out of focal surface */
	    else/*changed by francesco 15-01-2010*/
	      zz=FocalSurface(r,param);
	  }
	else
	  zz=z0[j];
      }while(fabs(r-r2)>FS_TOLERANCE && !flg && fsloop<FSLOOP_MAX);

      if(flg)
	{
#ifdef DEBUG
	  fprintf(stderr,"curpos(%.1f,%.1f,%.1f) vec(%.2f,%.2f,%.2f) "
		  "est pos on FS(%.1f,%.1f,%.1f) r=%.1f\n",
		  photon->x[X],photon->x[Y],photon->x[Z],
		  photon->px[X],photon->px[Y],photon->px[Z],
		  xx,yy,zz,r
		  );

#endif
	  return (SURFACE_FS+RT_OUT_OF_SURFACE1);
	}
      else if(fsloop>=FSLOOP_MAX)
	{
	  return (SURFACE_FS+RT_STRAY_LIGHT);
	}
      else
	{
	  photon->x[X]=xx;
	  photon->x[Y]=yy;
	  photon->x[Z]=zz;
	  photon->t+=dist*param->material[param->surface[3].matid].n[0]/EConst::Clight();

//DN      StorePhotonHistory(photon);
          if(CheckRcut(param,photon,SURFACE_FS))
            return (SURFACE_FS+RT_OUT_OF_SURFACE1);
	}
    }


  return 0;
}

/* -------------------------------------------------------------------------
 * int read_tel_param(char *filename, tel_param *param, FILE *fplog)
 *
 *  --- read optics parameters from a file
 *
 *  [input value]
 *  filename: the name of the file for optics parameters
 *  param   : pointer to tel_param structure (see also tracemain_optF1v2.h)
 *  fplog   : FILE* for (error) message logging
 *
 *  [return value]
 *  -1 : "(char *)filename" not found
 *   0 : normal end
 * -------------------------------------------------------------------------
 */
int NTraceLens::read_tel_param(char *filename, tel_param *param, FILE *fplog)
{
  FILE *fr;
  char buf[256],item[256],tname[256],tfile[256];
  double val;
  int i;

  /* initialize tel_param structure */
  param->flag_printray=0;

  param->nmat=2;
  param->material[0].name=strdup("VACUUM");
  param->material[0].ndata=1;
  param->material[0].n=(double *)malloc(sizeof(double));
  param->material[0].k=(double *)malloc(sizeof(double));
  param->material[0].n[0]=1.0;
  param->material[0].k[0]=0.0;

  param->material[1].name=strdup("AIR");
  param->material[1].ndata=1;
  param->material[1].n=(double *)malloc(sizeof(double));
  param->material[1].k=(double *)malloc(sizeof(double));
  param->material[1].n[0]=1.0003;
  param->material[1].k[0]=0.0;

  if(( fr=fopen(filename,"r") )==NULL)
    {
      if(fplog)
	fprintf(fplog,"Telescope parameter file %s cannot be opened.\n",filename);
      return -1;
    }
  while(fgets(buf,255,fr))
    {
      if(buf[0]!=0 || buf[0]!='#')
	{
	  sscanf(buf,"%255s",item);
	  switch(item[0])
	    {
	    case 'F':
	      sscanf(buf,"%*s %lf",&val);
	      switch(item[2])
		{
		case '_':
		  if(item[3]=='C')
		    param->FS_C=val;
		  else if(item[3]=='K')
		    param->FS_K=val;
		  else if(item[3]=='O')
		    param->FS_OFFSET=val;
		  break;

		case 'A':
		case 'B':
		case 'C':
		case 'D':
		  param->FS[item[2]-'A']=val;
		  break;
		}
	      break;

	    case 'l':
	      switch(item[1])
		{
		case 'e':
		  sscanf(buf,"%*s %255s",param->lens_dir);
		  break;
		case 'a':
		  sscanf(buf,"%*s %lf",&param->lambda0);
		  break;
		}
	      break;

	    case 'M':
	      if(isdigit(item[1]) && item[1]>='0' && item[1]<'0'+NSURFACES)
		{
		  sscanf(buf,"%*s %256s",tname);
		  for(i=0; i<param->nmat; i++)
		    if(!strcmp(tname,param->material[i].name))
		      {
			param->surface[item[1]-'0'].matid=i;
			break;
		      }
		}
	      else if(item[1]=='a')
		{
		  sscanf(buf,"%*s %256s %256s",tname,tfile);
		  for(i=0;i<param->nmat;i++)
		    if(!strcmp(tname,param->material[i].name))
		      {
			fprintf(stderr,"%s is already registered!\n",tname);
			break;
		      }
		  if(i==param->nmat)
		    {
		      param->material[param->nmat].name=strdup(tname);
		      param->material[param->nmat].filename=strdup(tfile);
		      param->material[param->nmat].zn=NULL;
		      param->material[param->nmat].zk=NULL;
		      param->nmat++;
		    }
		}
	      break;

	    case 'N':
	      sscanf(buf,"%*s %255s",tname);
	      param->name=strdup(tname);
	      break;

	    case 'P':
	      if(!strcmp(item,"PrintRay"))
		param->flag_printray=1;
	      break;

	    case 'R':
	      sscanf(buf,"%*s %lf",&val);
	      switch(item[1])
		{
		case '_':
		  switch(item[2])
		    {
		    case 'L':
              if(strncmp(item,"R_LENS",6)==0)
            {
              if(item[6]&& item[6]>='1' && item[6]<='7')
                param->surface[item[6]-'0'].r_lens=val;
              for(i=0;i<NSURFACES+1;i++)
                param->surface[i].r_lens=val;
            }
		      break;
		    case 'W':
		      param->r_wall=val;
		      break;
		    case 'F':
		      param->r_fs=val;
		      break;
                    case 'c':
		      if(item[5]-'0'<NSURFACES)
			param->surface[item[5]-'0'].r_cut=val;
		      else if(item[5]=='8') /* FS */
			param->r_cut_fs=val;
                      break;
		    }
		  break;

		case 'c':
		  if(isdigit(item[2]) && item[2]>'0' && item[2] < '0'+NSURFACES)
		    param->surface[item[2]-'0'].Rc=val;
		  else if(item[2]=='D') /* for DOE */
		    param->surface[SURFDOE].Rc=val;
		  break;

		case 's':
		  if(isdigit(item[2]) && item[2]>'0' && item[2] < '0'+NSURFACES)
		    param->surface[item[2]-'0'].r_lens=val;
		  else if(item[2]=='D')
		    {
		      param->surface[SURFDOE].r_lens=val;
		      sscanf(buf,"%*s %*f %256s",tfile);
		      param->surface[SURFDOE].filename=strdup(tfile);
		    }
		  break;
		}

	    case 'S':
	      sscanf(buf,"%*s %lf",&val);
	      if(item[1]=='z')
		{
		  if(isdigit(item[2]) && item[2]<'0'+NSURFACES)
		    param->surface[item[2]-'0'].Sz=val;
		  else if(item[2]=='D')
		    param->surface[SURFDOE].Sz=val;
		  if(item[2]>'0' && item[2]<'0'+NSURFACES)
		    {
		      sscanf(buf,"%*s %*f %s",tfile);
		      param->surface[item[2]-'0'].filename=strdup(tfile);
		    }
		}
	      break;

	    default:
	      ; /* skip */
	    }
	}
  }
  fclose(fr);

  for(i=1;i<=7;i++)
    param->surface[i].Sz += param->surface[0].Sz;

  for(i=1;i<8;i++)
    {
      param->surface[i].Zc = param->surface[i].Rc+param->surface[i].Sz;
    }

  param->FS_OFFSET += param->surface[0].Sz;

  param->surface[0].r_cut=3000.0;
  param->surface[5].r_cut=param->surface[4].r_cut;
  param->surface[2].r_cut=param->surface[1].r_cut;
  param->surface[7].r_cut=param->surface[6].r_cut;

  param->surface[2].r_lens=param->surface[1].r_lens;
  param->surface[5].r_lens=param->surface[4].r_lens;
  param->surface[7].r_lens=param->surface[6].r_lens;

  return 0;
}

/* -------------------------------------------------------------------------
 * int print_param(tel_param *p)
 *
 *  --- print optics parameters for debbuging purpose
 *
 *  [input value]
 *  p: pointer to tel_param structure
 *
 *  [return value]
 *  always 0
 *
 * -------------------------------------------------------------------------
 */
int NTraceLens::print_param(tel_param *p)
{
  int i;

  printf("*** JEM-EUSO Optics <%s> ***\n",p->name);
  for(i=1;i<=7;i++)
    printf("r_lens(%d)=%.3f\n",i,p->surface[i].r_lens);
  printf("r_lens(DOE)=%.3f\n\n",p->surface[SURFDOE].r_lens);

  printf("r_wall=%.3f ",p->r_wall);
  printf("r_fs=%.3f\n",p->r_fs);
  for(i=1;i<=7;i++)
    {
      printf("Rc%d=%.5f \t",i,p->surface[i].Rc);
      printf("Sz%d=%.5f \t",i,p->surface[i].Sz);
      printf("Zc%d=%.5f\n",i,p->surface[i].Zc);
    }
  printf("dir=%s\n",p->lens_dir);

  printf("FS_C=%g ",p->FS_C);
  printf("FS_K=%.2f\n",p->FS_K);
  for(i=0;i<4;i++)
    printf("FS%c=%g ",i+'A',p->FS[i]);

  printf("FS_offset=%.2f\n",p->FS_OFFSET);
  for(i=1;i<NSURFACES;i++)
    printf("Rcut%d=%.2f  Rlens=%.2f\n",i,p->surface[i].r_cut,p->surface[i].r_lens);

  printf("\n\n");
  for(i=1;i<7;i++)
    printf("material(S%d):%s\n",i,p->material[p->surface[i].matid].name);

  return 0;
}

/* -------------------------------------------------------------------------
 * double FocalSurface(double r, tel_param *p)
 *
 *    --- calculate the Focal surface height as a function of the radius
 *        from the center
 *        As photons move toward z>0, FS_C should always be negative.
 *
 *  [input value]
 *  r: radius in mm
 *  p: pointer to tel_param structure (optics parameter)
 *
 *  [return value]
 *  relative height of FS to that at r=0 in mm
 *
 * -------------------------------------------------------------------------
 */
/* photons move to the direction z>0 */
double NTraceLens::FocalSurface(double r /* mm */, tel_param *p)  /* mm */
{
  double r2;

  r2=r*r;
  return r2*(p->FS_C /(1.0+sqrt(1.0-(1.0 + p->FS_K)* p->FS_C * p->FS_C *r2))
	     +r2*(p->FS[0] +r2*(p->FS[1]+r2*(p->FS[2] + r2* p->FS[3]))))+ p->FS_OFFSET;
}
