// $Id: OpticsFactory.cc 2849 2010-11-11 11:59:41Z biktem $
// Author: M.Pallavicini, D.Demarco

/*****************************************************************************
 * ESAF: Euso Simulation and Analysis Framework                              *
 *                                                                           *
 *  Id: OpticsFactory                                                        *
 *  Package: Optics                                                          *
 *  Coordinator: Alessandro.Thea                                             *
 *                                                                           *
 *****************************************************************************/

//_____________________________________________________________________________
//
//   Optics Subsystem Factory
//   ========================
//
//   Factory used to create the optical parts of the detector (main optics,
//   photon generator, focalplane and walls).  Using OpticsFactory the
//   DetectorTransportManager is decoupled from the actual inplementation of
//   the various part of the simulation and these can be changed at runtime
//

#include "OpticsFactory.hh"
#include "PolarFocalPlane.hh"
#include "CartesianFocalPlane.hh"
#include "IdealFocalSurface.hh"
#include "JIdealFocalSurface.hh"
#include "KIdealFocalSurface.hh"
#include "LIdealFocalSurface.hh"
#include "MIdealFocalSurface.hh"
#include "NIdealFocalSurface.hh"
#include "TestOpticalAdaptor.hh"
#include "FakeOpticalAdaptor.hh"
#include "IdealOpticalAdaptor.hh"
#include "PipesOpticalAdaptor.hh"
#include "JOpticalSystem.hh"
#include "KOpticalSystem.hh"
#include "LOpticalSystem.hh"
#include "MOpticalSystem.hh"
#include "NOpticalSystem.hh"
#include "ConicBaffle.hh"
#include "ParamOpticalSystem.hh"
#include "DatabaseOpticalSystem.hh"

#include "VirtualDetectorTransportManager.hh"
#include "DetectorTransportManager.hh"
#include "ElecTestDetTransManager.hh"

#ifdef ___USING_GEANT4___
#include "G4TransportManager.hh"
#endif

ClassImp(OpticsFactory)

OpticsFactory *OpticsFactory::fgMe=0;

//______________________________________________________________________________
OpticsFactory *OpticsFactory::Get() {
    //
    // Static getter
    //
    if(fgMe==0)
        fgMe = new OpticsFactory;
    return fgMe;
}

//______________________________________________________________________________
OpticsFactory::OpticsFactory() : fOptics(0) {
    //
    // Constructor
    //

    Build();
}
//______________________________________________________________________________
OpticsFactory::~OpticsFactory() {
    //
    // Destructor
    //
    if(fOptics) delete fOptics;

}

//______________________________________________________________________________
void OpticsFactory::Build() {
    //
    // Build child-parent associations
    //
    Config::Get()->AssociateParent(Conf()->GetStr("OpticsFactory.fWalls"),"WallInteraction");
    Config::Get()->AssociateParent(Conf()->GetStr("OpticsFactory.fBaffle"),"Baffle");

    // Not necessary any more
    // Config::Get()->AssociateParent(Conf()->GetStr("OpticsFactory.fFocalPlane"),"FocalPlane");
    // Config::Get()->AssociateParent(Conf()->GetStr("OpticsFactory.fOpticalSystem"),"OpticalSystem");
    // Config::Get()->AssociateParent(Conf()->GetStr("OpticsFactory.fIdealFocalSurface"),"IdealFocalSurface");
    // Config::Get()->AssociateParent(Conf()->GetStr("OpticsFactory.fOpticalAdaptor"),"OpticalAdaptor");
}

//______________________________________________________________________________
OpticalSystem *OpticsFactory::MakeOpticalSystem() {
    // avoid double creation. For example with DiffusePhotonOnPupil+DetectorTransportManager
    if ( fOptics ) return fOptics;

    string g=Conf()->GetStr("OpticsFactory.fOpticalSystem");
    if( g == "TestOpticalSystem" )          fOptics = new TestOpticalSystem;
    else if ( g == "JOpticalSystem")        fOptics = new JOpticalSystem;
    else if ( g == "KOpticalSystem")        fOptics = new KOpticalSystem;
    else if ( g == "LOpticalSystem")        fOptics = new LOpticalSystem;
    else if ( g == "MOpticalSystem")        fOptics = new MOpticalSystem;
    else if ( g == "NOpticalSystem")        fOptics = new NOpticalSystem;
    else if ( g == "ParamOpticalSystem")    fOptics = new ParamOpticalSystem;
    else if ( g == "DatabaseOpticalSystem") fOptics = new DatabaseOpticalSystem;
    // other opticalsystems go here
    else throw invalid_argument("invalid opticalsystem: "+g);

    return fOptics;
}

//______________________________________________________________________________
FocalPlane *OpticsFactory::MakeFocalPlane() {

    FocalPlane* fp(0);
    string g=Conf()->GetStr("OpticsFactory.fFocalPlane");
    if( g == "TestFocalPlane")            fp = new TestFocalPlane;
    else if( g == "PolarFocalPlane" )     fp = new PolarFocalPlane;
    else if( g == "CartesianFocalPlane" ) fp = new CartesianFocalPlane;
    // other focalplanes go here
    else throw invalid_argument("invalid fFocalPlane: "+g);

    fFocalPlane = fp;
    return fp;
}

//______________________________________________________________________________
IdealFocalSurface *OpticsFactory::MakeIdealFocalSurface() {

    IdealFocalSurface* ifs(0);
    string g=Conf()->GetStr("OpticsFactory.fIdealFocalSurface");
    if(g=="TestIdealFocalSurface")   ifs=new TestIdealFocalSurface;
    else if(g=="JIdealFocalSurface") ifs=new JIdealFocalSurface;
    else if(g=="KIdealFocalSurface") ifs=new KIdealFocalSurface;
    else if(g=="LIdealFocalSurface") ifs=new LIdealFocalSurface;
    else if(g=="MIdealFocalSurface") ifs=new MIdealFocalSurface;
    else if(g=="NIdealFocalSurface") ifs=new NIdealFocalSurface;
    // other focalplanes go here
    else throw invalid_argument("invalid fIdealFocalSurface: "+g);

    return ifs;
}

//______________________________________________________________________________
WallInteraction *OpticsFactory::MakeWalls() {

    WallInteraction* w(0);

    string g=Conf()->GetStr("OpticsFactory.fWalls");
    if(g=="WallInteraction") w=new WallInteraction;
    // other fWalls go here
    else throw invalid_argument("invalid fWalls: "+g);

    return w;
}

//______________________________________________________________________________
Baffle *OpticsFactory::MakeBaffle() {

    Baffle* b(0);
    string g=Conf()->GetStr("OpticsFactory.fBaffle");
    if(g=="TestBaffle")  b = new TestBaffle;
    if(g=="ConicBaffle") b = new ConicBaffle;
    // other baffles go here
    else throw invalid_argument("invalid fBaffle: "+g);

    return b;
}

//______________________________________________________________________________
OpticalAdaptor *OpticsFactory::MakeOA() {

    OpticalAdaptor *oa = NULL;
    string g=Conf()->GetStr("OpticsFactory.fOpticalAdaptor");
    if(g=="TestOpticalAdaptor")       oa = new TestOpticalAdaptor();
    else if(g=="FakeOpticalAdaptor")  oa = new FakeOpticalAdaptor();
    else if(g=="IdealOpticalAdaptor") oa = new IdealOpticalAdaptor();
    else if(g=="PipesOpticalAdaptor") oa = new PipesOpticalAdaptor();
    // other adaptors go here
    else throw invalid_argument("invalid OpticalAdaptor: "+g);

    return oa;
}

//______________________________________________________________________________
VirtualDetectorTransportManager *OpticsFactory::MakeTransportManager() {

    VirtualDetectorTransportManager* tm = NULL;

        string g=Conf()->GetStr("OpticsFactory.fTransportManager");
        if (g=="Standard")            tm = new DetectorTransportManager();
        else if(g=="ElectronicsTest") tm = new ElecTestDetTransManager();
#ifdef ___USING_GEANT4___
        else if(g=="G4TransportManager") tm = new G4TransportManager();
#endif
        else
            throw invalid_argument("invalid DetectorTransportManager: "+g);

        return tm;
}
