// $Id$
// Author: Dmitry V.Naumov   2011/04/27

/*****************************************************************************
 * ESAF: Euso Simulation and Analysis Framework                              *
 *                                                                           *
 *  Id: PixelEffMapBuilder                                                   *
 *  Package: <packagename>                                                   *
 *  Coordinator: <coordinator>                                               *
 *                                                                           *
 *****************************************************************************/

//_____________________________________________________________________________
//
// PixelEffMapBuilder
//
// <extensive class description>
//
//   Config file parameters
//   ======================
//
//   <parameter name>: <parameter description>
//   -Valid options: <available options>
//

#include "PixelEffMapBuilder.hh"
#include "THnSparse.h"
#include "TStopwatch.h"
#include "TRandom.h"
#include "TCanvas.h"
#include "TFile.h"
#include "TStyle.h"
#include <TMath.h>
#include <TBenchmark.h>
#include <TH2F.h>
#include <TTree.h>
#include <TLegend.h>
#include <TLegendEntry.h>
#include <TSeqCollection.h>
#include <TPolyLine.h>

#include "EsafSpectrum.hh"
#include "Photon.hh"
#include "EsafRandom.hh"
#include "OpticalSystem.hh"
// #include "IdealFocalSurface.hh"
#include "Detector.hh"
#include "ElectronicsFactory.hh"
#include "DetectorTransportManager.hh"
#include "EusoElectronics.hh"
#include "DetectorGeometry.hh"
#include "Photomultiplier.hh"
#include "EEvent.hh"
#include "ERunParameters.hh"
using namespace sou;
using namespace std;
ClassImp(PixelEffMapBuilder)

//_____________________________________________________________________________
PixelEffMapBuilder::PixelEffMapBuilder() : EsafConfigurable(), fDetector(0), fTransporter(0), fElectronics(0),
                                           fFile(0), fHeader(0),  fPixelsMap(0), fOpticalSystem(0),
                                           fPhoton(0), fCanvas(0), fTreePrompt(0), fSpectrum(0), fThetaPhiHisto(0) {
  //
  // Constructor
  //
  if (!CheckMemory()) {
      Msg(EsafMsg::Error) << "There is not enough free memory in your system. Less than 100 Mb...Exit" << MsgDispatch;
      return;
  }
  if (!OpenRoot()) {
      return;
  }
  Msg(EsafMsg::Info) << "Building Detector" << MsgDispatch;

  EEvent* fEvent = new EEvent;
  fEvent->SetCurrent(fEvent);
  fHeader->fRunPars = (ERunParameters*)fEvent->GetRunPars()->Clone();
  delete fEvent;

  fDetector = ElectronicsFactory::Get()->MakeDetector();
  fTransporter = fDetector->GetDetectorTransportManager();
  fOpticalSystem = fTransporter->GetOptics();

  fSpectrum = new EsafSpectrum(Conf()->GetStr("PixelEffMapBuilder.fSpectrum").c_str());
  fElectronics = gEusoElectronics;

  Msg(EsafMsg::Debug) << "TransportManager Modules" << MsgDispatch;
  Msg(EsafMsg::Debug) << "Optics: " << fTransporter->GetOptics()->IsA()->GetName() << MsgDispatch;
  Msg(EsafMsg::Debug) << "FocalPlane: " << fTransporter->GetFocalPlane()->IsA()->GetName() << MsgDispatch;
  Msg(EsafMsg::Info) <<  "Building complete" << MsgDispatch;
  fPhoton = new Photon();
  fPixelsMap = new PixelsMap();

}

//_____________________________________________________________________________
PixelEffMapBuilder::~PixelEffMapBuilder() {
    //
    // Destructor
    //
    SafeDelete(fPhoton);
}

//_____________________________________________________________________________
Bool_t PixelEffMapBuilder::OpenRoot() {
  //
  //
  //
  fFileName = Conf()->GetStr("PixelEffMapBuilder.fFileName");

  fFile = new TFile(fFileName+".root","UPDATE");
  if (fFile->IsZombie()) {
      Msg(EsafMsg::Warning) << MsgDispatch;
      Msg(EsafMsg::Warning) << "Error opening file " << fFile->GetName() << MsgDispatch;
      Msg(EsafMsg::Warning) << "Probably it already exists or there is no " << MsgDispatch;
      Msg(EsafMsg::Warning) << "write access permission to this directory" << MsgDispatch;
      Msg(EsafMsg::Warning) << MsgDispatch;
      return kFALSE;
  }

  fTreePrompt = (TTree*)fFile->Get("PixelEffMap");
  if (!fTreePrompt) {
      fTreePrompt = new TTree("PixelEffMap","PixelEffMap");
      fTreePrompt->Branch("ThetaIndex",&fThetaIndexCurrent,"ThetaIndex/I");
      fTreePrompt->Branch("PhiIndex",&fPhiIndexCurrent,"PhiIndex/I");
      fTreePrompt->Branch("ThetaPhiToPixelsMap","PixelEffMapBuilder::PixelsMap",&fPixelsMap);
      fTreePrompt->Branch("ThetaPhiHisto","TH2F",&fThetaPhiHisto);
  }
  else {
    // setup branches
    fTreePrompt->SetBranchAddress("ThetaIndex",&fThetaIndexCurrent);
    fTreePrompt->SetBranchAddress("PhiIndex",&fPhiIndexCurrent);
    fTreePrompt->SetBranchAddress("ThetaPhiToPixelsMap",&fPixelsMap);
    fTreePrompt->SetBranchAddress("ThetaPhiHisto",&fThetaPhiHisto);
  }
  fHeader = (Header*)fTreePrompt->GetUserInfo()->FindObject("PixelEffMapBuilder::Header");

  // now check if header exists
  if ( !fHeader ) {
      Msg(EsafMsg::Info) << "No header found in the rootfile" << fFile->GetName() << MsgDispatch;
      Msg(EsafMsg::Info) << "I assume that this is the first run of PixelEffMapBuilder. If not than there is something wrong!!!"  << MsgDispatch;
      MakeHeader();
  }

  return kTRUE;
}

//_____________________________________________________________________________
Bool_t PixelEffMapBuilder::MakeHeader() {
  //
  //
  //
  fHeader = new Header();
  fHeader->fBinsThetaFoV = Conf()->GetNum("PixelEffMapBuilder.fBinsThetaFoV");
  fHeader->fBinsPhiFoV = Conf()->GetNum("PixelEffMapBuilder.fBinsPhiFoV");
  fHeader->fPhPerBunch = Conf()->GetNum("PixelEffMapBuilder.fPhPerBunch");
  fHeader->fEfficiencyThreshold  = Conf()->GetNum("PixelEffMapBuilder.fEfficiencyThreshold");
  fHeader->fThetaFoVMin = 0;
  fHeader->fThetaFoVMax = 35*TMath::DegToRad();
  fHeader->fPhiFoVMin = 0;
  fHeader->fPhiFoVMax = TMath::TwoPi();
  fTreePrompt->GetUserInfo()->Add(fHeader);

  Msg(EsafMsg::Info) << "Created Header object" << MsgDispatch;
  return kTRUE;
}


//_____________________________________________________________________________
Bool_t PixelEffMapBuilder::CloseRoot() {
  //
  //
  //
  fFile->Close();
  return kTRUE;
}

//_____________________________________________________________________________
Bool_t PixelEffMapBuilder::CheckMemory() {
  //
  //
  //
  gSystem->GetMemInfo(&meminfo);
  if (meminfo.fMemFree < 5) return kFALSE;
  else                        return kTRUE;
}

//_____________________________________________________________________________
Bool_t PixelEffMapBuilder::ProcessAll() {
  //
  // Process all bins over Theta, Phi
  //
  if (fHeader->fAllProcessed) {
      Msg(EsafMsg::Info) << "Everything was already processed. If you want you could delete the output file " << fFile->GetName() << MsgDispatch;
      Msg(EsafMsg::Info) << "and reprocess again. Now close root file and exit." << MsgDispatch;
      if (!CloseRoot()) {
	Msg(EsafMsg::Warning) << "Problem closing the root file " << fFile->GetName() << MsgDispatch;
	return kFALSE;
      }
      else
	return kTRUE;
  }
  else {
    // main loops here

    Float_t delta_theta  = 1.e0/fHeader->fBinsThetaFoV; // non uniform binning over Theta_FoV
    Float_t delta_phi    = (fHeader->fPhiFoVMax-fHeader->fPhiFoVMin)/fHeader->fBinsPhiFoV;

    // calculate here an amount of calls expected for this run
    Int_t ncalls = (fHeader->fBinsThetaFoV-fHeader->fLastProcessedBinTheta-1)*(fHeader->fBinsPhiFoV-fHeader->fLastProcessedBinPhi-1);
    TString jobname = Form ("Job with %g calls", Float_t(ncalls));
    Msg(EsafMsg::Info) << "Starting " << jobname << MsgDispatch;
    TBenchmark gB;
    gB.Start (jobname);

    Int_t bins_processed = 0;
    Int_t progress = 0;
    for (Int_t itheta=0; itheta< fHeader->fBinsThetaFoV; itheta++) {
      fThetaIndexCurrent = itheta;
      if (itheta<=fHeader->fLastProcessedBinTheta) continue;
      for (Int_t iphi=0; iphi<fHeader->fBinsPhiFoV; iphi++) {
	fPhiIndexCurrent = iphi;
	if (iphi<=fHeader->fLastProcessedBinPhi) continue;
	else                                     fHeader->fLastProcessedBinPhi=-1;
        if (iphi%10 == 0) {
	    if ( !CheckMemory()) {
	      Msg(EsafMsg::Warning) << "Low level of free memory. I'd better save info and exit " << MsgDispatch;
	      SetupLastBins(itheta,iphi);
	      SaveInfo();
	      return kTRUE;
	    }
	}
        Float_t t1 = itheta*delta_theta;
	Float_t t2 = (itheta+1)*delta_theta;
        Float_t theta_fov_min = 0.5*acos(1.e0-t1*(1.e0-cos(2*fHeader->fThetaFoVMax)));
	Float_t theta_fov_max = 0.5*acos(1.e0-t2*(1.e0-cos(2*fHeader->fThetaFoVMax)));

        Float_t phi_fov_min   = fHeader->fPhiFoVMin + iphi*delta_phi;
	Float_t phi_fov_max   = fHeader->fPhiFoVMin + (iphi+1)*delta_phi;

	// This is the main processing of photons!
        ProcessPortion(theta_fov_min,theta_fov_max,phi_fov_min,phi_fov_max);
	fTreePrompt->AutoSave();
	bins_processed++;
        if (100.*bins_processed/ncalls >= progress) {
            Msg(EsafMsg::Info).SetProgress(progress);
            Msg(EsafMsg::Info) << "Tracking Photons:  " << MsgCount;
            progress+=10;
        }

      }
    }
    // Everything is processed. Save info.
    SetupLastBins(fHeader->fBinsThetaFoV,fHeader->fBinsPhiFoV);
    SaveInfo();
    Msg(EsafMsg::Info) << "The number of processed pixels above threshold " << fHeader->fNpixels << MsgDispatch;
    gB.Stop(jobname);
    MsgForm (EsafMsg::Info, "Total time:  REAL=%6.2f s   CPU=%6.2f s",gB.GetRealTime (jobname), gB.GetCpuTime (jobname));
  }

  return kTRUE;
}

//_____________________________________________________________________________
Bool_t PixelEffMapBuilder::SaveInfo() {
  //
  // save processed information
  //
  if (fHeader->fLastProcessedBinTheta == fHeader->fBinsThetaFoV && fHeader->fLastProcessedBinPhi == fHeader->fBinsPhiFoV )
    fHeader->fAllProcessed = kTRUE;
  if (fFile->IsWritable()) {
      fFile->cd();
      fTreePrompt->Write();
      fFile->Write();
  }
  else
    return kFALSE;

  if (!CloseRoot()) {
       Msg(EsafMsg::Warning) << "Problem closing the root file " << fFile->GetName() << MsgDispatch;
       return kFALSE;
  }
  return kTRUE;
}

//_____________________________________________________________________________
Bool_t PixelEffMapBuilder::ProcessPortion(Float_t& theta_fov_min, Float_t& theta_fov_max,
					  Float_t& phi_fov_min, Float_t& phi_fov_max) {
  //
  //
  //
  fElectronics->EnableSimulation(kFALSE);
  TRandom* rndm = EsafRandom::Get();
  TVector3 pos;
  TVector3 dir;
  Double_t x,y,z;
  Float_t theta_fov_mean = 0.5*(theta_fov_min+theta_fov_max);
  Float_t phi_fov_mean   = 0.5*(phi_fov_min+phi_fov_max);
  if (fThetaPhiHisto != NULL)
    SafeDelete(fThetaPhiHisto);
  fThetaPhiHisto = new TH2F("photons","",30,theta_fov_min,theta_fov_max,30,phi_fov_min,phi_fov_max);


  for(Long_t i(0); i<fHeader->fPhPerBunch; i++) {
      // new photon position
      RandomPosOnLense(x,y,z);
      pos.SetXYZ(x,y,z);
      // photon direction and wavelength uniform within given narrow bins
      Float_t theta_fov   = theta_fov_min+ rndm->Rndm()*(theta_fov_max-theta_fov_min);
      Float_t phi_fov     = phi_fov_min + rndm->Rndm()*(phi_fov_max-phi_fov_min);
      fThetaPhiHisto->Fill(theta_fov,phi_fov);
      Float_t lambda      = fSpectrum->GetLambda();
      dir.SetMagThetaPhi(1.e0,theta_fov,phi_fov);
      fPhoton->time = 0;
      fPhoton->wl = lambda;
      fPhoton->pos = pos;
      fPhoton->dir = dir;
      fPhoton->iterations = 0;
      fPhoton->hit = false;
      fPhoton->hitIfs = false;
      fPhoton->MChit = false;
      fPhoton->madeSignal = false;
      fPhoton->pixelUid = -1;
      fPhoton->fate = 0;
      fPhoton->parent = NULL;
      fPhoton->history=0;
      fPhoton->id = fHeader->fPhotonsTraced++;
      fTransporter->Transport(fPhoton);
      // Fill info
      if (fPhoton->pixelUid>0) {
	  if (fPixelsMap->fHiredPixels.count(fPhoton->pixelUid) == 0) {
	    fPixelsMap->fHiredPixels[fPhoton->pixelUid] = new TH2F(Form("pixel%d",fPhoton->pixelUid),
								   Form("pixel%d",fPhoton->pixelUid),
								   30,theta_fov_min,theta_fov_max,
								   30,phi_fov_min,phi_fov_max);
	    fPixelsMap->fHiredPixels[fPhoton->pixelUid]->SetXTitle("#theta [rad]");
	    fPixelsMap->fHiredPixels[fPhoton->pixelUid]->SetYTitle("#phi [rad]");
	    fPixelsMap->fHiredPixels[fPhoton->pixelUid]->SetZTitle("efficiency");
	    fPixelsMap->fHiredPixels[fPhoton->pixelUid]->GetXaxis()->SetTitleOffset(2.5);
	    fPixelsMap->fHiredPixels[fPhoton->pixelUid]->GetYaxis()->SetTitleOffset(2.5);
	    fPixelsMap->fHiredPixels[fPhoton->pixelUid]->GetZaxis()->SetTitleOffset(1.5);
	  }
	  fPixelsMap->fHiredPixels[fPhoton->pixelUid]->Fill(theta_fov,phi_fov);
      }
  }
  // analyze the map. Remove low efficiency pixels from it
  map <Int_t, TH2F*>::iterator it;
  vector <Int_t> tobeERASED;
  for (it = fPixelsMap->fHiredPixels.begin(); it !=fPixelsMap->fHiredPixels.end(); it++) {
       TH2F* h = it->second;
       double n = h->GetEntries();
       h->Divide(fThetaPhiHisto);
       h->SetEntries(n);

       Float_t max_eff = h->GetBinContent(h->GetMaximumBin());

       if (max_eff < fHeader->fEfficiencyThreshold) {
	 SafeDelete(h);
	 tobeERASED.push_back(it->first);
       }
       else {
	 fHeader->fTreeMap[it->first].push_back(fTreePrompt->GetEntries());
	 Bool_t found = kFALSE;
	 for (UInt_t i=0; i<fHeader->fProcessedPixels.size(); i++) {
	      if (fHeader->fProcessedPixels[i] == it->first) {found = kTRUE; break;}
	 }
	 if (!found) {
	   fHeader->fProcessedPixels.push_back(it->first);
	 }
       }
  }
  for (UInt_t i(0); i<tobeERASED.size(); i++) {
      fPixelsMap->fHiredPixels.erase(tobeERASED[i]);
      fHeader->fTreeMap.erase(tobeERASED[i]);
  }
  fTreePrompt->Fill();
  map <Int_t, vector<Int_t> >::iterator ll;

  for (it = fPixelsMap->fHiredPixels.begin(); it !=fPixelsMap->fHiredPixels.end(); it++) {
	ll = fHeader->fTreeMap.find(it->first);
	if (ll == fHeader->fTreeMap.end()) {
	  Msg(EsafMsg::Error) << "MISMATCH: NOT FOUND UID: " << it->first << " SHOULD NOT HAPPEN " << MsgDispatch;
	}
  }

  fHeader->fNpixels = fHeader->fProcessedPixels.size();
//   DisplayPixelsFillingPolyLine(theta_fov_mean,phi_fov_mean);
  DisplayPixelsFilling(theta_fov_mean,phi_fov_mean);
  for (it = fPixelsMap->fHiredPixels.begin(); it !=fPixelsMap->fHiredPixels.end(); it++) {
       TH2F* h = it->second;
       SafeDelete(h);
  }
  fPixelsMap->fHiredPixels.clear();
  SafeDelete(fThetaPhiHisto);
  return kTRUE;
}

//______________________________________________________________________________
void PixelEffMapBuilder::SetupLastBins(Int_t& itheta, Int_t& iphi) {
  //
  //
  //
  fHeader->fLastProcessedBinPhi    = iphi;
  fHeader->fLastProcessedBinTheta  = itheta;
}

//______________________________________________________________________________
void PixelEffMapBuilder::RandomPosOnLense(Double_t &x, Double_t &y, Double_t &z) {
  //
  // This method returns a random position on a bottom lense of EUSO.
  //
  TRandom* rndm = EsafRandom::Get();
  Double_t LenseCurvature = fOpticalSystem->GetOpticalSystemParam()->surface[1].Rc;
  Double_t Ycut = fOpticalSystem->GetOpticalSystemParam()->surface[1].r_cut;
  Double_t fDiskRadius = fOpticalSystem->GetOpticalSystemParam()->surface[1].r_lens;
  Double_t zmin = fOpticalSystem->Bottom();
  Double_t CosThetaMin = TMath::Sqrt(pow(LenseCurvature,2) - pow(fDiskRadius,2))/LenseCurvature;

  while (1) {
      Double_t cosTheta = CosThetaMin + rndm->Rndm()*(1-CosThetaMin);
      Double_t phi = TMath::TwoPi()*rndm->Rndm();
      Double_t sinTheta = TMath::Sqrt(1-pow(cosTheta,2));
      y = LenseCurvature*sinTheta*TMath::Sin(phi);
      if (fabs(y) < Ycut) {
	x = LenseCurvature*sinTheta*TMath::Cos(phi);
	z = LenseCurvature*(1-cosTheta) + zmin;
	break;
      }
  }
}



//______________________________________________________________________________
void PixelEffMapBuilder::DisplayPixelsFilling(Float_t& theta_fov, Float_t& phi_fov) {
  //
  //
  //

  if (!fHeader->fDisplayPixels) {
       fHeader->fDisplayPixels = new TH2F("fDisplayPixels","Pixels map",2000,-1000,1000,2000,-1000,1000);
       fHeader->fDisplayPixels->SetStats(kFALSE);
       fHeader->fHistoLegend = new TLegend(0.65,0.75,0.9,0.98,"Info","brNDC");
       fHeader->fHistoLegend->SetBorderSize(1);
       fHeader->fHistoLegend->SetTextAlign(12);
       fHeader->fLegendEntryPhotons = fHeader->fHistoLegend->AddEntry((TObject*)NULL,Form("#gamma traced %d",fHeader->fPhotonsTraced),"F");
       fHeader->fLegendEntryPixels  = fHeader->fHistoLegend->AddEntry((TObject*)NULL,Form("Pixels above threshold %d",fHeader->fNpixels),"F");
       fHeader->fLegendEntryTheta   = fHeader->fHistoLegend->AddEntry((TObject*)NULL,Form("Theta_FoV %g (deg)",theta_fov*TMath::RadToDeg()),"F");
       fHeader->fLegendEntryPhi     = fHeader->fHistoLegend->AddEntry((TObject*)NULL,Form("Phi_FoV %g (deg)",phi_fov*TMath::RadToDeg()),"F");
  }
  if (!fCanvas) {
    fCanvas = new TCanvas();
    fCanvas->SetRightMargin(0.15);
    fCanvas->Update();
    fCanvas->Modified();
  }
  map <Int_t, TH2F*>::iterator it;
  for (it = fPixelsMap->fHiredPixels.begin(); it!= fPixelsMap->fHiredPixels.end(); it++) {
       Photomultiplier* pmt = fElectronics->PmtId(it->first);
       PmtGeometry* geo = pmt->Geometry();
       TVector3 center = geo->Center();
       Float_t max_eff = it->second->GetBinContent(it->second->GetMaximumBin());
       fHeader->fDisplayPixels->Fill(center.X(),center.Y(),max_eff);
  }
  if (fHeader->fPhotonsTraced%fHeader->fPhPerBunch==0) {
    fHeader->fDisplayPixels->Draw("colz");
    fHeader->fLegendEntryPhotons->SetLabel(Form("#gamma traced %d",fHeader->fPhotonsTraced));
    fHeader->fLegendEntryPixels->SetLabel(Form("Pixels above threshold %d",fHeader->fNpixels));
    fHeader->fLegendEntryTheta->SetLabel(Form("Theta_FoV %g (deg)",theta_fov*TMath::RadToDeg()));
    fHeader->fLegendEntryPhi->SetLabel(Form("Phi_FoV %g (deg)",phi_fov*TMath::RadToDeg()));
    fHeader->fHistoLegend->Draw();
    fCanvas->Modified();
    fCanvas->Update();
  }
}

//______________________________________________________________________________
void PixelEffMapBuilder::DisplayPixelsFillingPolyLine(Float_t& theta_fov, Float_t& phi_fov) {
  //
  //
  //
  if (!fCanvas) {
    fCanvas = new TCanvas();
    fCanvas->SetRightMargin(0.15);
  }

  if (!fHeader->fDisplayPixels) {
       Double_t fDiskRadius = fOpticalSystem->GetOpticalSystemParam()->surface[1].r_lens;
       fHeader->fDisplayPixels = new TH2F("fDisplayPixels","Pixels map",300,-fDiskRadius,fDiskRadius,300,-fDiskRadius,fDiskRadius);
       fHeader->fDisplayPixels->SetStats(kFALSE);
       fHeader->fDisplayPixels->Draw();
       fCanvas->Modified();
       fCanvas->Update();
       fHeader->fHistoLegend = new TLegend(0.65,0.75,0.9,0.98,"Events","brNDC");
       fHeader->fHistoLegend->SetBorderSize(1);
       fHeader->fHistoLegend->SetTextAlign(12);
       fHeader->fLegendEntryPhotons = fHeader->fHistoLegend->AddEntry((TObject*)NULL,Form("#gamma traced %d",fHeader->fPhotonsTraced),"F");
       fHeader->fLegendEntryPixels  = fHeader->fHistoLegend->AddEntry((TObject*)NULL,Form("Pixels above threshold %d",fHeader->fNpixels),"F");
       fHeader->fLegendEntryTheta   = fHeader->fHistoLegend->AddEntry((TObject*)NULL,Form("Theta_FoV %g (deg)",theta_fov*TMath::RadToDeg()),"F");
       fHeader->fLegendEntryPhi     = fHeader->fHistoLegend->AddEntry((TObject*)NULL,Form("Phi_FoV %g (deg)",phi_fov*TMath::RadToDeg()),"F");
  }
  map <Int_t, TH2F*>::iterator it;
  for (it = fPixelsMap->fHiredPixels.begin(); it!= fPixelsMap->fHiredPixels.end(); it++) {
      Bool_t drawn=kFALSE;
      TPolyLine* p = GetPixelLine(it->first, drawn);
      if (drawn) continue;
//       p->SetFillColor(38);
      p->SetLineColor(2);
      p->SetLineWidth(0.1);
/*      p->Draw("f");
      p->Draw();*/
  }
  if (fHeader->fPhotonsTraced%fHeader->fPhPerBunch==0) {
     fHeader->fDisplayPixels->Draw("colz");
     fHeader->fLegendEntryPhotons->SetLabel(Form("#gamma traced %d",fHeader->fPhotonsTraced));
     fHeader->fLegendEntryPixels->SetLabel(Form("Pixels above threshold %d",fHeader->fNpixels));
     fHeader->fLegendEntryTheta->SetLabel(Form("Theta_FoV %g (deg)",theta_fov*TMath::RadToDeg()));
     fHeader->fLegendEntryPhi->SetLabel(Form("Phi_FoV %g (deg)",phi_fov*TMath::RadToDeg()));
     fHeader->fHistoLegend->Draw();

     // now draw pixels
     map<Int_t, TPolyLine* >::iterator ip;
     for (ip = fHeader->fPixelLines.begin(); ip != fHeader->fPixelLines.end(); ip++) {
      ip->second->Draw();
     }
     fCanvas->Modified();
     fCanvas->Update();
  }
}

//_____________________________________________________________________________
TPolyLine* PixelEffMapBuilder::GetPixelLine( Int_t uid, Bool_t& drawn) {
    //
    // Returns a TPolyLine corresponding to uid if doesn't exist.
    // If doesn't exists yet in fPixelLines creates a new one.
    //

    if ( fHeader->fPixelLines.count(uid) != 0 ) {
        drawn = kTRUE;
        return fHeader->fPixelLines[uid];
    }
    drawn = kFALSE;
    // projections of the axis of the geometry

    Float_t x[5], y[5];
    for(Int_t iCorner(0); iCorner<4; iCorner++){
        TVector3 dummy = fHeader->fRunPars->PixelCorner(uid,iCorner);//-fViewCenter;
        x[iCorner] = dummy.X();//*fXaxis;
        y[iCorner] = dummy.Y();//*fYaxis;
    }
    x[4] = x[0];
    y[4] = y[0];

    TPolyLine *p = new TPolyLine(5,x,y);
    fHeader->fPixelLines[uid] = p;

    return p;
}



