// implementation of SimuApplication
// $Id: SimuApplication.cc 2863 2011-02-18 11:00:44Z biktem $
// M. Pallavicini - created 28-12-2001
//

#include <iostream>

#include "SimuApplication.hh"
#include "EsafMsgDispatcher.hh"
#include "Config.hh"

#include "Detector.hh"
#include "DetectorGeometry.hh"
#include "EEventGeoFiller.hh"
#include "LightToEuso.hh"
#include "EusoElectronics.hh"
#include "ElectronicsFactory.hh"
#include "LightFactory.hh"
#include "Telemetry.hh"
#include "PhotonsOnPupil.hh"
#include "EsafRandom.hh"
#include "EsafSys.hh"

#include "EEvent.hh"
#include "EDetector.hh"
#include "EHeader.hh"
#include "EAtmosphere.hh"
#include "EShower.hh"

#include <TBenchmark.h>
#include <TTree.h>
#include <TSysEvtHandler.h>
#include <TSystem.h>
#include <TTimeStamp.h>

#include <TObjectTable.h>

ClassImp (SimuApplication)
     using TMath::Max;

// static pointer to SimuApplication
     SimuApplication *gSimuApplication = NULL;

//----- Interrupt signal handler -----------------------------------------------
//______________________________________________________________________________
     class SimuInterruptHandler:public TSignalHandler, EsafMsgSource
     {
     public:
       SimuInterruptHandler ():TSignalHandler (kSigInterrupt, kFALSE)
       {
       }
       Bool_t Notify ();
     };

//______________________________________________________________________________
Bool_t
SimuInterruptHandler::Notify ()
{
  // TRint interrupt handler.

  Msg (EsafMsg::Panic) << "Esaf execution aborted" << MsgDispatch;

  return kTRUE;;
}

//______________________________________________________________________________
SimuApplication::SimuApplication (Int_t argc, char **argv):
TRint ("Euso Detector Simulation", &argc, argv, 0, 0, kTRUE),
fLightToEuso (0), fDetector (0), fLock (0), fEvCount (0), fEvMax (1),
fRunName ("Simu"), fManager (0)
{
  //
  // Constructor.
  // Parse inline parameters, build the detector and electronics
  // reload random seed if required
  //

  // Install interrupt handler
  SimuInterruptHandler *ih = new SimuInterruptHandler;
  ih->Add ();
  SetSignalHandler (ih);

  fWriteTelemetry = kFALSE;
  fWriteRoot = kTRUE;


  //
  // re-build command line and look for local
  //
  string cmd;
  Bool_t isLocal (kFALSE);	//FIXME: this should be handled by a CmdLineParser object.
  for (Int_t i (0); i < argc; i++)
    {
      string str = argv[i];
      cmd += str + " ";
      if (str == "--local")
	isLocal = kTRUE;
    }




  //
  // FIXME: wrong way to decide when the local/global path to esaf has to be used.
  // Is it possible? Or must be done via cmdline?
  //
  const char esafsys[] = "ESAFSYS";
  string esafCfgPath;

  TString esafPath;
  if (gSystem->Getenv (esafsys) && !isLocal)
    esafPath = gSystem->Getenv (esafsys);
  else
    esafPath = gSystem->WorkingDirectory ();

  gEsafSys = new EsafSys (esafPath);


  esafCfgPath = gEsafSys->GetCfgPath ();

  ConfigFileParser::ChangeConfigDir (esafCfgPath);

  // initialize message dispatcher
  EsafMsgDispatcher::Get ()->LoadConfig ();

  // load standard configs
  Config::Get ()->LoadStandardConfigs ();



  Msg (EsafMsg::Debug) << "Simu command line:" << MsgDispatch;
  Msg (EsafMsg::Debug) << cmd << MsgDispatch;

  ParseCmdLine (argc, argv);

  Init ();

  // force all elements of the chain to be built (if required by Config)
  SetChanged (kESAFLight);
  SetChanged (kESAFDetector);


  Int_t len = 80;
  TString head ('_', len);
  try
  {
    Build ();
  }
  catch (exception & e)
  {
    Msg (EsafMsg::Error) << head << MsgDispatch;
    Msg (EsafMsg::Error) << MsgDispatch;
    Msg (EsafMsg::
	 Error) << "An error has occurred during Build()" << MsgDispatch;
    Msg (EsafMsg::Error) << "Message:" << MsgDispatch;
    Msg (EsafMsg::Error) << MsgDispatch;
    Msg (EsafMsg::Error) << "  " << e.what () << MsgDispatch;
    Msg (EsafMsg::Error) << MsgDispatch;
    Msg (EsafMsg::Error) << "ESAF Stopped" << MsgDispatch;
    Msg (EsafMsg::Error) << head << MsgDispatch;
    Msg (EsafMsg::Error) << MsgDispatch;
    exit (999);
  }
  catch (EsafMsg & msg)
  {
    Msg (EsafMsg::Error) << head << MsgDispatch;
    Msg (EsafMsg::Error) << MsgDispatch;
    Msg (EsafMsg::Error) << "An error in " << msg.
      GetSender () << " has occurred during Build()" << MsgDispatch;
    Msg (EsafMsg::Error) << "Message:" << MsgDispatch;
    Msg (EsafMsg::Error) << MsgDispatch;
    Msg (EsafMsg::Error) << "  " << msg.GetText () << MsgDispatch;
    Msg (EsafMsg::Error) << MsgDispatch;
    Msg (EsafMsg::Error) << "ESAF Stopped" << MsgDispatch;
    Msg (EsafMsg::Error) << head << MsgDispatch;
    Msg (EsafMsg::Error) << MsgDispatch;
    exit (999);
  }


  gSimuApplication = this;
}

//______________________________________________________________________________
SimuApplication::~SimuApplication ()
{
  // destructor
  // close output file if open
  // save random engine status

  EndJob ();
  if (fManager)
    {
      delete fManager;
      Msg (EsafMsg::Debug) << "Manager deleted" << MsgDispatch;
    }
}

//______________________________________________________________________________
void
SimuApplication::ParseCmdLine (Int_t argc, char **argv)
{
  //
  // Parse the commandline and set the corresponding variables
  //

  for (Int_t i = 1; i < argc; i++)
    {				//FIXME parser is too stupid....
      if (strlen (argv[i]) >= 2 && (argv[i][0] == '-') && (argv[i][0] != '-'))
	{
	  MsgForm (EsafMsg::Warning, "Simu usage:");
	  MsgForm (EsafMsg::Warning, "");
	  MsgForm (EsafMsg::Warning,
		   "     Simu [-b --events=<n> --seed=<n> --ofile=<fname> \
                    --no-telemetry --no-root --no-files--stdcfg=<cfg> --usrcfg=<cfg> \
                    --ClassName.VarName=val] ");

	}
      else if (strncmp (argv[i], "--local", 7) == 0)
	{
	  Msg (EsafMsg::
	       Info) << "ESAF-Simu running from the Current Working Directory"
	    << MsgDispatch;
	}
      else if (strncmp (argv[i], "--events=", 9) == 0)
	{
	  // Number of events
	  sscanf (argv[i] + 9, "%d", &fEvMax);

	}
      else if (strncmp (argv[i], "--seed=", 7) == 0)
	{
	  // Set random seed
	  string key = "EsafRandom";
	  string value = key + ".fSeed=" + (argv[i] + 7);
	  fUsrKeys.push_back (key);
	  fUsrValues.push_back (value);

	}
      else if (strncmp (argv[i], "--ofile=", 8) == 0)
	{
	  // Superseed output file name

	  string key = "Run";
	  string value = key + ".fFilesPrefix=" + (argv[i] + 8);
	  fUsrKeys.push_back (key);
	  fUsrValues.push_back (value);

	  Msg (EsafMsg::Info) << "Output file name set to: " << (argv[i] +
								 8) <<
	    MsgDispatch;

	}
      else if (strncmp (argv[i], "--odir=", 7) == 0)
	{
	  // Set Custom output directory

	  string key = "Run";
	  string value = key + ".fOutputPath=" + (argv[i] + 7);
	  fUsrKeys.push_back (key);
	  fUsrValues.push_back (value);

	  Msg (EsafMsg::Info) << "Output directory set to: " << (argv[i] +
								 7) <<
	    MsgDispatch;


	}
      else if (strncmp (argv[i], "--runname=", 10) == 0)
	{
	  // Add run name to the truth
	  fRunName = argv[i] + 10;

	}
      else if (strncmp (argv[i], "--no-telemetry", 14) == 0)
	{
	  // Don't write telemery
	  fWriteTelemetry = kFALSE;

	}
      else if (strncmp (argv[i], "--no-root", 8) == 0)
	{
	  // Don't save rootfile
	  fWriteRoot = kFALSE;

	}
      else if (strncmp (argv[i], "--no-files", 8) == 0)
	{
	  // Don't write root AND telemetry
	  fWriteTelemetry = kFALSE;
	  fWriteRoot = kFALSE;

	}
      else if (strncmp (argv[i], "--stdcfg", 8) == 0)
	{
	  //Standard Configuration
	  string s = argv[i] + 8;
	  size_t pos = s.find ('=');
	  if (pos != string::npos)
	    {
	      string stdcfg = s.substr (pos + 1);
	      if (!Config::Get ()->UseConfig (stdcfg))
		FatalError (stdcfg + " standard config not defined");
	    }
	  else
	    FatalError ("Invalid parameter " + string (argv[i]));

	}
      else if (strncmp (argv[i], "--usrcfg", 8) == 0)
	{
	  // User Configuration
	  string s = argv[i] + 7;
	  size_t pos = s.find ('=');
	  if (pos != string::npos)
	    fUsrCfgs.push_back (s.substr (pos + 1));
	  else
	    FatalError ("Invalid parameter " + string (argv[i]));

	}
      else if (strncmp (argv[i], "--", 2) == 0)
	{
	  // Ok, it seems to be user-defined var
	  string s = argv[i] + 2;

	  size_t pos = s.find ('.');
	  if (pos == string::npos)
	    FatalError ("Invalid parameter " + string (argv[i]));
	  else
	    // user defined variable syntax
	    // ClassName::VarName
	    fUsrKeys.push_back (s.substr (0, pos));
	  fUsrValues.push_back (s);

	}
    }

  // load user definition file first
  vector < string >::const_iterator itCfg;
  for (itCfg = fUsrCfgs.begin (); itCfg != fUsrCfgs.end (); itCfg++)
    Config::Get ()->LoadUserConfig (itCfg->c_str ());

  // the apply custom parameters
  for (size_t i (0); i < fUsrKeys.size (); i++)
    {
      Config::Get ()->AddUserDefinition (fUsrKeys[i], fUsrValues[i]);
    }
}

//______________________________________________________________________________
void
SimuApplication::Init ()
{
  //
  // Initialize Config and the message system
  //

  // initialize the file names.
  fOutPrefix =
    Config::Get ()->GetCF ("General",
			   "Run")->GetStr ("Run.fFilesPrefix").c_str ();

  TString outdir =
    Config::Get ()->GetCF ("General",
			   "Run")->GetStr ("Run.fOutputPath").c_str ();

  gEsafSys->SetOutPath (outdir);

  fFileNames[kLog] =
    gEsafSys->AbsolutePath (gEsafSys->GetOutPath () + fOutPrefix + ".log");
  fFileNames[kDump] =
    gEsafSys->AbsolutePath (gEsafSys->GetOutPath () + fOutPrefix +
			    ".ConfigDump.cfg");
  fFileNames[kRoot] =
    gEsafSys->AbsolutePath (gEsafSys->GetOutPath () + fOutPrefix + ".root");
  fFileNames[kTelemetry] =
    gEsafSys->AbsolutePath (gEsafSys->GetOutPath () + fOutPrefix +
			    ".Telemetry.dat.gz");

  // open logfile
  EsafMsgDispatcher::Get ()->OpenLog (fFileNames[kLog]);

  Msg (EsafMsg::Info) << "Logging to file " + fFileNames[kLog] +
    " started" << MsgDispatch;

  // header data
  UInt_t len = 80;

  TString head1 ('_', 80);
  head1[0] = ' ';
  head1[len - 1] = ' ';
  TString head2 (' ', 80);
  head2[0] = '|';
  head2[len - 1] = '|';
  TString head3 ('_', 80);
  head3[0] = '|';
  head3[len - 1] = '|';
  TString line, msg;

  Msg (EsafMsg::Info) << head1 << MsgDispatch;
  Msg (EsafMsg::Info) << head2 << MsgDispatch;

  line = head2;
  msg = "ESAF Run Starting";
  line.Replace (3, msg.Length (), msg.Data (), msg.Length ());

  Msg (EsafMsg::Info) << line << MsgDispatch;

  TTimeStamp tStart;
  line = head2;
  msg = Form ("Local time: %s", tStart.AsString ("l"));
  line.Replace (5, msg.Length (), msg.Data (), msg.Length ());

  Msg (EsafMsg::Info) << line << MsgDispatch;

  Msg (EsafMsg::Info) << head3 << MsgDispatch;
  Msg (EsafMsg::Info) << MsgDispatch;


  MsgForm (EsafMsg::Info, "ESAF Environment:");
  Msg (EsafMsg::Info) << MsgDispatch;
  MsgForm (EsafMsg::Info, "   Current Directory: %s",
	   gSystem->WorkingDirectory ());
  MsgForm (EsafMsg::Info, "   Simu Config Path : %s",
	   gEsafSys->GetCfgPath ());
  MsgForm (EsafMsg::Info, "   Simu Auxilar Path: %s",
	   gEsafSys->GetAuxPath ());
  MsgForm (EsafMsg::Info, "   Simu Tmp Path    : %s",
	   gEsafSys->GetTmpPath ());
  MsgForm (EsafMsg::Info, "   Simu Output Path : %s",
	   gEsafSys->GetOutPath ());
  MsgForm (EsafMsg::Info, "   Host Machine     : %s", gSystem->HostName ());
  Msg (EsafMsg::Info) << MsgDispatch;


  // set config dump
  Config::Get ()->SetCfgDump (fFileNames[kDump].c_str ());

  vector < string >::const_iterator itCfg;
  for (itCfg = fUsrCfgs.begin (); itCfg != fUsrCfgs.end (); itCfg++)
    Msg (EsafMsg::Debug) << "Using custom config: " + (*itCfg) << MsgDispatch;

  // add the user defined variables to the logfile
  for (size_t i (0); i < fUsrKeys.size (); i++)
    Msg (EsafMsg::Debug) << "Cmd line option: class " +
      fUsrKeys[i] + ", value " + fUsrValues[i] << MsgDispatch;



}

//______________________________________________________________________________
void
SimuApplication::EndJob ()
{
  // final clean up

  // delete Euso detector if it exists
  if (Euso ())
    {
      delete fDetector;
      fDetector = NULL;
    }

  // delete LightToEuso
  if (Light ())
    {
      delete fLightToEuso;
      fLightToEuso = NULL;
    }

  // save trees and close file
  if (Manager ())
    Manager ()->Close ();

  // save standard config if required
  Config::Get ()->SaveConfig ();

  UInt_t len = 80;

  TString head1 ('_', 80);
  head1[0] = ' ';
  head1[len - 1] = ' ';
  TString head2 (' ', 80);
  head2[0] = '|';
  head2[len - 1] = '|';
  TString head3 ('_', 80);
  head3[0] = '|';
  head3[len - 1] = '|';
  TString line, msg;

  Msg (EsafMsg::Info) << head1 << MsgDispatch;
  Msg (EsafMsg::Info) << head2 << MsgDispatch;

  line = head2;
  msg = "ESAF Run Completed";
  line.Replace (3, msg.Length (), msg.Data (), msg.Length ());

  Msg (EsafMsg::Info) << line << MsgDispatch;

  TTimeStamp tEnd;
  line = head2;
  msg = Form ("Local time: %s", tEnd.AsString ("l"));
  line.Replace (5, msg.Length (), msg.Data (), msg.Length ());

  Msg (EsafMsg::Info) << line << MsgDispatch;

  Msg (EsafMsg::Info) << head3 << MsgDispatch;
  Msg (EsafMsg::Info) << MsgDispatch;

}


//______________________________________________________________________________
void
SimuApplication::Build ()
{
  // build (or rebuild) application chain

  //FIXME: When Config is resetted all ESAF modules must be rebuilt due to
  //   the ConfigFileParser pointers.
  //   Config::Reset must be modified to reset just one module at time.

  Config::Get ()->Reset (kFALSE);

  SafeDelete (fManager);
  if (fWriteRoot)
    fManager = new SimuRootFileManager ();


  // rebuild euso detector if required
  // set gEusoDetector to right value (global pointer defined in EusoDetector.hh)
  if (HasChanged (kESAFDetector))
    {
      if (fDetector)
	delete fDetector;
      fDetector = NULL;

      // parameters are filled
      if (Manager ())
	Manager ()->ResetRunPars ();
      fDetector = ElectronicsFactory::Get ()->MakeDetector ();

      DetectorGeometry *detGeo = fDetector->GetGeometry ();
      ConfigFileParser *p = Config::Get ()->GetCF ("General", "Euso");
      Double_t alt = p->GetNum ("Euso.fAltitude") * sou::km;
      TVector3 pos (0, 0, alt);
      detGeo->SetPos (pos);

      /*             // Detector's orientation
       *             // diabled until further tests.
       *
       *             Double_t phi, theta, psi;
       *             phi = p->GetNum("Euso.fXPhi");
       *             theta = p->GetNum("Euso.fXTheta");
       *             psi = p->GetNum("Euso.fXPsi");
       *             geo->SetXEulerRotation(phi, theta, psi);
       */

      SetChanged (kESAFDetector, kFALSE);
    }

  // rebuild shower generator if required
  if (HasChanged (kESAFLight))
    {
      if (fLightToEuso)
	delete fLightToEuso;
      fLightToEuso = NULL;
      fLightToEuso = LightFactory::Get ()->GetLightToEuso ();

      SetChanged (kESAFLight, kFALSE);
    }
}


//______________________________________________________________________________
void
SimuApplication::Help (const char *line)
{
  TApplication::Help (line);
}

//______________________________________________________________________________
Bool_t
SimuApplication::DoEvent ()
{
  // simulate and (if required) display a single event
  // WriteTelemetry=true     : writes in Telemetry (ASCII) file


  UInt_t len = 80;
  TString divider ('_', len);
  TString head1 ('_', len);
  head1[0] = ' ';
  head1[len - 1] = ' ';
  TString head2 (' ', len);
  head2[0] = '|';
  head2[len - 1] = '|';
  TString head3 ('_', len);
  head3[0] = '|';
  head3[len - 1] = '|';
  TString line, msg;

  msg = Form ("Event %d Started", fEvCount);
  line = head2;
  line.Replace (3, msg.Length (), msg.Data (), msg.Length ());

  Msg (EsafMsg::Info) << head1 << MsgDispatch;
  Msg (EsafMsg::Info) << head2 << MsgDispatch;
  Msg (EsafMsg::Info) << line << MsgDispatch;
  Msg (EsafMsg::Info) << head2 << MsgDispatch;
  Msg (EsafMsg::Info) << MsgDispatch;

  // time info
  TString s = Form ("Event %d", fEvCount);
  TBenchmark gB;
  gB.Start (s);

  if (isLocked ())
    {
      MsgForm (EsafMsg::Warning,
	       "Cannot run because application is locked\n");
      MsgForm (EsafMsg::Warning,
	       "Close configuration windows before running\n");
      return kFALSE;
    }

  try
  {
    Euso ()->ClearMemory ();
    if (Light ())
      Light ()->ClearMemory ();
    // clear root event
    if (Event ())
      {
	Event ()->ClearAndShrink ("C");
	Event ()->Build (GetRunNumber (), fEvCount, fRunName);
	// save the status of the random number generator before simulating the new event
	Event ()->GetHeader ()->SetRandom (EsafRandom::Get ());


	if (Euso ())
	  {
	    EEventGeoFiller a (Euso ()->GetGeometry ());
	    Event ()->Fill (a);
	  }
      }
  }
  catch (exception & e)
  {

    Msg (EsafMsg::Warning) << divider << endl;
    Msg (EsafMsg::Warning) << MsgDispatch;
    Msg (EsafMsg::
	 Error) <<
      "An exception without sender has been catched while cleaning memory." <<
      MsgDispatch;
    Msg (EsafMsg::Error) << "Rethrowing..." << MsgDispatch;
    Msg (EsafMsg::Warning) << MsgDispatch;
    throw e;


  }
  catch (EsafMsg & msg)
  {

    if (msg.GetSeverity () >= EsafMsg::Panic)
      throw runtime_error (msg.GetText ());

    string msgtxt = msg.GetText ();
    Msg (EsafMsg::Warning) << divider << endl;
    Msg (EsafMsg::Warning) << MsgDispatch;
    Msg (EsafMsg::Warning) << "An Exception message sent by " << msg.
      GetSender () << " has been catched while cleaning memory" <<
      MsgDispatch;
    Msg (EsafMsg::Warning) << "Message:" << MsgDispatch;
    Msg (EsafMsg::Warning) << MsgDispatch;
    Msg (EsafMsg::Warning) << "  " << msgtxt << MsgDispatch;
    Msg (EsafMsg::Warning) << MsgDispatch;

    Msg (EsafMsg::Warning) << "Event status set to failed" << MsgDispatch;
    if (Event ())
      Event ()->GetHeader ()->SetStatus (EHeader::kFailed, msgtxt.c_str ());
    MsgForm (EsafMsg::Warning, "Exception message saved into Event %d",
	     fEvCount);
    Msg (EsafMsg::Warning) << divider << endl;
    Msg (EsafMsg::Warning) << MsgDispatch;

  }

  Bool_t failed = kFALSE;

  // event generation and radiative transport through atmosphere
  MCTruth *truth = 0;
  PhotonsOnPupil *photons = 0;
  Telemetry *data = 0;

  try
  {
    if (Light () && Euso ())
      {
	Light ()->Reset ();
	photons = Light ()->Get (Euso ()->GetGeometry ());
	if (photons == NULL)
	  {
	    MsgForm (EsafMsg::Info, "PhotonsOnPupil object is null");
	    return kTRUE;
	  }
	truth = Light ()->GetTruth ();
      }
    else
      MsgForm (EsafMsg::Info, "LightToEuso is null.");

    // detector response simulation
    if ((Euso () && photons && Light ()) || (Euso () && (!Light ())))
      {
	Euso ()->Reset ();
	data = Euso ()->Get (photons);
      }
  }
  catch (exception & e)
  {

    failed = kTRUE;
    Msg (EsafMsg::Warning) << divider << endl;
    Msg (EsafMsg::Warning) << MsgDispatch;
    Msg (EsafMsg::
	 Error) << "An exception without sender has been catched." <<
      MsgDispatch;

    Msg (EsafMsg::Error) << "Message:" << MsgDispatch;
    Msg (EsafMsg::Error) << "  " << e.what () << MsgDispatch;
    Msg (EsafMsg::Error) << "Rethrowing..." << MsgDispatch;
    Msg (EsafMsg::Warning) << MsgDispatch;
    throw e;


  }
  catch (EsafMsg & msg)
  {

    if (msg.GetSeverity () >= EsafMsg::Panic)
      throw runtime_error (msg.GetText ());

    string msgtxt = msg.GetText ();
    failed = kTRUE;
    Msg (EsafMsg::Warning) << divider << endl;
    Msg (EsafMsg::Warning) << MsgDispatch;
    Msg (EsafMsg::Warning) << "An Exception message sent by " << msg.
      GetSender () << " has been catched" << MsgDispatch;
    Msg (EsafMsg::Warning) << "Message:" << MsgDispatch;
    Msg (EsafMsg::Warning) << MsgDispatch;
    Msg (EsafMsg::Warning) << "  " << msgtxt << MsgDispatch;
    Msg (EsafMsg::Warning) << MsgDispatch;

    Msg (EsafMsg::Warning) << "Event status set to failed" << MsgDispatch;
    if (Event ())
      Event ()->GetHeader ()->SetStatus (EHeader::kFailed, msgtxt.c_str ());
    MsgForm (EsafMsg::Warning, "Exception message saved into Event %d",
	     fEvCount);
    Msg (EsafMsg::Warning) << divider << endl;
    Msg (EsafMsg::Warning) << MsgDispatch;

  }

  // dump time info on screen
  gB.Stop (s);

  msg = Form ("Event %d %s, REAL=%6.2f s  CPU=%6.2f s",
	      fEvCount, failed ? "Failed" : "Completed",
	      gB.GetRealTime (s), gB.GetCpuTime (s));
  line.Replace (3, msg.Length (), msg.Data (), msg.Length ());

  Msg (EsafMsg::Info) << MsgDispatch;
  Msg (EsafMsg::Info) << head2 << MsgDispatch;
  Msg (EsafMsg::Info) << line << MsgDispatch;
  Msg (EsafMsg::Info) << head3 << MsgDispatch;
  Msg (EsafMsg::Info) << MsgDispatch;

  // write output data file (Telemetry format)
  if (data && fWriteTelemetry)
    {
      data->SetTruth (truth);

      if (!data->IsOpen ())
	data->Open (fFileNames[kTelemetry].c_str ());
      data->Write ();
    }

  // fill root file if enabled
  if (Manager ())
    Manager ()->FillEvent ();

  fEvCount++;
  return kTRUE;


}

//______________________________________________________________________________
Bool_t
SimuApplication::DoAll ()
{
  //
  // Simulate all events available from the data source
  // DoEvent returns kFalse in case of error so the run will stop
  // if fEvMax is -1 it never stops until error or end of file
  //

  Int_t ev = fEvMax;
  ResetEvCounter ();

  // OpenFile
  if (Manager ())
    Manager ()->Open (fFileNames[kRoot]);
  if (Event ())
    Event ()->GetHeader ()->SetRandom (EsafRandom::Get ());

  // Lock the Event
  if (Manager ())
    Manager ()->LockEvent ();

  ////////////////////////////////////////////////////////

  TString jobname = Form ("Job with %d events", fEvMax);
  TBenchmark gB;
  gB.Start (jobname);
  SetRunNumber ((Int_t) Config::Get ()->GetCF ("General", "Run")->
		GetNum ("Run.fRunNumber"));
  try
  {
    while (ev--)
      {
	if (!DoEvent ())
	  return kFALSE;
      }
  }
  catch (exception & e)
  {
    UInt_t len = 80;
    TString head ('_', len);

    Msg (EsafMsg::Error) << head << MsgDispatch;
    Msg (EsafMsg::Error) << MsgDispatch;
    Msg (EsafMsg::Error) << "A Critical error has occurred" << MsgDispatch;
    Msg (EsafMsg::Error) << "Message:" << MsgDispatch;
    Msg (EsafMsg::Error) << MsgDispatch;
    Msg (EsafMsg::Error) << "  " << e.what () << MsgDispatch;
    Msg (EsafMsg::Error) << MsgDispatch;
    Msg (EsafMsg::Error) << "EventLoop stopped" << MsgDispatch;
    Msg (EsafMsg::Error) << head << MsgDispatch;
    Msg (EsafMsg::Error) << MsgDispatch;
  }
  gB.Stop (jobname);
  MsgForm (EsafMsg::Info,
	   "Average time per event:  REAL=%6.2f s   CPU=%6.2f s",
	   gB.GetRealTime (jobname) / fEvMax,
	   (gB.GetCpuTime (jobname) / fEvMax));
  MsgForm (EsafMsg::Info, "Total time:  REAL=%6.2f s   CPU=%6.2f s",
	   gB.GetRealTime (jobname), gB.GetCpuTime (jobname));

  // Unlock Event
  if (Manager ())
    Manager ()->UnlockEvent ();
  // Close File
  if (Manager ())
    Manager ()->Close ();
  //
  return kTRUE;
}

//______________________________________________________________________________
void
SimuApplication::SetLock (Bool_t lk)
{
  //
  // Lock the application.
  // If locked, events are not processed
  //

  if (lk)
    fLock++;
  else
    fLock--;
  if (fLock < 0)
    fLock = 0;
}

//______________________________________________________________________________
Bool_t
SimuApplication::DumpLastEvent (string * buffer)
{
  //
  // dump event on stdout or into a buffer
  // buffer must be large enough
  // if buffer is NULL, dump on cout
  // return kFALSE if no data is available
  //

  //  if ( ! E()->Data() )
  //    return kFALSE;
  //  if ( buffer )
  //    *buffer = E()->Data()->Dump();
  //  else {
  //    cout << endl << endl;
  //    cout << "EUSO DETECTOR SIMULATION DUMP.   Event=" << EventCounter() << endl;
  //    cout << E()->Data()->Dump() << endl << endl;
  //    cout.flush();
  //  }
  return kTRUE;
}

//
////______________________________________________________________________________
//void SimuApplication::EnableModule(EsafElement e, Bool_t val) {
//    fEnModule[e]=val;
//    SetChanged(e);
//    //Build();
//}

//______________________________________________________________________________
void
SimuApplication::ResetCopies ()
{
  // reset list of event copies
  // used for re-rerunning the same event with different parameters

  for (Int_t i = 0; i < fEventCopies.GetEntriesFast (); i++)
    {
      EEvent *pEv = (EEvent *) (fEventCopies[i]);
      if (pEv)
	delete pEv;
    }
}
