// $Id: PhPRootFileLightToEuso.cc 2792 2008-02-11 07:42:14Z naumov $
// Author: Alessandro Thea   2005/02/14

/*****************************************************************************
 * ESAF: Euso Simulation and Analysis Framework                              *
 *                                                                           *
 *  Id: PhPRootFileLightToEuso                                               *
 *  Package: <packagename>                                                   *
 *  Coordinator: <coordinator>                                               *
 *                                                                           *
 *****************************************************************************/

//_____________________________________________________________________________
//
// PhPRootFileLightToEuso
//
// <extensive class description>
//
//   Config file parameters
//   ======================
//
//   <parameter name>: <parameter description>
//   -Valid options: <available options>
//

#include "PhPRootFileLightToEuso.hh"
#include "DetectorGeometry.hh"
#include "EsafRefFrame.hh"
#include "MCTruth.hh"
#include "ETruth.hh"
#include "EGeometry.hh"
#include "ListPhotonsOnPupil.hh"
#include "EPhoton.hh"
#include "EEvent.hh"
#include "EEventTruthAdder.hh"

#include <TTree.h>
#include <TFile.h>
#include <TSystem.h>
#include <TClonesArray.h>
#include <TMath.h>

// DEBUG
// #include "stdio.h"
// DEBUG

ClassImp(PhPRootFileLightToEuso)

using namespace TMath;

//_____________________________________________________________________________
PhPRootFileLightToEuso::PhPRootFileLightToEuso() : LightToEuso("PHPROOTFILE"),
    fRootTruth(0), fRootGeometry(0), fNumPhotons(0), fPhotons(0) {
    //
    // Constructor
    //

    fFileName = Conf()->GetStrPath("PhPRootFileLightToEuso.fFileName");
    fFirstEvent = (Long64_t)Conf()->GetNum("PhPRootFileLightToEuso.fFirstEvent") ;

    fMCTruth = new MCTruth;
    fPhotonsOnPupil = new ListPhotonsOnPupil();

    Init();
}

//_____________________________________________________________________________
PhPRootFileLightToEuso::~PhPRootFileLightToEuso() {
    //
    // Destructor
    //

    SafeDelete( fFile );
    SafeDelete( fPhotonsOnPupil );
    SafeDelete( fMCTruth );
    SafeDelete( fRootTruth );
    SafeDelete( fPhotons );
}


//______________________________________________________________________________
Bool_t PhPRootFileLightToEuso::ClearMemory() {
    //
    //
    //

    Reset();


    return fPhotonsOnPupil ? fPhotonsOnPupil->ClearMemory() : kTRUE;
}

//______________________________________________________________________________
void PhPRootFileLightToEuso::Init() {
    //
    // Open file and associate objects
    //
    
    string ext = ".php.root";
    if (fFileName.rfind(ext) != (fFileName.size()-ext.size()))
        fFileName += ext;

    fFile =  new TFile(fFileName.c_str());
    if ( fFile->IsZombie() )
        FatalError("Unable to open "+fFileName);

    if (!(fTree = (TTree*)fFile->Get("PhPTree"))) 
        FatalError("No PhPTree in "+fFileName);
    
    fTree->SetBranchAddress("EPhotons",&fPhotons);
    fTree->SetBranchAddress("ENumPhotons",&fNumPhotons);
    fTree->SetBranchAddress("ETruth",&fRootTruth);

  
    Bool_t forceFrame = Conf()->GetBool("PhPRootFileLightToEuso.fForceFrame");
    if ( !forceFrame ) {
        if ( !fTree->FindBranch("EGeometry") )
            Msg(EsafMsg::Panic) << "Geometry not saved in the PhPRootFile. Use fForceFrame" << MsgDispatch;
        else {
            fTree->SetBranchAddress("EGeometry",&fRootGeometry);
            EsafRefFrame *frame = new EsafRefFrame;
            fPhotonsOnPupil->AddFrame(frame);
        }

    } else {
        TVector3 pos;
        pos[2] = Conf()->GetNum("PhPRootFileLightToEuso.fAltitude")*sou::km;

        Double_t phi, theta, psi;
        //
        // angles in degrees or radians?
        //
        phi = Conf()->GetNum("PhPRootFileLightToEuso.fXPhi");
        theta = Conf()->GetNum("PhPRootFileLightToEuso.fXTheta");
        psi = Conf()->GetNum("PhPRootFileLightToEuso.fXPsi");

        EsafRefFrame *frame = new EsafRefFrame;
        frame->SetPos(pos);
        frame->SetXEulerRotation(phi,theta,psi);
        fPhotonsOnPupil->AddFrame(frame);
    }
    fEvCounter = fFirstEvent;
}

//______________________________________________________________________________
PhotonsOnPupil* PhPRootFileLightToEuso::Get(const DetectorGeometry* dg) {
    //
    // Reads file and creates the list of photons
    //

    fPhotonsOnPupil->Clear();
    fMCTruth->Clear();

    if ( fEvCounter >= fTree->GetEntriesFast() )
        return 0;

    fTree->GetEntry(fEvCounter);
    if ( fNumPhotons != fPhotons->GetEntries())
        MsgForm(EsafMsg::Warning,"Mismatch  between EPhotons size and ENumPhotons",
                fNumPhotons, fPhotons->GetEntries());

    TVector3 v(0,0,0);
    Int_t nTransmitted = 0;

    // DEBUG
    // FILE* flog = fopen("php-out.log","w");
    // DEBUG

    for( Int_t i(0); i<fNumPhotons; i++) {

        EPhoton* ph = (EPhoton*)fPhotons->At(i);
        if (ph->GetPos().Mag() > dg->GetRadius() ) continue;

        ParentPhoton* parent = new ParentPhoton( i, v, ph->GetPos(), ph->GetTheta(),
                ph->GetPhi(), ph->GetLambda(), ph->GetTime(), (ParentPhotonType)ph->GetType()); 
        fPhotonsOnPupil->Add(parent);
        nTransmitted++;

        //Photon& p=parent->GetPhoton();
        // DEBUG
        // fprintf(flog,"%d   %.3f   %.3f   %.3f   %.3f   %.3f   %.3f   %.3f\n",
        //         p.parent->Type(), p.dir.Theta()/deg, p.dir.Phi()/deg, p.wl/nm, p.time/ns, 
        //        p.pos.X(), p.pos.Y(), p.pos.Z());
        // DEBUG
    } 
    // DEBUG
    // fclose(flog);
    // DEBUG

    MsgForm(EsafMsg::Info,"PhP event %d read", fEvCounter);
    MsgForm(EsafMsg::Info,"   Energy: %.3g MeV ParticleCode: %d",
            fRootTruth->GetTrueEnergy(), fRootTruth->GetTrueParticleCode());
    MsgForm(EsafMsg::Info,"   Theta: %.3f Phi: %.3f", fRootTruth->GetTrueTheta()*RadToDeg(),
            fRootTruth->GetTruePhi()*RadToDeg());
    MsgForm(EsafMsg::Info,"   NumPhotons %d, Transmitted %d",
            fNumPhotons, nTransmitted);

    // copy ETruth in MCTruth
    fMCTruth->SetEnergy( fRootTruth->GetTrueEnergy() );
    fMCTruth->SetThetaPhi( fRootTruth->GetTrueTheta(), fRootTruth->GetTruePhi() );
    fMCTruth->SetParticle( fRootTruth->GetTrueParticleCode() );
    fMCTruth->SetFirstInt( fRootTruth->GetTrueInitPos(), fRootTruth->GetTrueX1() );
    fMCTruth->SetEarthImpact( fRootTruth->GetTrueEarthImpact(), fRootTruth->GetTrueEarthAge() );
    fMCTruth->SetShowerMax( fRootTruth->GetTrueShowerMaxPos(), fRootTruth->GetTrueShowerXMax() );
    fMCTruth->SetHclouds( fRootTruth->GetTrueHclouds() );
    fMCTruth->SetCloudsthick( fRootTruth->GetTrueCloudsthick() );
    fMCTruth->SetCloudsOD( fRootTruth->GetTrueCloudsOD() );
    fMCTruth->SetLatitude( fRootTruth->GetTrueLatitude() );
    fMCTruth->SetLongitude( fRootTruth->GetTrueLongitude() );
    fMCTruth->SetDate( fRootTruth->GetTrueDate() );

    if ( fRootGeometry ) {
        EsafRefFrame* frame = fPhotonsOnPupil->GetFrame();
        frame->SetPos(fRootGeometry->GetPos());
        const  TRotation rot = fRootGeometry->GetRotation();
        frame->SetXEulerRotation(rot.GetXPhi(), rot.GetXTheta(), rot.GetXPsi());
    }

    // save the truth
    EEvent* ev = EEvent::GetCurrent();
    if ( ev ) {
        EEventTruthAdder a( fMCTruth );
        ev->Fill(a);
    }

    fEvCounter++;

    return fPhotonsOnPupil;
}
