// $Id: SinglePhotonRootFileLightToEuso.cc 2779 2006-12-21 15:23:54Z moreggia $
// Author: Sylvain Moreggia   2006/06/01

/*****************************************************************************
 * ESAF: Euso Simulation and Analysis Framework                              *
 *                                                                           *
 *  Id: SinglePhotonRootFileLightToEuso                                      *
 *  Package: <packagename>                                                   *
 *  Coordinator: <coordinator>                                               *
 *                                                                           *
 *****************************************************************************/

//_____________________________________________________________________________
//
// SinglePhotonRootFileLightToEuso
//
// <extensive class description>
//
//   Config file parameters
//   ======================
//
//   <parameter name>: <parameter description>
//   -Valid options: <available options>
//

#include "SinglePhotonRootFileLightToEuso.hh"
#include "DetectorGeometry.hh"
#include "MCTruth.hh"
#include "EGeometry.hh"
#include "ETruth.hh"
#include "ListPhotonsOnPupil.hh"
#include "ESinglePhoton.hh"
#include "EAtmosphere.hh"
#include "EEvent.hh"
#include "EEventTruthAdder.hh"
#include "EsafRandom.hh"
#include "SinglePhoton.hh"
#include "Config.hh"

#include <TTree.h>
#include <TFile.h>
#include <TSystem.h>
#include <TClonesArray.h>
#include <TMath.h>

using namespace TMath;

ClassImp(SinglePhotonRootFileLightToEuso)

//_____________________________________________________________________________
SinglePhotonRootFileLightToEuso::SinglePhotonRootFileLightToEuso(): LightToEuso("SINGLEPHROOTFILE"),
    fEv(0), fRootTruth(0), fNumPhotons(0), fRootGeometry(0), fAtmosphere(0) {
    //
    // Constructor
    //

    fFileName = Conf()->GetStrPath("SinglePhotonRootFileLightToEuso.fFileName");
    fFirstEvent = (Long64_t)Conf()->GetNum("SinglePhotonRootFileLightToEuso.fFirstEvent") ;
    fNewPhFormat = Conf()->GetBool("SinglePhotonRootFileLightToEuso.fNewPhFormat");
    fSeveralEvents = Conf()->GetBool("SinglePhotonRootFileLightToEuso.fSeveralEvents");

    fEv = new EEvent();
    fMCTruth = new MCTruth;
    fPhotonsOnPupil = new ListPhotonsOnPupil();

    Init();
}

//_____________________________________________________________________________
SinglePhotonRootFileLightToEuso::~SinglePhotonRootFileLightToEuso() {
    //
    // Destructor
    //

    SafeDelete( fFile );
    SafeDelete( fPhotonsOnPupil );
    SafeDelete( fMCTruth );
    SafeDelete( fEv );
}


//______________________________________________________________________________
Bool_t SinglePhotonRootFileLightToEuso::ClearMemory() {
    //
    //
    //

    Reset();

    return fPhotonsOnPupil ? fPhotonsOnPupil->ClearMemory() : kTRUE;
}

//______________________________________________________________________________
void SinglePhotonRootFileLightToEuso::Init() {
    //
    // Open file and associate objects
    //
    
    string ext = ".root";
    if (fFileName.rfind(ext) != (fFileName.size()-ext.size()))
        fFileName += ext;

    fFile =  new TFile(fFileName.c_str());
    if ( fFile->IsZombie() )
        FatalError("Unable to open "+fFileName);

    if (!(fTree = (TTree*)fFile->Get("etree"))) 
        FatalError("No etree in "+fFileName);
    
    // tell which branch will be read
    fTree->SetBranchStatus("*",0);
    fTree->SetBranchStatus("fNumSingles",1);
    fTree->SetBranchStatus("fSingle*",1);
    fTree->SetBranchStatus("fTrue*",1);
    fTree->SetBranchStatus("fFoV",1);
    fTree->SetBranchStatus("fRadius",1);
    fTree->SetBranchStatus("fPosX",1);
    fTree->SetBranchStatus("fPosY",1);
    fTree->SetBranchStatus("fPosZ",1);
    
    // set branches
    fEv->SetBranches(fTree);
    fAtmosphere   = fEv->GetAtmosphere();
    fRootTruth    = fEv->GetTruth();
    fRootGeometry = fEv->GetGeometry();
    
    if(!fAtmosphere) Msg(EsafMsg::Panic) << "No Atmosphere in root input file" << MsgDispatch;
    if(!fRootTruth) Msg(EsafMsg::Panic) << "No Truth in root input file" << MsgDispatch;
    if(!fRootGeometry) Msg(EsafMsg::Panic) << "No Geometry in root input file" << MsgDispatch;
  
    fEvCounter = fFirstEvent;
    
    // check if atmosphere is recorded in output, it should not so far (viewer related pb)
    ConfigFileParser* pConfig = Config::Get()->GetCF("General","SimuRootFileManager");
    if(pConfig->GetBool("SimuRootFileManager.fSaveAtmosphere") || pConfig->GetBool("SimuRootFileManager.fSaveShower"))
        Msg(EsafMsg::Panic) << "Atmosphere and Shower SHOULD not be recorded in root output file" << MsgDispatch;
}

//______________________________________________________________________________
PhotonsOnPupil* SinglePhotonRootFileLightToEuso::Get(const DetectorGeometry* dg) {
    //
    // Reads file and creates the list of photons
    //

    // clear objects
    fPhotonsOnPupil->Clear();
    fMCTruth->Clear();
    TRandom* rndm = EsafRandom::Get();

    // stop simulation when last event of input root file is reached
    if ( fEvCounter < fTree->GetEntriesFast() ) {
        // read input root file
        fTree->GetEntry(fEvCounter);
        fNumPhotons = fAtmosphere->GetNumSingles();
    }
    else if(!fSeveralEvents) Msg(EsafMsg::Panic) << "No more event in root input file. Simu is stopped !" << MsgDispatch;
    

    // check if current geometry matches the input rootfile one
    EarthVector current = dg->GetPos();
    EarthVector input   = fRootGeometry->GetPos();
    if( (current - input).Mag() > kTolerance ) {
        Msg(EsafMsg::Warning) << "current = " << current <<", input = " << input << MsgDispatch;
        Msg(EsafMsg::Panic) << "Current detector position does NOT match the one used for input file" << MsgDispatch;
    }
    Double_t Current = dg->GetFoV();
    Double_t Input   = fRootGeometry->GetFoV();
    if( fabs(Current - Input) > kTolerance ) {
        Msg(EsafMsg::Warning) << "current = " << Current <<", input = " << Input << MsgDispatch;
        Msg(EsafMsg::Panic) << "Current detector FoV does NOT match the one used for input file" << MsgDispatch;
    }
    Current = dg->GetRadius();
    Input   = fRootGeometry->GetRadius();
    if( fabs(Current - Input) > kTolerance ) {
        Msg(EsafMsg::Warning) << "current = " << Current <<", input = " << Input << MsgDispatch;
        Msg(EsafMsg::Panic) << "Current detector Radius does NOT match the one used for input file" << MsgDispatch;
    }

    // generate photons on pupil from singlephoton information
    SinglePhoton* sph = 0;
    Double_t date(0.), wl(0.), r(0.), phi(0.), tof(0.);
    EarthVector pos, dir, local_dir;
    EarthVector EUSO(dg->GetPos());
    PhotonType type;
    PhotonStatus status;
    UInt_t bid = 0;
    Int_t nTransmitted = 0;
    
    for( Int_t i(0); i<fNumPhotons; i++) {
        // copy root singlephoton info, if photon has not been absorbed during last transfer in atmosphere
	ESinglePhoton* ph = fAtmosphere->GetSingle(i);
	if(fNewPhFormat && ph->IsAbsorbed() && !ph->IsCloudAbsorbed()) continue;
	else if(!fNewPhFormat && ph->IsAbsorbed()) continue;
	date = ph->GetDate();
	wl = ph->GetWl();
	pos = ph->GetShowerPos();
	dir = (EUSO - ph->GetPos()).Unit(); 
	type = PhotonType(ph->GetType());
	status = PhotonStatus(ph->GetHistory());
	tof = ph->GetTof();
	bid = ph->GetBunchId();
	if(!fNewPhFormat) sph = new SinglePhoton(date,wl,pos,dir,type,status,bid);
	else {
	    sph = new SinglePhoton(date,wl,pos,dir,type,status,bid,ph->GetShowerAge());
	    sph->SetCloudAbsorbed(ph->IsCloudAbsorbed());
	}
	sph->SetPosInAtmo(ph->GetPos());
	sph->SetPos(EUSO);
	sph->AddToTof(tof);
	sph->SetOutFoV(ph->IsOutFoV());
	local_dir = sph->Dir();
	for(Int_t ii=0; ii < ph->GetNbinter(); ii++) sph->AddInteraction();
	
	// find random position on pupil (not used in SimpleDetector simulation)
	// WARNING : decoupled mode is not handled here !!
        r   = dg->GetRadius() * Sqrt(rndm->Rndm());
        phi = TwoPi() * rndm->Rndm();
        pos.SetXYZ(r*Cos(phi),r*Sin(phi),0);
        local_dir *= *(dg->GetToLocal());
	
	// convert it into parentphoton, then add it to the list
	ConvertThenAdd(*sph,pos,local_dir);
	nTransmitted++;
	SafeDelete(sph);
    }


    // print info about read event
    MsgForm(EsafMsg::Info,"   SinglePhotonRootFile event %d read", fEvCounter);
    MsgForm(EsafMsg::Info,"   Energy: %.3g MeV ParticleCode: %d",
            fRootTruth->GetTrueEnergy(), fRootTruth->GetTrueParticleCode());
    MsgForm(EsafMsg::Info,"   Theta: %.3f Phi: %.3f", fRootTruth->GetTrueTheta()*RadToDeg(),
            fRootTruth->GetTruePhi()*RadToDeg());
    MsgForm(EsafMsg::Info,"   NumPhotons %d, Transmitted %d",
            fNumPhotons, nTransmitted);


    // copy ETruth in MCTruth
    fMCTruth->SetEnergy( fRootTruth->GetTrueEnergy() );
    fMCTruth->SetThetaPhi( fRootTruth->GetTrueTheta(), fRootTruth->GetTruePhi() );
    fMCTruth->SetParticle( fRootTruth->GetTrueParticleCode() );
    fMCTruth->SetFirstInt( fRootTruth->GetTrueInitPos(), fRootTruth->GetTrueX1() );
    fMCTruth->SetEarthImpact( fRootTruth->GetTrueEarthImpact(), fRootTruth->GetTrueEarthAge() );
    fMCTruth->SetTOAImpact( fRootTruth->GetTrueTOAImpact() );
    fMCTruth->SetShowerMax( fRootTruth->GetTrueShowerMaxPos(), fRootTruth->GetTrueShowerXMax() );
    fMCTruth->SetHclouds( fRootTruth->GetTrueHclouds() );
    fMCTruth->SetCloudsthick( fRootTruth->GetTrueCloudsthick() );
    fMCTruth->SetCloudsOD( fRootTruth->GetTrueCloudsOD() );
    fMCTruth->SetLatitude( fRootTruth->GetTrueLatitude() );
    fMCTruth->SetLongitude( fRootTruth->GetTrueLongitude() );
    fMCTruth->SetDate( fRootTruth->GetTrueDate() );
    fMCTruth->SetHGround( fRootTruth->GetTrueHGround() );
    fMCTruth->SetGroundAlbedo( fRootTruth->GetTrueGroundAlbedo() );
    fMCTruth->SetGroundSpecular( fRootTruth->GetTrueGroundSpecular() );


    // save the truth
    EEvent* ev = EEvent::GetCurrent();
    if ( ev ) {
        EEventTruthAdder a( fMCTruth );
        ev->Fill(a);
    }

    fEvCounter++;

    return fPhotonsOnPupil;
}

//______________________________________________________________________________
void SinglePhotonRootFileLightToEuso::ConvertThenAdd(const SinglePhoton& sph, const EarthVector& pos, const EarthVector& local_dir) {
    //
    // convert SinglePhoton into ParentPhoton, then add it to fPhotonsOnPupil
    //
    
    if(!sph.IsAbsorbed()) fPhotonsOnPupil->Add(sph,pos,local_dir);
}
