// ESAF : Euso Simulation and Analysis Framework
// $Id: FastRadiativeProcessesCalculator.cc 2764 2006-11-15 14:15:25Z moreggia $
// Sylvain Moreggia created Jan, 29 2004

#include "FastRadiativeProcessesCalculator.hh"
#include "EarthVector.hh"
#include "Atmosphere.hh"
#include <math.h>
#include "BunchOfPhotons.hh"
#include "SinglePhoton.hh"
#include "Config.hh"
#include "EConst.hh"
#include "EsafRandom.hh"

ClassImp(FastRadiativeProcessesCalculator)

using EConst::Avogadro;
using namespace sou;
using namespace TMath;

//_____________________________________________________________________________
FastRadiativeProcessesCalculator::FastRadiativeProcessesCalculator() : RadiativeProcessesCalculator() {
    //
    // ctor
    //
    Msg(EsafMsg::Info) <<"*** FastRadiativeProcessesCalculator built ***" << MsgDispatch;
    fName = "fast";
}

//_____________________________________________________________________________
FastRadiativeProcessesCalculator::~FastRadiativeProcessesCalculator() {
    //
    // dtor
    //
}

//_____________________________________________________________________________
Double_t FastRadiativeProcessesCalculator::Trans(const SinglePhoton& p, const EarthVector& finalpos, Double_t* TransArray) const {
    //
    // total Transmission between SinglePhoton position and finalpos
    //
    
    // 1. rayleigh transmission : uses Lowtran formula
    const Atmosphere* atmo = Atmosphere::Get();
    Double_t X = atmo->Grammage(p.Pos(),finalpos,"pos",50*km);
    Double_t wlterm = pow(p.Wl()/cm,-4) / (1 - 1.08426/9.38076 * 1e-9 * pow(p.Wl()/cm,-2));
    Double_t Trayl = exp( -8.24689e-22*X*cm2/gram * wlterm);
    // 1.1. analytical rayleigh
    //Double_t ro_0 = atmo->Air_Density(0.);
    //Double_t norm = 8*pow(pi,3)*4/3;
    //norm *= (6 + 3*0.0279) / (6 - 7*0.0279);
    //norm *= pow(2.83e-4,2)*28.966/Avogadro();
    //Double_t To = exp( -(norm*X/(gram/cm2)) / (pow(p.Wl()/cm,4)*pow(ro_0/(gram/cm3),2)) );
    // 1.2. these kevin Reil, HiRes
    //Trayl = exp(-X/(2970*gram/cm2)*pow(400*nm/p.Wl(),4));
    
    
    // 2. clouds transmission
    Double_t OD = atmo->GetClouds()->TotalOD(p.Pos(),p.Dir());
    Double_t Tclouds = exp(-OD);
    //TOFIX : so far is OK cause this method only used to go toward detector : should stop at finalpos actually
    if(finalpos.Zv() < 100*km) Msg(EsafMsg::Warning) <<"<Trans> Clouds algo MUST be ADAPTED if finalpos is not DETECTOR position" << MsgDispatch;
     
    // 3. ozone transmission
    Double_t OzoneAmount = atmo->OzoneAmountAlongPath(p.Pos(),finalpos,p.Wl());
    Double_t Tozone = exp(-OzoneAmount);
    
    // 4. Aerosol transmission
    Double_t dummy = 0.;
    Double_t Taerosol = atmo->GetAerosol()->Trans(p.Pos(),finalpos,p.Wl(),dummy);
    
    if(TransArray) {
        TransArray[0] = Trayl * Tozone * Taerosol * Tclouds;
	TransArray[1] = Trayl;
	TransArray[2] = Tozone;
	TransArray[3] = Taerosol;
    }
    	
    return Trayl * Tclouds * Tozone * Taerosol;
} 

//_____________________________________________________________________________
Double_t FastRadiativeProcessesCalculator::RandomGrammage_RaylScat(Double_t Wl) const {
    //
    // sample the rayleigh exponential distribution in grammage -> get grammage at which scattering occurs 
    //
    
    TRandom* rndm = EsafRandom::Get();
    Double_t wlterm = pow(Wl/cm,-4) / (1 - 1.08426/9.38076 * 1e-9 * pow(Wl/cm,-2));
    Double_t grammage = log(1 - rndm->Rndm()) / (-8.24689e-22*cm2/gram * wlterm);
    
    return grammage; // in esaf units
} 

//_____________________________________________________________________________
Int_t FastRadiativeProcessesCalculator::RandomOzoneAbsorbPosition(const EarthVector& pos, const EarthVector& dir, Double_t wl, EarthVector& rtn, Double_t TOA) const {
    //
    // Generate a random Ozone absorption position along the given path (pos,dir)
    // return the status of Atmosphere::InvertOzoneAmount() method
    //
    
    // init
    TRandom* rndm = EsafRandom::Get();
    Int_t status = -1;
    Double_t RNDM = -Log(1. - rndm->Rndm());
    
    // get position of interaction
    status = Atmosphere::Get()->InvertOzoneAmount(pos,dir,RNDM,wl,rtn,TOA);
    
    return status;
} 

//_____________________________________________________________________________
Double_t FastRadiativeProcessesCalculator::Trans(const BunchOfPhotons& b,const EarthVector& nextpos,vector<Double_t*>& coeff) const {
    //
    // total transmission between bunch and a final position.   
    // //TOFIX : must be adapted to include Ozone (so far useless cause Lowtran is used for bunch)
    //
    
    // init
    Double_t wl, wlterm;
    const Atmosphere* atmo = Atmosphere::Get();
    Double_t X = atmo->Grammage(b.GetPos(),nextpos,"pos",50*km);
    Int_t nb = b.GetWlSpectrum().GetAxis().GetNbins();
    
    // calculations
    for(Int_t i=0; i<nb; i++) {
        
	// 1. rayleigh transmission : uses Lowtran formula
        wl = b.GetWlSpectrum().GetLambdaEntry(i);
        wlterm = pow(wl/cm,-4) / (1 - 1.08426/9.38076 * 1e-9 * pow(wl/cm,-2));
        Double_t Trayl = exp( -8.24689e-22*X*cm2/gram * wlterm);
	// 1.1. analytical rayleigh
	//Double_t ro_0 = atmo->Air_Density(0.);
	//Double_t norm = 8*pow(pi,3)*4/3;
	//norm *= (6 + 3*0.0279) / (6 - 7*0.0279);
	//norm *= pow(2.83e-4,2)*28.966/Avogadro();
	//Double_t To = exp( -(norm*X/(gram/cm2)) / (pow(p.Wl()/cm,4)*pow(ro_0/(gram/cm3),2)) );
	// 1.2. these kevin Reil, HiRes
	//Trayl = exp(-X/(2970*gram/cm2)*pow(400*nm/p.Wl(),4));
    
	// 2. ozone transmission
	Double_t OzoneAmount = atmo->OzoneAmountAlongPath(b.GetPos(),nextpos,wl);
	Double_t Tozone = exp(-OzoneAmount);

	// 3. Aerosol transmission
	Double_t dummy = 0.;
	Double_t Taerosol = atmo->GetAerosol()->Trans(b.GetPos(),nextpos,wl,dummy);
    
	
        coeff[0][i] = Trayl * Tozone * Taerosol;
	coeff[1][i] = Trayl;
	coeff[2][i] = Tozone;
	coeff[3][i] = Taerosol;
    }
    
    // 4. clouds transmission
	 //TOFIX : work only with O1_propagator (no rayleigh scattering simu)
    Double_t OD = atmo->GetClouds()->TotalOD(b.GetPos(),(nextpos - b.GetPos()).Unit());
    Double_t Tclouds = exp(-OD);

    return Tclouds;
}





