// ESAF : Euso Simulation and Analysis Framework
// $Id: TestGround.cc 2733 2006-06-09 12:43:12Z moreggia $
// S. Moreggia created 27 October 2003

#include "euso.hh"
#include "TestGround.hh"
#include "utils.hh"
#include <stdlib.h>
#include "EConst.hh"
#include "EsafRandom.hh"
#include "SinglePhoton.hh"
#include "ListPhotonsInAtmosphere.hh"
#include <TF2.h>

ClassImp(TestGround)

using EConst::EarthRadius;
using namespace TMath;
using namespace sou;


//______________________________________________________________________________
Double_t Specular_phase_function(Double_t *x, Double_t *par) {
    //
    // for specular reflexion mode
    // calculate value of 2D phase function using theta-phi
    //
    // VECTORS ARE TAKEN LOCALLY, i.e. AT GROUND IMPACT
    // so that angles are define w.r.t. local vertical
    //
    // x[0]    =  theta out
    // x[1]    =  phi   out
    // par[0]  =  theta incident  (incident direction has been reversed !!)
    // par[1]  =  phi   incident  (incident direction has been reversed !!)
    // par[2]  =  gaussian sigma parameter
    //
    //
    
    // 1. init
    EarthVector incomdir, outgodir, Uz;
    incomdir.SetMagThetaPhi(1.,par[0],par[1]);
    outgodir.SetMagThetaPhi(1.,x[0],x[1]);
    Uz.SetXYZ(0,0,1.);
    
    // 1. define change of frame : outgodir expressed in the frame where incomdir is the Z-axis (incomdir has been reversed before)
    EarthVector rotaxis_spec = Uz.Cross(incomdir);
    Double_t rotangle_spec   = -Uz.Angle(incomdir);
    
    // 2. move outgoing direction into new frame and get the out angle
    outgodir.Rotate(rotangle_spec,rotaxis_spec);
    Double_t th_out_spec  = outgodir.Theta();
    
    // 3. if th_out_spec > 90deg, zero returned
    if(th_out_spec > PiOver2()) return 0.;
        
    return 1. / (Sqrt(TwoPi()) * par[2]) * Exp(-th_out_spec*th_out_spec / (2*par[2]*par[2]));
}

//_____________________________________________________________________________________________________
TestGround::TestGround() {
    //
    // ctor
    //
    
    Specular_distrib = 0;
    Configure();
    Msg(EsafMsg::Info) << "*** TestGround built ***" << MsgDispatch;
}

//_____________________________________________________________________________________________________
TestGround::~TestGround(){
    //
    // dtor
    //
    
    if(Specular_distrib) SafeDelete(Specular_distrib);
}

//_____________________________________________________________________________________________________
void TestGround::Configure() {
    //
    // Configure the object
    //
    
    fType = Conf()->GetStr("TestGround.fType");
    fSpec = Conf()->GetNum("TestGround.fSpec");
    fSigma = Conf()->GetNum("TestGround.fSigma")*DegToRad();
    fWindSpeed = Conf()->GetNum("TestGround.fWindSpeed");
    fAlbedo  = Conf()->GetNum("TestGround.fAlbedo");
    fAltitude = Conf()->GetNum("TestGround.fAltitude")*km;
    Msg(EsafMsg::Info) <<"Ground with Albedo = " << fAlbedo << MsgDispatch;
    
    // if specular component exists
    if(fType == "mixed") {
        if(fSigma < 0) {
            fSigma = ((fWindSpeed - 2) + 6)*DegToRad();
	    fSpec  = 0.75;
        }
        Msg(EsafMsg::Info) <<"Ground with mixed BRDF, specular component = " << fSpec <<", Gaussian width = "<<fSigma*RadToDeg()<< "deg" << MsgDispatch;
	Specular_distrib = new TF2("Specular_distrib",Specular_phase_function,-PiOver2(),PiOver2(),-Pi(),Pi(),3);
	
	// create tables for specular component normalization
        Msg(EsafMsg::Info) <<"Build specular component normalization tables" << MsgDispatch;
	NORMmax = 0;
	for(Int_t k=0; k< 91; k++) {
	    Specular_distrib->SetParameters(k*DegToRad(),0,fSigma);
	    // calculate normalization factor to achieve a self-normalized specular intensity 2D-distribution
	    // use trapezoidal integration (in theta and phi)
	    Double_t norm = 0.;
	    Double_t Integ_phi1(0.), Integ_phi2(0.);
	    Int_t nbsteps = 100;  // allow precision better than 0.001
	    Double_t ttheta(0.), pphi(0.), f1(0.), f2(0.), dOmega1(0.), dOmega2(0.);
	    Double_t dth = PiOver2() / Double_t(nbsteps);
	    Double_t dph = TwoPi() / Double_t(nbsteps);

	    // first value for Integ_phi
	    ttheta = 0.;
	    pphi = -Pi();
	    f2 = Specular_distrib->Eval(ttheta,pphi);
	    dOmega2 = 0.;     // zero cause Sin(th=0) term
	    Integ_phi2 = 0.;  // zero cause Sin(th=0) term
	    // loop on theta
	    for(Int_t i=0; i < nbsteps; i++) {
		ttheta += dth;
		pphi = -Pi();
		// first value for f at this theta
		f2 = Specular_distrib->Eval(ttheta,pphi);
		Integ_phi1 = Integ_phi2;
		Integ_phi2 = 0.;
		// loop on phi
		for(Int_t j=0; j< nbsteps; j++) {
		    f1 = f2;
	            pphi += dph;
		    f2 = Specular_distrib->Eval(ttheta,pphi);
		    Integ_phi2 += 0.5 * (f1 + f2);
		}
		dOmega1 = dOmega2;
		dOmega2 = Sin(ttheta) * dth * dph;
		norm += 0.5 * (Integ_phi1 * dOmega1 + Integ_phi2 * dOmega2);
	    }
            NORM[k] = norm;
	    if(norm > NORMmax) NORMmax = norm;
	}
	
        Msg(EsafMsg::Info) <<"Tables created" << MsgDispatch;
    }
}

//_____________________________________________________________________________________________________
void TestGround::RandomDir(const EarthVector& pos, EarthVector& incom_dir) const {
    //
    // set incom_dir to the outgoing direction of scattering, sampling Outgoing phase function
    // - pos       = position on ground (MES frame)
    // - incom_dir = incident direction (MES frame)  :  FINALLY SET TO THE OUTGOING DIRECTION
    //
    // Phase function is normalized when integrated over an hemisphere, i.e. theta within [0,Pi()/2]   (SOLID ANGLE integration !!)
    // Specular pattern is assumed cirular in shape (i.e. independent of phi angle around incomdir_local)
    // Specular pattern is coded as a gaussian for INTENSITY (NOT for radiance)
    //
    
    // init
    TRandom* rndm = EsafRandom::Get();
    Double_t theta_out(0.), th_in(0.), phi_out(0.), phi_in(0.);
    EarthVector incomdir_local = -incom_dir.Unit();   // WARNING : local incomdir is REVERSED !
    EarthVector outgodir_local(1,0,0);
    const EarthVector Uz(0,0,1);
    
    // 0. do some checks
        // 0.1. in weird cases (should not occur)
    if(incom_dir.Mag() == 0) {
        Msg(EsafMsg::Warning) << "<RandomDir> NO INCOMING dir, it SHOULD NOT, direction is not changed" << MsgDispatch;
	return;
    }
        // 0.2. check if pos is at ground level
    if((pos.Zv() - fAltitude) > kAltitudeTolerance) {
        Msg(EsafMsg::Warning) << "<RandomDir> Ask for BRDF but position is NOT at ground level, delta_h = "<<pos.Zv() - fAltitude<<", direction is not changed" << MsgDispatch;
	return;
    }

    
    // 1. move to local frame
    //    rotation axis  defined by       Uz x earthlocalaxis(pos)       (expressed in MES)
    //    rotation angle defined by       Angle(Uz,earthlocalaxis)       (expressed in MES)
    //
        // 1.0. check the very specific case where pos is (0,0,0)
	//      if it is the case, does not apply change of frame
    EarthVector rotaxis(1,0,0);
    Double_t rotangle = 0.;
    if(pos.Perp() > kAltitudeTolerance) {
        // 1.1. init
	EarthVector earthlocalaxis(pos);
	earthlocalaxis(2) +=  EarthRadius();
	earthlocalaxis.Unit();
        // 1.2. define rotation
	rotaxis = Uz.Cross(earthlocalaxis);
	rotangle   = Uz.Angle(earthlocalaxis);
	if(rotaxis.Mag() < kTolerance) {      // rotaxis should never be null as pos is not (0,0,0)
            Msg(EsafMsg::Warning) << "<RandomDir> Rotation axis is NULL (never should!), direction is not changed" << MsgDispatch;
	    return;
        }
        // 1.3. move direction into local frame
	incomdir_local.Rotate(rotangle,rotaxis);

        // 1.4. Define local angles (local incomdir is REVERSED !)
        th_in   = incomdir_local.Theta();
        phi_in  = incomdir_local.Phi();
    }
    else {
        th_in   = incomdir_local.Theta();
        phi_in  = incomdir_local.Phi();
    }


    // 2. checks if incoming local direction is locally going downward (WARNING : local incomdir is REVERSED !)
    if(th_in > PiOver2() + kTolerance) {
        Msg(EsafMsg::Warning) << "\n<RandomDir> theta_local/halfpi = "<<th_in/PiOver2()<<", theta_local = "<<th_in*RadToDeg()<<" deg" << MsgDispatch;
        Msg(EsafMsg::Warning) << "<RandomDir> pos = "<<pos << MsgDispatch;
        Msg(EsafMsg::Warning) << "<RandomDir> Incoming local direction is going upward (can occur if theta close to 90deg, round-off issue), direction put at 90deg (locally horizontal)" << MsgDispatch;
	th_in = PiOver2();
    }


    // 3. Get an outgoing direction in sampling Outgoing_phase_function
    if(fType == "isotropic") {  // theta between [0,pi/2] only
        theta_out = acos(rndm->Rndm());        // dP/dcos(theta) = 2pi * dP/dphi.dcos(theta) --- dP/dphi.dcos(theta) = 1/2pi, si cos(theta) = x -->  dProba/dx = 1
	phi_out   = rndm->Rndm()*TwoPi();
    }
    
    else if(fType == "lambertian") {
        theta_out = acos(sqrt(rndm->Rndm())); // dP/dcos(theta) = 2pi * dP/dphi.dcos(theta)  --- dP/dphi.dcos(theta) = cos(theta)/pi, si cos(theta) = x -->  dProba/dx = 2x
	phi_out   = rndm->Rndm()*TwoPi();
    }
    
    else if(fType == "mixed") {
	// photon is reflected diffusely
	if(fSpec <  rndm->Rndm()) {  // theta between [0,pi/2] only
            theta_out = acos(sqrt(rndm->Rndm())); // dP/dcos(theta) = 2pi * dP/dphi.dcos(theta)  --- dP/dphi.dcos(theta) = cos(theta)/pi, si cos(theta) = x -->  dProba/dx = 2x
	    phi_out   = rndm->Rndm()*TwoPi();
	}
	else {
	    // photon is reflected by specular process
	    // cos(theta) and phi are two independent variables
	    Double_t pfvalue = 0.;
	    Double_t u1(0.), u2(0.);
	    do {
		//  3.1. specular component value sample uniformly between 0 and max
        	u2 = rndm->Rndm() / (Sqrt(TwoPi()) * fSigma);

        	// 3.2. specular component at ASin(u1)
	            // 3.2.1. init function using incoming angles info
        	Specular_distrib->SetParameters(th_in,phi_in,fSigma);
	            // 3.2.3. set direction to get specular distribution needed value
        	u1 = rndm->Rndm();  // theta between [0,pi/2] only
		theta_out = ACos(u1);
                phi_out = TwoPi()*rndm->Rndm();
		outgodir_local.SetMagThetaPhi(1.,theta_out,phi_out);
        	pfvalue += Specular_distrib->Eval(theta_out,phi_out);
	    } while(u2 >= pfvalue);
	}
    }
    
    else Msg(EsafMsg::Panic) << "<RandomDir> Unvalid type of brdf = "<< fType << MsgDispatch;
    
    // 4. returns in the MES
        // check if outgoing direction is upward
    if(theta_out > PiOver2() + kTolerance) {
        Msg(EsafMsg::Warning) << "<RandomDir> Outgoing local direction is going downward (never should!),  direction is not changed" << MsgDispatch;
	return;
    }
    outgodir_local.SetMagThetaPhi(1.,theta_out,phi_out);
    if(pos.Perp() > kAltitudeTolerance) outgodir_local.Rotate(-rotangle,rotaxis);
    incom_dir = outgodir_local; // outgoing direction
}

//_____________________________________________________________________________________________________
Double_t TestGround::Outgoing_phase_function(const EarthVector& pos, const EarthVector& incom, const EarthVector& outgo) const {
    //
    // Phase function of the outgoing light (earth sphericity taken into account)
    // - pos      = position on ground (MES frame)
    // - incomdir = incident direction (MES frame)
    // - outgodir = outgoing position  (MES frame)
    // Phase function is normalized when integrated over an hemisphere, i.e. theta within [0,Pi()/2]   (SOLID ANGLE integration !!)
    // Specular pattern is assumed cirular in shape (i.e. independent of phi angle around incomdir_local)
    // Specular pattern is coded as a gaussian for INTENSITY (NOT radiance)
    //
    
    // init
    Double_t rtn(0);
    Double_t th_out(0.), th_in(0.), phi_out(0.), phi_in(0.);
    EarthVector incomdir_local = -incom.Unit();   // WARNING : local incomdir is REVERSED !
    EarthVector outgodir_local = outgo.Unit();
    const EarthVector Uz(0,0,1);
    
    // 0. do some checks
        // 0.1. in weird case incomdir comes from a fluo bunchofphotons (dir.mag=0)
    if(incom.Mag() == 0) {
        Msg(EsafMsg::Warning) << "<Outgoing_phase_function> NO INCOMING dir, it SHOULD NOT, ZERO RETURNED" << MsgDispatch;
	return 0.;
    }
        // 0.2. check if pos is at ground level
    if((pos.Zv() - fAltitude) > kAltitudeTolerance) {
        Msg(EsafMsg::Warning) << "<Outgoing_phase_function> Ask for BRDF but position is NOT at ground level, delta_h = "<<pos.Zv() - fAltitude<<"  ZERO RETURNED" << MsgDispatch;
	return 0.;
    }

    
    // 1. move to local frame
    //    rotation axis  defined by       Uz x earthlocalaxis(pos)       (expressed in MES)
    //    rotation angle defined by       Angle(Uz,earthlocalaxis)       (expressed in MES)
    //
        // 1.0. check the very specific case where pos is (0,0,0)
	//      if it is the case, does not apply change of frame
    if(pos.Perp() > kAltitudeTolerance) {
        // 1.1. init
	EarthVector earthlocalaxis(pos);
	earthlocalaxis(2) +=  EarthRadius();
	earthlocalaxis.Unit();
        // 1.2. define rotation
	EarthVector rotaxis = Uz.Cross(earthlocalaxis);
	Double_t rotangle   = Uz.Angle(earthlocalaxis);
	if(rotaxis.Mag() < kTolerance) {      // rotaxis should never be null as pos is not (0,0,0)
            Msg(EsafMsg::Warning) << "<Outgoing_phase_function> Rotation axis is NULL (never should!), ZERO RETURNED" << MsgDispatch;
	    return 0.;
        }
        // 1.3. move directions into local frame
	incomdir_local.Rotate(rotangle,rotaxis);
	outgodir_local.Rotate(rotangle,rotaxis);

        // 1.4. Define local angles (local incomdir is REVERSED !)
        th_out  = outgodir_local.Theta();
        th_in   = incomdir_local.Theta();
        phi_out = outgodir_local.Phi();
        phi_in  = incomdir_local.Phi();
    }
    else {
        th_out  = outgodir_local.Theta();
        th_in   = incomdir_local.Theta();
        phi_out = outgodir_local.Phi();
        phi_in  = incomdir_local.Phi();
    }

    // 2. checks if : 
    //    - outgoing local direction is locally going upward
    //    - incoming local direction is locally going downward (WARNING : local incomdir is REVERSED !)
    if(th_in > PiOver2() + kTolerance) {
        Msg(EsafMsg::Warning) << "\n<Outgoing_phase_function> theta_local/halfpi = "<<th_in/PiOver2()<<", theta_local = "<<th_in*RadToDeg()<<" deg" << MsgDispatch;
        Msg(EsafMsg::Warning) << "<Outgoing_phase_function> pos = "<<pos << MsgDispatch;
        Msg(EsafMsg::Warning) << "<Outgoing_phase_function> Incoming local direction is going upward (can occur if theta close to 90deg, round-off issue), direction put at 90deg (locally horizontal)" << MsgDispatch;
	th_in = 89.9*DegToRad();
    }
    if(th_out > PiOver2()) return 0.;


    // 3. Calculate value of reflected intensity (= outgoing phase function)
    if(fType == "isotropic")
        rtn = 1./TwoPi();
    
    else if(fType == "lambertian")
        rtn = Cos(th_out)/Pi();
    
    else if(fType == "mixed") {
        // diffuse intensity component (self-normalized)
	rtn = (1 - fSpec) / Pi() * Cos(th_out);
	// specular radiance component
	Specular_distrib->SetParameters(th_in,phi_in,fSigma); // init function using incoming angles info
	Int_t index = Int_t(th_in*RadToDeg());  // to retrieve factor that achieves a normalized 2D-distribution
	if(index > 90) {
            Msg(EsafMsg::Warning) << "<Outgoing_phase_function, norm factor> Int_t(theta) is too high, th = "<<th_in*RadToDeg()<<". Set to 90deg" << MsgDispatch;
	    index = 90;
        }
	Double_t norm = (NORM[index+1] - NORM[index]) * (th_in*RadToDeg() - Double_t(index)) + NORM[index];  // linear interpolation
	rtn += fSpec * Specular_distrib->Eval(th_out,phi_out) / norm;
    }
    
    else Msg(EsafMsg::Panic) << "<Outgoing_phase_function> Unvalid type of brdf = "<< fType << MsgDispatch;

    return rtn;
}

//______________________________________________________________________________
Double_t TestGround::GetMaxOutgoing_phase_function() const {
    //
    // maximum value of outgoing phase function
    // NB : Phase function is normalized when integrated over the upper hemisphere, i.e. outoing theta direction within [0,Pi()/2]
    //

    Double_t rtn = 0.;
    
    if(fType == "isotropic")
        rtn = 1./TwoPi();
    
    else if(fType == "lambertian")
        rtn = 1./Pi();
    
    else if(fType == "mixed")
        rtn = (1 - fSpec)/Pi() + fSpec / NORMmax / (Sqrt(TwoPi()) * fSigma);
    
    else Msg(EsafMsg::Panic) << "<Outgoing_phase_function> Unvalid type of brdf = "<< fType << MsgDispatch;
        
    return rtn;
}

//_____________________________________________________________________________________________________
EarthVector TestGround::GetImpact(const EarthVector& pos, const EarthVector& dir) const {
    //
    // Impact on ground form a given (position,direction) pair in atmosphere
    //

    // if pos is underground
    EarthVector rtn;
    if((pos.Zv() + kAltitudeTolerance) < fAltitude) {
        Msg(EsafMsg::Debug) << "Starting position is underground, so no ground impact calculated" << MsgDispatch;
        rtn.SetXYZ(0,0,-HUGE);
        return rtn;
    }

#ifdef DEBUG
   // Msg(EsafMsg::Debug) <<"Impact on ground = " << MsgDispatch;
#endif

    // to know if dir is locally going upward
    Double_t angle;
    EarthVector temp(pos);
    temp(2) += EarthRadius();  // pos expressed in earth-centered frame -> gives local vertical direction expressed in MES frame
    angle = dir.Angle(temp);   // angle between dir and local vertical
    if(angle < (PiOver2() - kTolerance)) {
      //Msg(EsafMsg::Debug) << MsgDispatch;
        //Msg(EsafMsg::Debug) <<"NO IMPACT ON GROUND, track is locally going upward" << MsgDispatch;
        rtn.SetXYZ(0,0,HUGE);
        return rtn;
    }

    // now impact calculation
    Double_t mag(0);
    EarthVector direc = dir.Unit();

/*  flat earth
if(fabs(direc.Z()) < kTolerance) rtn.SetXYZ(0,0,HUGE);
else {
    mag = -pos.Z() / direc.Z();
    if(mag < -kAltitudeTolerance) rtn.SetXYZ(0,0,HUGE);
    else rtn = pos + mag*direc;
}
*/

    // spherical earth
    Double_t b = pos*direc + direc.Z()*EarthRadius();
    Double_t c = pos.Mag2() + 2*EarthRadius()*(pos.Z() - fAltitude) - fAltitude*fAltitude;
    pair<Int_t,Double_t*>& p = findRoots(1.,2*b,c);
    if(p.first == 0) {
#ifdef DEBUG
        //Msg(EsafMsg::Debug) <<"NO IMPACT ON GROUND" << MsgDispatch;
#endif
        rtn.SetXYZ(0,0,HUGE);
        return rtn;
    }
    if(p.first == 1) mag = p.second[0];
    else if(p.first ==2) mag = min(p.second[0],p.second[1]);
    
    if(mag < -kAltitudeTolerance) Msg(EsafMsg::Debug) <<"SHOULD NOT HAVE IMPACT ON GROUND, mag = "<<mag << MsgDispatch;
    
    rtn = pos + mag*direc;

#ifdef DEBUG
    //Msg(EsafMsg::Debug) << rtn << MsgDispatch;
#endif


    return rtn;
}


